(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Extra operations on LLists\<close>

text \<open>
the operations  kfilter, lleast, lbutlast and lidx on coinductive lists are defined together with 
a library of lemmas. 
\<close>

theory LListKFilter 

 imports 
  "Coinductive.Coinductive_List"
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)
definition kfilter :: " ('a \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> 'a llist  \<Rightarrow> nat llist"
 where " kfilter P n xs  = lmap snd (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))"

definition lleast :: "('a \<Rightarrow> bool) \<Rightarrow> 'a llist \<Rightarrow> nat " 
 where "lleast P xs = (LEAST na. na < llength xs \<and> P (lnth xs na)) "

primcorec lbutlast :: "'a llist \<Rightarrow> 'a llist"
 where "lbutlast xs =
        (case xs of LNil \<Rightarrow> LNil |
          (LCons x xs') \<Rightarrow> (if xs' = LNil then LNil else (LCons x (lbutlast xs'))))"

simps_of_case lbutlast_simps [simp, code, nitpick_simp]: lbutlast.code

definition lidx :: "  nat llist \<Rightarrow> bool"
 where " lidx  xs \<longleftrightarrow>  (\<forall>n. (Suc n)<llength xs \<longrightarrow> lnth xs n < lnth xs (Suc n)) "

subsection \<open>Auxiliary lemmas\<close>

lemma enat_min:
  assumes "m \<ge> enat n'"
    and "enat n < m - enat n'"
  shows "enat n + enat n' < m" 
using assms 
by (metis add.commute enat.simps(3) enat_add_mono enat_add_sub_same le_iff_add)

lemma enat_min_eq:
  assumes "m \<ge> enat n'"
    and "enat n \<le> m - enat n'"
  shows "enat n + enat n' \<le> m" 
using assms
by (metis add.commute enat.simps(3) enat_add_sub_same enat_min le_iff_add le_less)
  
lemma llist_eq_lnth_eq:
 " (xs = ys) \<longleftrightarrow> (llength xs = llength ys \<and> (\<forall> i < llength xs. lnth xs i = lnth ys i))"
proof auto
 show "llength xs = llength ys \<Longrightarrow> \<forall>i. enat i < llength ys \<longrightarrow> lnth xs i = lnth ys i \<Longrightarrow> xs = ys" 
 proof (coinduction arbitrary: xs ys)
 case (Eq_llist xsa ysa)
 then show ?case 
    proof  -
     assume a: "llength xsa = llength ysa" 
     assume b: "\<forall>i. enat i < llength ysa \<longrightarrow> lnth xsa i = lnth ysa i" 
     have 1: " lnull xsa = lnull ysa" 
       using a by auto
     have 2: "\<not> lnull xsa \<longrightarrow>
              \<not> lnull ysa \<longrightarrow>
              lhd xsa = lhd ysa" 
       using b lhd_conv_lnth zero_enat_def by force
     have 3: "\<not> lnull xsa \<longrightarrow>
              \<not> lnull ysa \<longrightarrow>
                (\<exists>xs ys. ltl xsa = xs \<and> ltl ysa = ys \<and> llength xs = llength ys \<and> 
                         (\<forall>i. enat i < llength ys \<longrightarrow> lnth xs i = lnth ys i))"
       by (metis Extended_Nat.eSuc_mono a b eSuc_enat eSuc_epred llength_eq_0 llength_ltl lnth_ltl)
    show ?thesis using 1 2 3 by blast
   qed
 qed
qed

lemma exist_lset_lnth:
 " (\<exists> x \<in> lset xs. P x) \<longleftrightarrow> (\<exists> i<llength xs. P (lnth xs i))"
by (metis in_lset_conv_lnth)

lemma exist_llength_gr_zero:
 assumes "(\<exists> x \<in> lset xs. P x)"
 shows   " 0< llength xs"
 by (metis assms gr_implies_not_zero gr_zeroI in_lset_conv_lnth)


subsection \<open>lbutlast\<close>

lemma lbutlast_snoc [simp]:
 " lbutlast (lappend xs (LCons x LNil)) = xs"
proof (cases "lfinite xs")
case True
then show ?thesis 
  proof (induct rule: lfinite_induct)
  case (LNil xs)
  then show ?case using llist.collapse(1) by fastforce
  next
  case (LCons xs)
  then show ?case
  by (metis LNil_eq_lappend_iff lappend_code(2) lbutlast_simps(2) lhd_LCons_ltl llist.distinct(1)) 
  qed
next
case False
then show ?thesis 
 proof (coinduction arbitrary: xs)
 case (Eq_llist xsa)
 then show ?case 
     proof -
      have 1: "lnull (lbutlast (lappend xsa (LCons x LNil))) = lnull xsa"
        by (metis Eq_llist lappend_inf lbutlast.disc(2) lfinite.simps lhd_LCons_ltl lnull_imp_lfinite)
      have 2: "\<not> lnull (lbutlast (lappend xsa (LCons x LNil))) \<longrightarrow>
               \<not> lnull xsa \<longrightarrow>
               lhd (lbutlast (lappend xsa (LCons x LNil))) = lhd xsa" 
        by simp
      have 3: "\<not> lnull (lbutlast (lappend xsa (LCons x LNil))) \<longrightarrow>
               \<not> lnull xsa \<longrightarrow> 
              (\<exists>xs. ltl (lbutlast (lappend xsa (LCons x LNil))) = 
                           lbutlast (lappend xs (LCons x LNil)) \<and> ltl xsa = xs \<and> \<not> lfinite xs)" 
        by (metis Eq_llist lappend_inf lbutlast_simps(2) lfinite_LNil lfinite_ltl lhd_LCons_ltl 
            ltl_simps(2))
      show ?thesis using 1 2 3 by auto
   qed
 qed
qed

lemma lbutlast_ltl:
 " lbutlast (ltl xs) = ltl (lbutlast xs)"
proof (cases "lfinite xs")
case True
then show ?thesis 
  proof  (induct rule: lfinite_induct)
  case (LNil xs)
  then show ?case by (simp add: lbutlast.ctr(1))
  next
  case (LCons xs)
  then show ?case by (metis lbutlast.ctr(1) lbutlast.simps(4) llist.case_eq_if lnull_def lnull_ltlI)
  qed
next
case False
then show ?thesis 
   proof  (coinduction arbitrary: xs)
   case (Eq_llist xsa)
   then show ?case 
     proof -
      have 1: "lnull (lbutlast (ltl xsa)) = lnull (ltl (lbutlast xsa))" 
        by (metis Eq_llist lbutlast_simps(2) lfinite_ltl lhd_LCons_ltl llist.discI(1) 
            lnull_imp_lfinite ltl_simps(2))
      have 2: "\<not> lnull (lbutlast (ltl xsa)) \<longrightarrow>
               \<not> lnull (ltl (lbutlast xsa)) \<longrightarrow>
               lhd (lbutlast (ltl xsa)) = lhd (ltl (lbutlast xsa))" 
        by (metis lbutlast.disc_iff(1) lbutlast.simps(4) llist.simps(5) lnull_ltlI ltl_def 
             not_lnull_conv)
      have 3: "\<not> lnull (lbutlast (ltl xsa)) \<longrightarrow>
               \<not> lnull (ltl (lbutlast xsa)) \<longrightarrow>
               (\<exists>xs. ltl (lbutlast (ltl xsa)) = lbutlast (ltl xs) \<and> 
                     ltl (ltl (lbutlast xsa)) = ltl (lbutlast xs) \<and> \<not> lfinite xs)" 
        by (metis Eq_llist lbutlast.simps(4) lfinite_LNil lfinite_ltl llist.case_eq_if ltl_def)
      show ?thesis using 1 2 3 by auto
    qed
   qed
qed

lemma lbutlast_not_lfinite:
 assumes "\<not> lfinite xs"
 shows   "lbutlast xs = xs"
 using assms 
by (coinduction arbitrary: xs)
   (metis lbutlast.disc_iff(1) lbutlast.simps(3) lbutlast_ltl lfinite_LNil lfinite_ltl) 

lemma lbutlast_lfinite:
 "lfinite (lbutlast xs) \<longleftrightarrow> lfinite xs"
proof 
 show "lfinite (lbutlast xs) \<Longrightarrow> lfinite xs"
  proof (induct zs\<equiv>"lbutlast xs" rule: lfinite_induct)
  case LNil
  then show ?case using lbutlast_not_lfinite by fastforce
  next
  case LCons
  then show ?case using lbutlast_not_lfinite by fastforce
  qed
 show "lfinite xs \<Longrightarrow> lfinite (lbutlast xs) "
  proof (induct rule: lfinite_induct)
  case (LNil xs)
  then show ?case by simp
  next
  case (LCons xs)
  then show ?case by (simp add: lbutlast_ltl)
  qed 
qed

lemma llength_lbutlast [simp]:
 " llength (lbutlast xs) = epred (llength xs)"
by (coinduction arbitrary: xs rule: enat_coinduct)
   (simp,
   metis epred_enat_unfold lappend_code(1) lappend_lnull2 lbutlast_ltl llength_def llength_eq_0 
   llength_ltl llist.discI(1))

lemma lbutlast_lappend:
 " lbutlast (lappend xs ys) = (if ys = LNil then lbutlast xs else lappend xs (lbutlast ys))"
proof (cases "lfinite xs")
case True
then show ?thesis 
   proof (induct arbitrary: ys rule: lfinite_induct)
   case (LNil xs)
   then show ?case by (simp add: lnull_def)
   next
   case (LCons xs)
   then show ?case 
       proof (cases ys)
       case LNil
       then show ?thesis by simp
       next
       case (LCons x21 x22)
       then show ?thesis 
           proof (cases xs)
           case LNil
           then show ?thesis by simp
           next
           case (LCons x21 x22)
           then show ?thesis using LCons.hyps(3) lappend_eq_LNil_iff by fastforce
           qed
        qed
   qed
next
case False
then show ?thesis by (metis lappend_inf lbutlast_snoc)
qed

lemma lappend_lbutlast_llast_id_lfinite:
 assumes "lfinite xs"
         " \<not> lnull xs"
 shows   "(lappend (lbutlast xs) (LCons (llast xs) LNil)) = xs"
using assms 
proof (induct rule: lfinite_induct)
case (LNil xs)
then show ?case by simp
next
case (LCons xs)
then show ?case 
   proof (cases xs)
   case lfinite_LNil
   then show ?thesis using LCons by simp
   next
   case (lfinite_LConsI xs x)
   then show ?thesis using LCons by (simp add: llast_LCons)
qed
qed
    
lemma lappend_lbutlast_llast_id_not_lfinite:
 assumes "\<not>lfinite xs"
         " \<not> lnull xs"
 shows   "(lappend (lbutlast xs) (LCons (llast xs) LNil)) = xs"
using assms 
proof (coinduction arbitrary: xs)
case (Eq_llist xsa)
then show ?case 
    by (auto simp add: lbutlast_ltl llast_linfinite)
       (metis lfinite_LNil lfinite_ltl)
qed
     
lemma lappend_lbutlast_llast_id [simp]:
shows " \<not> lnull xs \<Longrightarrow> (lappend (lbutlast xs) (LCons (llast xs) LNil)) = xs"
using lappend_lbutlast_llast_id_lfinite lappend_lbutlast_llast_id_not_lfinite by blast

lemma lbutlast_conv_ltake:
 " lbutlast xs = ltake (epred (llength xs)) xs"
proof (cases "lfinite xs")
case True
then show ?thesis 
   proof (induct rule: lfinite_induct)
   case (LNil xs)
   then show ?case by simp
   next
   case (LCons xs)
   then show ?case
     by (metis enat_le_plus_same(2) gen_llength_def lappend_lbutlast_llast_id_lfinite llength_code 
          llength_lbutlast ltake_all ltake_lappend1) 
   qed
next
case False
then show ?thesis 
  proof (coinduction arbitrary: xs)
  case (Eq_llist xsa)
  then show ?case 
   by (auto simp add: not_lfinite_llength)
      (metis lfinite_LNil lfinite_ltl,
       metis lbutlast.simps(4) lbutlast_not_lfinite lfinite_code(2) lhd_LCons_ltl 
       llength_eq_infty_conv_lfinite ltake_epred_ltl)
  qed
qed

lemma lmap_lbutlast:
 " lmap f (lbutlast xs) = lbutlast (lmap f xs)"
by (simp add: lbutlast_conv_ltake)
 
lemma snocs_eq_iff_lbutlast:
 " lappend xs (LCons x LNil) = ys \<longleftrightarrow> 
   (( lfinite ys \<and> \<not> lnull ys \<and> lbutlast ys = xs \<and> llast ys = x) 
   \<or> (\<not> lfinite ys \<and> lbutlast xs = ys))"
by (metis lappend.disc(2) lappend_inf lappend_lbutlast_llast_id lbutlast.ctr(1) lbutlast_snoc 
     lfinite_LNil llast_lappend llast_singleton llist.discI(2))

lemma in_lset_lbutlastD:
 " x \<in> lset(lbutlast xs) \<Longrightarrow> x \<in> lset xs"
by (metis in_lset_lappend_iff lappend_ltake_ldrop lbutlast_conv_ltake)

lemma in_lset_lbutlast_lappendI:
 " x \<in> lset (lbutlast xs) \<or> (lfinite xs \<and> x \<in> lset(lbutlast ys)) \<Longrightarrow>
   x \<in> lset (lbutlast (lappend xs ys))"
by (metis empty_iff in_lset_lappend_iff in_lset_lbutlastD lbutlast_lappend lset_LNil)

lemma lnth_lbutlast:
 assumes "n< llength(lbutlast xs)"
 shows   " lnth (lbutlast xs) n = lnth xs n"
proof (cases xs)
case LNil
then show ?thesis by simp
next
case (LCons x21 x22)
then show ?thesis by (metis assms lbutlast_conv_ltake llength_lbutlast lnth_ltake)
qed

subsection \<open>kfilter\<close>

lemma kfilter_code [simp, code]:
  shows kfilter_LNil: "kfilter P n LNil  = LNil"
  and   kfilter_LCons: "kfilter P n (LCons x xs)  = 
                        (if P x then LCons n (kfilter P (Suc n) xs) else kfilter P (Suc n) xs  )"
by (auto simp add: kfilter_def lzip.ctr(2))

lemma lmap_fst_imp_a: 
assumes "lnull (lfilter P xs)"
 shows  " lnull (lmap fst (lfilter (P \<circ> fst) (lzip xs (iterates Suc n)))) "
using assms lset_lzipD1 by fastforce

lemma lmap_fst_imp_b:
assumes "lnull (lmap fst (lfilter (P \<circ> fst) (lzip xs (iterates Suc n))))"
shows   "  lnull (lfilter P xs)"
proof -
 have 0: "lnull (lmap fst (lfilter (\<lambda> (x,k). P x) (lzip xs (iterates Suc n))))" 
   using assms by auto
 have 1: "lnull (lfilter (\<lambda> (x,k). P x) (lzip xs (iterates Suc n))) " 
   using 0   by auto
 have 2: "(\<forall> (x,k) \<in> lset(lzip xs (iterates Suc n)).  \<not> P x) " 
   using 1 by (simp add: prod.case_eq_if)
 have 3: "(\<forall> (x,k) \<in> { (lnth xs i , n+i) | i. enat i < llength xs}. \<not> P x)"
   using 2 by (simp add: lset_lzip) 
 have 4: "(\<forall> x \<in> { lnth xs i  | i. enat i < llength xs}. \<not> P x)" 
   using 3 by blast
  from 4 show ?thesis by (simp add: lset_conv_lnth)
qed

lemma lmap_snd_lnull:
 "lnull (lmap snd (lfilter (P \<circ> fst) (lzip xs (iterates Suc n)))) = lnull(lfilter P xs)"
by (metis llist.collapse(1) llist.disc(1) lmap_eq_LNil lmap_fst_imp_a lmap_fst_imp_b)

lemma kfilter_lnull_conv:
 " lnull (kfilter P n xs) \<longleftrightarrow> (\<forall> x \<in> lset xs. \<not> P x)"
unfolding kfilter_def using lmap_snd_lnull[of P xs] lnull_lfilter[of P xs] by blast 

lemma kfilter_not_lnull_conv:
 " \<not>lnull (kfilter P n xs) \<longleftrightarrow> (\<exists> x \<in> lset xs.  P x)"
by (simp add: kfilter_lnull_conv)

lemma lmap_fst_lfilter:
 " lfilter P (lmap fst (lzip xs (iterates Suc n))) = 
   lmap fst (lfilter (P \<circ> fst) (lzip xs (iterates Suc n))) "
using lfilter_lmap by blast

lemma lmap_fst_lzip:
 " (lmap fst (lzip xs (iterates Suc n))) = xs"
by (coinduction arbitrary: xs) (auto, simp add: lmap_fst_lzip_conv_ltake) 

lemma lfilter_kfilter_1: 
 " (lmap fst (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) =
   (lfilter P xs)"
by (metis (mono_tags, lifting)  lmap_fst_lfilter lmap_fst_lzip)

lemma lfilter_kfilter_snd_llength:
 " llength (lmap fst (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) =
   llength (lmap snd (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) "
by simp

lemma kfilter_llength:
 " llength(kfilter P n xs) = llength(lfilter P xs)"
proof -
 have 1: "llength(kfilter P n xs) = llength (lmap snd (lfilter (P\<circ>fst) (lzip xs (iterates Suc n))))"
   by (simp add: kfilter_def) 
 have 2: "llength (lmap snd (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) =
          llength (lmap fst (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) " 
   using lfilter_kfilter_snd_llength by auto
 have 3: "llength (lmap fst (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))) =
          llength(lfilter P xs)"
   by (metis lfilter_kfilter_1)
 from 1 2 3 show ?thesis by auto
qed

lemma ldropWhile_LEAST:
assumes " \<exists> n<llength xs. (Not \<circ> P) (lnth xs n)"
shows " ldropWhile P xs = ldropn (LEAST n. n < llength xs \<and> (Not \<circ>P) (lnth xs n)) xs " 
proof -
 from assms obtain m where 
   "m< llength xs \<and> (Not \<circ> P) (lnth xs m)" 
   "\<And>n. n< llength xs \<longrightarrow> (Not \<circ>P) (lnth xs n) \<Longrightarrow> m \<le> n"
 and *: "(LEAST n. n < llength xs \<and> (Not \<circ> P) (lnth xs n)) = m" 
  by atomize_elim
    (metis (no_types, lifting) dual_order.strict_trans1 enat_ord_code(2) less_imp_le not_le_imp_less 
       not_less_Least wellorder_Least_lemma(1))
 thus ?thesis unfolding *
 proof (induct m arbitrary: xs)
 case 0
  then show ?case 
    by (cases xs) simp_all
 next
 case (Suc m)
 then show ?case 
    proof -
      have 1: "ldropWhile P (ltl xs) = ldropn m (ltl xs)" using Suc 
        by (cases xs)
           (simp,
            metis Suc_le_mono ldrop_eSuc_ltl ldropn_Suc_conv_ldropn ldropn_eq_LNil llist.simps(3) 
            lnth_Suc_LCons ltl_simps(2) not_le_imp_less)
      have 2: "P (lnth xs 0) "
        using Suc.prems(2) by auto
     show ?thesis 
      by (metis "1" "2" ldropWhile_LCons ldrop_eSuc_ltl lhd_LCons_ltl llist.collapse(1) 
          lnth_0_conv_lhd ltl_simps(1))
 qed
qed
qed

lemma ldropWhile_LEAST_not:
assumes " \<exists> n<llength xs. P (lnth xs n)"
shows " ldropWhile (Not \<circ> P) xs = ldropn (LEAST n. n < llength xs \<and> P (lnth xs n)) xs "
using assms ldropWhile_LEAST[of xs "Not \<circ> P"] by simp

lemma ltakeWhile_LEAST:
assumes " \<exists> n<llength xs. (Not \<circ> P) (lnth xs n)"
shows "(ltakeWhile P xs) = (ltake (lleast (Not \<circ> P) xs) xs)" 
proof-
 from assms obtain m where 
  "m< llength xs \<and> (Not \<circ> P) (lnth xs m)" 
  "\<And>n. n< llength xs \<longrightarrow> (Not \<circ>P) (lnth xs n) \<Longrightarrow> m \<le> n"
 and *: " (lleast (Not \<circ> P) xs) = m" unfolding lleast_def 
 by atomize_elim
 (metis (no_types, lifting) Least_le dual_order.trans enat_ord_code(2) not_less not_less_iff_gr_or_eq
   wellorder_Least_lemma(1))
 thus ?thesis unfolding *
 proof (induct m arbitrary: xs)
 case 0
 then show ?case 
 proof -
 have "ltakeWhile P (LCons (lnth xs 0) (ldropn (Suc 0) xs)) = LNil"
 using 0 by fastforce 
 then show ?thesis 
 by (metis (no_types) lappend.disc_iff(2) lappend_ltakeWhile_ldropWhile ldropn_0 
     ldropn_Suc_conv_ldropn llist.collapse(1) lnull_ldropn ltake.ctr(1) not_less zero_enat_def)
 qed
 next
 case (Suc m)
 then show ?case 
 proof -
   have 1: "ltakeWhile P (ltl xs) = ltake m (ltl xs)" 
     using Suc by (cases xs)
        (simp,
         metis Extended_Nat.eSuc_mono Suc_le_mono eSuc_enat llength_LCons lnth_Suc_LCons ltl_simps(2))
   have 2: "P (lnth xs 0) "
     using Suc by auto
   show ?thesis 
     by (metis "1" "2" eSuc_enat lhd_LCons_ltl lnth_0_conv_lhd ltake.ctr(1) ltakeWhile.code 
         ltake_eSuc_LCons)
  qed
 qed
qed

lemma llength_LEAST_not:
 assumes " \<exists> n < llength xs.  (Not \<circ> P) (lnth xs n)"
 shows "(lleast (Not \<circ> P) xs) < llength xs"
using assms  unfolding lleast_def
by (metis (no_types, lifting) LeastI)

lemma llength_LEAST:
 assumes " \<exists> n < llength xs.  P (lnth xs n)"
 shows "(lleast P xs) < llength xs"
using assms llength_LEAST_not[of xs "Not \<circ> P"] unfolding lleast_def by simp

lemma llength_ltakeWhile_LEAST:
assumes " \<exists> n< llength xs.  (Not \<circ> P) (lnth xs n)"
shows " (llength (ltakeWhile P xs)) = (lleast (Not \<circ> P) xs) "
using assms ltakeWhile_LEAST[of xs P] unfolding lleast_def
by (metis (no_types, lifting) dual_order.strict_trans1 enat_ord_simps(2) le_cases llength_ltake 
    min_def not_less_Least)

lemma llength_ltakeWhile_LEAST_not:
assumes " \<exists> n< llength xs.  P (lnth xs n)"
shows " (llength (ltakeWhile (Not \<circ>P) xs)) = (lleast P xs)"
using assms llength_ltakeWhile_LEAST[of xs "Not \<circ> P"]  unfolding lleast_def by simp

lemma lzip_ldropWhile_fst: 
assumes " llength (lzip xs ys) = llength xs"
shows " lzip (ldropWhile P xs) (lmap snd (ldropWhile (P \<circ> fst) (lzip xs ys)) ) = 
        ldropWhile (P \<circ> fst) (lzip xs ys)"
using assms 
proof -
have f1: "ldropWhile P xs = lmap fst (ldropWhile (P \<circ> fst) (lzip xs ys))"
by (metis (no_types) llength_lzip assms ldropWhile_lmap lmap_fst_lzip_conv_ltake ltake_all order_refl) 
have "ltake (min (llength (ldrop (llength (ltakeWhile (P \<circ> fst) (lzip xs ys))) xs))
                  (llength (ldrop (llength (ltakeWhile (P \<circ> fst) (lzip xs ys))) ys)))
            (ldrop (llength (ltakeWhile (P \<circ> fst) (lzip xs ys))) (lzip xs ys)) = 
      ldrop (llength (ltakeWhile (P \<circ> fst) (lzip xs ys))) (lzip xs ys)"
by (simp add: ltake_all) 
then show ?thesis
using f1 by (simp add: ldropWhile_eq_ldrop lmap_fst_lzip_conv_ltake lmap_snd_lzip_conv_ltake ltake_lzip) 
qed

lemma lzip_ldropWhile_fst_iterates:
 " ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)) = 
   lzip (ldropWhile (Not \<circ>P) xs) (lmap snd (ldropWhile (Not \<circ>P \<circ> fst) (lzip xs (iterates Suc n)))  )"
by (metis llength_lmap lmap_fst_lzip lzip_ldropWhile_fst)

lemma ldropWhile_iterates_split:
assumes " \<exists> n<llength xs. P (lnth xs n)"
shows " ( (ldropWhile (Not \<circ>P \<circ> fst) (lzip xs (iterates Suc n)))) =
     (lzip  (ldropWhile (Not \<circ>P) xs)
                (iterates Suc (n+(lleast P xs))) )"
proof -
 have 1: "\<exists>na. enat na < llength (lzip xs (iterates Suc n)) \<and> 
           (Not\<circ> (Not\<circ> P) \<circ> fst) (lnth (lzip xs (iterates Suc n)) na)" 
   using lmap_fst_lzip[of xs n] 
   by (metis assms comp_apply llength_lmap lnth_lmap) 
 have 2: "(ldropWhile ((Not \<circ>P) \<circ> fst) (lzip xs (iterates Suc n))) =
          (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and>  
          ((Not \<circ> (Not \<circ>P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (lzip xs (iterates Suc n)))" 
   using 1 ldropWhile_LEAST[of " (lzip xs (iterates Suc n))" "(Not \<circ>P) \<circ> fst"] by auto
 have 3: "(ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and>
          ((Not \<circ> (Not \<circ>P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (lzip xs (iterates Suc n))) =
          (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
          (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (lzip xs (iterates Suc n)))" 
   by auto
 have 4: "(ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                  (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (lzip xs (iterates Suc n))) =
          (lzip (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                        (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) xs)
                (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                        (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (iterates Suc n)))" 
   using ldropn_lzip by blast
 have 5: "(ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                  (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (iterates Suc n)) =
          (iterates Suc (n+(LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                  (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)))) " 
   by (metis (no_types, lifting) funpow_Suc_conv ldropn_iterates)
 have 6: "(LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
          (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) =
          (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
          (( (P \<circ>fst))) (lnth ( xs ) na, lnth (iterates Suc n) na))"
   by (metis (no_types, opaque_lifting) comp_def fst_conv lmap_fst_lzip lnth_lmap)
 have 7: "llength(lzip xs (iterates Suc n)) = llength xs" 
   by simp
 have 8: "(LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
          (( (P \<circ>fst))) (lnth ( xs ) na, lnth (iterates Suc n) na)) =
          (LEAST na. na < llength xs \<and> P (lnth xs na)  )"
   by auto 
 have 9: "(lzip (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                        (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) xs)
                (ldropn (LEAST na. na < llength(lzip xs (iterates Suc n)) \<and> 
                        (( (P\<circ> fst))) (lnth (lzip xs (iterates Suc n)) na)) (iterates Suc n))) =
          (lzip (ldropn (LEAST na. na < llength xs \<and> P (lnth xs na)  ) xs) 
                (iterates Suc (n+(LEAST na. na < llength xs  \<and>  P (lnth xs na) ))) )"
   using "5" "6" by auto
 show ?thesis unfolding lleast_def  
   using assms 2 3 4 9 ldropWhile_LEAST_not[of xs P] by presburger   
qed
 
lemma kfilter_ldropWhile :
 assumes "\<not> lnull(kfilter P n xs)"
 shows " kfilter P n xs =
         (LCons (n+  (lleast P xs)) 
                (lmap snd (lfilter (P\<circ>fst) 
                  (lzip (ldrop (Suc (lleast P xs)) ( xs))
                        (iterates Suc (Suc (n+(lleast P xs))))))))"
proof -
 let ?Least = "(lleast P xs)" 
 have 1: "lhd(lfilter (P\<circ>fst) (lzip xs (iterates Suc n))) =
          lhd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))"
   by simp
 have 2: "ltl(lfilter (P\<circ>fst) (lzip xs (iterates Suc n))) =
           (lfilter (P\<circ>fst) (ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))))"  
   by (simp add: ltl_lfilter)
 have 3: "lfilter (P\<circ>fst) (lzip xs (iterates Suc n)) =
          (LCons (lhd(lfilter (P\<circ>fst) (lzip xs (iterates Suc n))))
                 (ltl(lfilter (P\<circ>fst) (lzip xs (iterates Suc n))))) "
   by (metis assms kfilter_llength llength_eq_0 llength_lmap llist.disc(1) llist.exhaust_sel 
        lmap_fst_lfilter lmap_fst_lzip) 
 have 4: "kfilter P n xs =
          lmap  snd (lfilter (P\<circ>fst) (lzip xs (iterates Suc n)))"
   by (simp add: kfilter_def) 
 have 5: "ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))) =
         ltl((lzip (ldropWhile (Not \<circ>P) xs) 
                   (lmap snd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))))) " 
   by (metis lzip_ldropWhile_fst_iterates)
 have 6: "ltl((lzip (ldropWhile (Not \<circ>P) xs) 
                    (lmap snd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))))) =
         (lzip (ltl (ldropWhile (Not \<circ>P) xs)) 
               (ltl (lmap snd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))))))"
   using lzip_ldropWhile_fst_iterates[of P xs n] 
         ltl_lzip[of "(ldropWhile (Not \<circ> P) xs)" 
                     "(lmap snd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))))"]
   by force
 have 7: "lmap snd (
         (LCons (lhd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))))
                (lfilter (P\<circ>fst) (ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))))))) =
          (LCons (snd (lhd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))))
                 (lmap snd (lfilter (P\<circ>fst) (ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n)))))))"
   by simp 
 have 8: "\<exists>n. enat n < llength xs \<and> P (lnth xs n)" 
    by (metis assms in_lset_conv_lnth kfilter_llength llength_eq_0 lnull_lfilter)
 have 9: "(snd (lhd (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))))) =
          (snd (lhd (lzip  (ldropWhile (Not \<circ>P) xs)
                (iterates Suc (n+?Least)) )))"
   using 8 ldropWhile_iterates_split[of xs P n] by simp
 have 10: "(snd (lhd (lzip  (ldropWhile (Not \<circ>P) xs)
                (iterates Suc (n+?Least)) ))) =
             lhd (iterates Suc (n+?Least))"
   by (metis (no_types, lifting) "3" comp_assoc iterates.disc_iff lfilter_eq_LCons lhd_lzip 
       llist.disc(2) lzip.disc(1) lzip_ldropWhile_fst_iterates snd_conv)   
 have 11: "lhd (iterates Suc (n+?Least)) = (n+?Least)"
   by simp   
 have 12: "ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))) =
           ltl (lzip  (ldropWhile (Not \<circ>P) xs)
                (iterates Suc (n+?Least)) )" 
   using 8 ldropWhile_iterates_split[of xs P n] by simp
 have 13: "ltl (lzip  (ldropWhile (Not \<circ>P) xs) (iterates Suc (n+?Least)) ) =
          (lzip (ltl (ldropWhile (Not \<circ>P) xs)) (ltl (iterates Suc (n+?Least))))"
   by (metis (no_types, lifting) "3" comp_assoc iterates.disc_iff lfilter_eq_LCons llist.discI(2) 
        ltl_lzip lzip.disc_iff(2) lzip_ldropWhile_fst_iterates)
 have 14: "(ltl (iterates Suc (n+?Least))) = (iterates Suc (Suc (n+?Least))) "
   by simp
 have 15: "ltl (ldropWhile (Not \<circ>P) xs) =  ltl (ldrop (llength (ltakeWhile (Not \<circ>P) xs)) xs)" 
   by (simp add: ldropWhile_eq_ldrop)
 have 16: "ltl (ldrop (llength (ltakeWhile (Not \<circ>P) xs)) xs) =
           ldrop (eSuc (llength (ltakeWhile (Not \<circ>P) xs))) ( xs)"
   by (simp add: ldrop_eSuc_conv_ltl) 
 have 17: "(llength (ltakeWhile (Not \<circ>P) xs)) = (llength (ltake ?Least xs))"
   using 8 ltakeWhile_LEAST[of xs "Not \<circ>P" ] unfolding lleast_def by auto
 have 18: "(llength (ltake ?Least xs)) = enat ?Least"
   using "17" "8" llength_ltakeWhile_LEAST_not unfolding lleast_def by fastforce
 have 19: "ldrop (eSuc (llength (ltakeWhile (Not \<circ>P) xs))) ( xs) = ldrop (Suc ?Least) ( xs)" 
   using "17" "18" by (simp add: eSuc_enat)
 have 20: "ltl (ldropWhile (Not \<circ> P \<circ> fst) (lzip xs (iterates Suc n))) =
           (lzip (ldrop (Suc ?Least) ( xs)) (iterates Suc (Suc (n+?Least))))"
   using "12" "13" "15" "16" "19"  by auto
 show ?thesis 
 using "2" "3" "4" "10" "20" "9" by auto
qed
   
lemma kfilter_eq_LCons:
 " kfilter P n xs = LCons x xs' \<Longrightarrow>
        x  = n+(lleast P xs) \<and> 
       xs' = (lmap snd (lfilter (P\<circ>fst) 
                                (lzip (ldrop (Suc (lleast P xs)) ( xs))
                                      (iterates Suc (Suc (n+(lleast P xs)))))))  "
using kfilter_ldropWhile[of P n xs] by auto

lemma kfilter_eq_LCons_1:
 " kfilter P n xs = LCons x xs' \<Longrightarrow>
     x   = (n+(lleast P xs)) \<and>
     xs' = kfilter P (Suc (n+(lleast P xs))) (ldrop (Suc (lleast P xs)) xs)"
using kfilter_eq_LCons[of P n xs x xs'] 
      kfilter_def[of P "(Suc (n + (lleast P xs)))" 
                       "(ldrop (Suc (lleast P xs)) xs)"]
by auto  

lemma kfilter_eq_conv:
 "kfilter P n xs = LNil \<or> 
    kfilter P n xs =
      LCons (n+(lleast P xs)) (kfilter P (Suc (n+(lleast P xs))) (ldrop (Suc (lleast P xs)) xs))"
proof (cases "kfilter P n xs")
case LNil
then show ?thesis by simp
next
case (LCons x21 x22)
then show ?thesis 
    proof -
       let ?Least = "(lleast P xs)" 
       have 1: "x21 = n+?Least"
         using  kfilter_eq_LCons_1[of P n xs x21 x22] by (meson LCons)
       have 2: "x22= (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs))" 
         using  kfilter_eq_LCons_1[of P n xs x21 x22] by (meson LCons)
       show ?thesis 
         by (metis "1" "2" LCons)
    qed
qed

lemma kfilter_lnth_zero:
 assumes " \<not>lnull(kfilter P n xs)"
 shows   "lnth (kfilter P n xs) 0 = n+ (lleast P xs)"
using assms 
by (metis kfilter_eq_LCons_1 lhd_LCons_ltl lnth_0_conv_lhd)

lemma length_LEAST_a:
 assumes "\<not>lnull(kfilter P n xs)"
 shows   "lleast P xs< llength xs"
using assms exist_lset_lnth[of xs P]  kfilter_not_lnull_conv[of P n xs]  
llength_LEAST[of xs P]  by blast

lemma kfilter_upperbound:
assumes " i < llength(kfilter P n xs)"      
 shows " (lnth (kfilter P n xs) i) < n + llength xs"
proof (cases "lfinite (kfilter P n xs)")
case True
then show ?thesis using assms
   proof (induct zs\<equiv>"kfilter P n xs" arbitrary: xs n i rule: lfinite_induct)
   case (LNil xs)
   then show ?case 
   using gr_implies_not_zero llength_lnull by blast
   next
   case (LCons xs)
   then show ?case 
         proof -
          let ?Least = "(lleast P xs)" 
          have 1: "\<exists> x xs'. kfilter P n xs = LCons x xs'"
            using LCons.hyps(2) kfilter_ldropWhile by blast
          obtain x xs' where 2:"kfilter P n xs = LCons x xs'" 
            using 1 by auto
          have 3: "x = n+?Least"
            using kfilter_ldropWhile[of P n xs] by (simp add: "2")
          have 4: " i=0 \<Longrightarrow> (lnth (kfilter P n xs) i) = x"
            by (simp add: "2")      
          have 5: "enat n+ enat ?Least < enat n + llength xs "  
            using length_LEAST_a[of P n xs] LCons.hyps(2) enat_add_mono by blast
          have 6: "i=0 \<Longrightarrow> enat (lnth (kfilter P n xs) i) < enat n + llength xs"
            using "3" "4" "5" by auto   
          have 7: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
            using kfilter_ldropWhile[of P n xs] "2" kfilter_eq_LCons_1 by blast
          have 8: "i>0 \<Longrightarrow> enat (lnth (kfilter P n xs) i) = enat (lnth xs' (i-1))"
            by (simp add: "2" lnth_LCons') 
          have 9: "i>0 \<and> lnull xs' \<Longrightarrow> enat (lnth (kfilter P n xs) i) < enat n + llength xs"
            by (metis "2" LCons.prems(1) One_nat_def enat_ord_simps(2) llength_LCons llength_LNil 
                llist.collapse(1) not_less_eq one_eSuc one_enat_def)
          have 10: "i>0 \<and> \<not>lnull xs' \<Longrightarrow> (i-1) < llength xs'" 
            using "2" LCons.prems(1) Suc_ile_eq by fastforce
          have 11: "i>0 \<and> \<not>lnull xs' \<Longrightarrow> 
                     enat (lnth xs' (i-1)) < enat (Suc (n+?Least)) + llength (ldrop (Suc ?Least) xs)"
            by (metis (no_types, lifting) "10" "2" "7" LCons.hyps(3) ltl_simps(2))   
          have 111: "lappend (ltake (Suc (?Least)) xs) (ldrop (Suc (?Least)) xs) = xs"
            using lappend_ltake_ldrop by blast
          have 112: "enat (Suc (n+?Least)) = n + (Suc ?Least)"
            by auto
          have 113: " Suc ?Least \<le> (llength xs) "
            using "5" Suc_ile_eq enat_add_mono by blast      
          have 114: "llength (ltake (Suc (?Least)) xs) = Suc ?Least"
            using llength_ltake[of "(Suc (?Least))" xs]   "113" by linarith    
          have 12: "enat (Suc (n+?Least)) + llength (ldrop (Suc ?Least) xs) \<le> enat n + llength xs "
            using llength_lappend[of "(ltake (Suc (?Least)) xs)" "(ldrop (Suc (?Least)) xs)"] 
            by (metis (no_types, lifting) "111" "112" "114" eq_refl group_cancel.add1 plus_enat_simps(1)) 
          have 14: "i>0 \<and> \<not>lnull xs' \<Longrightarrow> enat (lnth (kfilter P n xs) i) < enat n + llength xs"
            using "11" "12" "8" by auto
          have 15: "i>0 \<Longrightarrow> enat (lnth (kfilter P n xs) i) < enat n + llength xs" 
            using "14" "9" dual_order.strict_implies_order by blast         
          show ?thesis 
            using "15" "6" by blast
       qed
   qed
next
case False
then show ?thesis 
by (metis enat_ord_simps(4) iadd_le_enat_iff kfilter_llength leD leI llength_eq_enat_lfiniteD 
     llength_eq_infty_conv_lfinite llength_lfilter_ile)
qed

lemma kfilter_lowerbound:
assumes " i < llength(kfilter P n xs)"
 shows " n \<le>(lnth (kfilter P n xs) i) "
using assms
proof (induct i arbitrary: xs n)
case 0
then show ?case
using kfilter_ldropWhile zero_enat_def by fastforce
next
case (Suc i)
then show ?case 
  proof -
   let ?Least = "lleast P xs" 
   have 7: "\<exists> x xs'. (kfilter P n xs) = LCons x xs'"
     using Suc.prems gr_implies_not_zero kfilter_ldropWhile llength_lnull by blast
   obtain x xs' where 8: "(kfilter P n xs) = LCons x xs'"
     using 7 by auto
   have 9: "lnth (kfilter P n xs) (Suc i) = lnth xs' i"
     by (simp add: "8") 
   have 10: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
     using kfilter_ldropWhile[of P n xs] "8" kfilter_eq_LCons_1 by blast
   have 11: "enat i < llength (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs))"
     by (metis "10" "8" Extended_Nat.eSuc_mono Suc.prems(1) eSuc_enat llength_LCons)
   have 12: "lnull xs'  \<Longrightarrow> n \<le> lnth (kfilter P n xs) (Suc i)"
     using "10" "11" gr_implies_not_zero llength_eq_0 by blast 
   have 13: "\<not>lnull xs' \<Longrightarrow> (Suc (n+?Least)) \<le> lnth xs' i"
     using "10" "11" Suc.hyps by blast 
   have 14: "\<not> lnull xs'  \<Longrightarrow> n \<le> lnth (kfilter P n xs) (Suc i)" 
     using "13" "9" by linarith
   show ?thesis using "12" "14" by blast
 qed
qed

lemma kfilter_mono:
 assumes " (Suc i) < llength(kfilter P n xs)"
 shows " (lnth (kfilter P n xs) i) < (lnth (kfilter P n xs) (Suc i))"
using assms
 proof (induct i arbitrary: xs n)
 case 0
 then show ?case 
  proof -
   let ?Least = "lleast P xs"
   have 1: "\<exists> x xs'. kfilter P n xs = LCons x xs'"
     using "0.prems" gr_implies_not_zero kfilter_ldropWhile llength_lnull by blast
   obtain x xs' where 2:"kfilter P n xs = LCons x xs'" 
     using 1 by auto
   have 3: "x = n+?Least"
     using kfilter_ldropWhile[of P n xs] by (simp add: "2")
   have 4: "lnth (kfilter P n xs) 0 = n+?Least"
     by (simp add: "2" "3")
   have 5: "lnth (kfilter P n xs) (Suc 0)  = lnth xs' 0"
     by (simp add: "2") 
   have 6: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
     using kfilter_ldropWhile[of P n xs] "2" kfilter_eq_LCons_1 by blast
   have 7: "n+?Least < lnth xs' 0" 
     by (metis "0.prems" "2" "6" Extended_Nat.eSuc_mono One_nat_def Suc_le_lessD kfilter_lowerbound 
         llength_LCons one_eSuc one_enat_def zero_enat_def)
   show ?thesis by (simp add: "4" "5" "7")
 qed
 next
 case (Suc i)
 then show ?case 
  proof -
   let ?Least = "lleast P xs"
   have 8: "\<exists> x xs'. kfilter P n xs = LCons x xs'"
     using Suc.prems gr_implies_not_zero kfilter_ldropWhile llength_lnull by blast
   obtain x xs' where 9:"kfilter P n xs = LCons x xs'" 
     using 8 by auto
   have 10: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
     using kfilter_ldropWhile[of P n xs] "9" kfilter_eq_LCons_1 by blast
   have 11: "lnth (kfilter P n xs) (Suc (Suc i))  = lnth xs' (Suc i)"
     by (simp add: "9") 
   have 12: "lnth (kfilter P n xs) (Suc i) < lnth xs' (Suc i)"
     by (metis (no_types, lifting) "10" "9" Extended_Nat.eSuc_mono Suc(1) Suc(2) eSuc_enat 
         llength_LCons  lnth_Suc_LCons) 
   show ?thesis by (simp add: "11" "12")  
 qed
qed

lemma lfilter_kfilter:
 assumes "  i < llength(kfilter P n xs)"
 shows " (lnth xs ((lnth (kfilter P n xs) i)-n)) = (lnth (lfilter P xs) i)"
using assms
proof (induct i arbitrary: xs n)
case 0
then show ?case 
  proof -
    let ?Least = "lleast P xs" 
    have 1: "lnth (kfilter P n xs) 0 = n +?Least "
    using "0.prems" kfilter_ldropWhile zero_enat_def by fastforce
    have 2: "(lnth xs ((lnth (kfilter P n xs) 0)-n)) =
             (lnth xs ?Least)"
      by (simp add: "1")
    have 3: "(lnth (lfilter P xs) 0) = lhd (ldropWhile (Not \<circ> P) xs)" 
      by (metis (full_types) "0.prems" kfilter_llength lhd_conv_lnth lhd_lfilter llength_eq_0 not_iless0)
    have 4: " \<exists> n<llength xs. P (lnth xs n)" 
      by (metis "0.prems" dual_order.irrefl in_lset_conv_lnth kfilter_llength llength_eq_0 lnull_lfilter 
          zero_enat_def)
    have 5: " (ldropWhile (Not \<circ> P) xs) = ldropn ?Least xs"
      using ldropWhile_LEAST_not[of xs "P" ] "4" unfolding lleast_def by blast
    have 6: "lhd (ldropn ?Least xs) = (lnth xs ?Least)" 
      by (simp add: "4" lhd_ldropn llength_LEAST)
    show ?thesis using "2" "3" "5" "6" by auto
  qed           
next
case (Suc i)
then show ?case 
  proof -
   let ?Least= "lleast P xs"  
   have 7: "\<exists> x xs'. (kfilter P n xs) = LCons x xs'"
     using Suc.prems gr_implies_not_zero kfilter_ldropWhile llength_lnull by blast
   obtain x xs' where 8: "(kfilter P n xs) = LCons x xs'"
     using 7 by auto
   have 9: "lnth (kfilter P n xs) (Suc i) = lnth xs' i"
     by (simp add: "8")  
   have 10: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
     using kfilter_ldropWhile[of P n xs] by (simp add: "8" kfilter_eq_LCons_1)
   have 11: "enat i < llength (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs))"
     by (metis "10" "8" Extended_Nat.eSuc_mono Suc.prems(1) eSuc_enat llength_LCons)
   have 12: "lnull xs'  \<Longrightarrow>
             lnth xs (lnth (kfilter P n xs) (Suc i) - n) = lnth (lfilter P xs) (Suc i)  "
     by (metis "10" "11" llength_LNil llist.collapse(1) not_less_zero) 
   have 13: "\<not>lnull xs' \<Longrightarrow>  
              lnth (ldrop (Suc ?Least) xs) (lnth xs' i - (Suc (n+?Least))) = 
           lnth (lfilter P (ldrop (Suc ?Least) xs)) i " 
     by (metis (no_types, lifting) "10" "11" Suc.hyps)
   have 14: "(lnth (lfilter P xs) (Suc i)) = (lnth (ltl(lfilter P xs)) i)"
     by (metis "8" kfilter_not_lnull_conv llist.disc(2) lnth_ltl lnull_lfilter)
   have 15: "(ltl(lfilter P xs)) = lfilter P (ltl (ldropWhile (Not \<circ> P) xs))"
     using ltl_lfilter by blast 
   have 16: "(ltl (ldropWhile (Not \<circ> P) xs)) = 
            (ltl (ldropn (LEAST n. n < llength xs \<and> (Not \<circ>(Not\<circ> P)) (lnth xs n)  ) xs)) "
     using ldropWhile_LEAST[of xs "Not \<circ> P"]
     by (metis (no_types, lifting) "8" comp_apply in_lset_conv_lnth kfilter_lnull_conv llist.disc(2))  
   have 17: " (ltl (ldropn (LEAST n. n < llength xs \<and> (Not \<circ>(Not\<circ> P)) (lnth xs n)  ) xs)) =
            (ltl (ldropn ?Least xs))"
     unfolding lleast_def by auto     
   have 18: "(ltl (ldropn ?Least xs)) =
            ldropn (Suc ?Least) xs"
     by (simp add: ldropn_ltl ltl_ldropn)  
   have 19: "ldropn (Suc ?Least) xs =
            ldrop (Suc ?Least) xs"  
     by (simp add: ldrop_enat) 
   have 20: "lnth (lfilter P xs) (Suc i) = lnth (lfilter P (ldrop (Suc ?Least) xs)) i "
     using "14" "15" "16" "18" "19" unfolding lleast_def by auto
   have 21: "lnth xs (lnth (kfilter P n xs) (Suc i) - n) = lnth xs (lnth xs' i -n)"
     by (simp add: "10" "9")     
   have 22: "n\<le> lnth xs' i"
     using "9" Suc.prems(1) kfilter_lowerbound by fastforce 
   have 23: " (Suc (n+?Least)) \<le> lnth (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)) i "
     using  kfilter_lowerbound[of i P "Suc (n+?Least)" "(ldrop (Suc ?Least) xs) " ] "11" by force
   have 24: "(Suc ?Least) > llength (ltake ?Least xs)"
     by (simp add: min.strict_coboundedI1) 
   have 25: "(ldrop (Suc ?Least) (lappend (ltake ?Least xs) (ldrop ?Least xs))) =
             ldrop ((Suc ?Least) - llength(ltake ?Least xs) ) (ldrop ?Least xs)  "
     by (simp add: ldrop_lappend) 
   have 26: "lappend (ltake (Suc ?Least) xs) (ldrop (Suc ?Least) xs) = xs"
     by (simp add: lappend_ltake_ldrop) 
   have 27: " (Suc ?Least) \<le> (lnth xs' i -n)  "
     using "10" "23" by auto
   have 271: " lnth xs' i -n - (Suc ?Least) = lnth xs' i - (Suc (n+?Least)) "
     by auto 
   have 28: "lnth (lappend (ltake (Suc ?Least) xs) (ldrop (Suc ?Least) xs)) (lnth xs' i -n) =
              lnth (ldrop (Suc ?Least) xs) (lnth xs' i - (Suc (n+?Least))) "
     using lnth_lappend[of "(ltake (Suc ?Least) xs)" "(ldrop (Suc ?Least) xs)" "(lnth xs' i -n)"] 
           27 271 11 19
     by (metis (no_types, lifting) gr_implies_not_zero kfilter_LNil ldropn_eq_LNil 
          le_cases llength_lnull llength_ltake llist.disc(1) lnth_lappend2 min_def)        
   have 29: "lnth xs (lnth xs' i -n) =
              lnth (ldrop (Suc ?Least) xs) (lnth xs' i - (Suc (n+?Least)))" 
     using 28 by (metis (no_types, lifting) "26")
   have 30: " \<not>lnull xs' \<Longrightarrow> 
             lnth xs (lnth (kfilter P n xs) (Suc i) - n) = lnth (lfilter P xs) (Suc i) "
     by (metis "13" "20" "21" "29")    
   show ?thesis 
     using "12" "30" by blast
 qed
qed

lemma in_kfilter_lset:
 shows " x \<in> lset (kfilter P n xs) \<longleftrightarrow> x \<in> { n+i | i. i<llength xs \<and> P (lnth xs i) }" 
 (is "?lhs = ?rhs")
proof 
 assume ?lhs
  thus ?rhs
   proof (induct zs\<equiv>"kfilter P n xs" arbitrary: n xs rule:llist_set_induct)
   case (find)
   then show ?case 
    proof -
      let ?Least = "lleast P xs" 
      have 1: "lhd (kfilter P n xs) = n+?Least"
        using kfilter_ldropWhile[of P n xs] find by auto
      have 2: "P (lnth xs ?Least)" unfolding lleast_def 
        by (metis (mono_tags, lifting) LeastI exist_lset_lnth find.hyps kfilter_lnull_conv)
      have 3: "?Least < llength xs"
        by (metis find.hyps length_LEAST_a )
      have 4: "n+?Least \<in> {n + i |i. enat i < llength xs \<and> P (lnth xs i)}" 
        using "2" "3" by blast
      show ?thesis using "1" "4" by auto
    qed
 next
 case (step y)
 then show ?case 
    proof -
     let ?Least = "lleast P xs" 
     have 5: "ltl (kfilter P n xs) = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)"
       using kfilter_ldropWhile[of P n xs] 
       by (metis kfilter_def ltl_simps(2) step.hyps(1))
     have 6: "lnull (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)) \<Longrightarrow>
              y \<in> {n + i |i. enat i < llength xs \<and> P (lnth xs i)} "
       by (metis "5" gr_implies_not_zero in_lset_conv_lnth llength_eq_0 step.hyps(2))
     have 7: "\<not> lnull (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)) \<Longrightarrow>
               y \<in> {(Suc (n+?Least)) + i |i. enat i < llength (ldrop (Suc ?Least) xs) \<and> 
                                             P (lnth (ldrop (Suc ?Least) xs) i)}"
       using step.hyps using "5" by blast
     have 8: "(Suc (n+?Least)) = n + Suc ?Least"
       by auto 
     have 9: "\<not> lnull(kfilter P (Suc (n + ?Least)) (ldrop (enat (Suc ?Least)) xs)) \<Longrightarrow>
               \<exists>i. y = n + Suc (?Least) + i \<and> enat i < llength (ldrop (enat (Suc ?Least)) xs) \<and> 
                   P (lnth (ldrop (enat (Suc ?Least)) xs) i) \<Longrightarrow> 
                \<exists>i. y = n + i \<and> enat i < llength xs \<and> P (lnth xs i)" 
       proof -
        assume a0: "\<not> lnull(kfilter P (Suc (n + ?Least)) (ldrop (enat (Suc ?Least)) xs))"
        assume a1: "\<exists>i. y = n + Suc (?Least) + i \<and> enat i < llength (ldrop (enat (Suc ?Least)) xs) \<and> 
                          P (lnth (ldrop (enat (Suc ?Least)) xs) i)"
        show "\<exists>i. y = n + i \<and> enat i < llength xs \<and> P (lnth xs i)"
        proof -
         obtain i where 10: "y = n + Suc (?Least) + i \<and> enat i < llength (ldrop (enat (Suc ?Least)) xs) \<and>
                                P (lnth (ldrop (enat (Suc ?Least)) xs) i)"
           using a1 by auto
         have 11: "Suc (?Least) + i < llength xs"
           by (metis "10" add.commute ldrop_enat ldrop_ldrop leD le_less_linear lnull_ldropn 
               plus_enat_simps(1))    
         have 12: "P (lnth xs (Suc (?Least) + i))"
           by (metis "10" "11" add.commute ldrop_enat lnth_ldropn) 
         show ?thesis 
           using "10" "11" "12" ab_semigroup_add_class.add_ac(1) by blast
        qed
       qed
     have 13: "\<not> lnull (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)) \<Longrightarrow>
              y \<in> {n + i |i. enat i < llength xs \<and> P (lnth xs i)}" 
       using "7" "9" by auto
     show ?thesis 
       using "13" "6" by blast    
   qed
 qed
next
 assume ?rhs
 then obtain i where 15: "x = n+i \<and> enat i < llength xs \<and> P (lnth xs i)"
   by blast  
 thus ?lhs
  proof (induct i arbitrary: xs n)
  case 0
  then show ?case 
     proof -
      have 16:  "lhd (kfilter P n xs) = n+(lleast P xs)"
        using kfilter_ldropWhile[of P n xs] using "0.prems"
        by (metis in_lset_conv_lnth kfilter_lnull_conv lhd_LCons) 
      show ?thesis 
        by (metis "0.prems" add.right_neutral kfilter_LCons ldropn_0 ldropn_Suc_conv_ldropn 
            llist.set_intros(1))
     qed
  next
  case (Suc i)
  then show ?case 
     proof - 
      have 18: "lnull xs \<Longrightarrow> x \<in> lset (kfilter P n xs)" 
        using Suc(2) by auto
      have 19: "\<not> lnull xs \<Longrightarrow> P (lnth xs (Suc i)) = P (lnth (ltl xs) ( i)) " 
        by (simp add: lnth_ltl)
      have 20: "\<not> lnull xs \<Longrightarrow> x = (Suc n) +i \<and> enat i < llength (ltl xs) \<and> P (lnth (ltl xs) ( i))" 
        by (metis "19" Extended_Nat.eSuc_mono Suc.prems(1) add_Suc_shift eSuc_enat lhd_LCons_ltl 
            llength_LCons) 
      have 21: "\<not> lnull xs \<Longrightarrow> lnull (kfilter P n (ltl xs)) \<Longrightarrow> x \<in> lset (kfilter P n xs) "
        by (metis "20" in_lset_conv_lnth kfilter_lnull_conv)
      have 22: "\<not> lnull xs \<Longrightarrow> \<not>lnull (kfilter P n (ltl xs)) \<Longrightarrow> x \<in> lset (kfilter P (Suc n) (ltl xs))"
        by (metis "20" Suc.hyps) 
      have 23: "\<not>lnull xs \<Longrightarrow> x \<in> lset (kfilter P n xs)" 
        using kfilter_LCons[of P n "lhd xs" "ltl xs"]
              in_lset_ltlD lhd_LCons_ltl[of "(kfilter P n (ltl xs))"] 
        using "21" "22" by fastforce
      show ?thesis 
        using "18" "23" by blast
    qed
  qed
qed

lemma kfilter_lset:
 shows "  lset (kfilter P n xs) = { n+i | i. i<llength xs \<and> P (lnth xs i) }" 
using  in_kfilter_lset by blast 

lemma lfilter_lnth_exist:
 assumes 
         " i < llength (lfilter P xs)" 
 shows " (\<exists> k < llength xs. lnth (lfilter P xs) i = lnth xs k)"
using assms lset_lfilter[of P xs] 
by (metis (no_types, opaque_lifting) add.left_neutral diff_zero kfilter_llength kfilter_upperbound 
lfilter_kfilter zero_enat_def)

lemma ldistinct_kfilter:
 " ldistinct(kfilter P n xs)"
proof (coinduction arbitrary: n xs)
case (ldistinct n1 xs1)
then show ?case 
   proof -
    have 1: "lhd (kfilter P n1 xs1) \<notin> lset (ltl (kfilter P n1 xs1))"
      proof -
       have f1: "kfilter P n1 xs1 =
                 LCons (n1 + lleast P xs1)
                       (lmap snd 
                           (lfilter (P \<circ> fst) 
                                    (lzip 
                                         (ldrop (enat (Suc (lleast P xs1))) xs1) 
                                         (iterates Suc (Suc (n1 + lleast P xs1))))))"
         by (meson kfilter_ldropWhile ldistinct)
       then have "lmap snd 
                    (lfilter (P \<circ> fst) 
                            (lzip (ldrop (enat (Suc (lleast P xs1))) xs1) 
                                  (iterates Suc (Suc (n1 + lleast P xs1))))) = 
              kfilter P (Suc (n1 + lleast P xs1)) (ldrop (enat (Suc (lleast P xs1))) xs1)"
       using kfilter_eq_LCons_1 by blast
       then show ?thesis
       using f1 by (metis (no_types) in_lset_conv_lnth kfilter_lowerbound leD lessI lhd_LCons ltl_simps(2))
      qed
    have 2: "((\<exists>n xs. ltl (kfilter P n1 xs1) = kfilter P n xs) \<or> ldistinct (ltl (kfilter P n1 xs1)))"
      by (metis kfilter_eq_LCons_1 ldistinct llist.discI(1) llist.exhaust_sel)
    show ?thesis
      using "1" "2" by auto         
 qed
qed
   

lemma kfilter_llength_ltake:
 " llength(kfilter P n (ltake k xs)) \<le> llength(kfilter P n xs)"
by (simp add: kfilter_llength lprefix_lfilterI lprefix_llength_le)

lemma kfilter_ldropn_lset:
 assumes " k< llength xs"
 shows   "  lset(kfilter P n (ldropn k xs)) = 
            { n+i | i. i< llength xs - k \<and> P (lnth xs (k+i)) } "
using assms 
kfilter_lset[of P n "(ldropn k xs)" ] 
by auto
  (metis (no_types, lifting) add.commute enat_min lnth_ldropn order.strict_implies_order 
   plus_enat_simps(1),
   metis (no_types, lifting) add.commute dual_order.strict_implies_order enat_min lnth_ldropn 
   plus_enat_simps(1))

lemma kfilter_ldropn_lset_a:
 assumes " k< llength xs"
 shows "  lset(kfilter P n (ldropn k xs)) = 
          { n+(i-k) | i. k\<le>i \<and> i< llength xs \<and> P (lnth xs i) } "
proof -
 have 1: "\<And>x. x\<in> { n+i | i. i< llength xs - k \<and> P (lnth xs (k+i)) } \<longleftrightarrow> 
            x \<in> { n+(i-k) | i. k\<le>i \<and> i< llength xs \<and> P (lnth xs i) }" 
    proof auto
     show "\<And>i. enat i < llength xs - enat k \<Longrightarrow> 
               P (lnth xs (k + i)) \<Longrightarrow>
                \<exists>ia. i = ia - k \<and> k \<le> ia \<and> enat ia < llength xs \<and> P (lnth xs ia)" 
       using assms 
       by (metis add.commute add_diff_cancel_left' enat_min le_add1 less_imp_le plus_enat_simps(1))
     show "\<And>i. k \<le> i \<Longrightarrow>
                enat i < llength xs \<Longrightarrow> 
                P (lnth xs i) \<Longrightarrow> 
                \<exists>ia. n + i - k = n + ia \<and> enat ia < llength xs - enat k \<and> P (lnth xs (k + ia)) " 
       using assms ldropn_Suc_conv_ldropn[of _ xs] ldropn_eq_LConsD[of _ "ldropn k xs"] 
             ldropn_ldropn[of _ _ xs]
       by (metis Nat.add_diff_assoc le_add_diff_inverse le_add_diff_inverse2 llength_ldropn)
     qed 
  show ?thesis  using assms 1 kfilter_ldropn_lset[of k xs P n] by auto 
qed

lemma kfilter_ldropn_lset_b:
 assumes " k< llength xs"
 shows "  lset(kfilter P n (ldropn k xs)) = 
          { n+i | i. i< llength xs -k \<and> P (lnth xs (i+k)) } "
proof -
 have 1: "\<And>x. x\<in>{ n+i | i. i< llength xs - k \<and> P (lnth xs (k+i)) }  \<longleftrightarrow>
               x\<in> { n+i | i. i< llength xs -k \<and> P (lnth xs (i+k)) }" 
   by (auto simp add: add.commute)
 show ?thesis  using assms 1 kfilter_ldropn_lset[of k xs P n] by auto 
qed
  
lemma kfilter_llength_n_zero: 
shows " llength(kfilter P n xs) = llength(kfilter P  0 xs)"
 by (simp add: kfilter_llength)

lemma kfilter_lnth_n_zero_a:
 assumes " k < llength (kfilter P n xs)"
 shows " n \<le> (lnth (kfilter P n xs) k)"
using assms by (simp add: kfilter_lnull_conv kfilter_lowerbound)

lemma kfilter_lnth_n_zero:
 assumes " k < llength (kfilter P n xs)"
 shows "(lnth (kfilter P n xs) k) -n = (lnth (kfilter P 0 xs) k)"
using assms
proof (induct k arbitrary: xs n)
case 0
then show ?case by (cases "(kfilter P n xs)")
     (simp, 
      metis add.left_neutral add_left_cancel kfilter_ldropWhile kfilter_lnull_conv kfilter_lowerbound 
      le_add_diff_inverse llength_eq_0 lnth_0 not_gr_zero zero_enat_def)
next
case (Suc k)
then show ?case 
     proof -
      have 1: "lnull(kfilter P n xs) \<Longrightarrow> lnth (kfilter P n xs) (Suc k) - n = lnth (kfilter P 0 xs) (Suc k) "
        using Suc.prems gr_implies_not_zero llength_lnull by blast 
      have 2: "\<not>lnull(kfilter P n xs) \<Longrightarrow> lnth (kfilter P n xs) (Suc k) -n = lnth (kfilter P 0 xs) (Suc k)"
        proof -
         assume a0: "\<not>lnull(kfilter P n xs)"
         show "lnth (kfilter P n xs) (Suc k) -n = lnth (kfilter P 0 xs) (Suc k)"
          proof -
           let ?Least = "lleast P xs" 
           have 3: "\<exists> x xs'.(kfilter P n xs) = LCons x xs' "
             using a0 kfilter_ldropWhile by blast 
           obtain x xs' where 4: "(kfilter P n xs) = LCons x xs'"
             using 3 by auto
           have 5: "x = n+?Least"
             using kfilter_ldropWhile[of P n xs]
             by (simp add: "4")
           have 6: "\<not>lnull(kfilter P n xs) \<longleftrightarrow> \<not>lnull(kfilter P 0 xs)"
             by (simp add: kfilter_not_lnull_conv) 
           have 7: "\<exists> y ys'.(kfilter P 0 xs) = LCons y ys' "
             using "6" a0 kfilter_ldropWhile by blast
           obtain y ys' where 8: "(kfilter P 0 xs) = LCons y ys'" 
             using 7 by auto
           have 9: "y = ?Least" 
             using kfilter_ldropWhile[of P 0 xs] by (simp add: "8")
           have 10: "x-n = y"
             using "5" "9" diff_add_inverse by blast 
           have 11: "xs' = kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs) "
             using kfilter_ldropWhile[of P n xs]
             by (metis (no_types, lifting) "4" a0 kfilter_def ltl_simps(2))
           have 12: "ys' = kfilter P (Suc (0+?Least)) (ldrop (Suc ?Least) xs)" 
             using kfilter_ldropWhile[of P 0 xs]
             by (metis (no_types, lifting) "6" "8" a0 kfilter_def ltl_simps(2))
           have 13: "lnth (kfilter P n xs) (Suc k) - n  = lnth xs' k - n"
             by (simp add: "4") 
           have 14: "lnth (kfilter P 0 xs) (Suc k)   = lnth ys' k "
             by (simp add: "8") 
           have 15: "\<not>(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow>
                      lnth (kfilter P n xs) (Suc k) -n = lnth (kfilter P 0 xs) (Suc k)  " 
             using 8 Suc by auto 
             (metis (full_types) gr_implies_not_zero kfilter_eq_conv kfilter_not_lnull_conv 
              ldropn_Suc_LCons ldropn_Suc_conv_ldropn ldropn_eq_LConsD llength_LNil llist.disc(2))   
           have 16: "enat k < llength (kfilter P (Suc (n+?Least)) (ldrop (Suc ?Least) xs)) "
             by (metis (no_types, lifting) "12" "8" Extended_Nat.eSuc_mono Suc.prems eSuc_enat 
                  kfilter_llength llength_LCons)
           have 17: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow>
                   lnth xs' k  - (Suc (n+?Least)) = lnth (kfilter P 0 (ldrop (Suc ?Least) xs)) k"
             using Suc.hyps using "11" "16" by blast
           have 18: "enat k < llength (kfilter P (Suc (?Least)) (ldrop (Suc ?Least) xs)) "  
             by (metis "16" kfilter_llength)  
           have 19: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow>
                     lnth ys' k - (Suc (0+?Least)) =
                     lnth (kfilter P 0 (ldrop (Suc ?Least) xs)) k" 
             using Suc.hyps by (simp add: "12" "18")
           have 20: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow>
                        lnth (kfilter P n xs) (Suc k) -n = lnth xs' k -n"
             using "13" by blast
           have 21: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow> 
                       lnth xs' k -n = lnth (kfilter P 0 (ldrop (Suc ?Least) xs)) k + (Suc (?Least)) "
             using "11" "16" "17" kfilter_lowerbound by fastforce 
           have 22: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow> 
                       lnth (kfilter P 0 (ldrop (Suc ?Least) xs)) k + (Suc (?Least)) = lnth ys' k"
             using "12" "18" "19" kfilter_lowerbound by fastforce
           have 23: "(\<exists> x \<in> lset (ldrop (Suc ?Least) xs). P x) \<Longrightarrow>
                       lnth (kfilter P n xs) (Suc k) -n = lnth (kfilter P 0 xs) (Suc k)  "                
             using "14" "20" "21" "22" by linarith        
           show ?thesis 
             using "15" "23" by blast
         qed
       qed
     show ?thesis 
     using "1" "2" by blast
  qed
qed

lemma kfilter_n_zero:
 shows " (kfilter P n xs) = lmap (\<lambda>i. i+n) (kfilter P 0 xs)"
proof -
 have 1: "llength(kfilter P n xs) = llength (lmap (\<lambda>i. i+n) (kfilter P 0 xs)) "
   by (simp add: kfilter_llength)
 have 2: "\<And>k. k < llength(kfilter P n xs) \<longrightarrow> 
           lnth (kfilter P n xs) k = lnth (lmap (\<lambda>i. i+n) (kfilter P 0 xs)) k " 
   using kfilter_lnth_n_zero kfilter_lnth_n_zero_a
   by (metis "1"  le_add_diff_inverse2 llength_lmap lnth_lmap )
 show ?thesis 
   by (simp add: "1" "2" llist_eq_lnth_eq)   
qed

lemma kfilter_n_zero_a:
 shows " (kfilter P 0 xs) = lmap (\<lambda>i. i-n) (kfilter P n xs)"
proof -
 have 1: "llength(kfilter P 0 xs) = llength (lmap (\<lambda>i. i-n) (kfilter P n xs)) "
   by (simp add: kfilter_llength)
 have 2:  "\<And>k. k < llength(kfilter P 0 xs) \<longrightarrow> 
           lnth (kfilter P 0 xs) k = lnth (lmap (\<lambda>i. i-n) (kfilter P n xs)) k "
   by (simp add:  kfilter_llength kfilter_lnth_n_zero)
 show ?thesis 
   using "1" "2" llist_eq_lnth_eq by blast
qed

lemma kfilter_holds:
 assumes "y \<in> lset(kfilter P n xs)" 
 shows " P (lnth xs (y-n))"
using assms in_kfilter_lset[of y P n xs]  kfilter_lnull_conv[of P n xs]  
using lset_lnull by fastforce

lemma kfilter_holds_not:
 assumes "y \<in> ({i+n| i. i < llength xs} - (lset (kfilter P n xs)))" 
 shows "  \<not> P (lnth xs (y-n))"
using assms kfilter_lset[of P n xs] kfilter_lnull_conv[of P n xs]
by auto 

lemma kfilter_holds_a:
 assumes " i < llength xs "
         "(i+n) \<in> lset(kfilter P n xs) " 
 shows " P (lnth xs i) "
using assms kfilter_holds[of "i+n" P n xs]  by simp 

lemma kfilter_holds_not_a:
 assumes " i < llength xs "
         " P (lnth xs i)"
 shows " (i+n) \<in> lset(kfilter P n xs)  "   
using assms 
by (simp add: in_kfilter_lset kfilter_lnull_conv)

lemma kfilter_holds_b:
 assumes " i < llength xs"
 shows "(i+n) \<in> lset(kfilter P n xs) =   P (lnth xs i)"
using assms 
by (meson kfilter_holds_a kfilter_holds_not_a)    

lemma kfilter_holds_c: 
  assumes " n \<le> i"
          " i -n < llength xs"
  shows " i \<in> lset(kfilter P n xs) = P (lnth xs (i-n))"
using assms 
by (metis diff_add idiff_enat_enat kfilter_holds kfilter_holds_not_a)  

lemma kfilter_holds_not_b:
  assumes " n \<le> i"
          " i -n < llength xs"
  shows  " i \<notin> lset(kfilter P n xs) = (\<not> P (lnth xs (i-n)))"
using assms by (simp add: kfilter_holds_c)

lemma kfilter_disjoint_lset_coset: 
   shows "({i+n| i. i < llength xs} - (lset (kfilter P n xs))) \<inter> lset (kfilter P n xs) = {}  "
by blast 

lemma lidx_less:
 assumes " lidx xs "
         " Suc(n+k) < llength xs"
 shows   " lnth xs n < lnth xs (Suc(n+k)) "
using assms  unfolding lidx_def
proof (induct k)
case 0
then show ?case by auto
next
case (Suc k)
then show ?case 
by (metis Suc_ile_eq add_Suc_right less_imp_le less_trans)
qed 
  
lemma lidx_less_eq:
 assumes " lidx  xs "
         " k \<le> j"
         " j < llength xs "
 shows " lnth xs k \<le> lnth xs j" 
using assms
proof (cases "k=j")
case True
then show ?thesis by simp
next
case False
then show ?thesis using assms 
by (metis less_imp_Suc_add less_imp_le lidx_less order.not_eq_order_implies_strict)
qed

lemma lidx_gr_first:
 assumes "lidx xs"
         " 0<i"
         " i < llength xs"
 shows "(lnth xs 0) < lnth xs i "
using assms lidx_less[of  xs 0 "i-1" ] unfolding lidx_def by simp

lemma lidx_expand_1:
 " lidx xs \<longleftrightarrow> lnull xs \<or> llength xs = 1 \<or> 
               (1 < llength xs \<and> (\<forall>n. (Suc n)<llength xs \<longrightarrow> lnth xs n < lnth xs (Suc n)))"
by (metis One_nat_def enat_0_iff(1) iless_Suc_eq ldropn_0 ldropn_Suc_conv_ldropn ldropn_eq_LNil 
    lidx_def llength_LCons llength_LNil lnull_ldropn not_le_imp_less not_less_zero 
    old.nat.distinct(2) one_eSuc one_enat_def order.not_eq_order_implies_strict)

lemma lidx_LCons:
 " lidx (LCons x xs) \<longleftrightarrow> 
     lnull xs \<or> 
     (0 < llength xs \<and> 
      (\<forall>n. (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n)))" 
by (metis ldropn_0 lidx_def lidx_expand_1 llength_LCons llength_LNil lnull_def lnull_ldropn not_less 
    one_eSuc zero_enat_def)

lemma lidx_LCons_conv:
 "(0 < llength xs \<and> (\<forall>n. (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n)))
  \<longleftrightarrow> (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 0\<le> n \<and> (Suc n)<  (llength xs)  \<longrightarrow> lnth xs n < lnth xs (Suc n)))"
proof -
 have 1: "(0 < llength xs \<and> 
           (\<forall>n. (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n))) \<longleftrightarrow>
          (0 < llength xs \<and> 
           (\<forall>n. 0\<le> n \<and> (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n))) " 
       by blast
 have 2: "(0 < llength xs \<and> 
           (\<forall>n. 0\<le> n \<and> (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n))) \<longleftrightarrow>
          (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n)))"
           by (metis Extended_Nat.eSuc_mono One_nat_def eSuc_enat gr_implies_not_zero ldropn_0 less_Suc_eq 
               lhd_ldropn lnth_0 lnth_Suc_LCons not_le_imp_less zero_enat_def) 
 have 3: "(0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (Suc n)< eSuc (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n))) \<longleftrightarrow>
         (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (n)<  (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n)))" 
    using Suc_ile_eq by auto
 have 4: "(0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (n)<  (llength xs)  \<longrightarrow> lnth (LCons x xs) n < lnth (LCons x xs) (Suc n))) \<longleftrightarrow>
          (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (n)<  (llength xs)  \<longrightarrow> lnth xs (n -1) < lnth xs ( n)))" 
    by (metis le_add_diff_inverse llist.disc(2) lnth_ltl ltl_simps(2) plus_1_eq_Suc)
 have 5: "(0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 1\<le> n \<and> (n)<  (llength xs)  \<longrightarrow> lnth xs (n -1) < lnth xs ( n))) \<longleftrightarrow>
          (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 0\<le> (n-1) \<and> (Suc (n-1))<  (llength xs)  \<longrightarrow> lnth xs (n -1) < lnth xs (Suc (n -1))))"
         by (metis diff_Suc_1 le0 le_add1 le_add_diff_inverse plus_1_eq_Suc) 
 have 6: "(0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 0\<le> (n-1) \<and> (Suc (n-1))<  (llength xs)  \<longrightarrow> lnth xs (n -1) < lnth xs (Suc (n -1)))) \<longleftrightarrow>
          (0 < llength xs \<and> 
           x< lhd xs \<and> 
           (\<forall>n. 0\<le> n \<and> (Suc n)<  (llength xs)  \<longrightarrow> lnth xs n < lnth xs (Suc n)))"
         by (metis diff_Suc_1) 
  show ?thesis 
  using "2" "3" "4" "5" "6" by auto  
qed


lemma lidx_LCons_1 [simp]:
 "lidx (LCons x xs) \<longleftrightarrow> lnull xs \<or> (0 < llength xs \<and> x< lhd xs \<and> lidx xs )"
unfolding lidx_def 
by (auto simp add: enat_0_iff(1))
   (metis Suc_ile_eq i0_less lhd_conv_lnth llength_eq_0 lnth_0 zero_enat_def,
    metis Suc_ile_eq lnth_Suc_LCons,
    metis gr_implies_not_zero iless_Suc_eq lidx_LCons_conv llength_eq_0 lnth_Suc_LCons 
    not_less_iff_gr_or_eq)

lemma lidx_kfilter_expand:
 assumes " (Suc na) < llength(kfilter P n xs)" 
 shows " lnth (kfilter P n xs) na < lnth (kfilter P n xs) (Suc na)"
using assms kfilter_mono by force

lemma lidx_kfilter:
 shows " lidx  (kfilter P n xs)"
unfolding lidx_def 
using lidx_kfilter_expand by blast

lemma lidx_kfilter_gr_eq:
  assumes 
          "k\<le> j"
          " j < llength(kfilter P n xs)"
 shows    " lnth (kfilter P n xs) k \<le> lnth (kfilter P n xs) j"
using assms  
using lidx_kfilter lidx_less_eq by blast

lemma lidx_kfilter_gr:
  shows    "\<forall> j . k<j \<and> j < llength(kfilter P n xs) \<longrightarrow>  
             lnth (kfilter P n xs) k < lnth (kfilter P n xs) j"
using less_imp_Suc_add lidx_kfilter lidx_less by blast

lemma kfilter_not_before:
 assumes "0<llength(kfilter P 0 xs)" 
         " i< lnth (kfilter P 0 xs) 0 "
  shows " \<not> P (lnth xs i)"
proof -
 have 0: " (lnth (kfilter P 0 xs) 0)  < llength xs"
       by (metis assms(1) gen_llength_def kfilter_upperbound llength_code zero_enat_def)
 have 1: "\<not> lnull (kfilter P 0 xs)" 
    using assms(1) by auto
 have 2: " i \<in> lset (kfilter P 0 xs) \<Longrightarrow> 
           \<not> lnull (kfilter P 0 xs) \<Longrightarrow> 
           i< lnth (kfilter P 0 xs) 0 \<Longrightarrow> 
           False"
    proof (induct zs\<equiv>"(kfilter P 0 xs)" arbitrary: xs rule: lset_induct)
    case (find xs)
    then show ?case by (metis less_not_refl3 lhd_LCons lnth_0_conv_lhd)
    next
    case (step x' xs)
    then show ?case
    proof -
    have "\<forall>ns n. \<exists>na. ((n::nat) \<notin> lset ns \<or> lnth ns na = n) \<and> (n \<notin> lset ns \<or> enat na < llength ns)"
    by (meson in_lset_conv_lnth)
    then show ?thesis using step
    by (metis (no_types) leD less_nat_zero_code lidx_kfilter_gr_eq llist.set_intros(2) not_le_imp_less)
    qed 
    qed
 have 3: " i \<notin> lset (kfilter P 0 xs) \<and> i < llength xs \<longrightarrow> \<not> P (lnth xs i) "
   by (simp add: "1" in_kfilter_lset)      
 have 4: " i < llength xs" 
    using "0" assms(2) dual_order.strict_trans enat_ord_simps(2) by blast
 show ?thesis 
 using "1" "2" "3" "4" assms(2) by blast
qed

lemma kfilter_n_not_before:
 assumes "0 < llength (kfilter P n (ldropn n xs) ) "
         " n < llength xs" 
         " n \<le> i"
         " i< lnth (kfilter P n (ldropn n xs) ) 0 "
  shows " \<not> P (lnth xs i)"
proof -
 have 00: "\<not> lnull (kfilter P n (ldrop n xs)) " 
     by (metis assms(1) ldrop_enat less_numeral_extra(3) llength_LNil llist.collapse(1))
 have 0: "lnth (kfilter P n (ldrop n xs) ) 0 < llength xs" 
   using assms kfilter_upperbound[of 0 P n "(ldropn n xs)"]
   by (metis lappend_ltake_enat_ldropn ldrop_enat llength_lappend llength_ltake min.strict_order_iff 
       zero_enat_def) 
 have 1: "i \<in> lset (kfilter P n (ldropn n xs)) \<Longrightarrow> 
          \<not> lnull (kfilter P n (ldropn n xs)) \<Longrightarrow> 
          i< lnth (kfilter P n (ldropn n xs) ) 0 \<Longrightarrow> 
          n < llength xs  \<Longrightarrow> 
          n \<le> i\<Longrightarrow> 
          False" 
   proof (induct zs\<equiv>"(kfilter P n (ldropn n xs))" arbitrary: n xs rule: lset_induct)
   case (find xs)
   then show ?case 
   by (metis lnth_0 nat_less_le)
   next
   case (step x' xs)
   then show ?case 
   by (metis (no_types, lifting) in_lset_conv_lnth kfilter_eq_LCons_1 kfilter_lowerbound leD 
       less_SucI  lnth_0 )
   qed
 have 2: "i \<notin> lset (kfilter P n (ldropn n xs)) \<and> n \<le> i \<and> i < llength xs \<longrightarrow> \<not> P (lnth xs i) "
    using assms kfilter_holds_not_a[of i] 00 
       by (simp add: kfilter_ldropn_lset_a ldrop_enat)
 have 3: " n \<le> i \<and> i < llength xs"
     using assms 00 kfilter_ldropWhile[of P n "ldropn n xs" ] 
     by (metis "0" enat_ord_simps(2) ldrop_enat less_trans)    
 show ?thesis 
 using "1" "2" "3" assms(1) assms(2) assms(4) kfilter_not_lnull_conv by auto
qed 


lemma kfilter_not_after:
 assumes "0<llength(kfilter P 0 xs)" 
         " lnth (kfilter P 0 xs) (k-1) < i"
         " llength (kfilter P 0 xs) = (enat k)" 
         " i < llength xs" 
  shows " \<not> P (lnth xs i)"
proof -
 have 0: "\<not> lnull (kfilter P 0 xs)" 
     using assms(1) by auto
 have 01: "0<k" 
    using "0" assms(3) gr0I zero_enat_def by fastforce
 have 02: "i \<notin> lset (kfilter P 0 xs) "
 by (metis "01" One_nat_def Suc_pred assms(2) assms(3) diff_less enat_ord_simps(2) 
     in_lset_conv_lnth leD less_Suc_eq_le lidx_kfilter_gr_eq zero_less_one)   
 from 0 01 02 show ?thesis 
 by (metis add.right_neutral assms(4) kfilter_holds_not_a)
qed

lemma kfilter_n_not_after:
 assumes "0 < llength (kfilter P n (ldropn n xs))" 
         " n < llength xs" 
         " lnth (kfilter P n (ldropn n xs) ) (k-1) < i"
         " llength (kfilter P n (ldropn n xs) ) = (enat k)" 
         " i < llength xs" 
  shows " \<not> P (lnth xs i)"
proof -
 have 0: "\<not> lnull (kfilter P n (ldropn n xs))" 
      using assms(1) by auto
 have 1: "0<k"
   using assms(1) assms(4) by (simp add: zero_enat_def)
 have 2: "i \<notin> lset(kfilter P n (ldropn n xs))"  
     using assms
     by (metis "1" Suc_diff_1 enat_ord_simps(2) in_lset_conv_lnth leD less_Suc_eq_le 
         lidx_kfilter_gr_eq order_refl)       
 from  0 1 2 show ?thesis 
using assms kfilter_ldropn_lset_a[of n xs P n] kfilter_lowerbound[of _ P n "(ldropn n xs)"] 
  by auto (metis One_nat_def diff_less le_trans less_imp_le zero_less_one) 
qed

lemma kfilter_not_between:
 assumes " lnth (kfilter P 0 xs) (k) < i"
         " i<lnth (kfilter P 0 xs) (Suc k)" 
         " (Suc k) < llength (kfilter P 0 xs) " 
  shows " \<not> P (lnth xs i)"
proof -
 have 0: "\<exists> x \<in> lset xs. P x" 
     using assms(3) gr_implies_not_zero kfilter_not_lnull_conv by fastforce
 have 1: "\<not> lnull (kfilter P 0 xs)" 
    by (simp add: 0 kfilter_not_lnull_conv)
 have 2: "lnth (kfilter P 0 xs) (Suc k) \<le> llength xs"
    using kfilter_upperbound[of "Suc k" P 0 xs] 
    using "1" assms(3) zero_enat_def by auto 
 have 3: "i \<notin> lset(kfilter P 0 xs)"
  using assms
  by (metis Suc_ile_eq dual_order.strict_implies_order in_lset_conv_lnth leD lidx_kfilter_gr_eq 
      not_less_eq_eq) 
 from 1 2 3 show ?thesis 
 by (metis add.right_neutral assms(2) enat_ord_simps(2) kfilter_holds_not_a less_le_trans)
qed

lemma lidx_imp_lsorted:
assumes "lidx xs"
shows   "lsorted xs"
using assms 
by (metis (no_types, lifting) less_imp_le lhd_LCons_ltl lidx_LCons_1 lsorted_coinduct')

lemma lidx_imp_ldistinct:
 assumes "lidx  xs"
 shows  "ldistinct xs"
using assms 
proof (coinduction arbitrary: xs)
case (ldistinct xsa)
then show ?case 
   proof -
     have 1: "lhd xsa \<notin> lset (ltl xsa)"
       by (metis empty_iff lappend_code(1) lappend_lnull2 ldistinct(1) ldistinct(2) leD lhd_LCons_ltl 
     lidx_LCons_1 lidx_imp_lsorted lmember_code(2) lset_LNil lset_lmember lsortedD)        
     have 2: "((\<exists>xs. ltl xsa = xs \<and> lidx  xs) \<or> ldistinct (ltl xsa))" 
       unfolding lidx_def 
       by (metis Extended_Nat.eSuc_mono eSuc_enat ldistinct(1) ldistinct(2) lhd_LCons_ltl lidx_def 
           llength_LCons lnth_ltl)
     show ?thesis 
       using "1" "2" by auto    
  qed
qed

lemma ldistinct_Ex1:
 assumes "ldistinct xs"
         " x \<in> lset xs"
 shows   " \<exists>!i. i< llength xs \<and> (lnth xs i) = x"
using assms 
by (metis in_lset_conv_lnth ldistinct_conv_lnth)

lemma lidx_lset_eq:
 assumes "lidx  xs"
         "lidx  ys"
         "lset xs = lset ys"
 shows " xs = ys"
using assms 
 by (simp add: lidx_imp_ldistinct lidx_imp_lsorted lsorted_ldistinct_lset_unique)
                     
lemma lfilter_kfilter_ltake_lidx_a:
 assumes " k < llength(lfilter P xs)"
 shows  " lidx (ltake k (kfilter P n xs))"
unfolding lidx_def
using assms 
by (metis Suc_ile_eq kfilter_mono less_imp_le  llength_ltake lnth_ltake min.strict_boundedE) 

lemma lfilter_kfilter_ldropn_lidx_a:
 assumes " k < llength(lfilter P xs)"
 shows  " lidx (ldropn k (kfilter P n xs))"
using assms unfolding lidx_def 
  proof auto
   fix na 
   assume a0: "enat k < llength (lfilter P xs)"
   assume a1: "enat (Suc na) < llength (kfilter P n xs) - enat k" 
   show "lnth (ldropn k (kfilter P n xs)) na < lnth (ldropn k (kfilter P n xs)) (Suc na) "
   proof -
    have 1: "enat (k + (Suc na)) < llength (kfilter P n xs)"
      proof -
       have "Suc na + k = k + Suc na"
         by presburger
       then show ?thesis
         by (metis (no_types) a1 ldropn_eq_LNil ldropn_ldropn leD leI llength_ldropn)
      qed
    have 2 : "enat (k + na) < llength (kfilter P n xs)"
      by (metis "1" Suc_ile_eq add_Suc_right less_imp_le)
    have 3: "lnth (kfilter P n xs) (k + (na)) <  lnth (kfilter P n xs) (k + (Suc na))"
      using "1" kfilter_mono by auto
    show ?thesis by (metis "1" "2" "3" add.commute lnth_ldropn) 
   qed
qed

lemma lfilter_kfilter_ldropn_lidx_b:
 assumes " k < llength(lfilter P xs)"
 shows  " lidx (kfilter P (lnth (kfilter P n xs) k) (ldropn (lnth (kfilter P n xs) k) xs))"
using assms using lidx_kfilter by blast

lemma ltake_lset:
 assumes "k < llength xs"
 shows   " lset (ltake k xs) = {(lnth xs i) | i. i < k}  "
using assms 
by (auto simp add: in_lset_conv_lnth lnth_ltake)
   (blast, meson less_trans lnth_ltake)

lemma ldropn_lset:
 assumes " k < llength xs"
 shows " lset (ldropn k xs) = {(lnth xs i) | i. k \<le> i \<and> i < llength xs}"
using assms 
proof (auto simp add: in_lset_conv_lnth )
  fix n :: nat
  assume a1: "enat n < llength xs - enat k"
  assume "enat k < llength xs"
  then have "enat k \<le> llength xs"
    by (meson dual_order.strict_implies_order)
  then have "enat n + enat k < llength xs"
    using a1 by (meson enat_min)
  then show "\<exists>na. lnth (ldropn k xs) n = lnth xs na \<and> k \<le> na \<and> enat na < llength xs"
    using a1 by (metis ldropn_ldropn le_add2 lhd_ldropn llength_ldropn plus_enat_simps(1))
next
  fix i :: nat
  assume a1: "enat i < llength xs"
  assume a2: "k \<le> i"
  have 1: " enat (i-k) < llength xs -enat k"
    by (metis a1 a2 diff_add ldropn_Suc_conv_ldropn ldropn_ldropn llength_ldropn llist.disc(2)
          lnull_ldropn not_le_imp_less) 
  have 2: "lnth (ldropn k xs) (i-k) = lnth xs i"
    by (simp add: a1 a2) 
  show "\<exists>n. enat n < llength xs - enat k \<and> lnth (ldropn k xs) n = lnth xs i"
    using 1 2 by blast
qed  
 
lemma lfilter_kfilter_ltake_lset_eq:
 assumes 
         " k < llength(lfilter P xs)"
 shows " lset (ltake k (kfilter P 0 xs)) =
         lset (kfilter P 0 (ltake ((lnth (kfilter P 0 xs) k)) xs))"
proof -
 have 1: "(lnth (kfilter P 0 xs) k) < llength xs"
   using kfilter_llength kfilter_upperbound
   by (metis assms gen_llength_def  kfilter_llength kfilter_upperbound llength_code )
 have 2: " \<exists> x \<in> lset xs . P x"
   using assms gr_implies_not_zero llength_eq_0 lnull_lfilter by blast
 have 3: "{i. i< llength(ltake ((lnth (kfilter P 0 xs) k)) xs) \<and>
           P (lnth (ltake ((lnth (kfilter P 0 xs) k)) xs) i)} =
          {i. i< (lnth (kfilter P 0 xs) k) \<and> P (lnth xs i)} "
   by (auto simp add: lnth_ltake "1" order_less_subst2)
 have 5: "{i. i< (lnth (kfilter P 0 xs) k) \<and> P (lnth xs i)} =
          {i. i< (lnth (kfilter P 0 xs) k) \<and> i \<in> lset(kfilter P 0 xs)} "
   by (auto simp add: "1" kfilter_holds_c order_less_subst2)
 have 6: "{i. i< (lnth (kfilter P 0 xs) k) \<and> i \<in> lset(kfilter P 0 xs)} =
          {i. i< (lnth (kfilter P 0 xs) k) \<and> i \<in> {(lnth (kfilter P 0 xs) j) | j. j < llength(kfilter P 0 xs)}} "
   by (simp add: lset_conv_lnth) 
 have 7: "{i. i< (lnth (kfilter P 0 xs) k) \<and> i \<in> {(lnth (kfilter P 0 xs) j) | j. j < llength(kfilter P 0 xs)}} =
            {(lnth (kfilter P 0 xs) j) | j. j<k}"
   by (auto simp add: assms lidx_kfilter_gr kfilter_llength)
      (metis kfilter_llength leD lidx_kfilter_gr_eq not_le_imp_less,
       metis assms enat_ord_simps(2) less_trans)        
 have 8: "k < llength(kfilter P 0 xs)"
   by (simp add: assms kfilter_llength) 
 have 9: "{(lnth (kfilter P 0 xs) j) | j. j<k} = lset(ltake k (kfilter P 0 xs))" 
   using ltake_lset[of k "(kfilter P 0 xs)"] 8  by auto
 have 10: "lset (kfilter P 0 (ltake ((lnth (kfilter P 0 xs) k)) xs)) =
           {i. i< (lnth (kfilter P 0 xs) k) \<and>
               P (lnth (ltake (lnth (kfilter P 0 xs) k) xs) i) }"
   by (auto simp add: in_kfilter_lset)
      (meson "1" enat_ord_simps(2) less_trans)
 have 11: "(lnth (kfilter P 0 xs) k) = llength(ltake (lnth (kfilter P 0 xs) k) xs) "
   by (simp add: "1" dual_order.strict_implies_order min_def)      
 have 12: "{i. i< (lnth (kfilter P 0 xs) k) \<and>
               P (lnth (ltake (lnth (kfilter P 0 xs) k) xs) i) } =
           {i. i< llength(ltake (lnth (kfilter P 0 xs) k) xs) \<and>
               P (lnth (ltake (lnth (kfilter P 0 xs) k) xs) i) }" 
   using "11" by auto
 show ?thesis 
   using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma lfilter_kfilter_ldropn_lset_eq:
 assumes "  k < llength(lfilter P xs)"
 shows   " lset(kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs)) =
           lset(ldropn  k (kfilter P 0 xs))"
proof -
 have 1: "(lnth (kfilter P 0 xs) k) < llength xs "
   using kfilter_llength kfilter_upperbound
   by (metis assms gen_llength_def  kfilter_llength kfilter_upperbound llength_code )
 have 2: " \<exists> x \<in> lset xs . P x"
   using assms gr_implies_not_zero llength_eq_0 lnull_lfilter by blast
 have 10: "lset(kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs)) =
           {(lnth (kfilter P 0 xs) k) + i | i.  i< llength xs - (lnth (kfilter P 0 xs) k) \<and>
              P (lnth xs (i+(lnth (kfilter P 0 xs) k))) }"
   using 1 kfilter_ldropn_lset_b[of "(lnth (kfilter P 0 xs) k)" xs P "(lnth (kfilter P 0 xs) k)"] 
   by linarith
 have 5: "{(lnth (kfilter P 0 xs) k) + i | i.  i< llength xs - (lnth (kfilter P 0 xs) k) \<and>
              P (lnth xs (i+(lnth (kfilter P 0 xs) k))) } =
          {(lnth (kfilter P 0 xs) k) + i | i.  i< llength xs - (lnth (kfilter P 0 xs) k) \<and>
              (i+(lnth (kfilter P 0 xs) k)) \<in> lset(kfilter P 0 xs) }"
   using kfilter_holds_b[of _ xs 0 P] using "1" "2" 
   by auto 
      (metis ldropn_eq_LNil ldropn_ldropn leD llength_ldropn not_le_imp_less,
      metis gen_llength_def in_lset_conv_lnth kfilter_upperbound llength_code)
 have 51: "{(lnth (kfilter P 0 xs) k) + i | i.  i< llength xs - (lnth (kfilter P 0 xs) k) \<and>
              (i+(lnth (kfilter P 0 xs) k)) \<in> lset(kfilter P 0 xs) } =
           {(lnth (kfilter P 0 xs) k) + i | i.  
               (lnth (kfilter P 0 xs) k) \<le> i + (lnth (kfilter P 0 xs) k) \<and>
               i + (lnth (kfilter P 0 xs) k) < llength xs  \<and>
              (i+(lnth (kfilter P 0 xs) k)) \<in> lset(kfilter P 0 xs) }" 
   by auto
      (metis "1" enat_min less_imp_le plus_enat_simps(1),
       metis ldropn_eq_LNil ldropn_ldropn leD llength_ldropn not_le_imp_less)      
 have 52: "{(lnth (kfilter P 0 xs) k) + i | i.  
               (lnth (kfilter P 0 xs) k) \<le> i + (lnth (kfilter P 0 xs) k) \<and>
               i + (lnth (kfilter P 0 xs) k) < llength xs  \<and>
              (i+(lnth (kfilter P 0 xs) k)) \<in> lset(kfilter P 0 xs) } =
            {j. (lnth (kfilter P 0 xs) k) \<le>j \<and> j < llength xs \<and> j \<in> lset(kfilter P 0 xs)}"
   by (metis (no_types, lifting)  add.commute le_iff_add) 
 have 53: "{j. (lnth (kfilter P 0 xs) k) \<le>j \<and> j < llength xs \<and> j \<in> lset(kfilter P 0 xs)} =
            {j. (lnth (kfilter P 0 xs) k) \<le>j \<and> j < llength xs \<and> 
                j \<in> { (lnth (kfilter P 0 xs) jj)| jj. jj< llength(kfilter P 0 xs)}}"
   by (simp add: lset_conv_lnth) 
 have 54: "{j. (lnth (kfilter P 0 xs) k) \<le>j \<and> j < llength xs \<and> 
                j \<in> { (lnth (kfilter P 0 xs) jj)| jj. jj< llength(kfilter P 0 xs)}} =
          { (lnth (kfilter P 0 xs) j) | j. k\<le>j \<and> j < llength (kfilter P 0 xs)}"
   by auto
      (metis assms kfilter_llength lidx_kfilter_gr not_less,
       meson lidx_kfilter_gr_eq,
       metis gen_llength_def kfilter_upperbound llength_code)
 have 8: "k < llength(kfilter P 0 xs)"
   by (simp add: assms kfilter_llength) 
 have 9: "{ (lnth (kfilter P 0 xs) j) | j. k\<le>j \<and> j < llength (kfilter P 0 xs)} =
          lset(ldropn k (kfilter P 0 xs))"
   using ldropn_lset[of k "(kfilter P 0 xs)"] using "8" by blast 
 show ?thesis 
   using "10" "5" "51" "52" "53" "54" "9" by auto
qed

lemma kfilter_kfilter_ltake:
 assumes " k < llength(lfilter P xs)"
 shows   " (ltake k (kfilter P 0 xs)) = 
           (kfilter P 0 (ltake ((lnth (kfilter P 0 xs) k)) xs))"
using assms 
by (simp add: lfilter_kfilter_ltake_lidx_a lfilter_kfilter_ltake_lset_eq lidx_kfilter lidx_lset_eq)

lemma kfilter_kfilter_ldropn:
 assumes " k < llength(lfilter P xs)"
 shows  "(ldropn  k (kfilter P 0 xs)) =
           (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs))" 
using assms 
by (simp add: lfilter_kfilter_ldropn_lidx_a lfilter_kfilter_ldropn_lidx_b 
    lfilter_kfilter_ldropn_lset_eq lidx_lset_eq)

lemma kfilter_lmap_lfilter:
 shows   " lmap (\<lambda>n. (lnth xs n)) (kfilter P 0 xs) = lfilter P xs"
using lfilter_kfilter[of _ P 0 xs] 
by (metis (no_types, lifting) diff_zero  kfilter_llength llength_lmap 
    llist_eq_lnth_eq lnth_lmap)

lemma lfilter_kfilter_ltake:
 assumes "  k < llength(lfilter P xs)"
 shows   " ltake k (lfilter P xs) = 
           (lfilter P (ltake (lnth (kfilter P 0 xs) k) xs))"
proof -
 have 2: "lmap (\<lambda>n. (lnth xs n)) (kfilter P 0 xs) = lfilter P xs"
   using kfilter_lmap_lfilter by blast 
 have 3: "ltake k (lfilter P xs) = 
          ltake k (lmap (\<lambda>n. (lnth xs n)) (kfilter P 0 xs))" 
   by (simp add: "2")
 have 4: "ltake k (lmap (\<lambda>n. (lnth xs n)) (kfilter P 0 xs)) =
           lmap (\<lambda>s. lnth xs s) (ltake k (kfilter P 0 xs))"
   using ltake_lmap by blast 
 have 6: "(lfilter P (ltake (lnth (kfilter P 0 xs) k) xs)) =
          lmap (\<lambda>s. (lnth (ltake (lnth (kfilter P 0 xs) k) xs) s)) 
               (kfilter P 0 (ltake (lnth (kfilter P 0 xs) k) xs))"
   by (simp add: kfilter_lmap_lfilter) 
 have 7: "lmap (\<lambda>s. lnth xs s) (ltake k (kfilter P 0 xs)) =
          lmap (\<lambda>s. lnth xs s) (kfilter P 0 (ltake (lnth (kfilter P 0 xs) k) xs))"
   by (simp add:  assms kfilter_kfilter_ltake) 
 have 8: "lmap (\<lambda>s. lnth xs s) (kfilter P 0 (ltake (lnth (kfilter P 0 xs) k) xs)) =
          lmap (\<lambda>s. (lnth (ltake (lnth (kfilter P 0 xs) k) xs) s)) 
               (kfilter P 0 (ltake (lnth (kfilter P 0 xs) k) xs))"
   using kfilter_kfilter_ltake[of k P xs] 
   by (metis gen_llength_def kfilter_upperbound lappend_ltake_enat_ldropn ldistinct_Ex1 
       ldistinct_kfilter llength_code llist.map_cong0 lnth_lappend)
 show ?thesis 
   using "2" "4" "6" "7" "8" by auto
qed

lemma kfilter_lmap_shift_ldropn:
 shows   " lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs)) =
           lmap (\<lambda>s. lnth xs s) 
                (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs))"
proof -
 have 1: "llength (lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                  (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs))) =
          llength (lmap (\<lambda>s. lnth xs s) 
                  (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs)))"
   by (simp add: kfilter_llength) 
 have 2: "\<And>i. i< llength (lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                               (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs))) \<Longrightarrow>
             lnth (lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                        (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs))) i =
             lnth xs ((lnth (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs)) i)+
                      (lnth (kfilter P 0 xs) k))  "
   by simp 
 have 3: "\<And>i. i< llength (lmap (\<lambda>s. lnth xs s) 
              (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs))) \<Longrightarrow>
              lnth (lmap (\<lambda>s. lnth xs s) 
              (kfilter P (lnth (kfilter P 0 xs) k) 
                         (ldropn (lnth (kfilter P 0 xs) k) xs))) i =
              lnth xs  (lnth (kfilter P (lnth (kfilter P 0 xs) k) 
                                        (ldropn (lnth (kfilter P 0 xs) k) xs)) i) "
   by simp 
 have 4: "\<And>i. i< llength (lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                         (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs))) \<Longrightarrow>
               lnth xs ((lnth (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs)) i)+
                        (lnth (kfilter P 0 xs) k)) =
               lnth xs  (lnth (kfilter P (lnth (kfilter P 0 xs) k) 
                                         (ldropn (lnth (kfilter P 0 xs) k) xs)) i) "
   using kfilter_lnth_n_zero[of _ P "(lnth (kfilter P 0 xs) k)" "(ldropn (lnth (kfilter P 0 xs) k) xs)" ]  
         kfilter_lowerbound[of _ P "(lnth (kfilter P 0 xs) k)" "(ldropn (lnth (kfilter P 0 xs) k) xs)"]  
   using 1 diff_add by fastforce
 show ?thesis 
   using llist_eq_lnth_eq[of "lmap (\<lambda>s. lnth xs (s+(lnth (kfilter P 0 xs) k))) 
                (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k) xs))"]
   using "1" "2" "3" "4" by presburger
qed

lemma lfilter_kfilter_ldropn:
 assumes "  k < llength(lfilter P xs)"
 shows " (ldropn k (lfilter P xs)) = 
         (lfilter P (ldropn (lnth (kfilter P 0 xs) k) xs))"
proof -
 have 1: "\<exists> x \<in> lset xs. P x"
   using assms gr_implies_not_zero llength_eq_0 lnull_lfilter by blast  
 have 2: "(lfilter P xs) = lmap (\<lambda>s. lnth xs s) (kfilter P 0 xs)" 
   by (simp add: kfilter_lmap_lfilter)
 have 3: "ldrop k (lfilter P xs) =
          ldrop k (lmap (\<lambda>s. lnth xs s) (kfilter P 0 xs)) "
   by (simp add: "2") 
 have 4: "(ldrop k (lmap (\<lambda>s. lnth xs s) (kfilter P 0 xs))) =
          (lmap (\<lambda>s. lnth xs s) (ldrop k (kfilter P 0 xs)))" 
   using ldrop_lmap by blast
 have 6: "(lfilter P (ldropn (lnth (kfilter P 0 xs) k) xs)) =
          lmap (\<lambda>s. lnth (ldropn (lnth (kfilter P 0 xs) k) xs) s) 
               (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs))"
   by (simp add: kfilter_lmap_lfilter) 
 have 61: "\<And>z. z \<in> lset ((kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs))) 
            \<Longrightarrow> (\<lambda>s. lnth (ldropn (lnth (kfilter P 0 xs) k) xs) s) z = 
          (\<lambda>s. lnth xs (s + (lnth (kfilter P 0 xs) k)) ) z" 
   using assms in_lset_conv_lnth[of _ "((kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs)))"] 
   by simp
      (metis add.commute add.left_neutral kfilter_upperbound ldropn_eq_LNil ldropn_ldropn leD
       lnth_ldropn not_le_imp_less zero_enat_def)
 have 7: "lmap (\<lambda>s. lnth (ldropn (lnth (kfilter P 0 xs) k) xs) s) 
               (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs)) =
          lmap (\<lambda>s. lnth xs (s + (lnth (kfilter P 0 xs) k)) )
              (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs)) "
   using "61" by auto
 have 8: " lmap (\<lambda>s. lnth xs (s + (lnth (kfilter P 0 xs) k)) )
              (kfilter P 0 (ldropn (lnth (kfilter P 0 xs) k ) xs)) =
          lmap (\<lambda>s. lnth xs s) 
              (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs)) "
   by (simp add:  kfilter_lmap_shift_ldropn) 
 have 9: "lmap (\<lambda>s. lnth xs s) 
              (kfilter P (lnth (kfilter P 0 xs) k) (ldropn (lnth (kfilter P 0 xs) k) xs)) =
          lmap (\<lambda>s. lnth xs s) (ldropn k (kfilter P 0 xs))"
   by (simp add:  assms kfilter_kfilter_ldropn)
 show ?thesis 
   by (metis "4" "7" "8" "9" kfilter_lmap_lfilter ldrop_enat) 
qed

lemma lfilter_lnth_aa:
assumes "n < llength (lfilter P xs)"
 shows  " P (lnth (lfilter P xs) n)"
using assms 
by (meson in_lset_conv_lnth lfilter_id_conv lfilter_idem)

lemma exist_one_conv:
 "(\<exists>! i. i<llength xs \<and> P (lnth xs i)) \<longleftrightarrow>
  (\<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j\<noteq>k \<longrightarrow> \<not> P (lnth xs j)))"
by blast

lemma lfilter_llength_one_conv_a:
 assumes "llength(lfilter P xs) =1"
 shows   " \<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j\<noteq>k \<longrightarrow> \<not> P (lnth xs j))"
proof -
 have 1: "P (lnth xs (lnth (kfilter P 0 xs) 0))"
   by (metis assms(1)  kfilter_llength kfilter_lmap_lfilter less_numeral_extra(1) lfilter_lnth_aa 
       lnth_lmap zero_enat_def) 
 have 2: "(lnth (kfilter P 0 xs) 0) < llength xs"
   by (metis assms(1) gen_llength_def kfilter_llength  kfilter_upperbound less_numeral_extra(1) 
       llength_code zero_enat_def) 
 have 3: "(\<forall> j < llength xs. j\<noteq> (lnth (kfilter P 0 xs) 0) \<longrightarrow> \<not> P (lnth xs j)) "
   using assms kfilter_not_after[of P xs] kfilter_not_before[of P xs]
   by (metis One_nat_def diff_Suc_1 kfilter_llength linorder_neqE_nat one_enat_def zero_less_one) 
 show ?thesis 
   using "1" "2" "3" by blast
qed

lemma lfilter_llength_one_conv_c:
 " (\<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j\<noteq>k \<longrightarrow> \<not> P (lnth xs j))) \<longleftrightarrow>
   ( \<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j<k \<or> k<j \<longrightarrow> \<not> P (lnth xs j)))"
using antisym_conv3 by  auto

lemma lfilter_llength_one_conv_d:
 " ( \<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j<k \<or> k<j \<longrightarrow> \<not> P (lnth xs j))) \<longleftrightarrow>
   ( \<exists> k < llength xs. P (lnth xs k) \<and> 
      (\<forall> j . j<k  \<longrightarrow> \<not> P (lnth xs j)) \<and>
      (\<forall> j < llength xs. k<j \<longrightarrow> \<not> P (lnth xs j)))"
by (meson enat_ord_simps(2) less_trans)

lemma exist_one_lfilter_llength_one:
assumes "(\<exists>! i. i<llength xs \<and> P (lnth xs i))"
shows " llength(lfilter P xs) \<le>1 "
using assms 
proof auto
 fix i :: nat
 assume a1: "\<forall>y y'. enat y < llength xs \<and> P (lnth xs y) \<and> enat y' < llength xs \<and> P (lnth xs y') \<longrightarrow> 
             y = y'"
 show "llength (lfilter P xs) \<le> 1"
 proof -
 have f1: "\<forall>e ea. (e::enat) \<le> ea \<or> ea < e"
   by (meson not_le_imp_less)
 have f3: "llength (kfilter P 0 xs) = gen_llength 0 (lfilter P xs)"
   by (metis kfilter_llength llength_code)
 have f4: "enat 0 + llength xs = llength xs"
   by (metis gen_llength_def llength_code)
 have "llength xs = enat 0 + llength xs"
   by (metis (full_types) gen_llength_def llength_code)
 then have f5: "\<not> 1 < llength (lfilter P xs)"
  proof -
   have g1: "Suc 0 = 0 + Suc 0"
     by auto
   have g2: "\<And>i. enat i < llength (kfilter P 0 xs) \<Longrightarrow> 
                lnth xs (lnth (kfilter P 0 xs) i - 0) = lnth (lfilter P xs) i"
     using lfilter_kfilter[of _ P 0 xs] by auto
   have g3: "\<And>k. enat k < llength (kfilter P 0 xs) \<Longrightarrow> 
                lnth (kfilter P 0 xs) k - 0 = lnth (kfilter P 0 xs) k"
     using kfilter_mono[of _ P 0 xs] by auto
   have g4:  "\<And>n. enat n < llength (lfilter P xs) \<Longrightarrow> P (lnth (lfilter P xs) n)"
     using lfilter_lnth_aa[of _ P xs]   by auto
   have g5: "enat (0 + Suc 0) = 1" 
     using one_enat_def by auto
   have "\<not> 1 < gen_llength 0 (lfilter P xs) \<or> enat 0 < gen_llength 0 (lfilter P xs)"
     using zero_enat_def by fastforce
   then show ?thesis
     by (metis g1 g2 g3 g4 g5 a1 f3 f4 kfilter_upperbound ldistinct_conv_lnth ldistinct_kfilter 
         llength_code n_not_Suc_n)
   qed
 show ?thesis 
 using f5 not_le by blast
 qed
qed

lemma lfilter_llength_one_conv_b:
assumes " \<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j\<noteq>k \<longrightarrow> \<not> P (lnth xs j))"
shows " llength(lfilter P xs) =1 "
proof -
 have 1: "llength(lfilter P xs) \<le>1"
   by (metis assms exist_one_lfilter_llength_one)
 have 2: "0<llength(lfilter P xs)"
   by (metis assms gr_zeroI in_lset_conv_lnth llength_eq_0 lnull_lfilter) 
 show ?thesis
   by (metis "1" "2" One_nat_def Suc_ile_eq dual_order.antisym one_enat_def zero_enat_def) 
qed
              
lemma lfilter_llength_one_conv:
shows "(\<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j\<noteq>k \<longrightarrow> \<not> P (lnth xs j))) \<longleftrightarrow>
          llength(lfilter P xs) =1 "
using lfilter_llength_one_conv_a[of P xs] lfilter_llength_one_conv_b[of xs P] by blast 

lemma lfilter_llength_one_conv_1:
shows "(\<exists> k < llength xs. P (lnth xs k) \<and> 
           (\<forall> j < llength xs. j<k \<or> k<j \<longrightarrow> \<not> P (lnth xs j))) \<longleftrightarrow>
          llength(lfilter P xs) =1 "
using lfilter_llength_one_conv[of xs P] lfilter_llength_one_conv_c[of xs P]
by blast

lemma lfilter_llength_one_conv_2:
shows "( \<exists> k < llength xs. P (lnth xs k) \<and> 
      (\<forall> j . j<k  \<longrightarrow> \<not> P (lnth xs j)) \<and>
      (\<forall> j < llength xs. k<j \<longrightarrow> \<not> P (lnth xs j))) \<longleftrightarrow>
          llength(lfilter P xs) =1 "
using lfilter_llength_one_conv_1[of xs P] lfilter_llength_one_conv_d[of xs P]
by blast

lemma lfilter_lappend_ltake:
 assumes " k < llength xs"
 shows   " lfilter P (ltake k xs) = ltake (llength(lfilter P (ltake k xs))) (lfilter P xs)"
proof -
 have 1: " lfilter P xs = 
           lappend (lfilter P (ltake (the_enat k) xs)) (lfilter P (ldropn (the_enat k) xs)) " 
   by (metis enat_ord_code(4) lappend_ltake_enat_ldropn lfilter_lappend_lfinite lfinite_ltake)
 have 2: "ltake (llength(lfilter P (ltake k xs))) 
                (lappend (lfilter P (ltake (the_enat k) xs)) (lfilter P (ldropn (the_enat k) xs))) =
           lfilter P (ltake k xs) "
   by (metis assms enat_the_enat lfilter_idem lfinite_ltake llength_eq_infty_conv_lfinite 
       llength_lfilter_ile llength_ltake ltake_all ltake_lappend1 min.strict_order_iff) 
 show ?thesis using 1 2 by simp
qed

lemma kfilter_lappend_lfinite:
 "lfinite xs \<Longrightarrow>
    kfilter P n (lappend xs ys) = 
    lappend (kfilter P n xs) (kfilter P (n+ (the_enat(llength xs))) ys)"
unfolding kfilter_def
proof (induct arbitrary: n rule: lfinite.induct)
case lfinite_LNil
then show ?case by simp
next
case (lfinite_LConsI xs x)
then show ?case
  proof -
   have 1: "(lzip (lappend (LCons x xs) ys) (iterates Suc n)) =
            (LCons (x , n) (lzip (lappend xs ys) (iterates Suc (Suc n)))) "
     by (simp add: lzip.ctr(2))
   have 3: "lmap snd (lfilter (P \<circ> fst) (lzip (lappend (LCons x xs) ys) (iterates Suc n))) =
            (if P x then (LCons n (lmap snd (lfilter (P \<circ> fst) (lzip (lappend xs ys) (iterates Suc (Suc n))))))
                    else lmap snd (lfilter (P \<circ> fst) (lzip (lappend xs ys) (iterates Suc (Suc n))))) "
     using "1" by auto 
   have 4: "(if P x then (LCons n (lmap snd (lfilter (P \<circ> fst) (lzip (lappend xs ys) (iterates Suc (Suc n))))))
                    else lmap snd (lfilter (P \<circ> fst) (lzip (lappend xs ys) (iterates Suc (Suc n))))) =
            (if P x then 
              (LCons n (lappend (lmap snd (lfilter (P \<circ> fst) (lzip xs (iterates Suc (Suc n))))) 
                                (lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc ((Suc n) + the_enat (llength xs))))))))
             else (lappend (lmap snd (lfilter (P \<circ> fst) (lzip xs (iterates Suc (Suc n))))) 
                           (lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc ((Suc n) + the_enat (llength xs))))))))"
     using lfinite_LConsI.hyps(2) by auto 
   have 5: "(lmap snd (lfilter (P \<circ> fst) (lzip (LCons x xs) (iterates Suc n)))) =
            (if P x then (LCons n (lmap snd (lfilter (P \<circ> fst) (lzip ( xs) (iterates Suc (Suc n))))))
                    else (lmap snd (lfilter (P \<circ> fst) (lzip ( xs) (iterates Suc (Suc n))))))" 
              by (simp add: lzip.ctr(2))
   have 6: "(lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc (n + the_enat (llength (LCons x xs))))))) =
            (lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc ((Suc n) + the_enat (llength (xs)))))))" 
     using eSuc_enat lfinite_LConsI.hyps(1) llength_eq_infty_conv_lfinite by force
   have 7: "lappend (lmap snd (lfilter (P \<circ> fst) (lzip (LCons x xs) (iterates Suc n))))
            (lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc (n + the_enat (llength (LCons x xs))))))) =
            lappend (if P x then (LCons n (lmap snd (lfilter (P \<circ> fst) (lzip xs (iterates Suc (Suc n))))))
                            else (lmap snd (lfilter (P \<circ> fst) (lzip xs (iterates Suc (Suc n))))))
                    (lmap snd (lfilter (P \<circ> fst) (lzip ys (iterates Suc ((Suc n) + ( (the_enat (llength  xs))))))))
                        " 
     using "5" "6" by auto
   show ?thesis using "3" "4" "7" by auto
 qed           
qed

lemma kfilter_lappend_ltake:
 assumes " k < llength xs"
 shows   " kfilter P n (ltake k xs) = ltake (llength(kfilter P n (ltake k xs))) (kfilter P n xs)"
proof -
 have 1: " kfilter P n xs = lappend (kfilter P n (ltake (the_enat k) xs)) 
                                    (kfilter P (n+ (the_enat k)) (ldropn (the_enat k) xs)) " 
   using kfilter_lappend_lfinite[of "(ltake (the_enat k) xs)" P n "(ldropn (the_enat k) xs)" ]
   by (metis assms enat_iless enat_ord_code(4) lappend_ltake_ldrop ldrop_enat lfinite_ltake 
       llength_ltake min.strict_order_iff neq_iff the_enat.simps) 
 have 2: "ltake (llength(kfilter P n (ltake k xs))) 
           (lappend (kfilter P n (ltake (the_enat k) xs)) 
                    (kfilter P (n+ (the_enat k)) (ldropn (the_enat k) xs))) =
            kfilter P n (ltake k xs)"
   by (metis assms enat_the_enat lfinite_ltake llength_eq_infty_conv_lfinite llength_ltake 
       ltake_all ltake_lappend1 min.strict_order_iff order_refl)
 show ?thesis using 1 2 by simp
qed

lemma lfilter_ldropn_llength:
 assumes "k < llength xs"
 shows " llength (lfilter P ( (ldropn k xs))) \<le> llength (lfilter P ( xs))"
using assms 
proof (induct k arbitrary: xs)
case 0
then show ?case by simp
next
case (Suc k)
then show ?case 
 proof (cases xs)
 case LNil
 then show ?thesis by simp
 next
 case (LCons x21 x22)
 then show ?thesis using Suc Suc_ile_eq by auto
  (meson Suc_ile_eq dual_order.trans ile_eSuc)
 qed
qed

lemma lfilter_nlength_imp:
 shows   " llength (lfilter (\<lambda>x. P x \<and> Q x) xs) \<le> llength (lfilter P xs)"
proof -
 have 0: "lfilter (\<lambda>x. P x \<and> Q x) xs = lfilter (\<lambda>x. Q x \<and> P x) xs" 
   by meson
 have 1: "lfilter (\<lambda>x. Q x \<and> P x) xs = lfilter Q (lfilter P xs)"
   using lfilter_lfilter[of Q P xs] by auto
 have 2: "llength (lfilter Q (lfilter P xs)) \<le> llength (lfilter P xs)"
   using llength_lfilter_ile by blast 
 show ?thesis by (simp add: 1 2 0)
qed

lemma lnths_kfilter_lfilter:
 " lnths xs (lset(kfilter P 0 xs)) = lfilter P xs"
using lfilter_conv_lnths[of P xs]  kfilter_lset[of P 0 xs]
by simp 


subsection \<open>Transfer rules\<close>

context includes lifting_syntax
begin

lemma lbutlast_transfer [transfer_rule]:
 " (llist_all2 A ===> llist_all2 A) lbutlast lbutlast"
by (auto simp add: rel_fun_def lbutlast_conv_ltake llist_all2_llengthD llist_all2_ltakeI)

lemma lleast_transfer [transfer_rule]:
 " ((A ===> (=)) ===> llist_all2 A ===> (=)) lleast lleast "
unfolding lleast_def[abs_def] 
by (auto simp add: rel_fun_def)
   (metis (full_types, opaque_lifting) llist_all2_conv_all_lnth)

end

end
