(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Finite and Infinite ITL theorems using Weak Chop\<close>

theory Theorems
 imports
   ITL
begin
(*
 sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                        isar_proofs=false,
                    provers="z3 vampire cvc4 e spass " ]
*)


text \<open>
We give the proofs of a list of Finite and Infinite ITL theorems. These proofs and theorems are from 
\cite{Ben:ImpReaPro} but adapted for infinite and finite intervals.
\<close>

subsection \<open>Propositional reasoning\<close>

text \<open>
This is a list of propositional logic theorems used in the proofs of the ITL theorems. 
\<close>

lemma IfThenElseImp:
 "\<turnstile> (if\<^sub>i  g  then  f  else f1) \<longrightarrow> (( g \<longrightarrow> f) \<and> (\<not>g \<longrightarrow> f1)) "
by (simp add: itl_def Valid_def) 

lemma Prop01:
 assumes "\<turnstile> f \<longrightarrow> \<not> g \<or> h"
 shows   "\<turnstile> g \<and> f \<longrightarrow> h"
using assms by auto

lemma Prop02:
 assumes "\<turnstile> f \<longrightarrow> g"
         "\<turnstile> f1 \<longrightarrow> g"
  shows  "\<turnstile> f \<or> f1  \<longrightarrow> g"
using assms by fastforce

lemma Prop03:
  assumes "\<turnstile> f = (g \<or> h) "
  shows   "\<turnstile> h \<longrightarrow> f"
using assms by auto  

lemma Prop04:
 assumes "\<turnstile> f = h"
         "\<turnstile> f = h1"
 shows   "\<turnstile> h1 = h"
using assms using int_eq by auto

lemma Prop05:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows "\<turnstile> f \<longrightarrow> h \<or> g "
using assms by auto

lemma Prop06:
 assumes "\<turnstile> f = (g \<or> h) "
         "\<turnstile> h = h1"
 shows "\<turnstile> f = (g \<or> h1) "
using assms by fastforce

lemma Prop07:
 assumes "\<turnstile> f \<longrightarrow> g \<or> h "
 shows   "\<turnstile> f \<and> \<not> g \<longrightarrow> h"
using assms by auto

lemma Prop08:
 assumes "\<turnstile> f \<longrightarrow> g \<or> h"
         "\<turnstile> h \<longrightarrow> h1"
 shows   "\<turnstile> f \<longrightarrow> g \<or> h1"
using assms by fastforce

lemma Prop09:
 assumes "\<turnstile> f \<and> g \<longrightarrow> h"
 shows   "\<turnstile> f \<longrightarrow> (g \<longrightarrow> h)"
using assms by auto

lemma Prop10:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> f = (f \<and> g)"
using assms by auto 

lemma Prop11:
 "(\<turnstile> f = f1) = ( (\<turnstile> f \<longrightarrow> f1) \<and>  (\<turnstile> f1 \<longrightarrow> f) )"
by (auto simp: Valid_def)

lemma Prop12:
 "(\<turnstile> f  \<longrightarrow> ( f1 \<and> f2)) = ( (\<turnstile> f \<longrightarrow> f1) \<and> (\<turnstile> f \<longrightarrow> f2))"
by (auto simp: Valid_def)

lemma Prop13:
 assumes "\<turnstile> f \<longrightarrow> g \<or> h"
 shows   "\<turnstile> f \<and> \<not>h \<longrightarrow> g"
using assms by (auto simp: Valid_def)


subsection \<open>State formulas\<close>

text \<open>
The \<open>init\<close> operator denotes state formulas, i.e., ITL formula that only constrain
the first state of an interval. 
\<close>

lemma Initprop :
  "\<turnstile> ((init f) \<and> (init g)) = init(f \<and> g)"
  "\<turnstile> (\<not> (init f)) = init ( \<not> f )"
  "\<turnstile> ((init f) \<or> (init g)) = init (f \<or> g)" 
  "\<turnstile> init #True" 
by (auto simp: itl_defs)

lemma Finprop :
 "\<turnstile> ((#True;(f \<and> empty)) \<and> (#True;(g \<and> empty))) = (#True;((f \<and> g) \<and> empty)) "
 "\<turnstile> ((#True;(f \<and> empty)) \<or> (#True;(g \<and> empty))) = (#True;((f \<or> g) \<and> empty)) "
 "\<turnstile> (#True;((#True) \<and> empty))"
 "\<turnstile> finite \<longrightarrow> (\<not> (#True;(f \<and> empty))) = (#True;(\<not>f \<and> empty))"
 "\<turnstile> (\<not> (#True;(f \<and> empty))) = ( (#True;(\<not>f \<and> empty)) \<and> finite)"
using nfinite_nlength_enat 
by (auto simp: finalt_defs finite_defs, 
    auto simp add: itl_defs nfinite_nlength_enat, force)
  
subsection \<open>finite and inf properties\<close>

lemma EmptyImpFinite:
 "\<turnstile> empty \<longrightarrow> finite"
using nlength_eq_enat_nfiniteD by (auto simp add: itl_defs)

lemma SkipChopFiniteImpFinite:
 "\<turnstile> skip;finite \<longrightarrow> finite"
using nfinite_ndropn nlength_eq_enat_nfiniteD
by (simp add: Valid_def itl_defs,force  )

lemma FiniteChopSkipImpFinite:
 "\<turnstile> finite;skip \<longrightarrow> finite"
using nlength_eq_enat_nfiniteD 
by (simp add: Valid_def itl_defs, force)

lemma FiniteChopSkipImpMore:
 "\<turnstile> finite;skip \<longrightarrow> more"
using nlength_eq_enat_nfiniteD 
by (simp add: Valid_def itl_defs, force)

lemma FiniteAndMoreImpFiniteChopSkip:
 "\<turnstile> finite \<and> more \<longrightarrow> finite;skip"
proof (auto simp add: Valid_def itl_defs)
 fix w :: "'a nellist"
 assume a1: "nlength w \<noteq> enat 0"
 assume a2: "nfinite w"
 have f3: "\<forall>n na. \<not> (n::nat) \<le> na \<or> n - na = 0"
   using diff_is_0_eq' by satx
 obtain nn :: "'a nellist \<Rightarrow> nat" where
   f4: "nlength w = enat (nn w)"
   using a2 by (meson nfinite_nlength_enat)
 have "\<forall>n na. \<not> (n::nat) \<le> na \<or> na - (na - n) = n"
   using diff_diff_cancel by satx
 then show "\<exists>n. enat n \<le> nlength w \<and> nlength w - enat n = enat (Suc 0)"
   using f4 f3 a1
   by (metis Suc_diff_eq_diff_pred Suc_diff_le diff_le_self enat_ord_simps(1) gr_zeroI 
      idiff_enat_enat order_refl)
qed
 
lemma FiniteChopSkipEqvFiniteAndMore:
 "\<turnstile> finite;skip = (finite \<and> more)"
 by (simp add: FiniteAndMoreImpFiniteChopSkip FiniteChopSkipImpFinite FiniteChopSkipImpMore 
     Prop12 int_iffI)
  
lemma FiniteChopSkipEqvSkipChopFinite:
 "\<turnstile> finite;skip = skip;finite"
by (auto simp add: Valid_def itl_defs)
   (metis enat.distinct(1) enat_add_sub_same enat_le_plus_same(2) le_iff_add ,
    metis eSuc_enat enat.simps(3) enat_add_sub_same idiff_enat_0_right iless_Suc_eq le_zero_eq 
    less_eqE min.orderE  nlength_eq_enat_nfiniteD not_le one_eSuc plus_1_eSuc(2),
    metis add.commute enat.simps(3) enat_add_sub_same idiff_enat_enat le_iff_add 
    nfinite_nlength_enat)

lemma FiniteAndEmptyEqvEmpty:
 "\<turnstile> (finite \<and> empty) = empty"
by (auto simp add: Valid_def itl_defs nlength_eq_enat_nfiniteD 
    zero_enat_def)

lemma FiniteChopFiniteEqvFinite:
 "\<turnstile> finite;finite = finite"
by (auto simp add: Valid_def itl_defs) (metis  zero_enat_def zero_le) 

lemma InfChopInfEqvInf:
 "\<turnstile> inf;inf = inf"
by (simp add: Valid_def itl_defs)

lemma InfChopFiniteEqvInf:
 "\<turnstile> inf;finite = inf"
by (simp add: Valid_def itl_defs)

lemma FiniteChopInfEqvInf:
 "\<turnstile> finite;inf = inf"
by (auto simp add: Valid_def itl_defs) (metis  zero_enat_def zero_le)

lemma InfEqvNotFinite:
 "\<turnstile> inf = (\<not> finite)"
by (simp add: Valid_def itl_defs)

lemma FiniteEqvNotInf:
 "\<turnstile> finite = (\<not> inf)"
by (simp add: Valid_def itl_defs)

lemma ChopTrueAndFiniteEqvAndFiniteChopFinite:
 "\<turnstile> ((f;#True) \<and> finite) = (f \<and> finite);finite"
by (auto simp add: Valid_def itl_defs)

lemma TrueChopAndFiniteEqvAndFiniteChopFinite:
 "\<turnstile> ((#True;f) \<and> finite) = finite;(f \<and> finite)"
by (auto simp add: Valid_def itl_defs)

lemma FiniteChopMoreEqvMore:
 "\<turnstile> finite;more = more"
by (auto simp add: Valid_def itl_defs)
   (metis idiff_0_right zero_enat_def zero_le)

lemma ChopAndFiniteDist:
 "\<turnstile> ((f;g) \<and> finite) = (f \<and> finite);(g \<and> finite)"
by (auto simp add: Valid_def itl_defs)

lemma FiniteOrInfinite:
 "\<turnstile> finite \<or> inf"
by (simp add: Valid_def itl_defs)

lemma FiniteImpAnd:
 assumes "\<turnstile> finite \<longrightarrow> f = g"
 shows "\<turnstile> (f \<and> finite) = (g \<and> finite)"
using assms by (auto simp add: Valid_def itl_defs)

lemma FmoreEqvSkipChopFinite:
 "\<turnstile> fmore = skip;finite"
by (metis FiniteChopSkipEqvFiniteAndMore FiniteChopSkipEqvSkipChopFinite 
          fmore_d_def inteq_reflection lift_and_com)
 
lemma FiniteImp:
 "\<turnstile> (f \<and> finite \<longrightarrow> g) = (f \<and> finite \<longrightarrow> g \<and> finite)"
by (simp add: itl_defs Valid_def)

lemma ChopAndInf:
 "\<turnstile> ((f;g) \<and> inf) = (f;(g \<and> inf))"
by (auto simp add: Valid_def itl_defs)

lemma ChopAndInfB:
 "\<turnstile> ((f;g) \<and> inf) = ((f \<and> inf) \<or> (f \<and> finite);(g \<and> inf)  )"
by (auto simp add: Valid_def itl_defs)

lemma MoreAndInfEqvInf:
 "\<turnstile> (more \<and> inf) = inf"
by (metis ChopAndInf EmptyImpFinite FiniteChopMoreEqvMore InfEqvNotFinite Prop11 Prop12 empty_d_def
          finite_d_def int_simps(32) inteq_reflection)

lemma AndInfChopAndInfEqvAndInf:
 "\<turnstile> (f \<and> inf);(f \<and> inf) = (f \<and> inf)"
by (auto simp add: Valid_def itl_defs)

lemma AndInfChopEqvAndInf:
 "\<turnstile> (f \<and> inf);g = (f \<and> inf)"
by (auto simp add: Valid_def itl_defs)

lemma AndMoreAndInfEqvAndInf:
 "\<turnstile> ((f \<and> more) \<and> inf) = (f \<and> inf)"
by (auto simp add: Valid_def itl_defs nlength_eq_enat_nfiniteD zero_enat_def)

lemma AndMoreAndFiniteEqvAndFmore:
 "\<turnstile> ((f \<and> more) \<and> finite) = (f \<and> fmore)"
by (auto simp add: Valid_def itl_defs)

lemma NotFmoreAndEmpty:
 "\<turnstile> \<not> (empty \<and> fmore)"
by (auto simp add: fmore_d_def empty_d_def)

lemma NotFmoreAndInf:
 "\<turnstile> \<not> ((f \<and> inf) \<and> fmore)"
by (auto simp add: fmore_d_def finite_d_def infinite_d_def)

lemma FmoreChopAnd:
 "\<turnstile> (((f \<and> more);g) \<and> fmore) = ((f \<and> fmore);(g \<and> finite))"
by (auto simp add: Valid_def itl_defs)

lemma NotEmptyAndInf:
 "\<turnstile> \<not>(empty \<and> inf)"
by (auto simp add: Valid_def itl_defs nlength_eq_enat_nfiniteD zero_enat_def)

lemma OrFiniteInf:
 "\<turnstile>  f = (( f \<and> finite) \<or> ( f \<and> inf))"
by (auto simp add: Valid_def itl_defs)

lemma AndInfEqvChopFalse:
 "\<turnstile> (f \<and> inf) = f;#False"
by (auto simp add: Valid_def itl_defs)


subsection \<open>Basic Theorems\<close>

lemma BiChopImpChop :
    " \<turnstile> bi (f \<longrightarrow> f1) \<longrightarrow> f;g \<longrightarrow> f1;g "
 proof -
 have  1: "\<turnstile> g \<longrightarrow> g" by auto
 hence 2: "\<turnstile> \<box> ( g \<longrightarrow> g) " by (rule BoxGen)
 have  3: "\<turnstile> bi ( f \<longrightarrow> f1) \<and> \<box>(g \<longrightarrow> g) \<longrightarrow> f;g \<longrightarrow> f1;g " by (rule BiBoxChopImpChop)
 from 2 3 show ?thesis by fastforce
qed 

lemma AndChopA:
    " \<turnstile> (f \<and> f1);g \<longrightarrow> f;g "
proof -
 have  1: "\<turnstile> f \<and> f1 \<longrightarrow> f" by auto
 hence 2: "\<turnstile> bi (f \<and> f1 \<longrightarrow> f)" by (rule BiGen)
 have  3: "\<turnstile> bi (f \<and> f1 \<longrightarrow> f) \<longrightarrow> (f  \<and> f1);g \<longrightarrow> f;g" by (rule BiChopImpChop)
 from 2 3 show ?thesis using MP by blast
qed

lemma AndChopB:
    "\<turnstile> (f \<and> f1);g \<longrightarrow> f1;g "
proof -
 have  1: "\<turnstile> f \<and> f1 \<longrightarrow> f1" by auto
 hence 2: "\<turnstile>  bi (f \<and> f1 \<longrightarrow> f1)" by (rule BiGen)
 have  3: "\<turnstile>  bi (f \<and> f1 \<longrightarrow> f1) \<longrightarrow> (f \<and> f1);g \<longrightarrow> f1;g" by (rule BiChopImpChop)
 from 2 3 show ?thesis using MP by blast
qed

lemma NextChop:
  "\<turnstile> (\<circle> f);g  = \<circle>(f;g)"
proof -
 have 1: "\<turnstile> skip;(f;g)  = (skip;f);g" by (rule ChopAssoc)
 show ?thesis by (metis "1" int_eq next_d_def)
qed

lemma BoxChopImpChop :
    " \<turnstile>  \<box> (g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f;g1 "
 proof -
 have  1: "\<turnstile> g \<longrightarrow> g" by auto
 hence 2: "\<turnstile> bi ( g \<longrightarrow> g) " by (rule BiGen)
 have  3: "\<turnstile> bi ( f \<longrightarrow> f) \<and> \<box>(g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f;g1 " by (rule BiBoxChopImpChop)
 from 2 3 show ?thesis by fastforce
qed 

lemma LeftChopImpChop:
 assumes "\<turnstile> f\<longrightarrow> f1" 
 shows   "\<turnstile> f;g \<longrightarrow> f1;g"
proof -
 have  1: "\<turnstile> f\<longrightarrow> f1" using assms by auto
 hence 2: "\<turnstile> bi (f \<longrightarrow> f1)" by (rule BiGen)
 have  3: "\<turnstile> bi (f \<longrightarrow> f1) \<longrightarrow> f;g \<longrightarrow> f1;g" by (rule BiChopImpChop)
 from 2 3 show ?thesis using MP by blast
qed

lemma RightChopImpChop:
 assumes "\<turnstile> g \<longrightarrow> g1" 
 shows   "\<turnstile> f;g \<longrightarrow> f;g1"
proof -
 have  1: "\<turnstile> g \<longrightarrow> g1" using assms by auto
 hence 2: "\<turnstile> \<box> (g \<longrightarrow> g1)" by (rule BoxGen)
 have  3: "\<turnstile> \<box> (g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f;g1" by (rule BoxChopImpChop)
 from 2 3 show ?thesis using MP by blast
qed
 
lemma RightChopEqvChop:
 assumes "\<turnstile> g = g1"
 shows   "\<turnstile> (f;g) = (f;g1)"
using assms  RightChopImpChop[of "g" "g1" "f"] RightChopImpChop[of "g1" "g" "f"]
by fastforce

lemma InfRightChopEqvChop:
assumes "\<turnstile> inf \<longrightarrow> g = g1" 
shows   "\<turnstile> inf \<longrightarrow> (f;g) = (f;g1)"
using assms 
by (auto simp add: Valid_def itl_defs)

lemma ChopOrEqv:
 " \<turnstile> f;(g \<or> g1) = (f;g \<or> f;g1) "
proof -
 have  1: "\<turnstile> g \<longrightarrow> g \<or> g1" by auto
 hence 2: "\<turnstile> f;g \<longrightarrow> f;(g \<or> g1)" by (rule RightChopImpChop)
 have  3: "\<turnstile> g1 \<longrightarrow> g \<or> g1" by auto 
 hence 4: "\<turnstile> f;g1 \<longrightarrow> f;(g \<or> g1)" by (rule RightChopImpChop)
 from 2 4 show ?thesis by (meson ChopOrImp Prop02 Prop11) 
qed

lemma OrChopEqv:
 " \<turnstile> (f \<or> f1);g = (f;g \<or> f1;g) "
proof -
 have  1: "\<turnstile> f \<longrightarrow> f \<or> f1" by auto
 hence 2: "\<turnstile> f;g \<longrightarrow> (f\<or> f1);g" by (rule LeftChopImpChop)
 have  3: "\<turnstile> f1 \<longrightarrow> f \<or> f1" by auto 
 hence 4: "\<turnstile> f1;g \<longrightarrow> (f\<or> f1);g" by (rule LeftChopImpChop)
 from 2 4 show ?thesis 
 by (meson OrChopImp int_iffI Prop02)
qed

lemma OrChopImpRule:
 assumes "\<turnstile> f \<longrightarrow> f1 \<or> f2"
 shows   "\<turnstile> f;g \<longrightarrow> (f1;g) \<or> (f2;g)"
proof -
 have  1: "\<turnstile> f \<longrightarrow> f1 \<or> f2" using assms by auto
 hence 2: "\<turnstile> f;g \<longrightarrow> (f1 \<or> f2);g" by (rule LeftChopImpChop)
 have  3: "\<turnstile> (f1 \<or> f2); g = (f1;g \<or> f2;g)" by (rule OrChopEqv) 
 from 2 3 show ?thesis by fastforce
qed

lemma  LeftChopEqvChop:
 assumes "\<turnstile> f = f1"
 shows   "\<turnstile> f;g = (f1;g)"
proof -
 have  1: "\<turnstile> f = f1" using assms by auto
 hence 2: "\<turnstile> f \<longrightarrow> f1" by auto
 hence 3: "\<turnstile> f;g \<longrightarrow> f1;g" by (rule LeftChopImpChop)
 have "\<turnstile> f1 \<longrightarrow> f" using "1" by auto
 hence 4: "\<turnstile> f1;g \<longrightarrow> f;g" by (rule LeftChopImpChop)
 from 3 4 show ?thesis by (simp add: int_iffI)
qed

lemma OrChopEqvRule:
 assumes "\<turnstile> f = (f1 \<or> f2)"
 shows   "\<turnstile> f;g = ((f1;g) \<or> (f2;g))"
proof -
 have  1: "\<turnstile> f = (f1 \<or> f2)" using assms by auto
 hence 2: "\<turnstile> f;g = ((f1 \<or> f2);g)" by (rule LeftChopEqvChop) 
 have  3: "\<turnstile> (f1 \<or> f2);g = (f1;g \<or> f2;g)" by (rule OrChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma NextImpNext:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> \<circle> f \<longrightarrow> \<circle> g"
proof -
 have  1: "\<turnstile> f \<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> \<box> (f \<longrightarrow> g)" by (rule BoxGen)
 have  3: "\<turnstile> \<box> (f \<longrightarrow> g) \<longrightarrow> (skip;f) \<longrightarrow> (skip;g)" by (rule BoxChopImpChop)
 have  4: "\<turnstile>(skip;f) \<longrightarrow> (skip;g)" by (metis 2 3 MP)
 from 4 show ?thesis by (metis next_d_def)
qed

lemma ChopOrImpRule:
 assumes " \<turnstile> g \<longrightarrow> g1 \<or> g2"
 shows   " \<turnstile> f;g \<longrightarrow> (f;g1) \<or> (f;g2)"
proof -
 have  1: "\<turnstile> g \<longrightarrow> g1 \<or> g2" using assms by auto
 hence 2: "\<turnstile> f;g \<longrightarrow> f;(g1 \<or> g2)" by (rule RightChopImpChop)
 have  3: "\<turnstile> f;(g1 \<or> g2) = (f;g1 \<or> f;g2)" by (rule ChopOrEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma NextImpDist:
 "\<turnstile> \<circle> (f \<longrightarrow> g) \<longrightarrow> \<circle> f \<longrightarrow> \<circle> g"
proof -
 have  1: "\<turnstile> (\<not> (f \<longrightarrow> g)) = (f \<and> \<not> g)" by auto
 hence 2: "\<turnstile> skip;(\<not> (f \<longrightarrow> g)) = skip;(f \<and> \<not> g)" by (rule RightChopEqvChop)
 have  3: "\<turnstile> f \<longrightarrow> g \<or> (f \<and> \<not> g)" by auto
 hence 4: "\<turnstile> skip;f \<longrightarrow> (skip;g) \<or> (skip;(f \<and> \<not> g))" by (rule ChopOrImpRule)
 hence 5: "\<turnstile> \<not> (skip;(f \<and> \<not> g)) \<longrightarrow> (skip;f) \<longrightarrow> (skip;g)" by auto
 have  6: "\<turnstile> \<not> (skip;(\<not>(f \<longrightarrow> g))) \<longrightarrow> (skip;f) \<longrightarrow> (skip;g)" using "2" "5" by fastforce
 hence 7: "\<turnstile> \<not> (\<circle>(\<not>(f \<longrightarrow> g))) \<longrightarrow> (\<circle> f) \<longrightarrow> (\<circle> g)" by (simp add: next_d_def)
 have  8: "\<turnstile> \<circle>(f \<longrightarrow> g) \<longrightarrow> \<not> (\<circle>(\<not>(f \<longrightarrow> g)))" by (rule NextImpNotNextNot)
 from 7 8 show ?thesis using lift_imp_trans by blast
qed

lemma FiniteChopImpDiamond:
 "\<turnstile> (f \<and> finite);g  \<longrightarrow> \<diamond> g"
proof -
 have  1: "\<turnstile> f \<and> finite \<longrightarrow> finite" by auto
 hence 2: "\<turnstile> (f \<and> finite);g \<longrightarrow> finite;g" by (rule LeftChopImpChop)
 from 2 show ?thesis by (simp add: sometimes_d_def)
qed


lemma NowImpDiamond:
 "\<turnstile> f \<longrightarrow> \<diamond> f"
proof -
 have  1: "\<turnstile> empty;f = f" by (rule EmptyChop)
 have  2: "\<turnstile> empty \<longrightarrow> finite" by (rule EmptyImpFinite)
 hence 3: "\<turnstile> empty;f \<longrightarrow> finite;f" by (rule LeftChopImpChop)
 have  4: " \<turnstile> f \<longrightarrow> finite;f" using "1" "3" by fastforce
 from 4 show ?thesis by (simp add: sometimes_d_def)
qed

lemma BoxElim:
 "\<turnstile> \<box> f \<longrightarrow> f "
proof -
  have  1: "\<turnstile> \<not> f \<longrightarrow> \<diamond> (\<not> f)" by (rule NowImpDiamond)
  hence 2: "\<turnstile> \<not> (\<diamond> (\<not> f)) \<longrightarrow> f " by auto
  from 2 show ?thesis by (metis always_d_def)
qed

lemma NextDiamondImpDiamond:
 "\<turnstile> \<circle> (\<diamond> f) \<longrightarrow> \<diamond> f"
proof -
 have  1: "\<turnstile> skip;(finite;f) = ((skip;finite);f)" by (rule ChopAssoc)
 hence 2: "\<turnstile> (skip;finite);f = skip;(finite;f)" by auto 
 hence 3: "\<turnstile> (skip;finite);f = \<circle>(\<diamond>f)" by (simp add: next_d_def sometimes_d_def)
 have  4: "\<turnstile> (skip;finite);f \<longrightarrow> \<diamond> f" 
 by (simp add: SkipChopFiniteImpFinite LeftChopImpChop sometimes_d_def)
 from 3 4 show ?thesis by fastforce
qed

lemma BoxImpNowAndWeakNext:
 " \<turnstile> \<box> f \<longrightarrow> (f \<and> wnext ( \<box> f) )"
proof -
  have  1: "\<turnstile> \<not> f \<longrightarrow> \<diamond> (\<not> f) " by (rule NowImpDiamond)
  hence 2: "\<turnstile> \<not> (\<diamond> (\<not> f)) \<longrightarrow> f" by auto
  hence 3: "\<turnstile> \<box> f \<longrightarrow> f" by (metis always_d_def)
  have  4: "\<turnstile> \<circle> ( \<diamond> (\<not> f)) \<longrightarrow> \<diamond> ( \<not> f ) " by (rule NextDiamondImpDiamond)
  have  5: "\<turnstile> \<not> \<not> (\<diamond> (\<not> f)) \<longrightarrow> \<diamond>( \<not> f ) " by auto
  hence 6: "\<turnstile> \<circle> ( \<not> \<not> (\<diamond> (\<not> f)) )  \<longrightarrow> \<circle> (\<diamond>( \<not> f )) " by (rule NextImpNext)
  have  7: "\<turnstile> \<circle> ( \<not> \<not> (\<diamond> (\<not> f)) )  \<longrightarrow> \<diamond> ( \<not> f )" using "4" "6" by auto  
  hence 8: "\<turnstile> \<circle> ( \<not>( \<box> f)) \<longrightarrow> \<diamond> ( \<not> f ) " by (simp add: always_d_def)
  hence 9: "\<turnstile> \<not> (\<diamond> ( \<not> f ))  \<longrightarrow> \<not> ( \<circle> ( \<not>( \<box> f)))" by auto
  hence 10:"\<turnstile> \<box>f \<longrightarrow> wnext ( \<box> f )" by (simp add: always_d_def wnext_d_def) 
  from 3 10 show ?thesis by fastforce
qed

lemma BoxImpBoxRule:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> \<box> f\<longrightarrow> \<box> g" 
proof - 
 have  1: "\<turnstile> f \<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> \<not> g \<longrightarrow>  \<not> f" by auto
 hence 3: "\<turnstile> \<box>(\<not> g \<longrightarrow>  \<not> f)" by (rule BoxGen)
 have  4: "\<turnstile>  \<box>(\<not> g \<longrightarrow>  \<not> f) \<longrightarrow> (finite;(\<not>g)) \<longrightarrow> (finite;(\<not>f))" by (rule BoxChopImpChop)
 have  5: "\<turnstile> (finite;(\<not>g)) \<longrightarrow> (finite;(\<not>f))" using "3" "4" MP by blast
 hence 6: "\<turnstile> \<diamond> (\<not>g)  \<longrightarrow> \<diamond>(\<not>f)" by (simp add: sometimes_d_def)
 hence 7: "\<turnstile>  \<not> (\<diamond>(\<not>f) ) \<longrightarrow> \<not>( \<diamond> (\<not>g))" by auto
 from 7 show ?thesis by (simp add: always_d_def)
qed

lemma BoxImpDist:
 " \<turnstile>  \<box>(f \<longrightarrow> g) \<longrightarrow> \<box> f \<longrightarrow> \<box> g"
proof -
 have  1: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> (\<not>  g \<longrightarrow> \<not>  f) " 
   by auto
 hence 2: "\<turnstile> \<box>(f \<longrightarrow> g) \<longrightarrow> \<box>(\<not>  g \<longrightarrow> \<not>  f) " 
   by (rule BoxImpBoxRule)
 have  3: "\<turnstile> \<box>((\<not>  g) \<longrightarrow> \<not>  f) \<longrightarrow>  (finite; (\<not>  g)) \<longrightarrow> (finite; (\<not>  f))" 
   by (rule BoxChopImpChop)
 have  4: "\<turnstile> \<box>(f \<longrightarrow> g) \<longrightarrow>   (finite; (\<not>  g)) \<longrightarrow> (finite; (\<not>  f))" 
   using "2" "3" lift_imp_trans by blast   
 hence 5: "\<turnstile> \<box>(f \<longrightarrow> g) \<longrightarrow>   \<diamond>(\<not>  g) \<longrightarrow> \<diamond>(\<not>  f) " 
   by (simp add: sometimes_d_def)
 hence 6: "\<turnstile> \<box>(f \<longrightarrow> g) \<longrightarrow>   \<not>( \<diamond>(\<not>  f)) \<longrightarrow> \<not>( \<diamond>(\<not>  g)) " 
   by auto
 from 6 show ?thesis by (simp add: always_d_def)
qed

lemma DiamondEmptyEqvFinite:
 " \<turnstile>   \<diamond> empty = finite"
proof -
 have 1: "\<turnstile> finite;  empty  = finite " by (rule ChopEmpty)
 from 1 show ?thesis by (simp add: sometimes_d_def)
qed

lemma NextEqvNext:
 assumes "\<turnstile>   f= g"
 shows   "\<turnstile> \<circle> f = \<circle> g"
proof -
 have  1: "\<turnstile> f= g" using assms by auto
 hence 2: "\<turnstile> skip;f = skip;g"  by (rule RightChopEqvChop) 
 from 1 show ?thesis by (metis "2" next_d_def)
qed
 
lemma NextAndNextImpNextRule:
 assumes "\<turnstile> (f \<and> g) \<longrightarrow> h " 
 shows   "\<turnstile> (\<circle> f \<and> \<circle> g) \<longrightarrow> \<circle> h "
using assms 
by  (simp add: Valid_def itl_defs)

lemma NextAndNextEqvNextRule:
 assumes "\<turnstile> (f \<and> g) = h "
 shows   "\<turnstile> (\<circle> f \<and> \<circle> g) = \<circle> h "
using assms
by (simp add: NextAndNextImpNextRule NextImpNext Prop11 Prop12) 

lemma WeakNextEqvWeakNext:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile> wnext  f = wnext  g "
using assms using inteq_reflection by force

lemma DiamondImpDiamond:
 assumes "\<turnstile> f \<longrightarrow> g "
 shows   "\<turnstile> \<diamond> f \<longrightarrow> \<diamond> g "
using assms by (simp add: RightChopImpChop sometimes_d_def)

lemma DiamondEqvDiamond:
 assumes "\<turnstile> f = g " 
 shows   "\<turnstile> \<diamond> f = \<diamond> g "
using assms using int_eq by force

lemma BoxEqvBox:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile> \<box> f = \<box> g "
using assms using inteq_reflection by force

lemma BoxAndBoxImpBoxRule:
 assumes "\<turnstile> f \<and> g \<longrightarrow> h "
 shows   "\<turnstile> \<box> f \<and> \<box> g \<longrightarrow> \<box> h "
using assms by (auto simp: itl_defs Valid_def)

lemma BoxAndBoxEqvBoxRule:
 assumes "\<turnstile> (f \<and> g) = h "
 shows   "\<turnstile> (\<box> f \<and> \<box> g) = \<box> h " 
using assms BoxAndBoxImpBoxRule BoxImpBoxRule by (metis int_iffD1 int_iffD2 int_iffI Prop12)

lemma ImpBoxRule:
 assumes "\<turnstile>   f \<longrightarrow> g "
 shows   "\<turnstile> \<box> f \<longrightarrow> \<box> g "
using assms by (simp add: BoxImpBoxRule)

lemma WnextEqvEmptyOrNext:
 "\<turnstile> wnext f = (empty \<or> \<circle> f)"
by (auto simp: Valid_def itl_defs)

lemma BoxIntro:
 assumes "\<turnstile>  f \<longrightarrow> g" 
         "\<turnstile>  more  \<and> f \<longrightarrow> \<circle> f "
 shows   "\<turnstile> f \<longrightarrow> \<box> g "
proof -
 have   1: "\<turnstile>  more  \<and> f \<longrightarrow>\<circle> f " 
   using assms by auto
 hence  2: "\<turnstile> f\<longrightarrow> (empty \<or> \<circle>f)"
   unfolding empty_d_def by fastforce 
 hence  3: "\<turnstile> f \<longrightarrow> wnext f"
   by (metis WnextEqvEmptyOrNext inteq_reflection) 
 hence  4: "\<turnstile> \<box>(f \<longrightarrow> wnext f)" 
   by (rule BoxGen)
 have   5: "\<turnstile> (\<box> (f \<longrightarrow> wnext f)) \<and> f \<longrightarrow> \<box> f  " 
   by (rule BoxInduct)
 hence  6: "\<turnstile> (\<box> (f \<longrightarrow> wnext f)) \<longrightarrow> (f \<longrightarrow> \<box>f)" 
   by fastforce
 have   7: "\<turnstile> f \<longrightarrow> \<box>f " 
   using "4" "6" MP by blast
 have   8: "\<turnstile> \<box>f \<longrightarrow> f" 
   by (rule BoxElim)
 have   9: "\<turnstile> f = \<box> f" 
   using "7" "8" by fastforce
 have  10: "\<turnstile> f \<longrightarrow> g" 
   using assms by auto
 hence 11: "\<turnstile> \<box>f \<longrightarrow> \<box> g" 
   by (rule ImpBoxRule)
 from 7 9 11 show ?thesis  
   using lift_imp_trans by blast 
qed

lemma NextLoop:
 assumes "\<turnstile> f  \<longrightarrow> \<circle> f" 
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f "
proof -
 have   1: "\<turnstile> f\<longrightarrow> \<circle> f " 
   using assms by auto
 hence  2: "\<turnstile> f \<longrightarrow> (more \<and> wnext f)"
   unfolding more_d_def wnext_d_def
   by (metis NextImpNext NextImpNotNextNot Prop05 Prop10 Prop12 int_iffD1 int_simps(29) lift_imp_trans)
 hence  3: "\<turnstile> f \<longrightarrow> wnext f " 
   by auto
 hence  4: "\<turnstile> \<box>(f \<longrightarrow> wnext f)" 
   by (rule BoxGen)
 have   5: "\<turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f  " 
   by (rule BoxInduct)
 hence  6: "\<turnstile> \<box> (f \<longrightarrow> wnext f) \<longrightarrow> (f \<longrightarrow> \<box>f)" 
   by fastforce
 have   7: "\<turnstile> f \<longrightarrow> \<box>f "  
   using "4" "6" MP by blast
 have   8: "\<turnstile> \<box>f \<longrightarrow> f" 
   by (rule BoxElim)
 have   9: "\<turnstile> f = \<box> f" 
   using "7" "8" by fastforce
 have  10: "\<turnstile> f \<longrightarrow> more" 
   using "2" by auto
 hence 11: "\<turnstile> \<box> f \<longrightarrow> \<box> more" 
   by (rule ImpBoxRule)
 have  12: "\<turnstile> finite = (\<not>(\<box> more)) "
   by (metis DiamondEmptyEqvFinite InfEqvNotFinite always_d_def empty_d_def finite_d_def int_eq) 
 from 7 9 11 12 show ?thesis 
   by fastforce
qed

lemma NotEmptyAndNext:
 "\<turnstile> \<not>(empty \<and> \<circle> f)"
by (auto simp: Valid_def itl_defs) 

lemma BoxEqvAndWnextBox:
 "\<turnstile> \<box> f = (f \<and> wnext ( \<box> f))"
proof -
 have 1: "\<turnstile> \<box> f \<longrightarrow> f \<and> wnext ( \<box> f)" 
   using BoxImpNowAndWeakNext by blast
 have 2: "\<turnstile> f \<and> wnext ( \<box> f) \<longrightarrow> f" 
   by auto
 have 3: "\<turnstile> more \<and> (f \<and> wnext ( \<box> f) ) \<longrightarrow> \<circle> (f \<and> wnext ( \<box> f) )" 
   using "1" NextImpNext WnextEqvEmptyOrNext empty_d_def int_iffD1 
   by (metis Prop01 Prop05 Prop08)
 have 4: "\<turnstile> f \<and> wnext ( \<box> f) \<longrightarrow> \<box> f" 
   using "2" "3" BoxIntro by blast
 from 1 4 show ?thesis by fastforce
qed

lemma BoxEqvAndEmptyOrNextBox:
 "\<turnstile> \<box>f = (f \<and> (empty  \<or>  \<circle>(\<box> f)))"
using BoxEqvAndWnextBox WnextEqvEmptyOrNext by (metis int_eq)

lemma BoxEqvBoxBox:
 "\<turnstile> \<box>f = \<box> (\<box> f)"
using BoxGen BoxInduct 
by (metis BoxImpNowAndWeakNext MP int_iffI Prop09 Prop12)

lemma BoxBoxImpBox:
 "\<turnstile> \<box>(\<box>h) \<longrightarrow> \<box> h"
by (simp add: BoxElim)

lemma BoxImpBoxBox:
 "\<turnstile> \<box> h  \<longrightarrow>  \<box>(\<box>h)"
 by (simp add: BoxEqvBoxBox int_iffD1)

lemma DiamondIntroC:
 assumes "\<turnstile> f \<longrightarrow> \<circle> g"
 shows   "\<turnstile> f \<longrightarrow> \<diamond> g"
using assms 
by (metis (no_types, lifting) ChopAssoc FiniteChopSkipEqvSkipChopFinite NextChop 
          NextDiamondImpDiamond NowImpDiamond inteq_reflection lift_imp_trans next_d_def 
          sometimes_d_def)

lemma DiamondIntro:
 assumes "\<turnstile> (f \<and> \<not>  g) \<longrightarrow> \<circle> f "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> \<diamond> g "
proof -
 have   1: "\<turnstile> f \<and> \<not>  g \<longrightarrow> \<circle> f " 
   using assms by auto
 hence  2: "\<turnstile> f \<and> \<not>  g \<and> (\<box> (\<not> g))\<longrightarrow> (\<circle> f) \<and> (\<box> (\<not> g))" 
   by auto
 have   3: "\<turnstile> (\<box> (\<not> g)) \<longrightarrow> \<not>  g" 
   by (rule BoxElim)
 hence  4: "\<turnstile> \<box> (\<not> g) = ((\<box> (\<not> g)) \<and> \<not>  g)" 
   using BoxImpBoxBox BoxBoxImpBox by fastforce
 have   5: "\<turnstile> f \<and> (\<box> (\<not> g))\<longrightarrow> \<circle> f \<and> \<box> (\<not> g) " 
   using "2" "4"  by fastforce
 have   6: "\<turnstile> \<box> (\<not> g) = ((\<not> g) \<and> wnext(\<box> (\<not> g)))  "  
   using BoxEqvAndWnextBox by metis
 have   7: "\<turnstile> \<circle> f \<and> \<box> (\<not> g) \<longrightarrow>  \<circle> f \<and> wnext(\<box> (\<not> g)) " 
   using "6" by auto
 have   8: "\<turnstile> f \<and> (\<box> (\<not> g))\<longrightarrow> \<circle> f \<and> wnext(\<box> (\<not> g)) " 
   using "5" "7" using lift_imp_trans by blast
 hence  9: "\<turnstile> f \<and> (\<box> (\<not> g))\<longrightarrow> more \<and> wnext f \<and> wnext(\<box> (\<not> g))" 
   by (auto simp: Valid_def itl_defs)
 hence 10: "\<turnstile> f \<and> (\<box> (\<not> g))\<longrightarrow>wnext f \<and> wnext(\<box> (\<not> g)) " 
   by auto
 hence 11: "\<turnstile> f \<and> (\<box> (\<not> g))\<longrightarrow> wnext (f \<and> \<box> (\<not> g)) " 
   by (auto simp: Valid_def itl_defs)
 hence 12: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))\<longrightarrow> wnext (f \<and> \<box> (\<not> g))) " 
   by (rule BoxGen)
 have  13: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))\<longrightarrow> wnext (f \<and> \<box> (\<not> g))) \<and> f \<and> (\<box> (\<not> g)) \<longrightarrow> \<box>(f \<and> (\<box> (\<not> g))) " 
   by (rule BoxInduct)
 hence 14: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))\<longrightarrow> wnext (f \<and> \<box>(\<not> g))) \<longrightarrow> ((f \<and> (\<box>(\<not> g))) \<longrightarrow> \<box>(f \<and> (\<box> (\<not> g)))) " 
   by fastforce
 have  15: "\<turnstile> ((f \<and> (\<box> (\<not> g))) \<longrightarrow> \<box>(f \<and> (\<box> (\<not> g)))) " 
   using "12" "14" MP by blast
 have  16: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))) \<longrightarrow> (f \<and> (\<box> (\<not> g))) " 
   by (rule BoxElim)
 have  17: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))) = (f \<and> (\<box> (\<not> g)))" 
   using "16" "15" by fastforce
 have  18: "\<turnstile> (f \<and> (\<box> (\<not> g))) \<longrightarrow> more" 
   using "9" by auto
 hence 19: "\<turnstile> \<box>(f \<and> (\<box> (\<not> g))) \<longrightarrow> \<box> more" 
   by (rule ImpBoxRule)
 have  20: "\<turnstile> finite = (\<not>(\<box> more)) "
   by (metis DiamondEmptyEqvFinite InfEqvNotFinite always_d_def empty_d_def finite_d_def int_eq)   
 have  21: "\<turnstile> finite \<longrightarrow> \<not>(f \<and> (\<box> (\<not> g)))" 
   using "17" "19" "20" by fastforce
 hence 22: "\<turnstile> finite \<longrightarrow> \<not> f \<or> \<not> (\<box> (\<not> g)) " 
   by auto
 have  23: "\<turnstile> (\<not> (\<box> (\<not> g))) = \<diamond> g" 
   by (auto simp: always_d_def)
 from 22 23 show ?thesis by fastforce
qed
 
lemma DiamondIntroB:
 assumes "\<turnstile> (f \<and> \<not>  g) \<longrightarrow> \<circle> (f \<and> \<not> g) "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> \<diamond> g "
proof -
 have  1: "\<turnstile> (f \<and> \<not>  g) \<longrightarrow> \<circle> (f \<and> \<not>g) " using assms by auto
 hence 2: "\<turnstile> finite \<longrightarrow> \<not>(f \<and> \<not> g) " by (rule NextLoop)
 hence 3: "\<turnstile> f \<and> finite \<longrightarrow> g" by auto
 have  4: "\<turnstile> g \<longrightarrow> \<diamond> g" by (rule NowImpDiamond)
 from 3 4 show ?thesis using lift_imp_trans by blast
qed

lemma NextContra :
 assumes "\<turnstile> (f \<and> \<not>  g) \<longrightarrow> (\<circle> f \<and> \<not>( \<circle> g)) "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g "
proof -
 have  1: "\<turnstile> (f\<and> \<not>  g)\<longrightarrow> (\<circle> f \<and> \<not>( \<circle> g)) " using assms by auto
 hence 2: "\<turnstile> \<not>( f \<longrightarrow> g) \<longrightarrow> \<circle> ( \<not>(f \<longrightarrow> g)) " by (auto simp: itl_defs Valid_def)
 hence 3: "\<turnstile> finite \<longrightarrow> \<not> \<not>( f \<longrightarrow> g) " by (rule NextLoop)
 from 3 show ?thesis by auto
qed

lemma DiamondDiamondEqvDiamond:
 " \<turnstile> \<diamond>(\<diamond> f) = \<diamond> f"
proof -
 have  1: "\<turnstile> finite;finite = finite" by (simp add: FiniteChopFiniteEqvFinite)
 hence 2: "\<turnstile> (finite;finite);f = finite;f " using LeftChopEqvChop by blast
 have  3: "\<turnstile> (finite;finite);f = finite;(finite;f) " using ChopAssoc by fastforce
 from 2 3 show ?thesis by (metis inteq_reflection sometimes_d_def)
qed

lemma WeakNextDiamondInduct:
 assumes "\<turnstile> wnext  (\<diamond> f) \<longrightarrow> f "
 shows   "\<turnstile> finite \<longrightarrow> f "
proof -
 have  1: "\<turnstile> wnext  (\<diamond> f) \<longrightarrow> f " using assms by blast
 hence 2: "\<turnstile> \<not> f \<longrightarrow> \<not>( wnext  (\<diamond> f))" by fastforce  
 hence 3: "\<turnstile> \<not> f \<longrightarrow> \<circle>( \<not> (\<diamond> f)) " by (simp add: wnext_d_def)
 have  4: "\<turnstile> f \<longrightarrow> \<diamond> f" by (rule NowImpDiamond)
 hence 5: "\<turnstile> \<not>(\<diamond> f) \<longrightarrow> \<not> f  " by auto
 have  6: "\<turnstile> \<not>f \<longrightarrow> \<circle>( \<not>f)" using "3" "5" using NextImpNext lift_imp_trans by blast  
 hence 7: "\<turnstile> finite \<longrightarrow> \<not>\<not> f" by (rule NextLoop)
 from 7 show ?thesis by auto
qed

lemma EmptyNextInducta:
 assumes "\<turnstile> empty  \<longrightarrow> f " 
         "\<turnstile> \<circle> f \<longrightarrow> f " 
 shows "\<turnstile> finite \<longrightarrow> f "
proof -
 have  1: "\<turnstile> empty  \<longrightarrow> f " using assms by auto
 have  2: "\<turnstile> \<circle> f \<longrightarrow> f " using assms by blast
 have  3: "\<turnstile> (empty \<or> \<circle> f) \<longrightarrow> f  " using "1" "2"  by fastforce
 have  4: "\<turnstile> wnext f = (empty \<or> \<circle> f)" by  (rule  WnextEqvEmptyOrNext)
 hence 5: "\<turnstile> wnext f \<longrightarrow> f" using "3" by fastforce
 hence 6: "\<turnstile> \<not>f \<longrightarrow> \<not> (wnext f)" by auto
 hence 7: "\<turnstile> \<not>f \<longrightarrow> \<circle>(\<not> f) " by (auto simp: wnext_d_def)
 hence 8: "\<turnstile> finite \<longrightarrow> \<not> \<not> f " by (rule NextLoop)
 from 8 show ?thesis by auto
qed

lemma EmptyNextInductb:
 assumes "\<turnstile> empty  \<and> f \<longrightarrow> g " 
         "\<turnstile> \<circle>(f\<longrightarrow> g) \<and> f \<longrightarrow> g "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g "
proof -
 have  1: "\<turnstile> empty  \<and> f \<longrightarrow> g " using assms by auto
 have  2: "\<turnstile> \<circle>(f\<longrightarrow> g) \<and> f \<longrightarrow> g " using assms by blast
 have  3: "\<turnstile> (empty \<or> \<circle>(f\<longrightarrow> g)) \<and> f \<longrightarrow> g " using "1" "2" by fastforce  
 hence 4: "\<turnstile> wnext (f\<longrightarrow> g)  \<and> f \<longrightarrow> g "   using WnextEqvEmptyOrNext by fastforce 
 hence 5: "\<turnstile> wnext (f\<longrightarrow> g) \<longrightarrow> ( f \<longrightarrow> g) "  by fastforce 
 hence 6: "\<turnstile> \<not> ( f \<longrightarrow> g) \<longrightarrow> \<not> ( wnext (f\<longrightarrow> g)) " by fastforce
 hence 7: "\<turnstile> \<not> ( f \<longrightarrow> g) \<longrightarrow> \<circle> ( \<not>(f\<longrightarrow> g)) " by (simp add: wnext_d_def)
 hence 8: "\<turnstile> finite \<longrightarrow> \<not> \<not> ( f \<longrightarrow> g)" by (rule NextLoop)
 from 8 show ?thesis by auto
qed


lemma FinImpFin:
 assumes "\<turnstile> f \<longrightarrow> g "
 shows   "\<turnstile>  fin  f \<longrightarrow>  fin  g "
using ImpBoxRule[of "LIFT (empty \<longrightarrow> f)" "LIFT (empty \<longrightarrow> g)"] assms 
      fin_d_def[of "f"] fin_d_def[of "g"] by fastforce 

lemma FinEqvFin:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  fin  f =  fin  g "
using assms by (simp add: FinImpFin Prop11) 

lemma FinAndFinImpFinRule:
 assumes "\<turnstile> f \<and> g \<longrightarrow> h "
 shows   "\<turnstile>  fin  f \<and>  fin  g \<longrightarrow>  fin  h "
using assms by (auto simp add: itl_defs Valid_def)  

lemma FinAndFinEqvFinRule:
 assumes "\<turnstile> (f \<and> g) = h "
 shows   "\<turnstile>  (fin  f \<and>  fin  g) =  fin  h "
using assms 
by (simp add: FinAndFinImpFinRule FinImpFin Prop11 Prop12)

lemma HaltEqvHalt:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  halt f =  halt g "
proof -
 have  1: "\<turnstile> f = g" using assms by auto
 hence 2: "\<turnstile> (empty = f)  = (empty = g) " by auto
 hence 3: "\<turnstile> \<box>(empty = f)  = \<box> (empty = g)" by (rule BoxEqvBox)
 from 3 show ?thesis by (simp add: halt_d_def)
qed

lemma BiImpDiImpDi:
  "\<turnstile> bi (f \<longrightarrow> g) \<longrightarrow>  di  f \<longrightarrow>  di  g "
proof -
 have 1: "\<turnstile> bi (f\<longrightarrow> g)  \<longrightarrow> (f; #True) \<longrightarrow> (g; #True) "  by (rule BiChopImpChop)
 from 1 show ?thesis   by (simp add: di_d_def)
qed

lemma DiImpDi:
 assumes "\<turnstile> f \<longrightarrow> g "
 shows   "\<turnstile>  di  f \<longrightarrow>  di  g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g " using assms by auto
 hence 2: "\<turnstile> f; #True \<longrightarrow> g; #True " by (rule LeftChopImpChop)
 from 2 show ?thesis by (simp add: di_d_def)
qed

lemma BiImpBiRule:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> bi f \<longrightarrow> bi g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> \<not>  g \<longrightarrow> \<not>  f " by auto
 hence 3: "\<turnstile>  di (\<not>  g) \<longrightarrow>  di (\<not>  f) " by (rule DiImpDi)
 hence 4: "\<turnstile> \<not> ( di (\<not>  f)) \<longrightarrow> \<not> ( di (\<not>  g)) " by auto
 from 4 show ?thesis by (simp add: bi_d_def)
qed

lemma DiEqvDi:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  di f =  di g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> f; #True = g; #True " by (rule LeftChopEqvChop)
 from 2 show ?thesis by (simp add: di_d_def)
qed

lemma BiEqvBi:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  bi f =  bi  g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> (\<not>  f) = (\<not>  g) " by auto
 hence 3: "\<turnstile>  di (\<not>  f) =  di (\<not>  g) " by (rule DiEqvDi)
 hence 4: "\<turnstile> (\<not>  (di (\<not>  f))) = (\<not> ( di (\<not>  g))) " by auto
 from 4 show ?thesis by (simp add: bi_d_def)
qed

lemma LeftChopChopImpChopRule:
 assumes "\<turnstile> (f; g) \<longrightarrow> g "
 shows   "\<turnstile> (f; g); h \<longrightarrow> (g; h) "
proof -
 have  1: "\<turnstile> (f; g) \<longrightarrow> g " using assms by blast
 hence 2: "\<turnstile> (f; g); h \<longrightarrow> g; h " by (rule LeftChopImpChop)
 have  3: "\<turnstile> f; (g; h) = (f; g); h   " by (rule ChopAssoc) 
 from 2 3 show ?thesis by auto
qed

lemma AndChopCommute :
 "\<turnstile> (f \<and> f1); g = (f1 \<and> f); g "
proof -
 have 1: "\<turnstile> (f \<and> f1) = (f1 \<and> f)  " by auto
 from 1 show ?thesis by (rule LeftChopEqvChop)
qed

lemma BiAndChopImport:
 "\<turnstile>  bi f \<and> (f1; g) \<longrightarrow> (f \<and> f1);g "
proof -
 have  1: "\<turnstile> f \<longrightarrow> (f1 \<longrightarrow> f\<and> f1) " by auto
 hence 2: "\<turnstile> bi  f \<longrightarrow> bi  (f1 \<longrightarrow> f\<and> f1) " by (rule BiImpBiRule)
 have  3: "\<turnstile> bi  (f1 \<longrightarrow> (f \<and> f1)) \<longrightarrow>   f1; g \<longrightarrow> (f \<and> f1); g " by (rule BiChopImpChop)
 from 2 3 show ?thesis using MP by fastforce
qed

lemma StateAndChopImport:
 "\<turnstile> (init w) \<and> (f;g) \<longrightarrow> ((init w) \<and> f);g "
proof -
 have  1: "\<turnstile> (init w)\<longrightarrow> bi  (init w) " by (rule StateImpBi)
 hence 2: "\<turnstile> (init w) \<and> (f; g) \<longrightarrow> bi (init w) \<and> (f; g) " by auto
 have  3: "\<turnstile> bi  (init w) \<and> (f; g) \<longrightarrow> ((init w) \<and> f); g "  by (rule BiAndChopImport)
 from 2 3 show ?thesis using MP by fastforce
qed

subsection \<open>Further Properties Di and Bi\<close>

lemma ImpDi:
 "\<turnstile> f \<longrightarrow>  di  f "
proof -
 have  1: "\<turnstile> f;  empty  = f " by (rule ChopEmpty)
 have  2: "\<turnstile> empty  \<longrightarrow> #True " by auto
 hence 3: "\<turnstile> f;  empty  \<longrightarrow> f; #True " by (rule RightChopImpChop)
 have 4 : "\<turnstile> f \<longrightarrow> f; #True " using "1" "3" by fastforce
 from 4 show ?thesis by (simp add: di_d_def)
qed

lemma DiState:
 "\<turnstile> di (init w) = (init w) "
proof -
 have  0: "\<turnstile> (init (\<not>w)) \<longrightarrow> bi   (init (\<not>w)) " using StateImpBi by fastforce
 hence 1: "\<turnstile> \<not>(init w) \<longrightarrow> bi ( \<not> (init w))" using Initprop(2) by (metis inteq_reflection)
 hence 2: "\<turnstile> (\<not>  (init w)) \<longrightarrow> \<not> ( di (\<not> \<not>  (init w))) " by (simp add: bi_d_def)
 have  3: "\<turnstile> (\<not>  (init w) \<longrightarrow>\<not>  (di (\<not> \<not>  (init w)))) \<longrightarrow> 
             ( di (\<not> \<not>  (init w)) \<longrightarrow> (init w)) " by auto
 have  4: "\<turnstile>  di (\<not> \<not>  (init w)) \<longrightarrow> (init w) " using "2" "3" MP by blast
 have  5: "\<turnstile> (init w) \<longrightarrow> \<not> \<not>  (init w) " by auto
 hence 6: "\<turnstile>  di  (init w) \<longrightarrow>  di (\<not> \<not>  (init w)) " by (rule DiImpDi) 
 have  7: "\<turnstile>  di  (init w) \<longrightarrow> (init w)" using "6" "4" using lift_imp_trans by metis
 have  8: "\<turnstile> (init w) \<longrightarrow>  di  (init w) " by (rule ImpDi)
 from 7 8 show ?thesis by fastforce
qed

lemma StateChop:
 "\<turnstile> (init w); f \<longrightarrow> (init w) "
 by (metis ChopAssoc Prop12 RightChopImpChop TrueW init_d_def int_simps(13) int_simps(17)
     inteq_reflection)

lemma StateChopExportA:
 "\<turnstile> ((init w) \<and> f); g \<longrightarrow> (init w) "
using DiState AndChopA StateChop by fastforce 

lemma StateAndChop:
 "\<turnstile> ((init w) \<and> f); g = ((init w) \<and> (f; g)) "
by (simp add: AndChopB StateAndChopImport StateChopExportA Prop11 Prop12)
 
lemma StateAndChopImpChopRule:
 assumes "\<turnstile> (init w) \<and> f \<longrightarrow> f1 " 
 shows   "\<turnstile> (init w) \<and> (f; g) \<longrightarrow> (f1; g) "
proof -
 have  1: "\<turnstile> (init w) \<and> f \<longrightarrow> f1 " using assms by auto
 hence 2: "\<turnstile> ((init w) \<and> f); g \<longrightarrow> f1; g " by (rule LeftChopImpChop)
 have  3: "\<turnstile> ((init w) \<and> f); g = ((init w) \<and> (f; g)) " by (rule StateAndChop)
 from 2 3 show ?thesis by fastforce
qed

lemma StateImpChopEqvChop :
 assumes "\<turnstile> (init w) \<longrightarrow> (f = f1)" 
 shows   "\<turnstile> (init w) \<longrightarrow> ((f; g) = (f1; g)) "
proof -
 have  1: "\<turnstile> (init w) \<longrightarrow> (f= f1)" using assms by auto
 hence 2: "\<turnstile> (init w) \<and> f \<longrightarrow> f1 " by auto
 hence 3: "\<turnstile> (init w) \<and> (f; g) \<longrightarrow> (f1; g) " by (rule StateAndChopImpChopRule)
 have  4: "\<turnstile> (init w) \<and> f1 \<longrightarrow> f " using "1" by auto
 hence 5: "\<turnstile> (init w) \<and> (f1; g) \<longrightarrow> (f; g)" by (rule StateAndChopImpChopRule)
 from 3 5 show ?thesis by fastforce
qed

lemma ChopEqvStateAndChop:
  assumes "\<turnstile> f = (init w) \<and> f1 "
  shows   "\<turnstile> (f; g) = ((init w) \<and> (f1; g)) "
proof -
 have  1: "\<turnstile> f = ((init w) \<and> f1) " using assms by auto
 hence 2: "\<turnstile> f; g = (((init w) \<and> f1); g) " by (rule LeftChopEqvChop)
 have  3: "\<turnstile> ((init w) \<and> f1); g = ((init w) \<and> (f1; g)) " by (rule StateAndChop)
 from 2 3 show ?thesis by fastforce
qed

lemma DiIntro:
 "\<turnstile> f \<longrightarrow>  di  f "
proof -
  have  1: "\<turnstile> f;empty  = f " by (rule ChopEmpty)
  have  2: "\<turnstile>  empty  \<longrightarrow> #True " by auto
  hence 3: "\<turnstile> \<box>( empty  \<longrightarrow> #True)" by (rule BoxGen)
  have  4: "\<turnstile> \<box>( empty  \<longrightarrow> #True) \<longrightarrow>  (f;  empty  \<longrightarrow> f; #True) " by (rule BoxChopImpChop)
  have  5: "\<turnstile> f;empty  \<longrightarrow> f;#True " using "3" "4" MP by fastforce
  hence 6: "\<turnstile> f;empty  \<longrightarrow>  di  f"  by (simp add: di_d_def)
  from 1 6 show ?thesis by fastforce
qed

lemma BiElim:
 "\<turnstile>   bi  f \<longrightarrow> f "
proof - 
  have  1: "\<turnstile> \<not>f \<longrightarrow>  di (\<not>f) " by (rule DiIntro)
  have  2: "\<turnstile> (\<not>f \<longrightarrow>  di (\<not>f)) \<longrightarrow> (\<not>(di (\<not>f)) \<longrightarrow> f) " by auto
  have  3: "\<turnstile> \<not>  (di (\<not>f)) \<longrightarrow> f " using 1 2 MP by blast
  from 3 show ?thesis by (metis bi_d_def)
qed

lemma BiContraPosImpDist:
  "\<turnstile> bi (\<not>g \<longrightarrow> \<not>f) \<longrightarrow> (bi f) \<longrightarrow> (bi g) "
proof -
  have  1: "\<turnstile> bi (\<not>g \<longrightarrow> \<not>f) \<longrightarrow> (di (\<not>g)) \<longrightarrow> (di (\<not>f)) " by (rule BiImpDiImpDi)
  hence 2: "\<turnstile> bi (\<not>g \<longrightarrow> \<not>f) \<longrightarrow> (\<not>( di (\<not>f))) \<longrightarrow> (\<not>(di (\<not>g))) " by auto
  from 2 show ?thesis by (metis bi_d_def)
qed

lemma BiImpDist:
 "\<turnstile> bi (f\<longrightarrow> g) \<longrightarrow> (bi f) \<longrightarrow> (bi g) "
proof -
 have  1: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> (\<not>  g \<longrightarrow> \<not>  f) " by auto
 hence 2: "\<turnstile> \<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g) " by auto
 hence 3: "\<turnstile> bi (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g)) " by (rule BiGen)
 have  4: "\<turnstile> bi (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g))
             \<longrightarrow>  
             bi (f \<longrightarrow> g) \<longrightarrow> bi (\<not>  g \<longrightarrow> \<not>  f) " by (rule BiContraPosImpDist)
 have  5: "\<turnstile> bi (f \<longrightarrow> g) \<longrightarrow> bi (\<not>  g \<longrightarrow> \<not>  f) " using 3 4 MP by blast
 have  6: "\<turnstile> bi (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (bi  f) \<longrightarrow> (bi  g) " by (rule BiContraPosImpDist)
 from 5 6 show ?thesis  using lift_imp_trans by blast 
qed

lemma IfChopEqvRule:
 assumes "\<turnstile> f =  if\<^sub>i  (init w)  then  f1  else  f2 "
 shows   "\<turnstile> f; g =  if\<^sub>i  (init w)  then  (f1; g)  else  (f2; g) "
proof -
 have  1: "\<turnstile> f =  if\<^sub>i  (init w)  then  f1  else  f2" 
   using assms by auto
 hence 2: "\<turnstile> f = (((init w) \<and> f1) \<or>  ( (init (\<not> w)) \<and> f2)) "
   by (metis Initprop(2) ifthenelse_d_def inteq_reflection) 
 hence 3: "\<turnstile> f; g = (((init w) \<and> f1); g \<or>  (  (init (\<not> w)) \<and> f2); g) " 
   by (rule OrChopEqvRule)
 have  4: "\<turnstile> ((init w) \<and> f1); g = ((init w) \<and> (f1; g)) " 
   by (rule StateAndChop)
 have  5: "\<turnstile> (  (init (\<not> w)) \<and> f2); g =   ((init (\<not> w)) \<and> (f2; g)) " 
   by (rule StateAndChop)
 have  6: "\<turnstile> f; g = (((init w) \<and> f1; g)  \<or>  (  (init (\<not> w)) \<and> f2; g)) " 
   using "3" "4" "5" by fastforce
 from 6 show ?thesis
   by (metis Initprop(2) ifthenelse_d_def inteq_reflection) 
qed

lemma ChopOrEqvRule:
 assumes "\<turnstile> g = (g1 \<or>  g2) "
 shows   "\<turnstile> f;g = ((f;g1) \<or>  (f;g2)) "
proof -
 have  1: "\<turnstile> g = (g1 \<or>  g2) " using assms by auto
 hence 2: "\<turnstile> f; g = (f; (g1 \<or>  g2)) " by (rule RightChopEqvChop)
 have  3: "\<turnstile> f; (g1 \<or>  g2) = (f; g1 \<or>  f; g2) " by (rule ChopOrEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma EmptyOrChopEqv: 
 "\<turnstile> ( empty  \<or>  f); g = (g \<or>  (f; g)) "
proof -
 have 1: "\<turnstile> ( empty  \<or>  f); g = (( empty ; g) \<or>  (f; g)) " by (rule OrChopEqv)
 have 2: "\<turnstile>  empty ; g = g " by (rule EmptyChop)
 from 1 2 show ?thesis by fastforce
qed

lemma EmptyOrNextChopEqv:
  "\<turnstile> ( empty  \<or>  \<circle> f); g = (g \<or>  \<circle>(f; g)) "
proof -
 have 1: "\<turnstile> ( empty  \<or>  \<circle> f); g = (g \<or>  ((\<circle> f); g)) " by (rule EmptyOrChopEqv)
 have 2: "\<turnstile> (\<circle> f); g = \<circle>(f; g) " by (rule NextChop)
 from 1 2 show ?thesis by fastforce
qed

lemma EmptyOrChopImpRule:
 assumes "\<turnstile>   f \<longrightarrow>  empty  \<or>  f1 " 
 shows   "\<turnstile> f; g \<longrightarrow> g \<or>  (f1; g) " 
proof -
 have  1: "\<turnstile> f \<longrightarrow>  empty  \<or>  f1 " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> ( empty  \<or>  f1); g " by (rule LeftChopImpChop)
 have  3: "\<turnstile> ( empty  \<or>  f1); g = (g \<or>  (f1; g)) " by (rule EmptyOrChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma EmptyOrChopEqvRule:
 assumes "\<turnstile>   f =  (empty  \<or>  f1) "
 shows   "\<turnstile> f; g = (g \<or>  (f1; g)) "
proof -
 have  1: "\<turnstile> f =  (empty  \<or>  f1)  " using assms by auto
 hence 2: "\<turnstile> f; g = (( empty  \<or>  f1); g) " by (rule LeftChopEqvChop)
 have  3: "\<turnstile> ( empty  \<or>  f1); g = (g \<or>  (f1; g))" by (rule EmptyOrChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma EmptyOrNextChopImpRule:
 assumes "\<turnstile> f \<longrightarrow>  empty  \<or>  \<circle> f1 "
 shows   "\<turnstile> f; g \<longrightarrow> g \<or>  \<circle>(f1; g) "
proof -
 have  1: "\<turnstile> f \<longrightarrow>  empty  \<or>  \<circle> f1 " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> ( empty  \<or>  \<circle> f1); g " by (rule LeftChopImpChop)
 have  3: "\<turnstile> ( empty  \<or>  \<circle> f1); g = (g \<or>  \<circle>(f1; g)) " by (rule EmptyOrNextChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma EmptyOrNextChopEqvRule:
 assumes "\<turnstile> f =  (empty  \<or>  \<circle> f1) "
 shows   "\<turnstile> f; g = (g \<or>  \<circle>(f1; g)) "
proof -
 have  1: "\<turnstile> f =  (empty  \<or>  \<circle> f1) " using assms by auto
 hence 2: "\<turnstile> f; g = (( empty  \<or>  \<circle> f1); g) " by (rule LeftChopEqvChop)
 have  3: "\<turnstile> ( empty  \<or>  \<circle> f1); g = (g \<or>  \<circle>(f1; g)) " by (rule EmptyOrNextChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma ChopEmptyOrImpRule:
 assumes "\<turnstile>  g \<longrightarrow>  empty  \<or>  g1 "
 shows "\<turnstile> f; g \<longrightarrow> f \<or>  (f; g1) "
proof -
 have  1: "\<turnstile> g \<longrightarrow>  empty  \<or>  g1 " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> (f;  empty ) \<or>  (f; g1) " by (rule ChopOrImpRule)
 have  3: "\<turnstile> f;  empty  = f " by (rule ChopEmpty)
 from 2 3 show ?thesis by fastforce
qed

lemma StateAndEmptyImpBoxState:
 "\<turnstile> (init w) \<and> empty \<longrightarrow> \<box> (init w)"
 using BoxEqvAndEmptyOrNextBox by fastforce

lemma BoxEqvAndBox:
 "\<turnstile> \<box> f = (f \<and> \<box> f)"
 using BoxElim by fastforce

lemma NotBoxImpNotOrNotNextBox:
 "\<turnstile> \<not>( \<box> f) \<longrightarrow> \<not>f \<or>  \<not>( \<circle> (\<box> f) )"
proof -
  have 1: "\<turnstile> f \<and> (\<circle> (\<box> f)) \<longrightarrow> \<box> f" using BoxEqvAndEmptyOrNextBox by fastforce
  hence 2: "\<turnstile> \<not>( \<box> f) \<longrightarrow> \<not>(f \<and> (\<circle> (\<box> f)) )  " by fastforce
  have 3: "\<turnstile> (\<not>(f \<and> (\<circle> (\<box> f)) )) = (\<not>f \<or> \<not>( \<circle> (\<box> f) )   ) " by auto
  from 2 3 show ?thesis by auto
qed

lemma BoxStateChopBoxAndInfImpBox:
 "\<turnstile> \<box> (init w); \<box> (init w) \<and> inf \<longrightarrow> \<box> (init w)"
by (auto simp add: Valid_def itl_defs ) 
   (metis dual_order.order_iff_strict is_NNil_ndropn less_eqE ndropn_ndropn ndropn_nfirst 
      ndropn_nlength nlength_eq_enat_nfiniteD not_le ntaken_nnth) 

lemma BoxStateChopBoxEqvBox:
 "\<turnstile>   \<box> (init w); \<box> (init w) = \<box> (init w) "
proof -
 have   1: "\<turnstile> (\<box> (init w)) = ((init w) \<and> ( empty  \<or>  \<circle>(\<box> (init w)))) " 
   by (rule BoxEqvAndEmptyOrNextBox)
 hence  2: "\<turnstile> (\<box> (init w); \<box> (init w)) =
               ((init w) \<and> (( empty  \<or>  \<circle>(\<box> (init w))); \<box> (init w))) " 
   by (metis StateAndChop inteq_reflection)
 have   3: "\<turnstile> (( empty  \<or>  \<circle>(\<box> (init w))); \<box> (init w)) =
              (\<box> (init w) \<or>  \<circle>(\<box> (init w); \<box> (init w))) " 
   by (rule EmptyOrNextChopEqv)
 have   4: "\<turnstile> (\<box> (init w); \<box> (init w)) = 
              ((init w) \<and> (\<box> (init w) \<or>  \<circle>(\<box> (init w); \<box> (init w)))) " 
   using "2" "3" by fastforce
 have   5: "\<turnstile> \<not> (\<box> (init w)) \<longrightarrow> \<not>  (init w) \<or>  \<not>( \<circle>(\<box> (init w))) " 
   by (rule NotBoxImpNotOrNotNextBox)
 have   6: "\<turnstile> (\<box> (init w); \<box> (init w)) \<and> \<not>( \<box> (init w)) \<longrightarrow>
              \<circle>(\<box> (init w); \<box> (init w)) \<and> \<not>( \<circle>(\<box> (init w)))  " 
   using "4" "5" by fastforce
 hence  7: "\<turnstile> \<box> (init w); \<box> (init w) \<and> finite \<longrightarrow> \<box> (init w) " 
   by (rule NextContra) 
 have   8: "\<turnstile> \<box> (init w); \<box> (init w) \<and> inf \<longrightarrow> \<box> (init w)"
   by (rule BoxStateChopBoxAndInfImpBox)
 have   9: "\<turnstile> \<box> (init w); \<box> (init w) \<and> (finite \<or> inf) \<longrightarrow> \<box> (init w)"
   using "7" "8"  by fastforce
 hence  10: "\<turnstile>  \<box> (init w); \<box> (init w) \<longrightarrow> \<box> (init w)"
   using FiniteOrInfinite by fastforce  
 have  11: "\<turnstile> \<box> (init w) = ((init w) \<and> \<box> (init w)) " 
   by (rule BoxEqvAndBox)
 have  12: "\<turnstile> empty ; \<box> (init w) = \<box> (init w)" 
   by (rule EmptyChop)
 have  13: "\<turnstile> ((init w) \<and>  empty ); \<box> (init w) = ((init w) \<and> ( empty ; \<box> (init w))) " 
   by (rule StateAndChop)
 have  14: "\<turnstile> \<box> (init w) = ((init w) \<and>  empty ); \<box> (init w) " 
   using "11" "12" "13" by fastforce
 have  15: "\<turnstile> (init w) \<and>  empty  \<longrightarrow> \<box> (init w) " 
   by (rule StateAndEmptyImpBoxState)
 hence 16: "\<turnstile> ((init w) \<and>  empty ); \<box> (init w) \<longrightarrow> \<box> (init w); \<box> (init w) " 
   by (rule LeftChopImpChop)
 have  17: "\<turnstile> \<box> (init w) \<longrightarrow> \<box> (init w); \<box> (init w)" 
   using "14" "16" by fastforce
 from 10 17 show ?thesis by fastforce
qed

lemma NotBoxStateImpBoxYieldsNotBox:
 "\<turnstile>   \<not>( \<box> (init w)) \<longrightarrow> (\<box> (init w)) yields (\<not>( \<box> (init w)))"
proof -
 have  1: "\<turnstile> \<box> (init w); \<box> (init w) = \<box> (init w) " by (rule BoxStateChopBoxEqvBox)
 have  2: "\<turnstile> \<box> (init w) = (\<not> \<not>( \<box> (init w))) " by auto
 hence 3: "\<turnstile> \<box> (init w); \<box> (init w) = \<box> (init w); (\<not> \<not>( \<box> (init w))) " by (rule RightChopEqvChop)
 have  4: "\<turnstile> \<not>( \<box> (init w)) \<longrightarrow> \<not> (\<box> (init w); (\<not> \<not> (\<box> (init w)))) " using "1" "3" by auto
 from 4 show ?thesis by (simp add: yields_d_def)
qed

lemma StateEqvBi:
 "\<turnstile>   (init w) = bi  (init w) "
proof -
  have 1: "\<turnstile> (init w) \<longrightarrow> bi  (init w) " by (rule StateImpBi)
  have 2: "\<turnstile> bi  (init w) \<longrightarrow> (init w) " by (rule BiElim)
  from 1 2 show ?thesis by fastforce
qed

lemma FiniteChopEqvDiamond:
 "\<turnstile>   finite; f = \<diamond> f"
by (simp add: sometimes_d_def)


subsection \<open>Properties of Da and Ba\<close>

lemma DaEqvDtDi:
 "\<turnstile>    da  f = \<diamond> (di  f) "
proof -
 have  1: "\<turnstile> finite; (f; #True) = finite; (f; #True) " by auto
 hence 2: "\<turnstile> finite; (f; #True) = finite;  di  f " by (simp add: di_d_def)
 have  3: "\<turnstile> finite;  di  f = \<diamond>( di  f)" by (rule FiniteChopEqvDiamond)
 have  4: "\<turnstile> finite; (f; #True) = \<diamond>( di  f) " using "2" "3" by fastforce
 from 4 show ?thesis by (simp add:da_d_def)
qed

lemma DaEqvDiDt:
 "\<turnstile>    da  f =  di (\<diamond> f) "
proof -
 have  1: "\<turnstile> finite; f = \<diamond> f " by (rule FiniteChopEqvDiamond)
 hence 2: "\<turnstile> (finite; f); #True = (\<diamond> f); #True " by (rule LeftChopEqvChop)
 hence 3: "\<turnstile> (finite; f); #True =  di( \<diamond> f) " by (simp add: di_d_def)
 have  4: "\<turnstile> finite; (f; #True) = (finite; f); #True   " by (rule ChopAssoc)
 have  5: "\<turnstile> finite; (f; #True) =  di (\<diamond> f) " using "3" "4" by fastforce
 from 5 show ?thesis by (simp add: da_d_def)
qed

lemma DtDiEqvDiDt:
 "\<turnstile>   \<diamond> (di  f) =  di (\<diamond> f) "
by (metis ChopAssoc di_d_def sometimes_d_def)

lemma DiamondNotEqvNotBox:
 "\<turnstile> \<diamond> (\<not>  f) = (\<not> (\<box> f))"
by (simp add: always_d_def)

lemma BaEqvBiBt:
 "\<turnstile>    ba f = bi( \<box> f) "
proof -
 have  1: "\<turnstile>  da (\<not>  f) =  di( \<diamond> (\<not>  f)) " by (rule DaEqvDiDt)
 have  2: "\<turnstile> \<diamond> (\<not>  f) = (\<not>( \<box> f)) " by (rule DiamondNotEqvNotBox)
 hence 3: "\<turnstile>  di (\<diamond>(\<not>  f)) =  di (\<not> (\<box> f)) " by (rule DiEqvDi)
 have  4: "\<turnstile>  da (\<not>  f) =  di (\<not>( \<box> f)) " using "1" "3" by fastforce
 hence 5: "\<turnstile> (\<not>  (da (\<not>  f))) = (\<not> ( di (\<not>( \<box> f)))) " by auto
 hence 6: "\<turnstile> (\<not> ( da (\<not>  f))) = bi( \<box> f) " by (simp add: bi_d_def)
 from 6 show ?thesis by (simp add: ba_d_def)
qed
 
lemma DiNotEqvNotBi:
 "\<turnstile>    di (\<not>  f) = (\<not>( bi  f)) "
proof -
 have 1: "\<turnstile> bi  f = (\<not> ( di (\<not>  f))) " by (simp add: bi_d_def)
 from 1 show ?thesis by auto
qed

lemma NotDiamondNotEqvBox:
 "\<turnstile> (\<not> (\<diamond>(\<not> f))) = \<box> f "
by (simp add: always_d_def)

lemma BaEqvBtBi:
 "\<turnstile>    ba  f = \<box> (bi  f) "
proof -
 have  1: "\<turnstile>  da (\<not>  f) = \<diamond> (di (\<not>  f)) " by (rule DaEqvDtDi)
 have  2: "\<turnstile>  di (\<not>  f) = (\<not> (bi  f)) " by (rule DiNotEqvNotBi)
 hence 3: "\<turnstile> \<diamond> (di (\<not>  f)) = \<diamond>(\<not> (bi  f)) " by (rule DiamondEqvDiamond)
 have  4: "\<turnstile> (\<not> (\<diamond>(\<not> (bi  f)))) = \<box>(bi  f)  " by (rule NotDiamondNotEqvBox)
 have  5: "\<turnstile> (\<not> ( da (\<not>  f))) = \<box>(bi  f) " using "1" "2" "3" "4" by fastforce
 from 5 show ?thesis by (simp add: ba_d_def)
qed

lemma BtBiEqvBiBt:
 "\<turnstile>   \<box> (bi  f) = bi( \<box> f) "
proof -
 have 1: "\<turnstile>    ba  f = \<box> (bi  f)  " by (rule BaEqvBtBi)
 have 2: "\<turnstile>    ba  f = bi( \<box> f) "  by (rule BaEqvBiBt)
 from 1 2 show ?thesis by fastforce
qed

lemma BoxStateEqvBaBoxState:
 "\<turnstile>   \<box> (init w) =  ba (\<box> (init w)) "
proof -
 have  1: "\<turnstile> (init w) = bi  (init w) " by (rule StateEqvBi)
 hence 2: "\<turnstile> \<box> (init w) = \<box> (bi  (init w)) " by (rule BoxEqvBox)
 have  3: "\<turnstile> \<box> (bi  (init w)) = bi( \<box> (init w))  " by (rule BtBiEqvBiBt)
 have  4: "\<turnstile> \<box> (init w) = \<box>(\<box> (init w)) " by (rule BoxEqvBoxBox)
 hence 5: "\<turnstile> bi( \<box> (init w)) = bi (\<box>(\<box> (init w))) " by (rule BiEqvBi)
 have  6: " \<turnstile>  ba( \<box> (init w)) = bi( \<box>(\<box> (init w))) "  by (rule BaEqvBiBt)
 from 2 3 5 6 show ?thesis  by fastforce 
qed

lemma BaImpBi:
 "\<turnstile>    ba  f \<longrightarrow> bi  f "
proof -
 have 1: "\<turnstile>  ba  f = \<box>(bi  f) " by (rule BaEqvBtBi)
 have 2: "\<turnstile> \<box>(bi  f) \<longrightarrow> bi  f " by (rule BoxElim)
 from 1 2 show ?thesis using lift_imp_trans by fastforce
qed

lemma BaImpBt:
 "\<turnstile>    ba  f \<longrightarrow> \<box> f "
proof -
 have 1: "\<turnstile>  ba  f = bi( \<box> f) " by (rule BaEqvBiBt)
 have 2: "\<turnstile> bi( \<box> f) \<longrightarrow> \<box> f " by (rule BiElim)
 from 1 2 show ?thesis using lift_imp_trans by fastforce
qed

lemma DiamondImpDa:
 "\<turnstile>   \<diamond> f \<longrightarrow>  da  f "
by (metis DiIntro DiamondImpDiamond da_d_def di_d_def sometimes_d_def)

lemma DiImpDa:
 "\<turnstile>    di  f \<longrightarrow>  da  f"
by (metis NowImpDiamond da_d_def di_d_def sometimes_d_def)

lemma BoxAndChopImport:
 "\<turnstile>   \<box> h \<and> f; g \<longrightarrow> f; (h \<and> g) " 
proof -
 have  1: "\<turnstile> h \<longrightarrow> g \<longrightarrow> (h\<and> g) " by auto
 hence 2: "\<turnstile> \<box> h \<longrightarrow> \<box>(g \<longrightarrow> (h\<and> g)) " by (rule ImpBoxRule)
 have  3: "\<turnstile> \<box>(g \<longrightarrow> (h\<and> g)) \<longrightarrow> f; g \<longrightarrow> f; (h\<and> g) " by (rule BoxChopImpChop)
 from 2 3 show ?thesis by fastforce
qed

lemma BaAndChopImport:
 " \<turnstile>    ba  f \<and> (g; g1) \<longrightarrow> (f \<and> g); (f \<and> g1) "
proof -
 have  1: "\<turnstile> ba  f \<longrightarrow> bi  f " by (rule BaImpBi)
 have  2: "\<turnstile> bi  f \<and> (g; g1) \<longrightarrow> (f \<and> g); g1 " by (rule BiAndChopImport)
 have  3: "\<turnstile> ba  f \<longrightarrow> \<box> f " by (rule BaImpBt)
 have  4: "\<turnstile> \<box> f \<and> (f \<and> g); g1 \<longrightarrow> (f \<and> g); (f \<and> g1) " by (rule BoxAndChopImport)
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma ChopAndCommute:
 "\<turnstile>   f; (g \<and> g1) = f; (g1 \<and> g) "
proof -
 have 1: "\<turnstile> (g \<and> g1) = (g1 \<and> g)" by auto
 from 1 show ?thesis by (rule RightChopEqvChop)
qed

lemma ChopAndA: 
 "\<turnstile>   f; (g \<and> g1) \<longrightarrow> f; g "
proof -
 have 1: "\<turnstile> (g \<and> g1) \<longrightarrow>  g " by auto
 from 1 show ?thesis by (rule RightChopImpChop)
qed

lemma ChopAndB:
 "\<turnstile>   f; (g \<and> g1) \<longrightarrow> f; g1 "
proof -
 have 1: "\<turnstile> (g \<and> g1) \<longrightarrow>  g1 " by auto
from 1 show ?thesis by (rule RightChopImpChop)
qed
 
lemma BoxStateAndChopEqvChop:
 "\<turnstile>  (\<box> (init w) \<and> (f; g)) = ((\<box> (init w) \<and> f); (\<box> (init w) \<and> g)) "
proof -
 have  1: "\<turnstile> \<box> (init w) =  ba( \<box> (init w)) " 
   by (rule BoxStateEqvBaBoxState)
 have  2: "\<turnstile> ba( \<box> (init w)) \<and> (f; g) \<longrightarrow> (\<box> (init w) \<and> f); (\<box> (init w) \<and> g) " 
   by (rule BaAndChopImport)
 have  3: "\<turnstile> \<box> (init w) \<and> (f; g) \<longrightarrow> (\<box> (init w) \<and> f); (\<box> (init w) \<and> g) " 
   using "1" "2"  by fastforce 
 have 11: "\<turnstile> (\<box> (init w) \<and> f); (\<box> (init w) \<and> g) \<longrightarrow> (\<box> (init w)); (\<box> (init w) \<and> g) " 
   by (rule AndChopA)
 have 12: "\<turnstile> (\<box> (init w)); (\<box> (init w) \<and> g) \<longrightarrow> (\<box> (init w)); (\<box> (init w)) " 
   by (rule ChopAndA)
 have 13: "\<turnstile> (\<box> (init w)); (\<box> (init w)) = \<box> (init w) " 
   by (rule BoxStateChopBoxEqvBox)
 have 14: "\<turnstile> (\<box> (init w) \<and> f); (\<box> (init w) \<and> g) \<longrightarrow> f; (\<box> (init w) \<and> g) " 
   by (rule AndChopB)
 have 15: "\<turnstile> f; (\<box> (init w) \<and> g) \<longrightarrow> f; g " 
   by (rule ChopAndB)
 have 16: "\<turnstile> (\<box> (init w) \<and> f); (\<box> (init w) \<and> g) \<longrightarrow> \<box> (init w) \<and> (f; g) " 
   using  "11" "12" "13" "14" "15"  by fastforce
 from 3 16 show ?thesis by fastforce
qed

lemma DiEqvNotBiNot:
 " \<turnstile>    di  f = (\<not>( bi (\<not>  f))) "
proof -
 have  1: "\<turnstile> bi (\<not>  f) = (\<not> ( di (\<not> \<not>  f))) " by (simp add: bi_d_def)
 hence 2: "\<turnstile>  di (\<not> \<not>  f) = (\<not>( bi (\<not>  f))) " by auto
 have  3: "\<turnstile> f = (\<not> \<not>  f) " by auto
 hence 4: "\<turnstile>  di  f =  di  (\<not> \<not>  f) " by (rule DiEqvDi)
 from 2 4 show ?thesis by auto
qed

lemma ChopAndBoxImport:
 "\<turnstile>   f; g \<and> \<box> h \<longrightarrow> f; (g \<and> h) "
proof -
 have 1: "\<turnstile> \<box> h \<and> f; g \<longrightarrow> f; (h \<and> g) " by (rule BoxAndChopImport)
 have 2: "\<turnstile> f; (h \<and> g) = f; (g \<and> h) " by (rule ChopAndCommute)
 from 1 2 show ?thesis by fastforce
qed

lemma AndChopAndCommute:
 "\<turnstile>   (f \<and> g); (f1 \<and> g1) = (g \<and> f); (g1 \<and> f1) "
proof -
 have 1: "\<turnstile> (f \<and> g); (f1 \<and> g1) = (g \<and> f); (f1 \<and> g1) " by (rule AndChopCommute)
 have 2: "\<turnstile> (g \<and> f); (f1 \<and> g1) = (g \<and> f); (g1 \<and> f1) "  by (rule ChopAndCommute)
 from 1 2 show ?thesis by fastforce
qed

lemma ChopImpChop:
 assumes "\<turnstile> f \<longrightarrow> f1 " 
         "\<turnstile> g \<longrightarrow> g1 "
 shows   "\<turnstile> f;g \<longrightarrow> f1;g1 "
proof -
 have  1: "\<turnstile> f \<longrightarrow> f1 " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> f1; g " by (rule LeftChopImpChop)
 have  3: "\<turnstile> g \<longrightarrow> g1 " using assms by auto
 hence 4: "\<turnstile> f1; g \<longrightarrow> f1; g1 " by (rule RightChopImpChop)
 from 2 4 show ?thesis by fastforce
qed

lemma ChopEqvChop:
 assumes "\<turnstile> f = f1" 
         "\<turnstile> g = g1 "
 shows   "\<turnstile> f;g = f1;g1 "
proof -
 have  1: "\<turnstile> f = f1 " using assms by auto
 hence 2: "\<turnstile> f; g = f1; g " by (rule LeftChopEqvChop)
 have  3: "\<turnstile> g = g1 " using assms by auto
 hence 4: "\<turnstile> f1; g = f1; g1 " by (rule RightChopEqvChop)
 from 2 4 show ?thesis by fastforce
qed

lemma BoxImpBoxImpBox:
 "\<turnstile> \<box> h \<longrightarrow> \<box>(g \<longrightarrow> \<box> h \<and> g ) "
proof -
 have 1: "\<turnstile> \<box> h \<longrightarrow> (g \<longrightarrow> \<box> h \<and> g ) " by auto
 hence 2: "\<turnstile> \<box>(\<box> h) \<longrightarrow> \<box>(g \<longrightarrow> \<box> h \<and> g ) " by (rule ImpBoxRule)
 have 3: "\<turnstile> \<box> h = \<box>(\<box>h) " by (rule BoxEqvBoxBox)
 from 2 3 show ?thesis by fastforce
qed

lemma BoxChopImpChopBox:
 "\<turnstile>   \<box> h \<longrightarrow> f; g \<longrightarrow> f; (\<box> h \<and> g) "
proof -
 have 1: "\<turnstile> \<box> h \<longrightarrow> \<box>(g \<longrightarrow> \<box> h \<and> g ) " by (rule BoxImpBoxImpBox)
 have 2: "\<turnstile> \<box>(g \<longrightarrow> \<box> h \<and> g ) \<longrightarrow> f; g \<longrightarrow> f; (\<box> h \<and> g) " by (rule BoxChopImpChop)
 from 1 2 show ?thesis by fastforce
qed

lemma NotChopEqvYieldsNot:
"\<turnstile>   (\<not> (f; g)) = f yields  (\<not>  g) "
proof -
 have  1: "\<turnstile> g = (\<not> \<not>  g) " by auto
 hence 2: "\<turnstile> f; g = f; (\<not> \<not>  g) " by (rule RightChopEqvChop)
 hence 3: "\<turnstile> (\<not> (f; g)) = (\<not> (f; (\<not> \<not>  g))) " by auto
 from 3 show ?thesis by (simp add: yields_d_def)
qed

lemma NotDiFalse:
 "\<turnstile>   \<not> ( di  #False) "
proof -
 have  1: "\<turnstile> (init #True) \<longrightarrow> bi (init #True) " by (rule StateImpBi)
 hence 2: "\<turnstile> #True \<longrightarrow> bi #True " by (simp add: BiGen) 
 have  3: "\<turnstile> #True" by auto
 have  4: "\<turnstile> bi #True" using "2" "3" MP by auto
 hence 5: "\<turnstile> \<not> ( di (\<not> #True)) " by (simp add: bi_d_def)
 have  6: "\<turnstile> (\<not> #True) =  #False " by auto
 hence 7: "\<turnstile>  di (\<not> #True) =  di  #False " by (rule DiEqvDi)
 from 5 7 show ?thesis by auto
qed

lemma StateAndEmptyChop:
 "\<turnstile>   ((init w) \<and>  empty ); f = ((init w) \<and> f) "
proof - 
 have 1: "\<turnstile> ((init w) \<and>  empty ); f = ((init w) \<and>  empty ; f) " by (rule StateAndChop)
 have 2: "\<turnstile>  empty ; f = f " by (rule EmptyChop)
 from 1 2 show ?thesis by fastforce
qed

lemma StateAndNextChop:
 "\<turnstile>   ((init w) \<and> \<circle> f); g = ((init w) \<and> \<circle>(f; g)) "
proof -
 have 1: "\<turnstile> ((init w) \<and> \<circle> f); g = ((init w) \<and> (\<circle> f); g) " by (rule StateAndChop)
 have 2: "\<turnstile> (\<circle> f); g = \<circle>(f; g) " by (rule NextChop)
 from 1 2 show ?thesis by fastforce
qed

lemma NextAndEqvNextAndNext:
 "\<turnstile> \<circle> (f \<and> g) = (\<circle> f \<and> \<circle> g)"
by (metis NextAndNextEqvNextRule int_eq lift_and_com)

lemma NextStateAndChop:
 "\<turnstile>   \<circle>(((init w) \<and> f); g) = (\<circle> (init w) \<and> \<circle>(f; g)) "
proof -
 have  1: "\<turnstile> ((init w) \<and> f); g = ((init w) \<and> f; g)  " by (rule StateAndChop)
 hence 2: "\<turnstile> \<circle>(((init w) \<and> f); g) = \<circle>((init w) \<and> f; g) " by (rule NextEqvNext)
 have  3: "\<turnstile> \<circle>((init w) \<and> f; g) = (\<circle> (init w) \<and> \<circle>(f; g)) " by (rule NextAndEqvNextAndNext)
 from 2 3 show ?thesis by fastforce
qed

lemma StateYieldsEqv:
 "\<turnstile>    ((init w) \<longrightarrow> (f yields  g)) = ((init w) \<and> f) yields  g " 
proof -
 have  1: "\<turnstile> ((init w) \<and> f); (\<not>  g) = ((init w) \<and> f; (\<not>  g))  " by (rule StateAndChop)
 hence 2: "\<turnstile> ((init w) \<longrightarrow> \<not> (f; (\<not>  g))) = (\<not> (((init w) \<and> f); (\<not>  g) )) " by auto
 from 2 show ?thesis by (simp add: yields_d_def)
qed

lemma StateAndDi:
 "\<turnstile>   ((init w) \<and>  di  f) =  di ((init w) \<and> f) "
proof - 
 have 1: "\<turnstile> ((init w) \<and> f); #True=  ((init w) \<and> f; #True)  " by (rule StateAndChop)
 from 1 show ?thesis by (metis di_d_def inteq_reflection)
qed

lemma DiNext: 
 "\<turnstile>    di( \<circle> f) = \<circle> (di  f) "
proof -
 have 1: "\<turnstile> (\<circle> f); #True = \<circle>(f; #True) " by (rule NextChop)
 from 1 show ?thesis by (simp add: di_d_def)
qed

lemma DiNextState:
 "\<turnstile>    di( \<circle> (init w)) = \<circle> (init w) "
proof -
 have  1: "\<turnstile>  di( \<circle> (init w)) = \<circle>( di  (init w)) " by (rule DiNext)
 have  2: "\<turnstile>  di  (init w) = (init w) " by (rule DiState)
 hence 3: "\<turnstile> \<circle>( di  (init w)) = \<circle> (init w) " by (rule NextEqvNext)
 from 1 3 show ?thesis by fastforce
qed

lemma StateImpBiGen:
 assumes "\<turnstile> (init w) \<longrightarrow> f "
 shows   "\<turnstile> (init w) \<longrightarrow> bi  f "
proof -
 have  1: "\<turnstile> (init w) \<longrightarrow> f " using assms by auto
 hence 2: "\<turnstile> \<not>  f \<longrightarrow> \<not>  (init w) " by auto
 hence 3: "\<turnstile>  di (\<not>  f) \<longrightarrow>  di (\<not>  (init w)) " by (rule DiImpDi)
 hence 4: "\<turnstile>  di (\<not>  f) \<longrightarrow>  di  (init (\<not>w)) " by (metis Initprop(2) inteq_reflection)
 have  5: "\<turnstile>  di (init (\<not> w)) =   (init (\<not> w)) " by (rule DiState)
 have  6: "\<turnstile>  di (\<not>  f) \<longrightarrow> \<not>  (init w) " using 4 5 using Initprop(2) by fastforce
 hence 7: "\<turnstile> (init w) \<longrightarrow> \<not> ( di (\<not>  f)) " by auto
 from 7 show ?thesis by (simp add: bi_d_def)
qed

lemma ChopAndNotChopImp:
 "\<turnstile>   f; g \<and> \<not> (f; g1) \<longrightarrow> f; (g \<and> \<not>  g1) "
proof -
 have  1: "\<turnstile> g \<longrightarrow> (g\<and> \<not>  g1) \<or>  g1 " by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> f; ((g\<and> \<not>  g1) \<or>  g1) " by (rule RightChopImpChop)
 have  3: "\<turnstile> f; ((g\<and> \<not>  g1) \<or>  g1) \<longrightarrow> (f; (g\<and> \<not>  g1)) \<or>  (f; g1) " by (rule ChopOrImp)
 have  4: "\<turnstile> f; g \<longrightarrow> f; (g\<and> \<not>  g1) \<or>  f; g1 " using "2" "3" MP by fastforce
 from 4 show ?thesis by auto
qed

lemma ChopAndYieldsImp:
 "\<turnstile>   f; g \<and> f yields  g1 \<longrightarrow> f; (g \<and> g1) "
proof -
 have  1: "\<turnstile> g \<longrightarrow> (g\<and> g1) \<or>  \<not>  g1 " by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> f; ((g\<and>   g1) \<or>  \<not> g1) "  by (rule RightChopImpChop)
 have  3: "\<turnstile> f; ((g\<and>   g1) \<or>  \<not> g1) \<longrightarrow> (f; (g\<and>   g1)) \<or>  (f; (\<not> g1)) " by (rule ChopOrImp)
 have  4: "\<turnstile> f; g \<longrightarrow> f; (g\<and>   g1) \<or>  f; (\<not> g1) " using "2" "3" MP by fastforce
 hence 5: "\<turnstile> f; g \<and> \<not>  (f; (\<not>  g1)) \<longrightarrow> f; (g \<and> g1) " by auto
 from 5 show ?thesis by (simp add: yields_d_def)
qed

lemma ChopAndYieldsMP:
 "\<turnstile>   f; g \<and> f yields (g\<longrightarrow> g1) \<longrightarrow> f; g1 "
proof -
 have  1: "\<turnstile> f; g \<and> f yields (g\<longrightarrow> g1) \<longrightarrow> f; (g \<and> (g \<longrightarrow> g1)) " by (rule ChopAndYieldsImp)
 have  2: "\<turnstile> g \<and> (g \<longrightarrow> g1) \<longrightarrow> g1 " by auto
 hence 3: "\<turnstile> f; (g \<and> (g \<longrightarrow> g1)) \<longrightarrow> f; g1 " by (rule RightChopImpChop)
 from 1 3 show ?thesis by fastforce
qed

lemma OrYieldsImp:
  "\<turnstile>   (f \<or>  f1) yields  g = ((f yields  g) \<and> (f1 yields  g)) "
proof -
 have  1: "\<turnstile> ((f\<or>  f1); (\<not>  g)) = ((f; (\<not>  g)) \<or>  (f1; (\<not>  g))) " by (rule OrChopEqv)
 hence 2: "\<turnstile> (\<not> ((f\<or>  f1); (\<not>  g))) = (\<not> (f; (\<not>  g)) \<and>  \<not>(f1; (\<not>  g))) " by auto
 from 2 show ?thesis by (simp add: yields_d_def)
qed

lemma LeftYieldsImpYields:
 assumes "\<turnstile> f \<longrightarrow> f1 "
 shows   "\<turnstile> (f1 yields  g) \<longrightarrow> (f yields  g) "
proof -
 have  1: "\<turnstile> f \<longrightarrow> f1 " using assms by auto
 hence 2: "\<turnstile> f; (\<not>  g) \<longrightarrow> f1; (\<not>  g)  " by (rule LeftChopImpChop)
 hence 3: "\<turnstile> \<not> (f1; (\<not>  g)) \<longrightarrow> \<not> (f; (\<not>  g)) " by auto
 from 3 show ?thesis by (simp add: yields_d_def)
qed

lemma LeftYieldsEqvYields:
 assumes "\<turnstile>  f = f1 "
 shows "\<turnstile> (f yields  g) = (f1 yields  g) "
proof -
 have  1: "\<turnstile> f = f1 " using assms by auto
 hence 2: "\<turnstile> f; (\<not>  g) = f1; (\<not>  g) " by (rule LeftChopEqvChop)
 hence 3: "\<turnstile> (\<not> (f; (\<not>  g))) = (\<not> (f1; (\<not>  g))) " by auto
 from 3 show ?thesis by (simp add: yields_d_def)
qed


subsection \<open>Properties of Fin\<close>

lemma FinEqvTrueChopAndEmpty:
  " \<turnstile>  fin f = #True;(f \<and> empty)"
proof -
 have 1: "\<turnstile> fin f = \<box>(empty \<longrightarrow> f)" 
   by (simp add: fin_d_def)
 have 2: "\<turnstile> \<box>(empty \<longrightarrow> f) = (\<not>(\<diamond>(\<not>(empty \<longrightarrow> f ) ) ))" 
   by (simp add: always_d_def)
 have 3: "\<turnstile> (\<not>(empty \<longrightarrow> f )) = (\<not> f \<and> empty)  " 
   by auto
 hence 4: "\<turnstile> \<diamond>(\<not>(empty \<longrightarrow> f )) = \<diamond>(\<not> f \<and> empty) " 
   using DiamondEqvDiamond by blast
 hence 5: "\<turnstile> \<not>(\<diamond>(\<not>(empty \<longrightarrow> f ))) = (\<not>(\<diamond>(\<not> f \<and> empty)))" 
   by auto
 have 51: "\<turnstile> finite;((\<not> f \<and> empty) \<and> finite)  \<longrightarrow> finite ;  (\<not> f \<and> empty)"
   by (simp add: ChopAndA) 
 have 52: "\<turnstile> (#True;(\<not> f \<and> empty) \<and> finite) \<longrightarrow> finite ;  (\<not> f \<and> empty)"
   by (metis "51" TrueChopAndFiniteEqvAndFiniteChopFinite int_eq) 
 have 53: "\<turnstile> \<not> (#True;(f \<and> empty))  \<longrightarrow> finite ;  (\<not> f \<and> empty)"
   by (metis "52" Finprop(5) int_eq) 
 have 54: " \<turnstile> \<not> finite;(\<not> f \<and> empty) \<longrightarrow> #True;(f \<and> empty)" 
   using 53 by auto
 have 6: "\<turnstile> (\<not>(\<diamond>(\<not> f \<and> empty))) \<longrightarrow> #True;(f \<and> empty)"
   unfolding sometimes_d_def using 54 by auto
 have 61:  "\<turnstile> \<not> f \<and> empty \<longrightarrow> finite"
   by (metis ChopAndB DiamondEmptyEqvFinite NowImpDiamond inteq_reflection lift_imp_trans sometimes_d_def)
 have 62: "\<turnstile>(\<not> #True;(f \<and> empty)) = finite;(\<not> f \<and> empty)"
   using 61 
   by (metis (no_types) Finprop(5) Prop10 TrueChopAndFiniteEqvAndFiniteChopFinite inteq_reflection)
 have 7: "\<turnstile> #True;(f \<and> empty) \<longrightarrow> (\<not>(\<diamond>(\<not> f \<and> empty)))"
   unfolding sometimes_d_def using TrueChopAndFiniteEqvAndFiniteChopFinite[of "LIFT(f \<and> empty)"]
   using "62" by auto
 have 8: "\<turnstile> (\<not>(\<diamond>(\<not> f \<and> empty))) = #True;(f \<and> empty)" 
   by (simp add: "6" "7" int_iffI)
 from 1 2 5 8 show ?thesis by fastforce
qed


lemma DiamondFin:
 "\<turnstile> \<diamond>(fin w) = fin w"
by (metis (no_types, lifting) ChopAssoc ChopOrEqv FinEqvTrueChopAndEmpty FiniteChopFiniteEqvFinite
           FiniteChopInfEqvInf FiniteOrInfinite int_eq_true inteq_reflection sometimes_d_def)

lemma FiniteChopFinExportA:
 "\<turnstile> (f \<and> finite);(g \<and> fin w) \<longrightarrow> fin w"
using DiamondFin 
by (metis ChopAndB FiniteChopImpDiamond inteq_reflection lift_imp_trans)

lemma FinImpBox:
 "\<turnstile> fin w \<longrightarrow> \<box>(fin w)" 
by (metis BoxImpBoxBox fin_d_def)

lemma FinAndChopImport:
 "\<turnstile> (fin w) \<and> (f;g) \<longrightarrow> f;((fin w) \<and> g)"
proof -
 have 1: "\<turnstile> fin w \<longrightarrow> \<box>(fin w)" by (rule FinImpBox)
 hence 2: "\<turnstile> fin w \<and> f;g \<longrightarrow> \<box>(fin w) \<and> (f;g)" by auto
 have 3: "\<turnstile> \<box>(fin w) \<and> (f;g) \<longrightarrow> f;((fin w) \<and> g)" using BoxAndChopImport by blast
 from 2 3 show ?thesis using MP by fastforce
qed

lemma FinAndChop:
 "\<turnstile> ((f \<and> finite);(g \<and> fin w)) = (fin w \<and> (f \<and> finite);g)  "
using FinAndChopImport FiniteChopFinExportA ChopAndA ChopAndCommute 
by fastforce

lemma ChopAndEmptyEqvEmptyChopEmpty:
 "\<turnstile> ((f;g) \<and> empty) = (f \<and> empty);(g \<and> empty) "
by (auto simp: itl_defs min_absorb1)

lemma FinAndEmpty:
 "\<turnstile> ((fin w) \<and> empty) = (w \<and> empty)"
proof -
 have 1: "\<turnstile> ((fin w) \<and> empty) = (#True;(w \<and> empty) \<and> empty)" 
   using FinEqvTrueChopAndEmpty by fastforce
 have 2: "\<turnstile> (#True;(w \<and> empty) \<and> empty) = ((#True \<and> empty);(w \<and> empty))" 
   using ChopAndEmptyEqvEmptyChopEmpty[of "LIFT(#True)" "LIFT(w \<and> empty)"] 
   by (metis AndChopA ChopAndA ChopAndEmptyEqvEmptyChopEmpty Prop11 Prop12 int_eq) 
 have 3: "\<turnstile> (#True \<and> empty);(w \<and> empty) = (empty;(w \<and> empty))" 
   using LeftChopEqvChop by fastforce
 have 4: "\<turnstile> (empty;(w \<and> empty)) = (w \<and> empty)" 
   using EmptyChop by blast
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma AndFinEqvChopAndEmpty:
 "\<turnstile>   ((f \<and> finite) \<and> fin g) = (f \<and> finite); (g \<and>  empty ) "
proof -
 have 1: "\<turnstile> ((f \<and> finite) \<and> fin g) = ((f \<and> finite) ;empty \<and> fin g)" 
   using ChopEmpty by (metis inteq_reflection) 
 have 2: "\<turnstile> (fin g \<and> (f \<and> finite);empty) = ((f \<and> finite);(empty \<and> fin g))  " 
   using FinAndChop by fastforce
 have 3: "\<turnstile> (empty \<and> fin g) =  (fin g \<and> empty)" 
   by auto
 have 4: "\<turnstile> (fin g \<and> empty) = (g \<and> empty)" 
   using FinAndEmpty by metis
 have 5: "\<turnstile> (empty \<and> fin g) = (g \<and> empty)" 
   using "3" "4" by auto
 hence 6: "\<turnstile> (f \<and> finite);(empty \<and> fin g) = (f \<and> finite);(g \<and> empty)" 
   using RightChopEqvChop by blast
 from 1 2 5 show ?thesis by (metis inteq_reflection lift_and_com)
qed

lemma AndFinEqvChopStateAndEmpty:
 "\<turnstile>   ((f \<and> finite) \<and>  fin  (init w)) = (f \<and> finite); ((init w) \<and>  empty ) "
using AndFinEqvChopAndEmpty by blast

lemma FinStateEqvStateAndEmptyOrNextFinState:
 "\<turnstile>  fin  (init w) = (((init w) \<and>  empty ) \<or>  \<circle>( fin  (init w))) "
proof -
 have 1: "\<turnstile> fin (init w) = \<box>( empty \<longrightarrow> init w)" 
   by (simp add: fin_d_def)
 have 2 : "\<turnstile> \<box>(empty \<longrightarrow> init w) = 
               ((empty \<longrightarrow> init w) \<and> wnext (\<box> (empty \<longrightarrow> init w))) " 
   by (rule BoxEqvAndWnextBox)
 have 3: "\<turnstile> fin (init w) =  ((empty \<longrightarrow> init w) \<and> wnext (fin (init w))) " 
   using "1" "2" by (simp add: fin_d_def) 
 have 4: "\<turnstile> wnext (fin (init w)) = (empty \<or> \<circle> (fin (init w))) " 
   by (rule WnextEqvEmptyOrNext)
 have 5: "\<turnstile> fin (init w) =  ((empty \<longrightarrow> init w) \<and> (empty \<or> \<circle> (fin (init w))))" 
   using "3" "4" by fastforce
 have 6: "\<turnstile> ((empty \<longrightarrow> init w) \<and> (empty \<or> \<circle> (fin (init w)))) =
            (((empty \<longrightarrow> init w) \<and> empty) \<or> ((empty \<longrightarrow> init w) \<and> \<circle> (fin (init w))))" 
   by auto
 have 7: "\<turnstile> ((empty \<longrightarrow> init w) \<and> empty) = ((init w) \<and> empty)" 
   by auto
 have 8: "\<turnstile> ((empty \<longrightarrow> init w) \<and> \<circle> (fin (init w))) =  \<circle> (fin (init w))"
   by (metis (no_types, lifting) "5" DiamondFin NextDiamondImpDiamond Prop10 Prop12 int_eq 
       lift_and_com) 
 have 9: "\<turnstile> (((empty \<longrightarrow> init w) \<and> empty) \<or> ((empty \<longrightarrow> init w) \<and> \<circle> (fin (init w)))) =
            ((init w) \<and>  empty ) \<or>  \<circle>( fin  (init w))" 
   using "7" "8" by auto
 from 5 6 8 9 show ?thesis  by fastforce
qed

lemma FinChopEqvOr:
 " \<turnstile>   ( fin  (init w)); f = (((init w) \<and> f) \<or>  \<circle>(( fin  (init w)); f)) "
proof -
 have  1: "\<turnstile>  fin  (init w) = (((init w) \<and>  empty ) \<or>  \<circle>( fin  (init w))) " 
   by (rule FinStateEqvStateAndEmptyOrNextFinState)
 hence 2: "\<turnstile> ( fin  (init w)); f = (((init w) \<and>  empty )\<or>  \<circle>( fin  (init w))); f" 
   by (rule LeftChopEqvChop)
 have  3: "\<turnstile> (((init w) \<and>  empty )\<or>  \<circle> (fin  (init w))); f
            = (((init w) \<and>  empty ); f \<or>  (\<circle> (fin  (init w))); f) " 
   by (rule OrChopEqv)
 have  4: "\<turnstile> ((init w) \<and>  empty ); f = ((init w) \<and> f) " 
   by (rule StateAndEmptyChop)
 have  5: "\<turnstile> (\<circle> (fin  (init w))); f = \<circle>(( fin  (init w)); f) " 
   by (rule NextChop)
 from 2 3 4 5 show ?thesis by fastforce
qed

lemma FinChopEqvDiamond:
 "\<turnstile>   ( fin  (init w) \<and> finite); f = \<diamond> ((init w) \<and> f) "
proof -
 have  1: "\<turnstile> ( fin  (init w) \<and> finite) = (finite;((init w) \<and> empty))"
   by (metis AndFinEqvChopAndEmpty int_simps(17) inteq_reflection lift_and_com) 
 hence 2: "\<turnstile> (fin  (init w) \<and> finite);f = (finite;((init w) \<and> empty));f" 
   by (rule LeftChopEqvChop)
 have  3: "\<turnstile> finite;(( (init w) \<and> empty);f)  = (finite;((init w) \<and> empty));f " 
   by (rule ChopAssoc)
 have  4: "\<turnstile> finite;(( (init w) \<and> empty);f)= \<diamond> ( ( (init w) \<and> empty);f) " 
   by (simp add: sometimes_d_def)
 have  5: "\<turnstile> ( (init w) \<and> empty);f = ((init w) \<and> f) " 
   using StateAndEmptyChop by blast 
 hence 6: "\<turnstile> \<diamond> ( ( (init w) \<and> empty);f) = \<diamond> ( (init w) \<and> f) " 
   by (rule DiamondEqvDiamond)
 from 2 3 4 6 show ?thesis by fastforce
qed

lemma NotDiamondAndNot:
 "\<turnstile> \<not>( \<diamond> ( f \<and> \<not> f))"
proof -
 have 1: "\<turnstile> (\<not>( \<diamond> ( f \<and> \<not> f))) = \<box>(\<not>(f \<and> \<not>f)) " using NotDiamondNotEqvBox by fastforce
 have 2: "\<turnstile> \<not>(f \<and> \<not>f)" by simp
 have 3: "\<turnstile> \<box>(\<not>(f \<and> \<not>f))" using "2" by (simp add: BoxGen)
 from 1 3 show ?thesis by fastforce
qed

lemma FinYields:
 "\<turnstile>    ( fin  (init w) \<and> finite)  yields  (init w) "
proof -
 have 1: "\<turnstile> (fin (init w) \<and> finite); (\<not>(init w)) = \<diamond>((init w) \<and> \<not>(init w)) " 
   by (rule FinChopEqvDiamond)
 have 2: "\<turnstile>  \<not>( \<diamond>((init w) \<and> \<not>  (init w))) " 
   by (rule NotDiamondAndNot)
 have 3: "\<turnstile> \<not> (( fin  (init w) \<and> finite); (\<not>  (init w))) " 
   using "1" "2" by fastforce
 from 3 show ?thesis by (simp add: yields_d_def)
qed

lemma ImpAndFinStateOrFinNotState:
 "\<turnstile> f \<longrightarrow> (f \<and> fin (init w)) \<or> (f \<and> fin (\<not> (init w))) "
by (simp add: itl_defs Valid_def)

lemma AndFinChopEqvStateAndChop:
 "\<turnstile>   ((f \<and> finite) \<and> fin  (init w)); g = (f \<and> finite); ((init w) \<and> g) "
proof -
 have   1: "\<turnstile> ( fin  (init w) \<and> finite) yields  (init w) " 
   by (rule FinYields)
 have   2: "\<turnstile> (f \<and> finite) \<and>  fin  (init w) \<longrightarrow>  fin  (init w) " 
   by auto
 hence  3: "\<turnstile> ( fin  (init w) \<and> finite) yields  (init w) \<longrightarrow> 
              ((f \<and> finite) \<and>  fin  (init w)) yields  (init w) " 
   using LeftYieldsImpYields 
   by (metis AndFinEqvChopAndEmpty Prop11 Prop12 inteq_reflection)
 have   4: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)) yields  (init w) " 
   using "1" "3" MP by fastforce
 have   5: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); g \<and> ((f \<and> finite) \<and>  fin  (init w)) yields (init w)
              \<longrightarrow> ((f \<and> finite) \<and>  fin  (init w)); (g \<and> (init w)) " 
   by (rule ChopAndYieldsImp)
 have   6: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); g \<longrightarrow> 
               ((f \<and> finite) \<and>  fin  (init w)); (g \<and> (init w)) " 
   using "4" "5" by fastforce
 have   7: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); (g \<and> (init w)) \<longrightarrow> (f \<and> finite); (g \<and> (init w)) " 
   by (rule AndChopA)
 have   8: "\<turnstile> g \<and> (init w) \<longrightarrow> (init w) \<and> g " 
   by auto
 hence  9: "\<turnstile> (f \<and> finite); (g \<and> (init w)) \<longrightarrow> (f \<and> finite); ((init w) \<and> g) " 
   by (rule RightChopImpChop)
 have  10: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); g \<longrightarrow> (f \<and> finite); ((init w) \<and> g) " 
   using "6" "7" "9" by fastforce
 have  11: "\<turnstile> (f \<and> finite) \<longrightarrow> ((f \<and> finite) \<and> fin (init w)) \<or> ((f \<and> finite) \<and> fin (\<not> (init w))) " 
   using ImpAndFinStateOrFinNotState by blast
 hence 12: "\<turnstile> (f \<and> finite); ((init w) \<and> g) \<longrightarrow>
              (((f \<and> finite) \<and>  fin  (init w)) \<or> 
                ((finite \<and> f) \<and> fin (\<not> (init w)) )); ((init w) \<and> g) " 
   using LeftChopImpChop 
   by (metis inteq_reflection lift_and_com)
 have  13: "\<turnstile> (((f \<and> finite) \<and>  fin  (init w)) \<or> ((f \<and> finite) \<and> fin (\<not> (init w)))); ((init w) \<and> g)
              =  
              (((f \<and> finite) \<and>  fin  (init w)); 
               ((init w) \<and> g) \<or>  ((f \<and> finite) \<and> fin (\<not>  (init w))); ((init w) \<and> g)) " 
   by (rule OrChopEqv)
 have  14: "\<turnstile> ( (f \<and> finite) \<and> fin   (init (\<not> w))); ((init w) \<and> g) \<longrightarrow> 
               \<diamond>(  (init (\<not> w)) \<and> ((init w) \<and> g)) " 
   using FinChopEqvDiamond
   by (metis AndFinEqvChopAndEmpty ChopEmpty FiniteChopImpDiamond LeftChopImpChop int_eq) 
 have 141: "\<turnstile> \<not>( \<diamond>(  (init (\<not> w)) \<and> ((init w) \<and> g))) \<longrightarrow> 
              \<not> ( ( (f \<and> finite) \<and> fin   (init (\<not> w))); ((init w) \<and> g))" 
   using "14" by fastforce
 have 142: "\<turnstile> ((init (\<not> w)) \<and> ((init w) \<and> g)) = #False"
   using Initprop(2) by fastforce
 have  15: "\<turnstile> \<not>( \<diamond>(  (init (\<not> w)) \<and> ((init w) \<and> g))) "
   by (metis "142" NotDiamondAndNot int_simps(21) inteq_reflection) 
 have 151: "\<turnstile> \<not> ( ( (f \<and> finite) \<and> fin   (init (\<not> w))); ((init w) \<and> g))" 
   using "15" "141" by fastforce
 have 1511: "\<turnstile> ( (f \<and> finite) \<and> fin (\<not>  (init w))); ((init w) \<and> g) \<longrightarrow> #False" 
   using "151"  by (metis Initprop(2) int_simps(14) inteq_reflection) 
 have 152: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); 
                ((init w) \<and> g) \<or>  ( (f \<and> finite) \<and> fin (\<not>  (init w))); ((init w) \<and> g) \<longrightarrow>
              ((f \<and> finite)\<and>  fin  (init w)); ((init w) \<and> g)" 
   using "1511" by fastforce
 have  16: "\<turnstile> (f \<and> finite); ((init w) \<and> g) \<longrightarrow> ((f \<and> finite)\<and>  fin  (init w)); ((init w) \<and> g) " 
   using "12" "13" "152"
   proof -
    have "\<turnstile> (f \<and> finite);(init w \<and> g) \<longrightarrow> 
                ((f \<and> finite) \<and> fin (init w) \<or> (f \<and> finite) \<and> fin (\<not> init w));(init w \<and> g)"
      by (metis "12" inteq_reflection lift_and_com)
    then show ?thesis
      using "13" "152" by fastforce
   qed 
 have  17: "\<turnstile> ((f \<and> finite)\<and>  fin  (init w)); ((init w) \<and> g) \<longrightarrow> ((f \<and> finite) \<and> fin (init w)); g " 
   by (rule ChopAndB)
 have  18: "\<turnstile> (f \<and> finite); ((init w) \<and> g) \<longrightarrow> ((f \<and> finite) \<and>  fin  (init w)); g " 
   using "16" "17" by fastforce
 from 10 18 show ?thesis by fastforce
qed

lemma DiAndFinEqvChopState:
 "\<turnstile>    di ((f \<and> finite) \<and>  fin  (init w)) = (f \<and> finite); (init w) "
proof -
 have  1: "\<turnstile> ((f \<and> finite) \<and> fin(init w)); #True = (f \<and> finite);((init w) \<and> #True) " 
   by (rule AndFinChopEqvStateAndChop)
 have  2: "\<turnstile> ((init w) \<and> #True) = (init w) " 
   by auto
 hence 3: "\<turnstile> ((f \<and> finite); ((init w) \<and> #True)) = ((f \<and> finite); (init w))" 
   by (rule RightChopEqvChop)
 have  4: "\<turnstile> ((f \<and> finite) \<and>  fin  (init w)); #True = (f \<and> finite); (init w) " 
   using "1" "3" by auto
 from 4 show ?thesis by (simp add: di_d_def)
qed

lemma FinNotStateEqvNotFinState:
 "\<turnstile>   (\<not>( fin   (init w)) \<and> finite ) = (fin   (init (\<not> w)) \<and> finite  )"
using FinEqvTrueChopAndEmpty Finprop(4) Initprop(2) FiniteImpAnd by (metis inteq_reflection)

lemma BiImpFinEqvYieldsState:
 "\<turnstile>   bi (f \<and> finite \<longrightarrow>  fin  (init w)) = (f  \<and> finite)yields  (init w) "
proof -
 have  1: "\<turnstile>  di ((f \<and> finite) \<and> fin (init (\<not> w))) = (f \<and> finite);   (init (\<not> w)) " 
   by (rule DiAndFinEqvChopState)
 have  2: "\<turnstile> ((f \<and> finite) \<and> fin(init (\<not> w)))  = ((f \<and> finite) \<and> \<not>(fin(init w))) " 
   using FinNotStateEqvNotFinState by fastforce
 have  3: "\<turnstile> ((f \<and> finite) \<and> \<not> (fin(init w))) = (\<not> (f \<and> finite \<longrightarrow>  fin  (init w))) " 
   by auto
 have  4: "\<turnstile> ((f \<and> finite) \<and> fin(init  (\<not> w))) = (\<not> (f \<and> finite\<longrightarrow>  fin(init w)))" 
   using "2" "3" by fastforce
 hence 5: "\<turnstile>  di ((f \<and> finite) \<and>  fin   (init (\<not> w))) =  di (\<not> (f \<and> finite\<longrightarrow> fin(init w))) " 
   by (rule DiEqvDi)
 have  6: "\<turnstile>  di (\<not> (f \<and> finite \<longrightarrow>  fin  (init w))) = (\<not>( bi (f \<and> finite\<longrightarrow> fin(init w))))" 
   by (rule DiNotEqvNotBi)
 have  7: "\<turnstile> \<not> (bi (f \<and> finite \<longrightarrow>  fin  (init w))) = (f \<and> finite);(init (\<not> w)) " 
   using "1"  "5" "6" Initprop by fastforce
 hence 8: "\<turnstile> bi (f \<and> finite \<longrightarrow>  fin  (init w)) = (\<not> ((f \<and> finite); (\<not>  (init w)))) " 
   by (metis Initprop(2) int_eq int_simps(7))
 from 8 show ?thesis by (simp add: yields_d_def)
qed

lemma StateImpYields:
 assumes " \<turnstile>   (init w) \<and> f \<and> finite \<longrightarrow>  fin  (init w1) "
 shows   "\<turnstile> (init w) \<longrightarrow> ((f \<and> finite) yields  (init w1)) "
proof -
 have  1: "\<turnstile> (init w) \<and> f \<and> finite \<longrightarrow>  fin  (init w1) " 
   using assms by auto
 hence 2: "\<turnstile> (init w) \<longrightarrow> (f \<and> finite \<longrightarrow>  fin  (init w1)) " 
   by auto
 hence 3: "\<turnstile> (init w) \<longrightarrow> bi (f \<and> finite \<longrightarrow>  fin  (init w1)) " 
   by (rule StateImpBiGen)
 have  4: "\<turnstile> bi (f \<and> finite \<longrightarrow>  fin  (init w1)) = (f \<and> finite)yields  (init w1) " 
   by (rule BiImpFinEqvYieldsState)
 from 3 4 show ?thesis by fastforce
qed

lemma StateAndYieldsImpYields:
 assumes "\<turnstile>   (init w) \<and> f \<longrightarrow> f1 "
 shows   "\<turnstile> (init w) \<and> (f1 yields  g) \<longrightarrow> (f yields  g) "
proof -
 have  1: "\<turnstile> (init w) \<and> f \<longrightarrow> f1 " using assms by auto
 hence 2: "\<turnstile> (init w) \<and> (f; (\<not>  g)) \<longrightarrow> f1; (\<not>  g) " by (rule StateAndChopImpChopRule)
 hence 3: "\<turnstile> (init w) \<and> \<not> (f1; (\<not>  g)) \<longrightarrow> \<not> (f; (\<not>  g)) " by auto
 from 3 show ?thesis by (simp add: yields_d_def)
qed

lemma AndYieldsA:
 "\<turnstile>   f yields  g \<longrightarrow> (f \<and> f1) yields  g "
proof -
 have 1: "\<turnstile> f \<and> f1 \<longrightarrow> f " by auto
 from 1 show ?thesis by (rule LeftYieldsImpYields)
qed

lemma AndYieldsB:
 " \<turnstile>   f1 yields  g \<longrightarrow> (f \<and> f1) yields  g "
proof -
 have 1: "\<turnstile> f \<and> f1 \<longrightarrow> f1 " by auto
 from 1 show ?thesis by (rule LeftYieldsImpYields)
qed

lemma RightYieldsImpYields:
 assumes "\<turnstile>   g\<longrightarrow> g1 "
 shows   "\<turnstile> (f yields  g) \<longrightarrow> (f yields  g1) "
proof -
 have  1: "\<turnstile> g \<longrightarrow> g1 " using assms by auto
 hence 2: " \<turnstile> \<not>  g1 \<longrightarrow> \<not>  g" by auto
 hence 3: "\<turnstile> f; (\<not>  g1) \<longrightarrow> f; (\<not>  g) " by (rule RightChopImpChop)
 hence 4: "\<turnstile> \<not> (f; (\<not>  g)) \<longrightarrow> \<not> (f; (\<not>  g1)) " by auto
 from 4 show ?thesis by (simp add: yields_d_def)
qed

lemma RightYieldsEqvYields:
 assumes "\<turnstile>   g = g1 "
 shows   "\<turnstile> (f yields  g) = (f yields  g1) "
proof -
 have  1: "\<turnstile> g = g1 " using assms by auto
 hence 2: "\<turnstile> (\<not>  g) = (\<not>  g1) " by auto
 hence 3: "\<turnstile> f; (\<not>  g) = f; (\<not>  g1) " by (rule RightChopEqvChop)
 hence 4: "\<turnstile> (\<not> (f; (\<not>  g))) = (\<not> (f; (\<not>  g1))) " by auto
 from 4 show ?thesis by (simp add: yields_d_def)
qed

lemma BoxImpYields:
 "\<turnstile>   \<box> g \<longrightarrow> (f \<and> finite) yields  g "
proof -
 have  1: "\<turnstile> (f \<and> finite); (\<not>  g) \<longrightarrow> \<diamond>(\<not>  g) " by (rule FiniteChopImpDiamond)
 hence 2: "\<turnstile> \<not> (\<diamond>(\<not>  g)) \<longrightarrow> \<not> ((f \<and> finite); (\<not>  g)) " by auto
 from 2 show ?thesis by (simp add: yields_d_def always_d_def)
qed

lemma BoxEqvFiniteYields:
 "\<turnstile>   \<box> f = finite yields  f "
proof -
 have  1: "\<turnstile> finite; (\<not>  f) = \<diamond> (\<not>  f) " by (rule FiniteChopEqvDiamond)
 hence 2: "\<turnstile> (\<not> (finite; (\<not>  f))) = (\<not>( \<diamond> (\<not>  f))) " by auto
 have  3: "\<turnstile> \<box> f = (\<not> ( \<diamond> (\<not>  f))) " by (simp add: always_d_def)
 have  4: "\<turnstile> \<box> f = (\<not> (finite; (\<not>  f))) " using "2" "3" by fastforce
 from 4 show ?thesis by (simp add: yields_d_def)
qed

lemma YieldsGen:
 assumes "\<turnstile>   g "
 shows   "\<turnstile> (f \<and> finite) yields  g "
proof -
  have  1: "\<turnstile> g " using assms by auto
  hence 2: "\<turnstile> \<box> g " by (rule BoxGen)
  have  3: "\<turnstile> \<box> g \<longrightarrow> (f \<and> finite) yields  g " by (rule BoxImpYields)
  from 2 3 show ?thesis using MP by fastforce
qed

lemma YieldsAndYieldsEqvYieldsAnd:
  "\<turnstile>   ((f yields  g) \<and> (f  yields  g1)) = f yields (g \<and> g1) "
proof -
 have  1: "\<turnstile> f; (\<not>  g \<or>  \<not>  g1) = ((f; (\<not>  g)) \<or>  (f; (\<not>  g1))) " 
   by (rule ChopOrEqv)
 hence 2: "\<turnstile> ((f; (\<not>  g)) \<or>  (f; (\<not>  g1))) = f; (\<not>  g \<or>  \<not>  g1) " 
   by auto
 have  3: "\<turnstile> (\<not>  g \<or>  \<not>  g1)  =  (\<not> (g \<and> g1)) " 
   by auto
 hence 4: "\<turnstile> f; (\<not>  g \<or>  \<not>  g1)  =  f; (\<not> (g \<and> g1)) " 
   by (rule RightChopEqvChop)
 have  5: "\<turnstile> (f; (\<not>  g)) \<or>  (f; (\<not>  g1)) = f; (\<not> (g \<and> g1)) " 
   using "2" "4" by fastforce
 hence 6: "\<turnstile> (\<not> (f; (\<not>  g)) \<and> \<not> (f; (\<not>  g1))) = (\<not> (f; (\<not> (g \<and> g1)))) "
    by (metis "1" "3" int_simps(14) int_simps(33) inteq_reflection) 
 from 6 show ?thesis by (simp add: yields_d_def)
qed

lemma YieldsAndYieldsImpAndYieldsAnd:
 "\<turnstile>   (f yields  g) \<and> (f1 yields  g1) \<longrightarrow> (f\<and> f1) yields (g\<and> g1) "
proof -
 have 1: "\<turnstile> f yields  g \<longrightarrow> (f \<and> f1) yields  g" 
   by (rule AndYieldsA)
 have 2: "\<turnstile> f1 yields  g1 \<longrightarrow> (f \<and> f1) yields  g1 " 
   by (rule AndYieldsB)
 have 3: "\<turnstile> ((f \<and> f1) yields  g \<and> (f \<and> f1) yields  g1) = (f \<and> f1) yields  (g \<and> g1) " 
   by (rule YieldsAndYieldsEqvYieldsAnd)
 from 1 2 3 show ?thesis by fastforce
qed

lemma YieldsYieldsEqvChopYields:
 "\<turnstile>   f yields (g yields  h) = (f; g) yields h  " 
proof -
 have  1: "\<turnstile> f; (g; (\<not>  h)) = (f; g); (\<not>  h)   " by (rule ChopAssoc)
 hence 2: "\<turnstile> f; (g; (\<not>  h)) = (f; g); (\<not>  h) " by auto
 have  3: "\<turnstile> g; (\<not>  h) = (\<not> \<not> (g; (\<not>  h))) " by auto
 hence 4: "\<turnstile> f; (g; (\<not>  h)) = f; (\<not> \<not> (g; (\<not>  h))) " by (rule RightChopEqvChop)
 have  5: "\<turnstile> f; (\<not> \<not> (g; (\<not>  h))) = (f; g); (\<not>  h) " using "2" "4" by auto
 hence 6: "\<turnstile> f; (\<not> (g yields  h)) = (f; g); (\<not>  h) " by (simp add: yields_d_def)
 hence 7: "\<turnstile> (\<not> (f; (\<not> (g yields  h)))) = (\<not> ((f; g); (\<not>  h))) " by auto
 from 7 show ?thesis by (simp add: yields_d_def)
qed

lemma EmptyYields:
 "\<turnstile>    empty   yields  f = f "
proof -
 have  1: "\<turnstile>  empty ; (\<not>  f) = (\<not>  f) " by (rule EmptyChop)
 hence 2: "\<turnstile> (\<not> ( empty ; (\<not>  f))) = f " by auto
 from 2 show ?thesis by (simp add: yields_d_def)
qed

lemma NextYields:
 "\<turnstile>   (\<circle> f) yields  g = wnext (f yields  g) "
proof -
 have  1: "\<turnstile> (\<circle> f); (\<not>  g) = \<circle>(f; (\<not>  g)) " by (rule NextChop)
 hence 2: "\<turnstile> (\<not> ((\<circle> f); (\<not>  g))) = (\<not> (\<circle>(f; (\<not>  g)))) " by auto
 hence 3: "\<turnstile> (\<circle> f) yields  g = (\<not> (\<circle>(f; (\<not>  g)))) " by (simp add: yields_d_def)
 have  4: "\<turnstile> (\<not>( \<circle>(f; (\<not>  g)))) = wnext (\<not> (f; (\<not>  g))) " by (auto simp: wnext_d_def)
 have  5: "\<turnstile> (\<circle> f) yields  g = wnext (\<not> (f; (\<not>  g))) " using "3" "4" by fastforce
 from 5 show ?thesis by (simp add: yields_d_def)
qed

lemma SkipChopEqvNext:
 " \<turnstile>    skip ; f = \<circle> f "
by (simp add: next_d_def)

lemma SkipYieldsEqvWeakNext:
 "\<turnstile>    skip  yields  f = wnext  f "
proof -
 have  1: "\<turnstile>  skip ; (\<not>  f) = \<circle>(\<not>  f) " by (rule SkipChopEqvNext)
 hence 2: "\<turnstile> (\<not> ( skip ; (\<not>  f))) = (\<not>( \<circle>(\<not>  f))) " by auto
 have  3: "\<turnstile> (\<not> (\<circle>(\<not>  f))) = wnext  f " by (auto simp: wnext_d_def)
 have  4: "\<turnstile> (\<not> ( skip ; (\<not>  f))) = wnext  f " using "2" "3" by fastforce
 from 4 show ?thesis by (simp add: yields_d_def)
qed

lemma NextImpSkipYields:
 "\<turnstile>   \<circle> f \<longrightarrow>  skip  yields  f "
proof -
 have 1: "\<turnstile> \<circle> f \<longrightarrow> wnext  f " using WnextEqvEmptyOrNext by fastforce
 have 2: "\<turnstile>  skip  yields  f = wnext  f " by (rule SkipYieldsEqvWeakNext)
 from 1 2 show ?thesis by fastforce
qed

lemma MoreEqvSkipChopTrue:
 "\<turnstile>    more  =  skip ; #True"
proof -
 have  1: "\<turnstile>  skip ; #True = \<circle>#True " by (rule SkipChopEqvNext)
 hence 2: "\<turnstile> \<circle>#True =  skip ; #True " by auto
 from 2 show ?thesis by (simp add: more_d_def)
qed

lemma MoreChopImpMore:
 "\<turnstile>    more ; f \<longrightarrow>  more "
proof -
 have 1: "\<turnstile> (\<circle>#True); f = \<circle>(#True; f)" by (rule NextChop)
 have 2: "\<turnstile> \<circle>(#True; f) \<longrightarrow>  more " by (simp add: NextImpNext more_d_def) 
 have 3: "\<turnstile> (\<circle>#True; f) \<longrightarrow>  more " using "1" "2" by fastforce
 from 3 show ?thesis by (metis more_d_def)
qed

lemma FmoreChopImpFmore:
 "\<turnstile>    fmore ; (f \<and> finite) \<longrightarrow>  fmore "
proof -
 have 1: "\<turnstile> fmore; (f \<and> finite)= \<circle>(finite; (f \<and> finite))" 
   using FmoreEqvSkipChopFinite by (metis NextChop inteq_reflection next_d_def)
 have 2: "\<turnstile> \<circle>(finite; (f \<and> finite)) \<longrightarrow>  fmore "
   by (metis ChopAndB FiniteChopFiniteEqvFinite FmoreEqvSkipChopFinite RightChopImpChop 
       inteq_reflection next_d_def) 
 have 3: "\<turnstile> (\<circle>finite; (f \<and> finite)) \<longrightarrow>  fmore " using "1" "2" 
   by (metis FmoreEqvSkipChopFinite inteq_reflection next_d_def)
 from 1 2 3 show ?thesis by fastforce
qed

lemma ChopMoreImpMore:
 "\<turnstile>   f;  more  \<longrightarrow>  more  "
proof -
 have 1: "\<turnstile> (f \<and> finite) ;  more  \<longrightarrow> \<diamond> more " 
   by (rule FiniteChopImpDiamond)
 have 11: "\<turnstile> (f \<and> inf) ;  more  \<longrightarrow> more"
   by (metis AndChopA MoreAndInfEqvInf PowerstarEqvSemhelp2 Prop12 inteq_reflection lift_and_com) 
 have 2: "\<turnstile> \<diamond> more  \<longrightarrow>  more "
   by (metis FiniteChopMoreEqvMore NowImpDiamond inteq_reflection sometimes_d_def) 
 have 3: "\<turnstile> (f \<and> finite) ;  more  \<longrightarrow> more " 
   using 1 2 by fastforce
 have 4: "\<turnstile> f  = ( (f \<and> finite)  \<or> (f \<and> inf))  "
   by (simp add: OrFiniteInf)
 hence 5: "\<turnstile> f;more  = ( (f \<and> finite);more  \<or> (f \<and> inf);more) " 
   by (simp add: OrChopEqvRule) 
 from  11 3 5 show ?thesis  by fastforce
qed

lemma MoreChopEqvNextDiamond:
 "\<turnstile>    fmore ; f = \<circle>(\<diamond> f) "
proof -
 have 1: "\<turnstile>  fmore ; f = (\<circle> finite); f "
   by (simp add: FmoreEqvSkipChopFinite LeftChopEqvChop next_d_def)
 have 2: "\<turnstile> (\<circle> finite); f = \<circle>(finite; f) " 
   by (rule NextChop)
 have 3: "\<turnstile>  fmore ; f = \<circle>(finite; f) " 
   using "1" "2" by fastforce
 from 3 show ?thesis by (simp add: sometimes_d_def)
qed

lemma WeakNextBoxImpMoreYields:
 "\<turnstile>    fmore  yields  f = wnext( \<box> f) "
proof - 
 have 1: "\<turnstile> fmore ; (\<not> f) = \<circle>(\<diamond> (\<not>f)) " by (rule MoreChopEqvNextDiamond)
 have 2: "\<turnstile> \<circle>(\<diamond> (\<not>f)) = \<circle>(\<not>(\<box>f))" by (auto simp: always_d_def)
 have 3: "\<turnstile> \<circle>(\<not>(\<box>f)) = (\<not> ( wnext( \<box> f) ))" by (auto simp: wnext_d_def)
 have 4: "\<turnstile> fmore ; (\<not> f)  = (\<not>(fmore  yields  f))" by (simp add: yields_d_def) 
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma NotEqvYieldsMore:
 "\<turnstile>   (\<not>  f) = f yields  more  "
proof -
 have  1: "\<turnstile> f;  empty  = f " by (rule ChopEmpty)
 hence 2: "\<turnstile> (\<not> (f;  empty )) = (\<not>  f) " by auto
 have  3: "\<turnstile>  empty  = (\<not>  more) " by (auto simp: empty_d_def)
 hence 4: "\<turnstile> f;  empty  = f; (\<not>  more) " by (rule RightChopEqvChop)
 hence 5: "\<turnstile> (\<not> (f;  empty )) = (\<not> (f; (\<not>  more ))) " by auto
 have  6: "\<turnstile> (\<not>  f) = (\<not> (f; (\<not>  more) )) " using "2" "5" by fastforce
 from 6 show ?thesis by (metis yields_d_def)
qed

lemma LeftChopImpMoreRule:
 assumes "\<turnstile> f \<longrightarrow>  more "
 shows   "\<turnstile> f; g \<longrightarrow>  more "
proof -
 have  1: "\<turnstile> f \<longrightarrow>  more " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow>  more ; g " by (rule LeftChopImpChop)
 have 3: " \<turnstile>  more ; g \<longrightarrow>  more " by (rule MoreChopImpMore)
 from 2 3 show ?thesis using lift_imp_trans by blast
qed

lemma LeftChopImpFMoreRule:
 assumes "\<turnstile> f \<longrightarrow>  fmore "
 shows   "\<turnstile> f; (g \<and> finite) \<longrightarrow>  fmore "
proof -
 have  1: "\<turnstile> f \<longrightarrow>  fmore " using assms by auto
 hence 2: "\<turnstile> f; (g \<and> finite) \<longrightarrow>  fmore ; (g \<and> finite) " by (rule LeftChopImpChop)
 have 3: " \<turnstile>  fmore ; (g \<and> finite) \<longrightarrow>  fmore " using FmoreChopImpFmore by fastforce
 from 2 3 show ?thesis using lift_imp_trans by blast
qed

lemma RightChopImpMoreRule:
 assumes "\<turnstile>   g \<longrightarrow>  more "
 shows   "\<turnstile> f; g \<longrightarrow>  more "
proof -
 have  1: "\<turnstile> g \<longrightarrow>  more " using assms by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> f;  more " by (rule RightChopImpChop)
 have  3: "\<turnstile> f;  more  \<longrightarrow>  more " by (rule ChopMoreImpMore)
 from 2 3 show ?thesis using lift_imp_trans by blast
qed

lemma NotDiEqvBiNot:
 "\<turnstile>   (\<not> ( di  f)) = bi (\<not>  f) "
proof -
 have  1: "\<turnstile> f = (\<not> \<not>  f) " by auto
 hence 2: "\<turnstile>  di  f =  di (\<not> \<not>  f)" by (rule DiEqvDi)
 hence 3: "\<turnstile> (\<not>  ( di  f)) = (\<not> ( di (\<not> \<not>  f)))" by auto
 from 3 show ?thesis by (simp add: bi_d_def)
qed

lemma ChopImpDi:
 "\<turnstile>   f; g \<longrightarrow>  di  f "
proof -
 have  1: "\<turnstile> g \<longrightarrow> #True " by auto
 hence 2: "\<turnstile> f; g \<longrightarrow> f; #True " by (rule RightChopImpChop)
 from 2 show ?thesis by (simp add: di_d_def)
qed

lemma TrueEqvTrueChopTrue:
 "\<turnstile>   #True = #True; #True "
proof -
 have  1: "\<turnstile> #True; #True \<longrightarrow> #True" by auto
 have  2: "\<turnstile> #True \<longrightarrow>  di #True " by (rule DiIntro)
 hence 3: "\<turnstile> #True \<longrightarrow> #True; #True " by (simp add: di_d_def)
 from 1 3 show ?thesis by auto
qed

lemma DiEqvDiDi:
 "\<turnstile>    di  f =  di (  di  f) "
proof -
 have  1: "\<turnstile> #True = #True; #True " by (rule TrueEqvTrueChopTrue)
 hence 2: "\<turnstile> f; #True = f; (#True; #True) " by (rule RightChopEqvChop)
 have  3: "\<turnstile> f; (#True; #True)= (f; #True); #True   " by (rule ChopAssoc)
 have  4: "\<turnstile> f; #True = (f; #True); #True " using "2" "3"  by fastforce 
 from 4 show ?thesis by (metis di_d_def)
qed

lemma BiEqvBiBi:
 "\<turnstile>   bi  f = bi( bi  f) "
proof -
 have  1: "\<turnstile>  di (\<not>  f) =  di(  di (\<not>  f)) " by (rule DiEqvDiDi)
 have  2: "\<turnstile>  di (\<not>  f) = (\<not> ( bi  f)) " by (rule DiNotEqvNotBi)
 hence 3: "\<turnstile>  di ( di (\<not>  f)) =  di (\<not> ( bi  f)) " by (rule DiEqvDi)
 have  4: "\<turnstile>  di (\<not>  f) =  di (\<not>(  bi  f)) " using "1" "3"  by fastforce 
 hence 5: "\<turnstile> (\<not> ( di (\<not>  f))) = (\<not> ( di (\<not> ( bi  f)))) " by fastforce
 from 5 show ?thesis by (metis bi_d_def)
qed

lemma DiOrEqv:
 "\<turnstile>    di (f \<or>  g) =  (di  f \<or>   di  g) " 
proof -
 have 1: "\<turnstile> (f\<or>  g); #True = (f; #True \<or>  g; #True) " by (rule OrChopEqv)
 from 1 show ?thesis by (simp add: di_d_def)
qed

lemma DiAndA:
 "\<turnstile>    di (f \<and> g) \<longrightarrow>  di  f "
proof -
 have 1: "\<turnstile> (f \<and> g); #True \<longrightarrow> f; #True " by (rule AndChopA)
 from 1 show ?thesis by (simp add: di_d_def)
qed

lemma DiAndB:
 "\<turnstile>    di (f \<and> g) \<longrightarrow>  di  g "
proof -
 have 1: "\<turnstile> (f \<and> g); #True \<longrightarrow> g; #True " by (rule AndChopB)
 from 1 show ?thesis by (simp add: di_d_def)
qed

lemma DiAndImpAnd:
 "\<turnstile>    di (f \<and> g) \<longrightarrow>  di  f \<and>  di  g "  
proof -
 have 1: "\<turnstile>  di (f \<and> g) \<longrightarrow>  di  f " by (rule DiAndA)
 have 2: "\<turnstile>  di (f \<and> g) \<longrightarrow>  di  g " by (rule DiAndB)
 from 1 2 show ?thesis by fastforce
qed

lemma DiSkipEqvMore:
 "\<turnstile>    di  skip  =  more "
proof -
 have 1: "\<turnstile>  skip ; #True = \<circle>#True " by (rule SkipChopEqvNext)
 have 2: "\<turnstile> \<circle>#True =  more " by (auto simp: more_d_def)
 have 3: "\<turnstile>  skip ; #True =  more " using "1" "2" by fastforce
 from 3 show ?thesis by (simp add: di_d_def)
qed

lemma DiMoreEqvMore:
 "\<turnstile>    di  more  =  more "
proof -
 have  1: "\<turnstile>  di (\<circle> #True ) = \<circle>( di #True) "
   by (rule DiNext)
 have  2: "\<turnstile> \<circle>( di  #True)  \<longrightarrow>  more "
   by (metis "1" ChopImpDi TrueEqvTrueChopTrue di_d_def int_eq more_d_def) 
 have  3: "\<turnstile>  di( \<circle> #True)  \<longrightarrow>  more " 
   using "1" "2" by fastforce
 hence 4: "\<turnstile>  di  more  \<longrightarrow>  more " 
   by (simp add: more_d_def)
 have  5: "\<turnstile>  more  \<longrightarrow>  di  more " 
   by (rule ImpDi)
 from 4 5 show ?thesis by fastforce
qed

lemma DiIfEqvRule:
 assumes "\<turnstile>   f =  if\<^sub>i  (init w)  then  g  else  h "
 shows   "\<turnstile>  di  f =  if\<^sub>i  (init w)  then  ( di  g)  else   (di  h) "
proof -
 have  1: "\<turnstile> f =  if\<^sub>i  (init w)  then  g  else  h " using assms by auto
 hence 2: "\<turnstile> f; #True =  if\<^sub>i (init w) then (g; #True) else (h; #True) " by (rule IfChopEqvRule)
 from 2 show ?thesis by (simp add: di_d_def)
qed

lemma DiEmpty:
 "\<turnstile>    di  empty "
proof -
 have 1: "\<turnstile> #True  " by auto
 have 2: "\<turnstile>  empty ; #True = #True " by (rule EmptyChop) 
 have 3: "\<turnstile>  empty ; #True " using "1" "2" by auto
 from 3 show ?thesis by (simp add: di_d_def)
qed

lemma DaNotEqvNotBa:
 "\<turnstile>    da (\<not>  f) = (\<not> ( ba  f)) "
proof -
 have 1: "\<turnstile>  ba  f = (\<not> ( da (\<not>  f))) " by (simp add: ba_d_def)
 from 1 show ?thesis by fastforce
qed

lemma DaEqvDa:
 assumes "\<turnstile>   f = g "
 shows "\<turnstile> da  f =  da  g "
using assms using int_eq by force

lemma DaEqvNotBaNot:
 "\<turnstile>    da  f = (\<not> ( ba (\<not>  f))) "
proof -
 have  1: "\<turnstile>  ba (\<not>  f) = (\<not> ( da (\<not> \<not>  f)))" by (simp add: ba_d_def)
 hence 2: "\<turnstile>  da (\<not> \<not>  f) = (\<not>(  ba (\<not>  f))) " by fastforce
 have  3: "\<turnstile> f = (\<not> \<not>  f) " by simp
 hence 4: "\<turnstile>  da  f =  da  (\<not> \<not>  f) " by (rule DaEqvDa)
 from 2 4 show ?thesis by simp
qed

lemma BaElim:
 "\<turnstile>    ba  f \<longrightarrow> f "
proof -
 have  1: "\<turnstile>  ba  f = \<box>(bi  f) " by (rule BaEqvBtBi)
 have  2: "\<turnstile> bi  f \<longrightarrow> f " by (rule BiElim)
 hence 3: "\<turnstile> \<box>(bi  f \<longrightarrow> f) " by (rule BoxGen)
 have  4: "\<turnstile> \<box>(bi  f \<longrightarrow> f) \<longrightarrow> \<box>(bi  f) \<longrightarrow> \<box> f " by (rule BoxImpDist)
 have  5: "\<turnstile> \<box>(bi  f) \<longrightarrow> \<box> f " using "3" "4" MP by fastforce
 have  6: "\<turnstile> \<box> f \<longrightarrow> f " by (rule BoxElim)
 from 1 5 6 show ?thesis  using BaImpBt lift_imp_trans by metis
qed

lemma DaIntro:
 "\<turnstile>   f \<longrightarrow>  da  f "
proof -
 have  1: "\<turnstile>  ba  (\<not>  f) \<longrightarrow> (\<not>  f) "  by (rule BaElim)
 hence 2: "\<turnstile> \<not> \<not>  f \<longrightarrow> \<not> (  ba  (\<not>  f)) " by fastforce
 have  3: "\<turnstile> f = (\<not> \<not>  f) " by simp
 have  4: "\<turnstile>  da  f = (\<not>  ( ba  (\<not>  f))) " by (rule DaEqvNotBaNot)
 from 2 3 4 show ?thesis by fastforce
qed

lemma BaGen:
 assumes "\<turnstile>   f "
 shows   "\<turnstile>  ba  f "
proof -
 have  1: "\<turnstile>   f " using assms by auto
 hence 2: "\<turnstile> \<box> f " by (rule BoxGen)
 hence 3: "\<turnstile> bi( \<box> f) " by (rule BiGen)
 have  4: "\<turnstile>  ba  f = bi (\<box> f) " by (rule BaEqvBiBt)
 from 3 4 show ?thesis by fastforce
qed

lemma BaImpDist:
 "\<turnstile>  ba (f \<longrightarrow> g) \<longrightarrow>  ba  f \<longrightarrow>  ba  g "
proof -
 have  1: "\<turnstile> bi (f \<longrightarrow> g) \<longrightarrow> (bi  f \<longrightarrow> bi  g) " 
   by (rule BiImpDist)
 hence 2: "\<turnstile> \<box>(bi (f \<longrightarrow> g) \<longrightarrow> (bi  f \<longrightarrow> bi  g)) " 
   by (rule BoxGen)
 have  3: "\<turnstile> \<box>(bi (f \<longrightarrow> g) \<longrightarrow> (bi  f \<longrightarrow> bi  g))
            \<longrightarrow>  
            (\<box> (bi (f \<longrightarrow> g)) \<longrightarrow> (\<box>(bi  f) \<longrightarrow> \<box>(bi  g))) " 
   by (meson "2" BoxImpDist MP lift_imp_trans Prop01 Prop05 Prop09)
 have  4: "\<turnstile> \<box>(bi (f \<longrightarrow> g)) \<longrightarrow> (\<box>(bi  f) \<longrightarrow> \<box>(bi  g)) " 
   using "2" "3" MP by fastforce
 have  5: "\<turnstile>  ba (f\<longrightarrow> g) = \<box>(bi (f \<longrightarrow> g)) " 
   by (rule BaEqvBtBi)
 have  6: "\<turnstile>  ba  f = \<box>(bi  f) "
   by (rule BaEqvBtBi)
 have  7: "\<turnstile>  ba  g = \<box>(bi  g) " 
   by (rule BaEqvBtBi)
 from 4 5 6 7 show ?thesis by fastforce
qed

lemma BiAndEqv:
 "\<turnstile> bi (f \<and> g) = (bi f \<and> bi g)" 
proof -
 have 1: "\<turnstile> di (\<not> f \<or> \<not> g) = (di (\<not> f) \<or> di (\<not> g))"
   by (simp add: DiOrEqv)
 have 2: "\<turnstile> (\<not> (di (\<not> f \<or> \<not> g))) = (\<not> di (\<not> f) \<and> \<not> di (\<not> g)) "
   using "1" by auto
 have 3: "\<turnstile> (f \<and> g) = (\<not> (\<not> f \<or> \<not> g))"
   by fastforce
 have 4: "\<turnstile> bi (f \<and> g) = (\<not> (di (\<not> f \<or> \<not> g)))"
   unfolding bi_d_def using 3 by (metis int_simps(4) inteq_reflection)
 from 2 4 show ?thesis unfolding bi_d_def by (metis inteq_reflection)
qed  

lemma BaAndEqv:
 "\<turnstile>    ba (f \<and> g) =  (ba  f \<and>  ba  g) "
proof -
 have  1: "\<turnstile>    ba (f \<and> g) =   \<box>(bi (f \<and> g)) " 
   by (rule BaEqvBtBi)
 have  2: "\<turnstile>  bi (f \<and> g) = (bi f \<and> bi g) "
   by (simp add: BiAndEqv)
 hence 3: "\<turnstile> \<box>(bi (f \<and> g)) = \<box>(bi f \<and> bi g) " 
   using BoxEqvBox by blast
 have  4: "\<turnstile> \<box>(bi f \<and> bi g)= (\<box>(bi f)  \<and> \<box>(bi g)) " 
   by (metis "2" BoxAndBoxEqvBoxRule inteq_reflection)
 have  5: "\<turnstile>ba  f = \<box>(bi  f) " 
   by (rule BaEqvBtBi)
 have  6: "\<turnstile>  ba  g = \<box>(bi  g) " 
   by  (rule BaEqvBtBi)
 from 1 3 4 5 6 show ?thesis by  fastforce
qed

lemma BaImpBaEqvBa:
 "\<turnstile>   ba (f = g) \<longrightarrow>  (ba  f =  ba  g) "
proof -
 have  1: "\<turnstile>  ba (f \<longrightarrow> g) \<longrightarrow>  ba  f \<longrightarrow>  ba  g " by (rule BaImpDist)
 have  2: "\<turnstile>  ba (g \<longrightarrow> f) \<longrightarrow>  ba  g \<longrightarrow>  ba  f " by (rule BaImpDist)
 have  25: "\<turnstile> (f = g) = ((f \<longrightarrow> g) \<and> (g \<longrightarrow> f))" by fastforce
 have  3: "\<turnstile> ba (f = g) = ba ((f \<longrightarrow> g) \<and> (g \<longrightarrow> f)) " by (metis "25" BaAndEqv inteq_reflection) 
 have  4: "\<turnstile> ba ((f \<longrightarrow> g) \<and> (g \<longrightarrow> f)) = (ba((f \<longrightarrow> g)) \<and> ba((g \<longrightarrow> f))) " by (rule BaAndEqv) 
 have  5: "\<turnstile> ((ba  f \<longrightarrow>  ba  g) \<and> (ba  g \<longrightarrow>  ba  f)) = (ba  f =  ba  g) " by auto
 from 1 2 3 4 5 show ?thesis  by fastforce
qed

lemma BaImpBa:
 assumes "\<turnstile>   f \<longrightarrow> g"
 shows   "\<turnstile>ba  f \<longrightarrow>  ba  g"
using BaGen BaImpDist MP assms by metis

lemma BaEqvBa: 
 assumes "\<turnstile>   f = g"
 shows   "\<turnstile> ba  f =  ba  g"
using BaGen BaImpBaEqvBa MP assms by metis

lemma DaImpDa:
 assumes "\<turnstile>   f \<longrightarrow> g"
 shows   "\<turnstile> da  f \<longrightarrow>  da  g"
using assms by (metis DaEqvDtDi DiAndB DiamondImpDiamond inteq_reflection Prop10)

lemma DiamondEqvDiamondDiamond:
 "\<turnstile> \<diamond> f = \<diamond> (\<diamond> f)"
proof -
 have 1: "\<turnstile> \<diamond> (\<diamond> f) = finite;(finite;f)" 
   by (simp add: sometimes_d_def)
 have 2: "\<turnstile> finite;(finite;f)  = (finite;finite);f" 
   by (rule ChopAssoc)
 have 3: "\<turnstile> (finite;finite);f = finite;f" 
   by (simp add: LeftChopEqvChop FiniteChopFiniteEqvFinite)
 have 4: "\<turnstile> finite;f  = \<diamond>f" 
   by (simp add: sometimes_d_def)
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma DaEqvDaDa:
  "\<turnstile>   da  f =  da (  da  f)"
proof -
 have  1: "\<turnstile>  da  f = \<diamond>( di  f)" 
   by (rule DaEqvDtDi)
 have  2: "\<turnstile>  di  f =  (di ( di  f))"  
   by (rule DiEqvDiDi)
 hence 3: "\<turnstile> \<diamond> ( di  f) = \<diamond> (di  (di  f))" 
   by (rule DiamondEqvDiamond)
 have  4: "\<turnstile> \<diamond>  (di  f) = \<diamond>(\<diamond> (di  (di  f)))" 
   using DiamondEqvDiamondDiamond DiEqvDiDi using "3" by fastforce
 have  5: "\<turnstile> \<diamond>  (di  (di  f)) =  di (\<diamond>  (di  f))"  
   by (rule DtDiEqvDiDt)
 hence 6: "\<turnstile> \<diamond>(\<diamond>  (di  (di  f))) = \<diamond> (di (\<diamond>  (di  f)))" 
   by (rule DiamondEqvDiamond)
 have  7: "\<turnstile>  da  f = \<diamond> (di( \<diamond> ( di  f)))" 
   using "1" "3" "4" "6" by fastforce
 have  8: "\<turnstile>  da (\<diamond> ( di  f)) = \<diamond>( di (\<diamond>  (di  f)))" 
   by (rule DaEqvDtDi)  
 have  9: "\<turnstile>  da ( da  f) =  da (\<diamond>  (di  f))" 
   using 1 by (rule DaEqvDa)
 from 7 8 9 show ?thesis by fastforce
qed

lemma BaEqvBaBa:
 "\<turnstile>    ba  f =  ba  (ba  f)"
proof -
 have 1: "\<turnstile>  da (\<not>  f) =  da  (da (\<not>  f))" by (rule DaEqvDaDa)  
 have 2: "\<turnstile>  da  (da (\<not>  f)) = (\<not>  (ba (\<not>  (da (\<not>  f)))))"  by (rule DaEqvNotBaNot)  
 have 3: "\<turnstile> (\<not>  (da  (da (\<not>  f)))) =  ba (\<not>  (da (\<not>  f)))" by (auto simp: ba_d_def)  
 have 4: "\<turnstile> (\<not>  (da (\<not>  f))) =  ba (\<not>  (da (\<not>  f)))" using "1" "2" "3" by fastforce
 from 4 show ?thesis by (metis ba_d_def)
qed

lemma BaLeftChopImpChop:
 "\<turnstile>    ba (f\<longrightarrow> f1) \<longrightarrow> f; g \<longrightarrow> f1; g"
proof -
 have 1: "\<turnstile>  ba (f \<longrightarrow> f1) \<longrightarrow> bi (f \<longrightarrow> f1)" by (rule BaImpBi)  
 have 2: "\<turnstile> bi (f \<longrightarrow> f1) \<longrightarrow> f; g \<longrightarrow> f1; g" by (rule BiChopImpChop)  
 from 1 2 show ?thesis by fastforce
qed

lemma BaRightChopImpChop: 
 "\<turnstile>    ba (g\<longrightarrow> g1) \<longrightarrow> f; g \<longrightarrow> f; g1"
proof -
 have 1: "\<turnstile>  ba (g \<longrightarrow> g1) \<longrightarrow> \<box>(g \<longrightarrow> g1)" by (rule BaImpBt)  
 have 2: "\<turnstile> \<box>(g \<longrightarrow> g1) \<longrightarrow> f; g \<longrightarrow> f; g1" by (rule BoxChopImpChop)  
 from 1 2 show ?thesis by fastforce
qed

lemma ChopAndBaImport:
  "\<turnstile>   (f; f1) \<and>  ba  g \<longrightarrow> (f \<and> g); (f1 \<and> g)"
proof -
 have 1: "\<turnstile>  ba  g \<and> (f; f1) \<longrightarrow> (g \<and> f); (g \<and> f1)" by (rule BaAndChopImport)  
 have 2: "\<turnstile> (g \<and> f); (g \<and> f1) = (f \<and> g); (f1 \<and> g)" by (rule AndChopAndCommute) 
 from 1 2 show ?thesis by fastforce
qed

lemma BaAndChopImportA: 
 "\<turnstile> ba f \<and> g;g1 \<longrightarrow> (f \<and> g);g1"
by (meson BaAndChopImport ChopAndB lift_imp_trans)

lemma BaAndChopImportB:
 "\<turnstile> ba f \<and> g;g1 \<longrightarrow> (f \<and> g);(ba f \<and> g1)" 
proof -
 have 1: "\<turnstile> ba f = ba (ba f)" 
   by (simp add: BaEqvBaBa)
 have 2: "\<turnstile> ba (ba f) \<and> g;g1 \<longrightarrow> g;(ba f \<and> g1)"
   by (metis  AndChopB BaAndChopImport lift_imp_trans)
 have 3: "\<turnstile> ba f \<and> g;(ba f \<and> g1) \<longrightarrow> (f \<and> g);(ba f \<and> g1)"
   by (simp add: BaAndChopImportA)
 from 1 2 3 show ?thesis by fastforce
qed
 
lemma BaImpBaImpBaAnd:
 "\<turnstile> ba h \<longrightarrow> ba(g \<longrightarrow> ba h \<and> g ) "
proof -
 have  1: "\<turnstile> ba h \<longrightarrow> (g \<longrightarrow> ba h \<and> g ) " by fastforce
 hence 2: "\<turnstile> ba(ba h) \<longrightarrow> ba(g \<longrightarrow> ba h \<and> g ) " by (rule BaImpBa)
 have  3: "\<turnstile> ba h = ba(ba h) " by (rule BaEqvBaBa)
 from 2 3 show ?thesis by fastforce
qed

lemma BaChopImpChopBa:
 "\<turnstile>    ba  f \<longrightarrow> g; g1 \<longrightarrow> g; (( ba  f) \<and> g1)"
proof - 
 have 1: "\<turnstile>  ba  f \<longrightarrow>  ba (g1 \<longrightarrow> (ba f) \<and> g1 )" by (rule BaImpBaImpBaAnd)
 have 2: "\<turnstile>  ba (g1 \<longrightarrow>  ba  f \<and> g1 ) \<longrightarrow> g; g1 \<longrightarrow> g; ( ba  f \<and> g1)" by (rule BaRightChopImpChop)  
 from 1 2 show ?thesis by fastforce
qed

lemma DiNotBaImpNotBa:
  "\<turnstile>   di (\<not> ( ba  f)) \<longrightarrow> \<not> ( ba  f)" 
proof -
 have  1: "\<turnstile>  ba  f =  ba(  ba  f)" by (rule BaEqvBaBa)  
 have  2: "\<turnstile>  ba ( ba  f) \<longrightarrow> bi ( ba  f)" by (rule BaImpBi)  
 have  3: "\<turnstile>  ba  f \<longrightarrow> bi ( ba  f)" using "1" "2" by fastforce
 hence 4: "\<turnstile>  ba  f \<longrightarrow> \<not> ( di (\<not> ( ba  f)))" by (simp add: bi_d_def)    
 from 4 show ?thesis by fastforce
qed

lemma NotBaChopImpNotBa:
 "\<turnstile>   (\<not> ( ba  f)); g \<longrightarrow> \<not> ( ba  f)"
proof -
 have 1: "\<turnstile> (\<not> ( ba  f)); g \<longrightarrow>  di (\<not> ( ba  f))" by (rule ChopImpDi)  
 have 2: "\<turnstile>  di (\<not> ( ba  f)) \<longrightarrow> \<not> ( ba  f)"  by (rule DiNotBaImpNotBa) 
 from 1 2 show ?thesis using lift_imp_trans by blast  
qed

lemma DiamondFinImpFin:
 " \<turnstile> \<diamond> (fin f) \<longrightarrow> fin f"
proof -
 have  1: "\<turnstile> fin f = #True;(f \<and> empty)" 
   by (rule FinEqvTrueChopAndEmpty)
 hence 2: "\<turnstile> \<diamond> (fin f) = finite;(#True;(f \<and> empty)) " 
   by (metis FiniteChopFiniteEqvFinite LeftChopEqvChop inteq_reflection sometimes_d_def)      
 have  3: "\<turnstile> finite;(#True;(f \<and> empty)) = (finite;#True);(f \<and> empty)" 
   by (rule ChopAssoc)
 have  4: "\<turnstile> (finite;#True);(f \<and> empty) \<longrightarrow> #True;(f \<and> empty)" 
   using "1" "2" "3" DiamondFin by fastforce
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma ChopFinImpFin:
 " \<turnstile>   (f \<and> finite);  fin  (init w) \<longrightarrow>  fin  (init w)"
proof -
 have 1: "\<turnstile> (f \<and> finite);  fin  (init w) \<longrightarrow> \<diamond> ( fin  (init w))"  by (rule  FiniteChopImpDiamond) 
 have 2: "\<turnstile> \<diamond>  (fin  (init w)) \<longrightarrow>  fin  (init w)" by (rule DiamondFinImpFin)
 from  1 2  show ?thesis using lift_imp_trans by blast   
qed

lemma FiniteRightChopEqvChop:
 assumes "\<turnstile> finite \<longrightarrow> g = g1"
 shows   "\<turnstile> finite \<longrightarrow> f;g = f;g1"
using assms by (auto simp add: Valid_def itl_defs)

lemma FinImpYieldsFin:
 "\<turnstile>    fin  (init w) \<and> finite \<longrightarrow> (f \<and> finite) yields (fin  (init w) \<and> finite)"
proof -
 have  1: "\<turnstile> (f \<and> finite);  (fin   (init (\<not> w)) \<and> finite) \<longrightarrow>  (fin   (init (\<not> w)) \<and> finite)" 
   by (metis (no_types, lifting)  ChopAndB FiniteChopEqvDiamond FiniteChopFinExportA 
       FiniteChopFiniteEqvFinite FiniteChopImpDiamond Prop12 inteq_reflection
       lift_and_com lift_imp_trans) 
 have  2: "\<turnstile> finite \<longrightarrow> (\<not> ( fin (init w) \<and> finite)) = (fin (init (\<not> w)) \<and> finite)     " 
   using FinNotStateEqvNotFinState by fastforce
 hence 3: "\<turnstile>  finite \<longrightarrow> (f \<and> finite); (\<not> ( fin (init w) \<and> finite)) = 
              (f \<and> finite); ( fin (init (\<not> w)) \<and> finite)"  
   using FiniteRightChopEqvChop[of "LIFT(\<not> ( fin (init w) \<and> finite))" 
                                  "LIFT(fin (init (\<not> w)) \<and> finite)" "LIFT(f \<and> finite)"]
   by blast
 have  4: "\<turnstile> (f \<and> finite); (\<not> ( fin  (init w) \<and> finite)) \<longrightarrow> (\<not> ( fin  (init w) \<and> finite))" 
   using "1" "2" "3" by fastforce
 hence 5: "\<turnstile>  fin  (init w) \<and> finite \<longrightarrow> \<not> ((f \<and> finite); (\<not> ( fin  (init w) \<and> finite)))" 
   by fastforce     
 from 5 show ?thesis by (simp add: yields_d_def)
qed

lemma ChopAndFin:
 " \<turnstile>  ((f; g) \<and>  (fin  (init w) \<and> finite)) = (f \<and> finite); (g \<and>  (fin  (init w) \<and> finite))"
proof -
 have  1: "\<turnstile>  fin  (init w) \<and> finite \<longrightarrow> (f \<and> finite) yields ( fin  (init w) \<and> finite)" 
   by (rule FinImpYieldsFin)  
 have 10: "\<turnstile> ((f; g) \<and>  (fin  (init w) \<and> finite)) = 
             (((f \<and> finite); (g \<and> finite)) \<and>  fin  (init w) \<and> finite)"
   using ChopAndFiniteDist[of "f" "g"] by auto 
 have 2: "\<turnstile> (f; g) \<and>  (fin  (init w) \<and> finite) \<longrightarrow> 
            ((f \<and> finite); (g \<and> finite)) \<and> (f \<and> finite) yields  ( fin  (init w) \<and> finite)" 
   using 1 10 by fastforce  
 have  3: "\<turnstile> ((f \<and> finite); (g \<and> finite)) \<and> (f \<and> finite) yields  ( fin  (init w) \<and> finite) \<longrightarrow>
              (f \<and> finite); ((g \<and> finite) \<and>  (fin  (init w) \<and> finite))" 
   using ChopAndYieldsImp by blast
 have 30: "\<turnstile> ((g \<and> finite) \<and>  (fin  (init w) \<and> finite)) = (g \<and> fin  (init w) \<and> finite)"
   by auto 
 have  4: "\<turnstile> (f; g) \<and>  (fin  (init w) \<and> finite)\<longrightarrow> (f \<and> finite); (g \<and>  fin  (init w) \<and> finite)" 
   using "2" "3" "30"  
   by (metis (mono_tags, lifting) inteq_reflection lift_imp_trans)
 have 11: "\<turnstile> (f \<and> finite); (g \<and>  fin  (init w) \<and> finite) \<longrightarrow> (f \<and> finite); (g \<and> finite)"  
   using ChopAndA by (metis "30" inteq_reflection)  
 have 12: "\<turnstile> (f \<and> finite); (g \<and>  fin  (init w) \<and> finite) \<longrightarrow> 
             (f \<and> finite);  (fin  (init w) \<and> finite)" 
   by (rule ChopAndB)  
 have 13: "\<turnstile> (f \<and> finite);  (fin  (init w) \<and> finite)\<longrightarrow> \<diamond> ( fin  (init w) \<and> finite)" 
   using FiniteChopImpDiamond  by blast
 have 14: "\<turnstile> \<diamond>( fin  (init w) \<and> finite) \<longrightarrow>  fin  (init w)"   
   by (metis ChopAndA DiamondFin inteq_reflection sometimes_d_def)
 have 15: "\<turnstile> (f \<and> finite); (g \<and>  fin  (init w) \<and> finite) \<longrightarrow> 
              ((f \<and> finite); (g \<and> finite)) \<and>  fin  (init w)" 
   using "11" "12" "13" "14" by fastforce
 from 4 15 show ?thesis by (metis ChopAndFiniteDist Prop12 int_iffI inteq_reflection)
qed

lemma ChopAndNotFin:
  " \<turnstile>   (f; g \<and> \<not> ( fin  (init w)) \<and> finite) = (f \<and> finite); (g \<and> \<not> ( fin  (init w)) \<and> finite)"
proof -
 have  1: "\<turnstile> (f; g \<and>  fin   (init (\<not> w)) \<and> finite) = 
             (f \<and> finite); (g \<and>  fin   (init (\<not> w)) \<and> finite)" 
   by (rule ChopAndFin)  
 have  2: "\<turnstile>  (fin   (init (\<not> w)) \<and> finite)= ( (\<not>  ( fin  (init w) )) \<and> finite)"   
   using FinNotStateEqvNotFinState by fastforce
 show ?thesis by (metis "1" "2" int_eq) 
qed

lemma FinChopChain:
 "\<turnstile>   (((init w )  \<and> finite \<longrightarrow>  fin  (init w1))); 
      (((init w1)  \<and> finite \<longrightarrow>  fin  (init w2)))
     \<and> finite
  \<longrightarrow> (((init w )  \<and> finite \<longrightarrow>  fin  (init w2)))"
proof -
 have  1: "\<turnstile>  (init w) \<and> finite \<and>
              ((init w) \<and> finite \<longrightarrow>  fin  (init w1)); ((init w1) \<and> finite \<longrightarrow>  fin  (init w2))
           \<longrightarrow>   
              ((init w) \<and> finite \<and>  ((init w) \<and> finite \<longrightarrow> fin  (init w1))); 
              (((init w1) \<and> finite\<longrightarrow>  fin  (init w2)) \<and> finite)"
   using ChopAndFiniteDist StateAndChopImport
   by (metis (no_types, opaque_lifting) inteq_reflection lift_and_com)          
 have  2: "\<turnstile> (init w) \<and> finite \<and> ((init w) \<and> finite \<longrightarrow>  fin(init w1)) \<longrightarrow> fin (init w1) \<and> finite" 
   by auto
 have  3: "\<turnstile> ((init w) \<and> finite \<and> ((init w) \<and> finite \<longrightarrow>  fin  (init w1))); 
             (((init w1) \<and> finite \<longrightarrow>  fin  (init w2)) \<and> finite)
           \<longrightarrow>  
           ( fin  (init w1) \<and> finite); (((init w1) \<and> finite \<longrightarrow>  fin  (init w2)) \<and> finite)" 
   using  "2" LeftChopImpChop by blast
 have  4: "\<turnstile>  ( fin  (init w1) \<and> finite); (((init w1) \<and> finite \<longrightarrow>  fin  (init w2)) \<and> finite) =
            \<diamond>((init w1)  \<and> ((init w1) \<and> finite \<longrightarrow>  fin  (init w2)) \<and> finite)" 
   using FinChopEqvDiamond by blast
 have 41: "\<turnstile> ((init w1) \<and> finite \<and> ((init w1) \<and> finite \<longrightarrow>  fin (init w2))) \<longrightarrow> fin (init w2)" 
   by auto
 have 42: "\<turnstile> \<diamond>((init w1) \<and> finite \<and> ((init w1) \<and> finite \<longrightarrow>  fin  (init w2))) \<longrightarrow> \<diamond>(fin (init w2))" 
   using "41" DiamondImpDiamond   by blast
 have  5: "\<turnstile> \<diamond>( fin( init w2)) \<longrightarrow>  fin  (init w2)" 
   using DiamondFinImpFin by blast
 have  6: "\<turnstile> (init w) \<and> finite \<and> ((init w) \<and> finite \<longrightarrow>  fin  (init w1)); 
             ((init w1) \<and> finite \<longrightarrow>  fin  (init w2))
               \<longrightarrow>  fin  (init w2)" 
   using   "1" "3" "4" "5" "42"
   using ChopAndCommute FinChopEqvDiamond by fastforce 
 from 6 show ?thesis by fastforce 
qed

lemma ChopRule:
 assumes "\<turnstile>   (init w) \<and> f \<and> finite \<longrightarrow>  fin  (init w1)"  
         "\<turnstile>   (init w1)\<and> f1 \<and> finite \<longrightarrow>  fin  (init w2)"  
 shows   "\<turnstile>   (init w) \<and> (f; f1) \<and> finite \<longrightarrow>  fin  (init w2)"
proof -
 have  1: "\<turnstile> (init w) \<and> (f; f1) \<and> finite \<longrightarrow> ((init w) \<and> f \<and> finite); (f1 \<and> finite)" 
   using  StateAndChopImport 
   by (metis ChopAndFiniteDist inteq_reflection)
 have  2: "\<turnstile> (init w) \<and> f \<and> finite \<longrightarrow>  fin  (init w1) \<and> finite" 
   using assms by auto
 hence 3: "\<turnstile> ((init w) \<and> f \<and> finite); (f1 \<and> finite) \<longrightarrow> ( fin  (init w1) \<and> finite); (f1 \<and> finite)"
   by (rule LeftChopImpChop)  
 have  4: "\<turnstile> ( fin  (init w1) \<and> finite); (f1 \<and> finite) =  \<diamond>((init w1) \<and> f1 \<and> finite)" 
   by (rule FinChopEqvDiamond)  
 have  5: "\<turnstile> (init w1) \<and> f1 \<and> finite \<longrightarrow>  fin  (init w2)" 
   using assms by auto  
 hence 6: "\<turnstile> \<diamond>((init w1) \<and> f1 \<and> finite) \<longrightarrow> \<diamond> (fin  (init w2))" 
   by (rule DiamondImpDiamond)  
 have  7: "\<turnstile> \<diamond>( fin  (init w2)) \<longrightarrow>  fin  (init w2)"  
   using DiamondFinImpFin by blast
 from 1 3 4 6 7 show ?thesis by fastforce
qed

lemma ChopRep:
 assumes "\<turnstile>  (init w) \<and> f \<and> finite \<longrightarrow> f1 \<and>  fin  (init w1)"  
         "\<turnstile>   (init w1) \<and> g \<and> finite \<longrightarrow> g1"  
 shows   "\<turnstile>   (init w) \<and> (f; g) \<and> finite \<longrightarrow> ((f1 \<and> finite); g1)"
proof -
 have  1: "\<turnstile> (init w) \<and> f \<and> finite \<longrightarrow> ((f1 \<and> finite) \<and>  fin  (init w1))" 
   using assms by auto  
 hence 2: "\<turnstile> (init w) \<and> ((f \<and> finite); (g \<and> finite))  \<longrightarrow> 
            ((f1 \<and> finite)\<and>  fin  (init w1)); (g \<and> finite)" 
   using StateAndChopImpChopRule by blast
 have  3: "\<turnstile> ((f1 \<and> finite) \<and>  fin  (init w1)); (g \<and> finite) = 
            (f1 \<and> finite); ((init w1) \<and> (g \<and> finite))" 
   using AndFinChopEqvStateAndChop   by blast
 have  4: "\<turnstile> (init w1)\<and> g \<and> finite \<longrightarrow> g1" 
   using assms by auto   
 hence 5: "\<turnstile> (f1 \<and> finite); ((init w1) \<and> g \<and> finite) \<longrightarrow> (f1 \<and> finite); g1" 
   using RightChopImpChop by blast
 from 2 3 5 show ?thesis using ChopAndFiniteDist by fastforce
qed

lemma ChopRepAndFin:
 assumes "\<turnstile>   (init w) \<and> f \<and> finite \<longrightarrow> f1 \<and>  fin  (init w1)"  
         "\<turnstile>   (init w1) \<and> g \<and> finite \<longrightarrow> g1 \<and>  fin  (init w2)"  
 shows   "\<turnstile>   (init w) \<and> (f; g) \<and> finite \<longrightarrow> ((f1 \<and> finite); g1) \<and>  fin  (init w2)"
proof -
 have  1: "\<turnstile> (init w) \<and> f \<and> finite \<longrightarrow> f1 \<and>  fin  (init w1)" 
   using assms by auto  
 have  2: "\<turnstile> (init w1) \<and> g \<and> finite \<longrightarrow> g1 \<and>  fin  (init w2)" 
   using assms by auto   
 have  3: "\<turnstile> (init w) \<and> (f; g) \<and> finite \<longrightarrow> (f1 \<and> finite); (g1 \<and>  fin  (init w2))" 
   using "1" "2" by (rule ChopRep)  
 have  4: "\<turnstile> (f1 \<and> finite); (g1 \<and>  fin  (init w2)) \<longrightarrow> (f1 \<and> finite); g1" 
   by (rule ChopAndA)  
 have  5: "\<turnstile> (f1 \<and> finite); (g1 \<and>  fin  (init w2)) \<longrightarrow> (f1 \<and> finite);  fin  (init w2)" 
   by (rule ChopAndB)  
 have  6: "\<turnstile> (f1 \<and> finite);  fin  (init w2) \<longrightarrow>  fin  (init w2)" 
   by (rule ChopFinImpFin)  
 from 1 2 3 4 5 6 show ?thesis by (meson Prop12 lift_imp_trans)
qed

lemma TrueChopMoreEqvMore:
 "\<turnstile> #True ; more = more"
by (metis ChopMoreImpMore EmptyImpFinite FiniteAndEmptyEqvEmpty FiniteChopMoreEqvMore 
          LeftChopImpChop Prop09 int_eq_true int_iffI inteq_reflection)

lemma FiniteChopFmoreEqvFmore:
 "\<turnstile> finite;fmore = fmore"
by (metis TrueChopAndFiniteEqvAndFiniteChopFinite TrueChopMoreEqvMore fmore_d_def inteq_reflection)

lemma MoreChopLoop:
 assumes "\<turnstile>   f \<longrightarrow>  fmore ; f"
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f"
proof -
 have   1: "\<turnstile> f \<longrightarrow>  fmore ; f" 
   using assms by auto 
 hence 11: "\<turnstile> \<diamond> (f) \<longrightarrow> \<diamond> (fmore;f)" 
   using DiamondImpDiamond by blast
 have  12: "\<turnstile> \<diamond> (fmore;f) = finite;(fmore;f)" 
   by (simp add: sometimes_d_def)
 have  13: "\<turnstile> finite;(fmore;f) = (finite;fmore);f" 
   by (rule ChopAssoc)
 have  14: "\<turnstile>  \<diamond> (fmore;f) = fmore;f" 
   using FiniteChopFmoreEqvFmore "12" "13"  by (metis int_eq)
 have   2: "\<turnstile>  fmore ; f = \<circle>(\<diamond> f)" 
   using MoreChopEqvNextDiamond by blast  
 have   3: "\<turnstile>  \<diamond> (f) \<longrightarrow> \<circle>(\<diamond> f)"  
   using "11" "14" "2" by fastforce
 hence  4: "\<turnstile> finite \<longrightarrow> \<not> (\<diamond> f)" 
   using NextLoop by blast
 have   5: "\<turnstile> \<not> (\<diamond> f) \<longrightarrow> \<not> f" 
   using NowImpDiamond by fastforce
 from 4 5 show ?thesis using lift_imp_trans by blast
qed

lemma MoreChopContra:
 assumes "\<turnstile>   f \<and> \<not>  g \<longrightarrow> ( fmore ; (f \<and> \<not>  g))"
 shows   "\<turnstile> f  \<and> finite \<longrightarrow> g "
proof -
 have  1: "\<turnstile> f \<and> \<not>  g \<longrightarrow> ( fmore ; (f \<and> \<not>  g))" using assms by auto
 hence 2: "\<turnstile> finite \<longrightarrow> \<not> (f \<and> \<not>  g)" by (rule MoreChopLoop)  
 from 2 show ?thesis by auto 
qed

lemma MoreChopLoopFinite:
 assumes "\<turnstile>   f \<and> finite \<longrightarrow>  fmore ; f"
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f"
proof -
 have   1: "\<turnstile> f \<and> finite \<longrightarrow>  fmore ; f" 
   using assms by auto 
 hence 11: "\<turnstile> \<diamond> (f \<and> finite) \<longrightarrow> \<diamond> (fmore;f)" 
   using DiamondImpDiamond by blast
 have  12: "\<turnstile> \<diamond> (fmore;f) = finite;(fmore;f)" 
   by (simp add: sometimes_d_def)
 have  13: "\<turnstile> finite;(fmore;f) = (finite;fmore);f" 
   by (rule ChopAssoc)
 have  14: "\<turnstile>  \<diamond> (fmore;f) = fmore;f" 
   using FiniteChopFmoreEqvFmore "12" "13"  by (metis int_eq)
 have   2: "\<turnstile>  fmore ; f  = \<circle>(\<diamond> f)" 
   using MoreChopEqvNextDiamond by blast  
 have   3: "\<turnstile>  \<diamond> (f \<and> finite)  \<longrightarrow> \<circle>(\<diamond> f)"  
   using "11" "14" "2" by fastforce
 have 31: "\<turnstile> \<diamond> (f \<and> finite) = ((\<diamond> f) \<and> finite)"
   by (metis (no_types, lifting) "3" ChopAndB ChopAndNotChopImp DiamondDiamondEqvDiamond 
       DiamondIntroC FiniteChopFiniteEqvFinite FiniteChopInfEqvInf Prop11 Prop12 finite_d_def 
       inteq_reflection sometimes_d_def)
 have 32: "\<turnstile> (\<diamond> f) \<and> finite \<longrightarrow> \<circle>(\<diamond> f)"
   using "3" "31" by fastforce
 hence  4: "\<turnstile> finite \<longrightarrow> \<not> (\<diamond> f)" 
   by (metis (no_types, lifting) DiamondIntro FiniteChopInfEqvInf InfEqvNotFinite Prop09 
      finite_d_def int_simps(15) int_simps(32) inteq_reflection sometimes_d_def)
 have   5: "\<turnstile> \<not> (\<diamond> f) \<longrightarrow> \<not> f" 
   by (simp add: NowImpDiamond)
 from 4 5 show ?thesis using lift_imp_trans by fastforce
qed

lemma MoreChopEqvFmoreOrInf:
 "\<turnstile> more ; f = ( (fmore;f) \<or> inf)"
 by (metis AndInfChopEqvAndInf MoreAndInfEqvInf OrChopEqv OrFiniteInf fmore_d_def int_eq)

lemma MoreChopLoopFiniteB:
 assumes "\<turnstile>   f  \<longrightarrow>  more ; f"
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f"
proof -
 have   1: "\<turnstile> f \<longrightarrow>  more ; f" 
   using assms by auto 
 have  10: "\<turnstile> f  \<longrightarrow> (fmore;f) \<or> inf" 
   using MoreChopEqvFmoreOrInf  assms by fastforce
 hence 100: "\<turnstile> f  \<and> finite \<longrightarrow> (fmore;f)"
   by (simp add: Prop13 finite_d_def)
 hence 11: "\<turnstile> \<diamond> (f \<and> finite) \<longrightarrow> \<diamond> (fmore;f)" 
   using DiamondImpDiamond by blast
 have  12: "\<turnstile> \<diamond> (fmore;f) = finite;(fmore;f)" 
   by (simp add: sometimes_d_def)
 have  13: "\<turnstile> finite;(fmore;f) = (finite;fmore);f" 
   by (rule ChopAssoc)
 have  14: "\<turnstile>  \<diamond> (fmore;f) = fmore;f" 
   using FiniteChopFmoreEqvFmore "12" "13"  by (metis int_eq)
 have   2: "\<turnstile>  fmore ; f  = \<circle>(\<diamond> f)" 
   using MoreChopEqvNextDiamond by blast  
 have   3: "\<turnstile>  \<diamond> (f \<and> finite)  \<longrightarrow> \<circle>(\<diamond> f)"  
   using "11" "14" "2" by fastforce
 have 31: "\<turnstile> \<diamond> (f \<and> finite) = ((\<diamond> f) \<and> finite)"
   by (metis (no_types, opaque_lifting) ChopAndA ChopAndB ChopAndNotChopImp FiniteChopFiniteEqvFinite
       FiniteChopInfEqvInf Prop11 Prop12 finite_d_def inteq_reflection sometimes_d_def)
 have 32: "\<turnstile> (\<diamond> f) \<and> finite \<longrightarrow> \<circle>(\<diamond> f)"
   using "3" "31" by fastforce
 hence  4: "\<turnstile> finite \<longrightarrow> \<not> (\<diamond> f)" 
   by (metis (no_types, lifting) DiamondIntro FiniteChopInfEqvInf InfEqvNotFinite Prop09 
       finite_d_def int_simps(15) int_simps(32) inteq_reflection sometimes_d_def)
 have   5: "\<turnstile> \<not> (\<diamond> f) \<longrightarrow> \<not> f" 
   by (simp add: NowImpDiamond)
 from 4 5 show ?thesis using lift_imp_trans by fastforce
qed

lemma MoreChopContraFinite:
 assumes "\<turnstile>   (f \<and> \<not>  g) \<and> finite \<longrightarrow> ( fmore ; (f \<and> \<not>  g))"
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g "
proof -
 have  1: "\<turnstile> (f \<and> \<not>  g) \<and> finite \<longrightarrow> ( fmore ; (f \<and> \<not>  g))" using assms by auto
 hence 2: "\<turnstile> finite \<longrightarrow> \<not> (f \<and> \<not>  g)" using MoreChopLoopFinite by (simp add: MoreChopLoopFinite)
 from 2 show ?thesis by (simp add: Valid_def)
qed

lemma MoreChopContraFiniteB:
 assumes "\<turnstile>   (f \<and> \<not>  g)  \<longrightarrow> ( more ; (f \<and> \<not>  g))"
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g "
proof -
 have  1: "\<turnstile> (f \<and> \<not>  g)  \<longrightarrow> ( more ; (f \<and> \<not>  g))" using assms by auto
 hence 2: "\<turnstile> finite \<longrightarrow> \<not> (f \<and> \<not>  g)" using MoreChopLoopFinite by (simp add: MoreChopLoopFiniteB)
 from 2 show ?thesis by (simp add: Valid_def )
qed

lemma ChopLoop:
 assumes "\<turnstile>   f \<longrightarrow> g;f" 
         "\<turnstile> g \<longrightarrow>  fmore"
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f"
proof -
have  1: "\<turnstile> f \<longrightarrow> g; f" using assms by auto  
have  2: "\<turnstile> g \<longrightarrow>  fmore" using assms by auto  
hence 3: "\<turnstile> g; f \<longrightarrow>  fmore ; f" by (rule LeftChopImpChop)  
have  4: "\<turnstile> f \<longrightarrow>  fmore ; f" using "1" "3" by fastforce 
from 4 show ?thesis using MoreChopLoop by auto
qed

lemma ChopLoopB:
 assumes "\<turnstile>   f \<longrightarrow> g;f" 
         "\<turnstile> g \<longrightarrow>  more"
 shows   "\<turnstile> finite \<longrightarrow> \<not>  f"
proof -
have  1: "\<turnstile> f \<longrightarrow> g; f" using assms by auto  
have  2: "\<turnstile> g \<longrightarrow>  more" using assms by auto  
hence 3: "\<turnstile> g; f \<longrightarrow>  more ; f" by (rule LeftChopImpChop)  
have  4: "\<turnstile> f \<longrightarrow>  more ; f" using "1" "3" by fastforce 
from 4 show ?thesis using MoreChopLoopFiniteB by auto
qed

lemma ChopContra:
 assumes  "\<turnstile>   f \<and> \<not>  g \<longrightarrow> h; f \<and> \<not> (h; g)"
          "\<turnstile> h \<longrightarrow>  fmore" 
 shows    "\<turnstile> f  \<and> finite \<longrightarrow> g "
proof -
have  1: "\<turnstile> f \<and> \<not>  g \<longrightarrow> h; f \<and> \<not> (h; g)" using assms by auto  
have  2: "\<turnstile> h \<longrightarrow>  fmore" using assms by auto   
have  3: "\<turnstile> h; f \<and> \<not> (h; g) \<longrightarrow> h; (f \<and> \<not>  g)" by (rule ChopAndNotChopImp)  
have  4: "\<turnstile> h; (f \<and> \<not>  g) \<longrightarrow>  fmore ; (f \<and> \<not>  g)" using "2" by (rule LeftChopImpChop)  
have  5: "\<turnstile> f \<and> \<not>  g \<longrightarrow>  fmore ; (f \<and> \<not>  g)" using "1" "3" "4" by fastforce  
from 5 show ?thesis using MoreChopContra by auto
qed

lemma ChopContraB:
 assumes  "\<turnstile> f \<and> \<not>  g \<longrightarrow> h; f \<and> \<not> (h; g)"
          "\<turnstile> h \<longrightarrow>  more" 
 shows    "\<turnstile> f \<and> finite \<longrightarrow> g "
proof -
have  1: "\<turnstile> f \<and> \<not>  g \<longrightarrow> h; f \<and> \<not> (h; g)" using assms by auto  
have  2: "\<turnstile> h \<longrightarrow>  more" using assms by auto   
have  3: "\<turnstile> h; f \<and> \<not> (h; g) \<longrightarrow> h; (f \<and> \<not>  g)" by (rule ChopAndNotChopImp)  
have  4: "\<turnstile> h; (f \<and> \<not>  g) \<longrightarrow>  more ; (f \<and> \<not>  g)" using "2" by (rule LeftChopImpChop)  
have  5: "\<turnstile> f \<and> \<not>  g \<longrightarrow>  more ; (f \<and> \<not>  g)" using "1" "3" "4" by fastforce  
from 5 show ?thesis using MoreChopContraFiniteB by auto
qed


subsection \<open>Properties of Chopstar and Chopplus\<close>

lemma FPowerstardef:
 "\<turnstile> fpowerstar f = (\<exists> n. power f n)"
by (simp add: fpowerstar_d_def)

lemma Powerstardef:
 "\<turnstile> powerstar f = (fpowerstar f);(empty \<or> (f \<and> inf))"
by (simp add: fpowerstar_d_def powerstar_d_def)

lemma Chopstardef:
 "\<turnstile> chopstar f = powerstar (f \<and> more)"
by (simp add: chopstar_d_def)

lemma AndEmptyChopAndEmptyEqvAndEmpty:
 "\<turnstile> (f \<and> empty);(f \<and> empty) = (f \<and> empty)"
proof -
 have 1: "\<turnstile> (f \<and> empty);(f \<and> empty) \<longrightarrow> (f \<and> empty)"
   by (metis ChopAndB ChopEmpty int_eq)
 have 2: "\<turnstile> (f \<and> empty) \<longrightarrow> (f \<and> empty);(f \<and> empty)"
   by (auto simp add: Valid_def itl_defs)
     (metis iless_Suc_eq less_numeral_extra(1) ntake_0 ntake_all one_eSuc zero_enat_def)
 show ?thesis
   by (simp add: "1" "2" int_iffI) 
qed

lemma PowerCommute:
 "\<turnstile> (f \<and> finite) ;power f n = power f n;(f \<and> finite)" 
proof
 (induct n)
 case 0
 then show ?case 
 by (metis ChopEmpty EmptyChop inteq_reflection power_d.pow_0)
 next
 case (Suc n)
 then show ?case 
 by (metis ChopAssoc inteq_reflection power_d.pow_Suc)
qed

lemma ChopInductL:
 assumes "\<turnstile> g \<or> f;h \<longrightarrow> h"
 shows   "\<turnstile> (power f n);g \<longrightarrow> h"
proof 
 (induct n)
 case 0
 then show ?case using EmptyChop assms
 by (metis MP Prop12 int_eq int_iffD1 int_simps(33) pow_0) 
 next
 case (Suc n)
 then show ?case using assms
 by (metis AndChopA ChopAssoc Prop05 Prop11 RightChopImpChop lift_imp_trans pow_Suc) 
qed

lemma ChopInductFiniteL:
 assumes "\<turnstile> g \<or> (f \<and> finite);h \<longrightarrow> h"
 shows   "\<turnstile> (power f n);g \<longrightarrow> h"
proof 
 (induct n)
 case 0
 then show ?case using EmptyChop assms
 by (metis MP Prop12 int_eq int_iffD1 int_simps(33) pow_0) 
 next
 case (Suc n)
 then show ?case using assms 
  by (metis ChopAndB ChopAssoc Prop05 Prop10 inteq_reflection lift_imp_trans pow_Suc)
qed

lemma ChopInductFiniteMoreL:
 assumes "\<turnstile> g \<or> ((f \<and> more) \<and> finite);h \<longrightarrow> h"
 shows   "\<turnstile> (power f n);g \<longrightarrow> h"
proof 
 (induct n)
 case 0
 then show ?case using assms by (metis ChopInductFiniteL pow_0)
 next
 case (Suc n)
 then show ?case 
 proof -
 have 1: "\<turnstile>  power f (Suc n);g  = ((f \<and> finite);power f n);g " 
   by simp
 have 2: "\<turnstile> ((f \<and> finite);power f n);g = (f \<and> finite);((power f n);g)"
   by (meson ChopAssoc Prop11)
 have 3: "\<turnstile> (f \<and> finite);((power f n);g) \<longrightarrow> (f \<and> finite);h"
   by (simp add: RightChopImpChop Suc.hyps)
 have 31: "\<turnstile> f = ((f \<and> more) \<or> (f \<and> empty))"
   unfolding empty_d_def by fastforce
 have 32: "\<turnstile> (f \<and> finite) = ( ( (f \<and> more) \<and> finite) \<or> ((f \<and> empty) \<and> finite))"
   using 31 by fastforce
 have 4: "\<turnstile> (f \<and> finite);h = (( (f \<and> more) \<and> finite); h \<or> ((f \<and> empty) \<and> finite);h)"
   using 32 OrChopEqvRule by blast
 have 5: "\<turnstile> ( (f \<and> more) \<and> finite); h \<longrightarrow> h" 
   using assms by auto
 have 6: "\<turnstile> ( (f \<and> empty) \<and> finite); h \<longrightarrow> h" 
   by (metis AndChopA AndChopB EmptyChop inteq_reflection lift_imp_trans)
 from "5" "6" "4" "3" "2" "1" show ?thesis by fastforce
 qed
qed

lemma ChopInductInfL:
 assumes "\<turnstile> g \<or> f;h \<longrightarrow> h"
 shows   "\<turnstile> ((power f n);(f \<and> inf));g \<longrightarrow> h"
proof
 (induct n)
 case 0
 then show ?case using assms
 by (metis (no_types, lifting) AndInfChopEqvAndInf ChopAssoc ChopInductFiniteL PowerstarEqvSemhelp3
     Prop03 Prop10 Prop12 inteq_reflection)
 next
 case (Suc n)
 then show ?case using assms
 proof -
 have "\<turnstile>(f \<and> finite);(power f n;((f \<and> inf);g)) = (power f (Suc n);(f \<and> inf));g"
   by  (metis ChopAssoc inteq_reflection pow_Suc)
 then show ?thesis
   by (metis (no_types, lifting) AndChopA ChopAndB ChopAssoc Prop03 Prop10 Suc assms int_eq lift_imp_trans) 
 qed 
qed

lemma ChopInductInfMoreL:
 assumes "\<turnstile> g \<or> f;h \<longrightarrow> h"
 shows   "\<turnstile> ((power f n);((f \<and> more) \<and> inf));g \<longrightarrow> h"
using ChopInductInfL 
by (metis AndMoreAndInfEqvAndInf assms inteq_reflection)

lemma ChopInductR:
 assumes "\<turnstile> g \<or> h;f \<longrightarrow> h"
 shows   "\<turnstile> g;(power f n) \<longrightarrow> h"
proof
 (induct n)
 case 0
 then show ?case using ChopEmpty assms
 by (metis MP Prop12 int_iffD2 int_simps(33) inteq_reflection pow_0)
 next
 case (Suc n)
 then show ?case using assms 
 proof -
 have 1: "\<turnstile> power f n;(f \<and> finite) = power f (Suc n)"
   using PowerCommute by fastforce 
 have 2: "\<turnstile> h;(finite \<and> f) \<longrightarrow> h"
   by (metis (no_types) ChopAndA Prop05 assms inteq_reflection lift_and_com lift_imp_trans)
 then show ?thesis
   by (metis "1" AndChopB ChopAssoc Prop10 Suc int_eq lift_and_com lift_imp_trans) 
 qed
qed

lemma ChopInductInfR:
 assumes "\<turnstile> g \<or> h;f \<longrightarrow> h"
 shows   "\<turnstile> g;((power f n);(f \<and> inf)) \<longrightarrow> h"
using assms
by (metis ChopAndA ChopAssoc ChopInductR LeftChopImpChop Prop05 int_iffD1 lift_imp_trans) 

lemma ChopExistPower:
 "\<turnstile> (g;(\<exists>n. power f n)) = (\<exists> n. g;power f n)"
using ChopExist by fastforce

lemma ExistChopPower:
 "\<turnstile> (\<exists>n. (power f n);g) = (\<exists>n. power f n);g"
using ExistChop by fastforce

lemma PowerStarCommute:
 "\<turnstile> (f \<and> finite);(\<exists> n. power f n) = (\<exists> n. power f n);(f \<and> finite)"
proof - 
 have 1: "\<turnstile> (f \<and> finite) ;(\<exists> n. power f n) =
            (\<exists> n. (f \<and> finite) ;power f n)  "
   using ChopExistPower by blast
 have 2: "\<turnstile> (\<exists> n. (f \<and> finite) ;power f n) = 
            (\<exists> n. (power f n);(f \<and> finite) )"
   using PowerCommute by fastforce
 have 3: "\<turnstile> (\<exists> n. (power f n);(f \<and> finite)) =
            (\<exists> n. (power f n));(f \<and> finite) "
   using ExistChopPower by blast
 from 1 2 3  show ?thesis by fastforce
qed

lemma PowerSucAndEmptyEqvAndEmpty:
 "\<turnstile> (power (f \<and> empty) (Suc n)) = (f \<and> empty)"
proof
 (induct n)
 case 0
 then show ?case using ChopEmpty
 by (metis (no_types, lifting) FiniteAndEmptyEqvEmpty Prop10 Prop12 int_iffD1 inteq_reflection 
     pow_0 pow_Suc) 
 next
 case (Suc n)
 then show ?case 
  by (metis AndEmptyChopAndEmptyEqvAndEmpty EmptyImpFinite Prop10 Prop12 int_iffD1 
      inteq_reflection pow_Suc)
qed

lemma FiniteOr:
 "\<turnstile> ((f \<or> g) \<and> finite) = ((f \<and> finite) \<or> (g \<and> finite))"
by auto

lemma PowerOr:
 "\<turnstile>  (power (f \<or> g) (Suc n)) = ( ((f \<and> finite);power (f \<or> g) n) \<or> 
                                 ((g \<and> finite);power (f \<or> g) n))"
by (simp add: FiniteOr OrChopEqvRule)

lemma PowerEmptyOrMore:
 "\<turnstile> (power ( (f \<and> empty) \<or> (f \<and> more)) (Suc n)) =
    ((f \<and> empty);(power ( (f \<and> empty) \<or> (f \<and> more)) n) \<or> 
     (f \<and> fmore );(power ( (f \<and> empty) \<or> (f \<and> more)) n))"
proof -
have f2: "\<turnstile>(f \<and> empty) = (empty \<and> f)"
  by (meson lift_and_com) 
have f3: "\<turnstile>((f \<and> empty) \<and> finite) = (finite \<and> f \<and> empty)"
  by (meson lift_and_com) 
have f4: "\<turnstile>(empty \<and> f) = (finite \<and> f \<and> empty)"
  using FiniteAndEmptyEqvEmpty by auto
have "\<turnstile>((f \<and> more) \<and> finite) = (f \<and> fmore)"
  by (meson AndMoreAndFiniteEqvAndFmore) 
then show ?thesis
  using f4 f3 f2 by (metis PowerOr inteq_reflection)
qed

lemma PSEqvEmptyOrChopPS:
 "\<turnstile> powerstar f = (empty \<or> f;powerstar f)"
using PowerstarEqvSem Valid_def by blast

lemma EmptyImpCS:
 "\<turnstile> empty  \<longrightarrow> f\<^sup>\<star>"
proof -
 have 1: "\<turnstile> f\<^sup>\<star> = (empty \<or> (f \<and> more);f\<^sup>\<star>)" by (rule ChopstarEqv) 
 have 2: "\<turnstile> empty \<longrightarrow> empty \<or> (f \<and> more);f\<^sup>\<star>" by auto
 from 1 2 show ?thesis by fastforce
qed

lemma CSEqvOrChopCS:
 "\<turnstile>   f\<^sup>\<star> =  (empty  \<or>  (f; f\<^sup>\<star>)) " 
proof -
 have 1: "\<turnstile> f\<^sup>\<star> = (empty \<or> (f \<and> more);f\<^sup>\<star>)" by (rule ChopstarEqv)
 have 2: "\<turnstile> (f \<and> more);f\<^sup>\<star> \<longrightarrow> f;f\<^sup>\<star>" by (rule AndChopA)
 have 3: "\<turnstile> f\<^sup>\<star> \<longrightarrow>  empty  \<or>  f; f\<^sup>\<star> "using "1" "2" by (metis int_iffD1 Prop08)
 have 4: "\<turnstile> empty  \<longrightarrow> f\<^sup>\<star>" by (rule EmptyImpCS)
 have 5: "\<turnstile> f \<longrightarrow> empty \<or> (f \<and> more)" by (auto simp: empty_d_def) 
 have 6: "\<turnstile> f; f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> \<or>  (f \<and>  more ); f\<^sup>\<star>" using "5" by (rule EmptyOrChopImpRule)
 have 7: "\<turnstile> f\<^sup>\<star> \<longrightarrow> empty \<or> (f \<and> more);f\<^sup>\<star>" using "1" by fastforce
 have 8: "\<turnstile> f; f\<^sup>\<star> \<longrightarrow> empty \<or>  (f \<and>  more ); f\<^sup>\<star>"  using "6" "7" by fastforce
 hence 9: "\<turnstile> f; f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>" using "1" by fastforce
 have 10: "\<turnstile> empty  \<or>  f; f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> " using  "9" "4" by fastforce
 from 3 10 show ?thesis by fastforce
qed

lemma PowerChopCommute:
 "\<turnstile> ((f \<and> more) \<and> finite);power (f \<and> more) n = power (f \<and> more) n;((f \<and> more) \<and> finite)" 
using PowerCommute by auto

lemma ChopExist:
 "\<turnstile> (g;(\<exists>n. power (f \<and> more) n)) = (\<exists> n. g;power (f \<and> more) n)"
using ChopExistPower by auto

lemma ExistChop:
 "\<turnstile> (\<exists>n. (power (f \<and> more) n);g) = (\<exists>n. power (f \<and> more) n);g"
using ExistChopPower by auto

lemma FPowerstarInductL:
 assumes "\<turnstile> g \<or> (f \<and> finite);h \<longrightarrow> h"
 shows   "\<turnstile> (fpowerstar f);g \<longrightarrow> h"
proof -
 have 1: "\<turnstile> (fpowerstar f);g = (\<exists> n. power f n);g"
   by (simp add: fpowerstar_d_def LeftChopEqvChop)
 have 2: "\<turnstile> (\<exists> n. power f n);g = 
            (\<exists> n. (power f n);g) "
   using ExistChopPower by fastforce
 have 3: "\<And> n. \<turnstile> (power f n);g \<longrightarrow> h" 
   using ChopInductFiniteL assms by blast
 have 4: "\<turnstile> (\<exists> n. ((power f n));g) \<longrightarrow> h"
   using "3"  by (simp add: Valid_def) blast
 from 1  2 4 show ?thesis by (metis inteq_reflection)
qed 

lemma FPowerstarInductMoreL:
 assumes "\<turnstile> g \<or> ((f \<and> more) \<and> finite);h \<longrightarrow> h"
 shows   "\<turnstile> (fpowerstar f);g \<longrightarrow> h"
proof -
 have 1: "\<turnstile> (fpowerstar f);g = (\<exists> n. power f n);g"
   by (simp add: fpowerstar_d_def LeftChopEqvChop)
 have 2: "\<turnstile> (\<exists> n. power f n);g = 
            (\<exists> n. (power f n);g) "
   using ExistChopPower by fastforce
 have 3: "\<And> n. \<turnstile> (power f n);g \<longrightarrow> h" 
   using  ChopInductFiniteMoreL assms by blast
 have 4: "\<turnstile> (\<exists> n. ((power f n));g) \<longrightarrow> h"
   using "3"  by (simp add: Valid_def) blast
 from 1  2 4 show ?thesis by (metis inteq_reflection)
qed

lemma PowerstarInductL:
 assumes "\<turnstile> g \<or> f;h \<longrightarrow> h"
 shows   "\<turnstile> (powerstar f);g \<longrightarrow> h"
proof -
 have 1: "\<turnstile> (powerstar f);g = ((\<exists> n. power f n);(empty \<or> (f \<and> inf)));g"
   by (simp add: powerstar_d_def LeftChopEqvChop)
 have 11: "\<turnstile> ((\<exists> n. power f n);(empty \<or> (f \<and> inf)));g =
             (\<exists> n. power f n);((empty \<or> (f \<and> inf));g)"
   by (meson ChopAssoc Prop11)
 have 2: "\<turnstile> (\<exists> n. power f n);((empty \<or> (f \<and> inf));g) = 
            (\<exists> n. ((power f n);((empty \<or> (f \<and> inf));g))) "
   using ExistChopPower by fastforce
 have 3: "\<And> n. \<turnstile> (power f n);g \<longrightarrow> h" 
   using ChopInductL assms by blast
 have 31:"\<And> n. \<turnstile> ((power f n);(f \<and> inf));g \<longrightarrow> h"
   using ChopInductInfL assms by blast
 have 33: "\<And>n. \<turnstile> ((power f n);g \<or> ((power f n);(f \<and> inf));g) =
                   ((power f n);(empty \<or> (f \<and> inf)));g"
   by (metis ChopEmpty ChopOrEqv OrChopEqv inteq_reflection)
 have 32: "\<And> n. \<turnstile> ((power f n);(empty \<or> (f \<and> inf)));g \<longrightarrow> h "  
   by (metis "3" "31" "33" Prop02 inteq_reflection) 
 have 4: "\<turnstile> (\<exists> n. ((power f n);(empty \<or> (f \<and> inf)));g) \<longrightarrow> h"
   using "32" by fastforce
 from 1 11 2 4 show ?thesis
   by (metis Semantics.ExistChop inteq_reflection)
qed 

lemma ChopstarInductL:
 assumes "\<turnstile> g \<or> f;h \<longrightarrow> h"
 shows   "\<turnstile> (chopstar f);g \<longrightarrow> h"
proof -
 have 1: "\<turnstile> (chopstar f);g = ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g"
   by (simp add: chopstar_d_def powerstar_d_def LeftChopEqvChop)
 have 11: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g =
             (\<exists> n. power (f \<and> more) n);((empty \<or> ((f \<and> more) \<and> inf));g)  "
   by (meson ChopAssoc Prop11)
 have 2: "\<turnstile> (\<exists> n. power (f \<and> more) n);((empty \<or> ((f \<and> more) \<and> inf));g) = 
            (\<exists> n. (power (f \<and> more) n);(((empty \<or> ((f \<and> more) \<and> inf));g)))"
   using ExistChopPower by fastforce
 have 21: "\<turnstile> g \<or> (f \<and> more);h \<longrightarrow> h" 
   using AndChopA Prop03 Prop10 assms int_simps(33) inteq_reflection by fastforce
 have 3: "\<And> n. \<turnstile> (power (f \<and> more) n);g \<longrightarrow> h" 
   using "21" ChopInductL[of "g" "LIFT(f \<and> more)" "h"] assms by auto
 have 31: "\<And> n. \<turnstile> ((power (f \<and> more) n);((f \<and> more) \<and> inf));g \<longrightarrow> h" 
   using assms 
   by (metis (no_types, lifting) AndChopA ChopInductInfL Prop03 Prop10 int_eq_true int_simps(33) 
            inteq_reflection lift_imp_trans)
 have 32: "\<And> n. \<turnstile> ((power (f \<and> more) n);g \<or> ((power (f \<and> more) n);((f \<and> more) \<and> inf));g) =
                   ((power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g"
   by (meson ChopAssoc ChopOrEqvRule EmptyOrChopEqv Prop04 Prop06)
 have 41: "\<And>n. \<turnstile> ((power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g \<longrightarrow> h"
   by (metis "3" "31" "32" Prop02 inteq_reflection)
 have 4: "\<turnstile> (\<exists> n. ((power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g) \<longrightarrow> h"
   using "41" by fastforce
 from 1 11 2 4 show ?thesis 
   by (metis Semantics.ExistChop inteq_reflection)
qed 

lemma ChopstarInductMoreL:
 assumes "\<turnstile> g \<or> (f \<and> more);h \<longrightarrow> h"
 shows   "\<turnstile> (chopstar f);g \<longrightarrow> h"
proof -
 have 1: "\<turnstile> (chopstar f);g = ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g"
   by (simp add: chopstar_d_def powerstar_d_def LeftChopEqvChop)
 have 11: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g =
             (\<exists> n. power (f \<and> more) n);((empty \<or> ((f \<and> more) \<and> inf));g)  "
   by (meson ChopAssoc Prop11)
 have 2: "\<turnstile> (\<exists> n. power (f \<and> more) n);((empty \<or> ((f \<and> more) \<and> inf));g) = 
            (\<exists> n. (power (f \<and> more) n);((empty \<or> ((f \<and> more) \<and> inf));g))"
   using ExistChopPower by fastforce
 have 3: "\<And> n. \<turnstile> (power (f \<and> more) n);g \<longrightarrow> h" 
   using ChopInductL assms by (metis)
 have 31: " \<turnstile> (\<exists> n. (power (f \<and> more) n);g) \<longrightarrow> h"
   using 3 by fastforce
 have 32: "\<And> n. \<turnstile> ((power (f \<and> more) n);((f \<and> more) \<and> inf));g \<longrightarrow> h"
   using assms 
   by (metis (no_types, lifting) AndChopA ChopInductInfL  int_eq_true inteq_reflection)
 have 33: "\<turnstile> (\<exists> n. ((power (f \<and> more) n);((f \<and> more) \<and> inf));g) \<longrightarrow> h"
   using "32" by fastforce
 have 34: "\<turnstile> (\<exists> n. (power (f \<and> more) n);g) \<or> 
             (\<exists> n. ((power (f \<and> more) n);((f \<and> more) \<and> inf));g) \<longrightarrow> h"
   using "31" "33" by fastforce
 have 35: "\<turnstile> (\<exists> n. (power (f \<and> more) n);g \<or> ((power (f \<and> more) n);((f \<and> more) \<and> inf));g)
              \<longrightarrow> h"
   using "34" by fastforce
 have 36: "\<And> n . \<turnstile> ((power (f \<and> more) n);g \<or> ((power (f \<and> more) n);((f \<and> more) \<and> inf));g)  =
            ((power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g "
   by (meson ChopAssoc ChopOrEqvRule EmptyOrChopEqv Prop04 Prop06)
 have 4: "\<turnstile>  (\<exists> n. ((power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)));g) \<longrightarrow> h"
   using "36"  "35" by fastforce
 from 1 11 2 4 show ?thesis 
   by (metis Semantics.ExistChop inteq_reflection)
qed 

lemma FPowerstarInductR:
 assumes "\<turnstile> g \<or> h;f \<longrightarrow> h"
 shows   "\<turnstile> g;(fpowerstar f) \<longrightarrow> h"
proof -
 have 1: "\<turnstile> g;(fpowerstar f) = g;(\<exists> n. power f n)"
   by (simp add:  fpowerstar_d_def)
 have 2: "\<turnstile> (g;(\<exists> n. power f n)) = (\<exists> n. g;(power f n))"
   using ChopExistPower by blast
 have 3: "\<And> n. \<turnstile> g;(power f n) \<longrightarrow> h" 
   using ChopInductR assms by blast
 have 4: "\<turnstile> (\<exists> n. g;(power f n)) \<longrightarrow> h"
   using "3"  by fastforce
 from 1  2   4 show ?thesis by (metis  inteq_reflection)
qed 

lemma PowerstarInductR:
 assumes "\<turnstile> g \<or> h;f \<longrightarrow> h"
 shows   "\<turnstile> g;(powerstar f) \<longrightarrow> h"
proof -
 have 1: "\<turnstile> g;(powerstar f) = g;((\<exists> n. power f n);(empty \<or> (f \<and> inf)))"
   by (simp add: chopstar_d_def powerstar_d_def)
 have 11: "\<turnstile> g;((\<exists> n. power f n);(empty \<or> (f \<and> inf))) =
             (g;(\<exists> n. power f n));(empty \<or> (f \<and> inf))"
   using ChopAssoc by blast
 have 2: "\<turnstile> (g;(\<exists> n. power f n)) = (\<exists> n. g;(power f n))"
   using ChopExistPower by blast
 hence 21: "\<turnstile>  (g;(\<exists> n. power f n));(empty \<or> (f \<and> inf)) = 
               (\<exists> n. g;(power f n));(empty \<or> (f \<and> inf))"
   using LeftChopEqvChop by blast
 have 3: "\<And> n. \<turnstile> g;(power f n) \<longrightarrow> h" 
   using ChopInductR assms by blast
 have 31: "\<And> n. \<turnstile> g;((power f n);(f \<and> inf)) \<longrightarrow> h  "
   using ChopInductInfR assms by blast
 have 32: " \<And>n. \<turnstile> g;((power f n);(empty \<or> (f \<and> inf))) \<longrightarrow> h"
   using "3" "31"
   by (metis ChopEmpty ChopOrEqv Prop02 inteq_reflection) 
 have 4: "\<turnstile> (\<exists> n. g;((power f n);(empty \<or> (f \<and> inf)))) \<longrightarrow> h"
   using "32" by fastforce
 from 1 11 2 21  4 show ?thesis
   by (metis Semantics.ChopExist Semantics.ExistChop int_eq) 
qed 

lemma ChopstarInductR:
 assumes "\<turnstile> g \<or> h;f \<longrightarrow> h"
 shows   "\<turnstile> g;(chopstar f) \<longrightarrow> h"
proof -
 have 1: "\<turnstile> g;(chopstar f) = 
            g;((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)))"
   by (simp add: chopstar_d_def powerstar_d_def)
 have 11: "\<turnstile> g;((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf))) =
             (g;(\<exists> n. power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf))"
   using ChopAssoc by blast
 have 2: "\<turnstile> (g;(\<exists> n. power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf)) = 
            ((\<exists> n. g;power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf))"
   by (simp add: ChopExistPower LeftChopEqvChop)
 have 21: "\<turnstile> g \<or> h;(f \<and> more) \<longrightarrow> h" 
   using ChopAndA assms by fastforce
 have 3: "\<And> n. \<turnstile> g;(power (f \<and> more) n) \<longrightarrow> h" 
   using "21" ChopInductR[of "g" "h" "LIFT(f \<and> more)"] assms by auto
 have 31: "\<And> n. \<turnstile> g;((power (f \<and> more) n);(f \<and> inf)) \<longrightarrow> h  "
   using  assms "3" 
   by (metis "21" AndMoreAndInfEqvAndInf ChopInductInfR inteq_reflection)
 have 32: "\<And>n. \<turnstile> g;((power (f \<and> more) n);(empty \<or> (f \<and> inf)))  \<longrightarrow> h" 
   using "3" "31"
   by (metis ChopEmpty ChopOrEqv Prop02 inteq_reflection) 
 have 4: "\<turnstile> (\<exists> n. g;((power (f \<and> more) n);(empty \<or> (f \<and> inf))) ) \<longrightarrow> h"
   using "32" by fastforce 
 from 1 11 2  4 show ?thesis 
   by (metis Semantics.ChopExist Semantics.ExistChop AndMoreAndInfEqvAndInf
       inteq_reflection)
qed 

lemma ChopstarInductMoreR:
 assumes "\<turnstile> g \<or> h;(f \<and> more) \<longrightarrow> h"
 shows   "\<turnstile> g;(chopstar f) \<longrightarrow> h"
proof -
 have 1: "\<turnstile> g;(chopstar f) = g;((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)))"
   by (simp add: chopstar_d_def powerstar_d_def)
 have 11: "\<turnstile> g;((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf))) =
             (g;(\<exists> n. power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf))"
   using ChopAssoc by blast
 have 2: "\<turnstile> (g;(\<exists> n. power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf)) = 
            ((\<exists> n. g;power (f \<and> more) n));(empty \<or> ((f \<and> more) \<and> inf))"
   by (simp add: ChopExistPower LeftChopEqvChop)
 have 3: "\<And> n. \<turnstile> g;(power (f \<and> more) n) \<longrightarrow> h" 
   using ChopInductR assms by (metis)
 have 31: "\<And> n. \<turnstile> g;((power (f \<and> more) n);(f \<and> inf)) \<longrightarrow> h "
   using assms 
   by (metis ChopInductInfR AndMoreAndInfEqvAndInf inteq_reflection)
 have 32: "\<And>n. \<turnstile> g;((power (f \<and> more) n);(empty \<or> (f \<and> inf)))  \<longrightarrow> h"
   using "3" "31"
   by (metis ChopEmpty ChopOrEqv Prop02 inteq_reflection)
 have 4: "\<turnstile> (\<exists> n. g;((power (f \<and> more) n);(empty \<or> (f \<and> inf))) ) \<longrightarrow> h"
   using  "32" by fastforce 
 from 1 11 2 4 show ?thesis 
   by (metis Semantics.ChopExist Semantics.ExistChop AndMoreAndInfEqvAndInf 
       inteq_reflection)
qed 

lemma FPSEqvEmptyOrFiniteChopFPS:
 "\<turnstile> fpowerstar f = (empty \<or> (f \<and> finite);fpowerstar f)"
using FPowerstarEqvSem Valid_def by blast

lemma FPSAndMoreImpFPS:
 "\<turnstile> fpowerstar (f \<and> more) \<longrightarrow> fpowerstar f"
proof -
 have 1: "\<turnstile> ((f \<and> more) \<and> finite) \<longrightarrow> (f \<and> finite)"
   by auto
 have 11: "\<turnstile> ((f \<and> more) \<and> finite);fpowerstar f \<longrightarrow> (f \<and> finite); fpowerstar f"
   using "1" LeftChopImpChop by blast
 have 2: "\<turnstile> empty \<or> ((f \<and> more) \<and> finite);fpowerstar f \<longrightarrow> fpowerstar f"
   using "11" FPSEqvEmptyOrFiniteChopFPS[of "f"] 
   by fastforce
 have 3: "\<turnstile> fpowerstar (f \<and> more);empty \<longrightarrow> fpowerstar f"
   using "2" FPowerstarInductL by blast
 from 2 3 show ?thesis by (metis ChopEmpty int_eq)
qed

lemma FPSImpAndMoreFPS:
 "\<turnstile> fpowerstar f \<longrightarrow> fpowerstar (f \<and> more)"
by (meson ChopEmpty FPSEqvEmptyOrFiniteChopFPS FPowerstarInductMoreL int_iffD2 lift_imp_trans)

lemma FPSAndMoreEqvFPS:
 "\<turnstile> fpowerstar (f \<and> more) = fpowerstar f"
using FPSAndMoreImpFPS FPSImpAndMoreFPS by fastforce

lemma ChopstarImpPowerstar:
 "\<turnstile>  f\<^sup>\<star> \<longrightarrow> powerstar f"
by (metis ChopEmpty ChopstarInductL PSEqvEmptyOrChopPS int_eq int_iffD2)

lemma PowerstarImpChopstar:
 "\<turnstile> powerstar f \<longrightarrow>  f\<^sup>\<star> "
by (metis CSEqvOrChopCS ChopEmpty PowerstarInductL int_iffD2 inteq_reflection)

lemma ChopstarEqvPowerstar: 
 "\<turnstile> f\<^sup>\<star> = powerstar f"
using ChopstarImpPowerstar PowerstarImpChopstar by fastforce 

lemma CSAndMoreEqvAndMoreChop:
 " \<turnstile>  (f\<^sup>\<star> \<and>  more)  = (f \<and>  more ); f\<^sup>\<star> "
proof -
 have  1: "\<turnstile> ( empty  \<or>  (f \<and>  more ); f\<^sup>\<star>) \<and>  more \<longrightarrow> (f \<and>  more ); f\<^sup>\<star> " 
   by (auto simp: empty_d_def)
 have  2: "\<turnstile> f\<^sup>\<star> = (empty \<or> (f \<and> more);f\<^sup>\<star>)" 
   by (rule ChopstarEqv)
 have  3: "\<turnstile> f\<^sup>\<star> \<and>  more  \<longrightarrow> (f \<and>  more ); f\<^sup>\<star> " 
   using "1" "2" by fastforce
 have  4: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> " 
   using "2" by fastforce
 have  5: "\<turnstile> (f \<and>  more ) \<longrightarrow> more" 
   by auto
 hence 6: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<longrightarrow>  more " 
   by (rule LeftChopImpMoreRule)
 have  7: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> \<and>  more " 
   using "4" "6" by fastforce
 from 3 7 show ?thesis by fastforce
qed

lemma AndMoreCSEqvAndFmoreOrInf:
 "\<turnstile> (f \<and> more);f\<^sup>\<star> = ((f \<and> fmore);f\<^sup>\<star> \<or> (f \<and> inf))"
proof -
 have 1: "\<turnstile> (f \<and> more) = ( (f \<and> fmore) \<or> (f \<and> inf)) "
   by (metis AndMoreAndFiniteEqvAndFmore AndMoreAndInfEqvAndInf OrFiniteInf int_eq)
 hence 2: "\<turnstile> (f \<and> more);f\<^sup>\<star> = ( (f \<and> fmore) \<or> (f \<and> inf));f\<^sup>\<star> "
   by (simp add: LeftChopEqvChop)
 have 3: "\<turnstile> ( (f \<and> fmore) \<or> (f \<and> inf));f\<^sup>\<star> = ((f \<and> fmore);f\<^sup>\<star> \<or> (f \<and> inf);f\<^sup>\<star> ) "
   by (simp add: OrChopEqv)
 have 4: "\<turnstile> (f \<and> inf);f\<^sup>\<star> = (f \<and> inf) "
   using AndInfChopEqvAndInf by blast
 have 5: "\<turnstile>((f \<and> fmore);f\<^sup>\<star> \<or> (f \<and> inf);f\<^sup>\<star> ) = ((f \<and> fmore);f\<^sup>\<star> \<or> (f \<and> inf) )"
   using "3" "4" by auto
 from 2 3 5 show ?thesis by fastforce
qed
    

lemma PowerAndMoreAndFinite:
 "\<turnstile> ((power (f \<and> more) n) \<and> finite) = (power ((f \<and> finite) \<and> more) n)"
proof
 (induct n)
 case 0
 then show ?case using FiniteAndEmptyEqvEmpty by auto
 next
 case (Suc n)
 then show ?case 
   proof -
    have 1: "\<turnstile> (power (f \<and> more) ( Suc n) \<and> finite) =
               (((f \<and> more) \<and> finite) ; power (f \<and> more) n \<and> finite)
            " 
      by simp
    have 2: "\<turnstile> (((f \<and> more) \<and> finite) ; power (f \<and> more) n \<and> finite) =
                ((((f \<and> more) \<and> finite) \<and> finite) ; (power (f \<and> more) n \<and> finite))"
      by (simp add: ChopAndFiniteDist)
    have 3: "\<turnstile> (((f \<and> more) \<and> finite) \<and> finite) = (((f \<and> finite) \<and> more) \<and> finite)"
      by auto
    have 4: "\<turnstile> (((f \<and> finite) \<and> more) \<and> finite); (power ((f \<and> finite) \<and> more) n) =
              power ((f \<and> finite) \<and> more) (Suc n) 
            " 
      by simp
    show ?thesis 
      by (metis "1" "2" "3" "4" Suc.hyps inteq_reflection)
  qed
qed

lemma CSAndFiniteDist:
 "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite) =
    ((\<exists> n. power (f \<and> more) n) \<and> finite)"
proof -
 have 1: "\<turnstile>((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite) =
           (((\<exists> n. power (f \<and> more) n) \<and> finite);
            ((empty \<or> ((f \<and> more) \<and> inf)) \<and> finite)) "
   using ChopAndFiniteDist by blast
 have 11: "\<turnstile> (((f \<and> more) \<and> inf) \<and> \<not> inf) = #False"
   by fastforce 
 have 2: "\<turnstile> ((empty \<or> ((f \<and> more) \<and> inf)) \<and> finite) = empty"
   unfolding finite_d_def using 11 NotEmptyAndInf by auto
 from 1 2 show ?thesis 
   by (metis ChopEmpty inteq_reflection)
qed
 

lemma CSAndFinite:
 "\<turnstile> (f\<^sup>\<star> \<and> finite) = (f \<and> finite)\<^sup>\<star>"
proof -
 have 1: "\<turnstile> (f\<^sup>\<star> \<and> finite) = ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite)"
   by (simp add: chopstar_d_def powerstar_d_def intI)
 have 11: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite) =
             ((\<exists> n. power (f \<and> more) n) \<and> finite)"
   using CSAndFiniteDist by blast
 have 2: "\<turnstile> ((\<exists> n. power (f \<and> more) n) \<and> finite) =
            (\<exists> n. power (f \<and> more) n \<and> finite) "
   by (simp add: Valid_def)
 have 3: "\<turnstile> (\<exists> n. power (f \<and> more) n \<and> finite) =
          (\<exists> n. (power ((f \<and> finite) \<and> more) n)) "
   using PowerAndMoreAndFinite by fastforce
 have 12: "\<turnstile> (((f ) \<and> inf) \<and> \<not> inf) = #False"
       by fastforce 
 have 31: "\<turnstile> (empty \<or> ((f \<and> finite) \<and> inf)) = empty" 
   unfolding finite_d_def using 12 NotEmptyAndInf by auto
 have 4: "\<turnstile> (\<exists> n. (power ((f \<and> finite) \<and> more) n)) = (f \<and> finite)\<^sup>\<star> "
   using "31" 
   by (metis ChopEmpty AndMoreAndInfEqvAndInf chopstar_d_def inteq_reflection powerstar_d_def)
 have 5: "\<turnstile> (\<exists>n. power (f \<and> more) n \<and> finite) = (\<exists>n. (power ((f \<and> finite) \<and> more) n))"
   by (metis "3" ) 
 have 6: "\<turnstile> (f\<^sup>\<star> \<and> finite) = (\<exists>n. (power (f \<and> more) n) \<and> finite)"
   using "1" "11" by fastforce
 show ?thesis  
   using "4" 5 6 by fastforce
qed 
   
lemma PowerchopAndMore:
 "\<turnstile> ((power ((f \<and> finite) \<and> more) (Suc n)) \<and> more) =
            (power ((f \<and> finite) \<and> more) (Suc n))"
proof (induct n)
case 0
then show ?case 
   proof -
    have 01: " \<turnstile> (power ((f \<and> finite) \<and> more) (Suc 0) \<and> more) =
                 ((((f \<and> finite) \<and> more) \<and> finite);empty \<and> more)
             " 
      by simp
    have 02: "\<turnstile> ((((f \<and> finite) \<and> more) \<and> finite);empty \<and> more) =
                ((((f \<and> finite) \<and> more) \<and> finite) \<and> more)
             " 
      by (metis ChopEmpty inteq_reflection)
    have 03: "\<turnstile> ((((f \<and> finite) \<and> more) \<and> finite) \<and> more) =
                ((((f \<and> finite) \<and> more) \<and> finite))"
      by auto
    have 04: "\<turnstile> ((((f \<and> finite) \<and> more) \<and> finite)) = 
                ((((f \<and> finite) \<and> more) \<and> finite));empty
             " 
      by (simp add: ChopEmpty int_iffD1 int_iffD2 int_iffI)
    have 05: "\<turnstile> ((((f \<and> finite) \<and> more) \<and> finite));empty=
                 power ((f \<and> finite) \<and> more) (Suc 0)"
      by simp
    show ?thesis 
      by (metis "03" "04" "05" inteq_reflection)
  qed
next
case (Suc n)
then show ?case 
  by (metis LeftChopImpMoreRule Prop10 Prop11 Prop12 lift_and_com pow_Suc)
qed

lemma PowerchopAndFmore:
 "\<turnstile> ((power (f \<and> more) (Suc n) ) \<and> fmore) = (power ((f \<and> finite) \<and> more) (Suc n))"
proof -
 have 1: "\<turnstile> ((power (f \<and> more) (Suc n) ) \<and> fmore) =
            (((power (f \<and> more) (Suc n) ) \<and> finite) \<and> more) "
   by (auto simp add: fmore_d_def)
 have 2: "\<turnstile> (((power (f \<and> more) (Suc n) ) \<and> finite) \<and> more) =
            ((power ((f \<and> finite) \<and> more) (Suc n)) \<and> more)" 
   by (metis PowerAndMoreAndFinite inteq_reflection lift_and_com)
 have 3: "\<turnstile> ((power ((f \<and> finite) \<and> more) (Suc n)) \<and> more) =
            (power ((f \<and> finite) \<and> more) (Suc n))"
   using PowerchopAndMore by blast
 show ?thesis
   by (metis "1" "2" "3" inteq_reflection) 
qed


lemma ExistPowerAndMoreExpand:
 "\<turnstile> (\<exists> n. power (f \<and> more) n) = ( empty \<or> (\<exists> n. (power (f \<and> more) (Suc n))))"
using powersem1[of "LIFT(f \<and> more)"] by auto 

lemma CSAndFmoreDist:
 "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> fmore) =
             ((\<exists> n. power (f \<and> more) n) \<and> fmore)"
proof -
 have 1: "\<turnstile>((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> fmore) =
           ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite \<and> more) "
   by (metis fmore_d_def inteq_reflection lift_and_com)
 have 2: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite) =
            ((\<exists> n. power (f \<and> more) n) \<and> finite)"
   using CSAndFiniteDist by blast
 have 3: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> finite \<and> more) =
            ((\<exists> n. power (f \<and> more) n) \<and> finite \<and> more)"
   using "2" by fastforce
 have 4: "\<turnstile> ((\<exists> n. power (f \<and> more) n) \<and> finite \<and> more) =
            ((\<exists> n. power (f \<and> more) n) \<and> fmore)"
   by (metis fmore_d_def inteq_reflection lift_and_com)
 show ?thesis
   by (metis "1" "3" "4" inteq_reflection) 
qed

lemma CSAndMoreEqvAndFMoreChop:
 " \<turnstile>  (f\<^sup>\<star> \<and>  fmore)  = (f \<and>  fmore ); (f \<and> finite)\<^sup>\<star> "
proof -
 have 1: "\<turnstile> (f\<^sup>\<star> \<and>  fmore) = ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> fmore)"
   by (simp add: chopstar_d_def powerstar_d_def intI)
 have 11: "\<turnstile> ((\<exists> n. power (f \<and> more) n);(empty \<or> ((f \<and> more) \<and> inf)) \<and> fmore) =
             ((\<exists> n. power (f \<and> more) n) \<and> fmore)"
   using CSAndFmoreDist by fastforce
 have 2: "\<turnstile> ((\<exists> n. power (f \<and> more) n) \<and> fmore) =
            (\<exists> n. power (f \<and> more) n \<and> fmore)"
   by (simp add: Valid_def)
 have 3: "\<turnstile> (\<exists> n. power (f \<and> more) n \<and> fmore) =
            ((power (f \<and> more) 0 \<or> (\<exists> n. (power (f \<and> more) (Suc n)))) \<and> fmore)"
   using ExistPowerAndMoreExpand by fastforce
 have 4: "\<turnstile> ((power (f \<and> more) 0 \<or> (\<exists> n. (power (f \<and> more) (Suc n)))) \<and> fmore) =
            (((power (f \<and> more) 0  \<and> fmore) \<or> ((\<exists> n. (power (f \<and> more) (Suc n))) \<and> fmore)))"
   by auto
 have 5: "\<turnstile> (((power (f \<and> more) 0  \<and> fmore) \<or> ((\<exists> n. (power (f \<and> more) (Suc n))) \<and> fmore))) =
             ((\<exists> n. (power (f \<and> more) (Suc n))) \<and> fmore)" 
   using NotFmoreAndEmpty by fastforce
 have 6: "\<turnstile> ((\<exists> n. (power (f \<and> more) (Suc n))) \<and> fmore) =
             (\<exists> n. (power ((f \<and> finite) \<and> more) (Suc n)))"
   using PowerchopAndFmore by fastforce
 have 7: "\<turnstile> (\<exists> n. (power ((f \<and> finite) \<and> more) (Suc n))) =
            (\<exists> n. ((((f \<and> finite) \<and> more) \<and> finite);(power ((f \<and> finite) \<and> more) n))) "
   by (simp)
 have 8: " \<turnstile> (\<exists> n. ((((f \<and> finite) \<and> more) \<and> finite);(power ((f \<and> finite) \<and> more) n))) =
            (((f \<and> finite) \<and> more) \<and> finite);(\<exists> n. (power ((f \<and> finite) \<and> more) n)) "
   by (auto simp add: Valid_def itl_defs) 
 have 9: "\<turnstile> (((f \<and> finite) \<and> more) \<and> finite) = ( f \<and> fmore)"
   by (auto simp add: fmore_d_def)
 have 10: "\<turnstile> (((f \<and> finite) \<and> more) \<and> finite);(\<exists> n. (power ((f \<and> finite) \<and> more) n)) =
             (f \<and> fmore);(\<exists> n. (power ((f \<and> finite) \<and> more) n)) "
   using 8 9 by (simp add: LeftChopEqvChop)
 have 1011: "\<turnstile> ((f \<and> finite) \<and> inf) = #False"
   unfolding finite_d_def by fastforce
 have 101: "\<turnstile> (empty \<or> ((f \<and> finite) \<and> inf)) = empty"
   unfolding finite_d_def using 1011 NotEmptyAndInf by auto
 have 11: "\<turnstile> (\<exists> n. (power ((f \<and> finite) \<and> more) n)) =
             (f \<and> finite)\<^sup>\<star> "
   using 101 
   by (metis ChopEmpty AndMoreAndInfEqvAndInf chopstar_d_def inteq_reflection powerstar_d_def)
 hence 12: "\<turnstile> (f \<and> fmore);(\<exists> n. (power ((f \<and> finite) \<and> more) n)) =
             (f \<and> fmore);(f \<and> finite)\<^sup>\<star>"
   by (simp add: RightChopEqvChop)
 from 1 11 2 3 4 5 6 7 8 10 12 show ?thesis 
   by (metis CSAndFmoreDist inteq_reflection)
qed


lemma CSAndMoreImpChopCS:
 "\<turnstile>   f\<^sup>\<star> \<and>  more  \<longrightarrow> f; f\<^sup>\<star> "
proof -
 have 1: "\<turnstile> (f\<^sup>\<star> \<and>  more)  = (f \<and>  more ); f\<^sup>\<star> " by (rule CSAndMoreEqvAndMoreChop)
 have 2: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<longrightarrow> f; f\<^sup>\<star> " by (rule AndChopA)
 from 1 2 show ?thesis by fastforce
qed

lemma NotAndMoreEqvEmptyOr:
 "\<turnstile> \<not> (f \<and> more) = (empty \<or> \<not>f)"
by (auto simp: empty_d_def)

lemma MoreAndEmptyOrEqvMoreAnd:
 "\<turnstile> (more \<and> (empty \<or> \<not>f)) = (more \<and> \<not> f) "
by (auto simp: empty_d_def)

lemma CSMoreNotImpChopCSAndMore:
"\<turnstile>   f\<^sup>\<star> \<and>  more  \<and> \<not>  f \<longrightarrow> (f \<and>  more ); (f\<^sup>\<star>\<and>  more ) "
proof -
 have  1: "\<turnstile> (f\<^sup>\<star> \<and>  more)  = (f \<and>  more ); f\<^sup>\<star>" 
   by (rule CSAndMoreEqvAndMoreChop)
 have  2: "\<turnstile>  empty  \<or>   more" 
   by (auto simp: empty_d_def)  
 hence 3: "\<turnstile> f\<^sup>\<star> \<longrightarrow>  empty  \<or>  (f\<^sup>\<star> \<and>  more )" 
   by auto   
 hence 4: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<longrightarrow> (f \<and>  more ) \<or>  ((f \<and>  more ); (f\<^sup>\<star> \<and>  more ))" 
   by (rule ChopEmptyOrImpRule)  
 hence 5: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<and> \<not>(f \<and> more) \<longrightarrow>   ((f \<and>  more ); (f\<^sup>\<star> \<and>  more )) " 
   by fastforce
 have  6: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> =  ((f \<and>  more ); f\<^sup>\<star> \<and> more)  " using "1" 
   by auto
 have  7: "\<turnstile> ((f \<and>  more ); f\<^sup>\<star> \<and> \<not>(f \<and> more)) = ((f \<and>  more ); f\<^sup>\<star> \<and> more \<and> \<not>(f \<and> more)) " 
   using "6" by auto
 have  8: "\<turnstile> (f \<and>  more ); f\<^sup>\<star> \<and>  more  \<and> \<not>  f\<longrightarrow> (f \<and>  more ); (f\<^sup>\<star> \<and>  more )" 
   using "5" "7" by auto 
 have  9: "\<turnstile> (f\<^sup>\<star> \<and>  more  \<and> \<not>  f) = ((f\<^sup>\<star> \<and>  more) \<and> (more  \<and> \<not>  f)) " 
   by auto
 have  10: "\<turnstile> ((f\<^sup>\<star> \<and>  more) \<and> (more  \<and> \<not>  f)) = ((f \<and>  more ); f\<^sup>\<star> \<and>  (more  \<and> \<not>  f))" 
   using "1" by fastforce
 from 1 8 9 10 show ?thesis by fastforce
qed

lemma ChopplusCommuteImpA:
 "\<turnstile> f\<^sup>\<star>;f \<longrightarrow> f;f\<^sup>\<star>"
by (metis CSEqvOrChopCS ChopAndB ChopEmpty ChopstarInductL EmptyImpCS Prop02 Prop03 Prop10
          inteq_reflection) 

lemma ChopplusCommuteImpB:
 "\<turnstile> f;f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>;f"

proof -
have f2: "\<turnstile> (f\<^sup>\<star>;f);f \<longrightarrow> f\<^sup>\<star>;f"
  by (metis CSEqvOrChopCS ChopplusCommuteImpA LeftChopImpChop Prop05 inteq_reflection)
have "\<turnstile> f \<longrightarrow> f\<^sup>\<star>;f"
  by (metis AndChopB EmptyChop EmptyImpCS Prop10 inteq_reflection)
then show ?thesis
  using f2 ChopstarInductR Prop02 by blast (* 2 ms *)
qed


lemma ChopplusCommute:
 "\<turnstile> f;f\<^sup>\<star> = f\<^sup>\<star>;f"
using ChopplusCommuteImpA ChopplusCommuteImpB   by fastforce

lemma CSEqvOrChopCSB:
 "\<turnstile>   f\<^sup>\<star> =  (empty  \<or>  (f\<^sup>\<star>;f)) " 
by (meson CSEqvOrChopCS ChopplusCommute Prop06)

lemma CSAndMoreImpCSChop:
 "\<turnstile>   f\<^sup>\<star> \<and>  more  \<longrightarrow> f\<^sup>\<star>; f "
using CSAndMoreEqvAndMoreChop  ChopplusCommute CSAndMoreImpChopCS by fastforce

lemma PowerChopPower:
 "\<turnstile> (power (f \<and> more) n); (power (f \<and> more) k) = (power (f \<and> more) (n+k))"
proof
 (induct n arbitrary: k)
 case 0
 then show ?case using EmptyChopSem by auto
 next
 case (Suc n)
 then show ?case 
 by (metis (no_types, lifting) ChopAssoc add_Suc inteq_reflection pow_Suc)
qed

lemma CSChopCS:
 "\<turnstile> f\<^sup>\<star> ; f\<^sup>\<star> = f\<^sup>\<star>"
proof -
 have 1: "\<turnstile> f\<^sup>\<star> ; f\<^sup>\<star>  \<longrightarrow>  f\<^sup>\<star>"
   by (meson CSEqvOrChopCSB ChopstarImpPowerstar ChopstarInductR PowerstarImpChopstar Prop02 Prop03 
      lift_imp_trans)
 have 2: "\<turnstile> f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> ; f\<^sup>\<star>"
   by (metis ChopEmpty EmptyImpCS RightChopImpChop inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma NotEmptyEqvMore:
 "\<turnstile> (\<not> empty) = more"
by (simp add: empty_d_def)

lemma NotCSImpMore:
 "\<turnstile>    \<not>  (f\<^sup>\<star>) \<longrightarrow>  more "
proof -
 have  1: "\<turnstile> empty \<longrightarrow> (f\<^sup>\<star>)  " using EmptyImpCS by blast
 hence 2: "\<turnstile>  \<not> empty \<or> (f\<^sup>\<star>)" by fastforce
 from 2 show ?thesis  using "1" NotEmptyEqvMore by fastforce 
qed

lemma PowerAndInfB:
 "\<turnstile> ( ((f \<and> more);(power (f \<and> more) n)) \<and> inf) = 
      ((f \<and> inf) \<or> (f \<and> fmore);((power (f \<and> more) n) \<and> inf))"
proof -
 have 1: "\<turnstile> ( ((f \<and> more);(power (f \<and> more) n)) \<and> inf) =
            ( ((f \<and> more) \<and> inf) \<or> ((f \<and> more) \<and> finite);((power (f \<and> more) n) \<and> inf)) "
   using ChopAndInfB by blast
 have 2: "\<turnstile> ( ((f \<and> more) \<and> inf) \<or> ((f \<and> more) \<and> finite);((power (f \<and> more) n) \<and> inf))  =
             ( (f \<and> inf) \<or> (f \<and> fmore);((power (f \<and> more) n) \<and> inf))"
   using AndMoreAndInfEqvAndInf AndMoreAndFiniteEqvAndFmore 
   by (metis "1" inteq_reflection)
 show ?thesis
   by (metis "1" "2" int_eq) 
qed


lemma CSAndInf:
 "\<turnstile> (f\<^sup>\<star> \<and> inf) = f\<^sup>\<star>;(f \<and> inf)"
by (meson AndChopA AndInfChopEqvAndInf CSEqvOrChopCSB ChopAndA ChopAndInf Prop03 Prop11 
          Prop12 lift_imp_trans)

lemma CSChopCSImpCS:
 "\<turnstile> (f\<^sup>\<star>; f\<^sup>\<star>)  \<longrightarrow> f\<^sup>\<star>"
by (simp add: CSChopCS int_iffD1)

lemma ImpChopPlus:
 "\<turnstile>   f \<longrightarrow> f;f\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f\<^sup>\<star> = (empty \<or>  f; f\<^sup>\<star>)  " by (rule CSEqvOrChopCS)
 hence 2: "\<turnstile> f;f\<^sup>\<star> = (f;empty \<or>  f; (f;f\<^sup>\<star>)) " using ChopOrEqvRule by blast
 have  3: "\<turnstile> f;empty = f" using ChopEmpty by blast
 from 2 3 show ?thesis by fastforce
qed

lemma ImpCS:
 "\<turnstile>   f \<longrightarrow> f\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f \<longrightarrow> f;f\<^sup>\<star> " by (rule ImpChopPlus)
 hence 2: "\<turnstile> f \<longrightarrow>  empty  \<or>  f;f\<^sup>\<star> " by auto
 from 2 show ?thesis using CSEqvOrChopCS by fastforce
qed 

lemma CSChopImpCS:
 "\<turnstile>   f\<^sup>\<star>; f  \<longrightarrow> f\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f \<longrightarrow> f\<^sup>\<star> " by (rule ImpCS)
 hence 2: "\<turnstile> f\<^sup>\<star>; f \<longrightarrow> f\<^sup>\<star>; f\<^sup>\<star> " by (rule RightChopImpChop)
 hence  3: "\<turnstile> f\<^sup>\<star>; f \<longrightarrow> f\<^sup>\<star>; f\<^sup>\<star> " by auto
 have  4: "\<turnstile> f\<^sup>\<star>; f\<^sup>\<star>  \<longrightarrow> f\<^sup>\<star> " by (rule CSChopCSImpCS)
 from 2 3 4 show ?thesis using lift_imp_trans by blast
qed

lemma ChopPlusImpCS:
 " \<turnstile>   f;f\<^sup>\<star>  \<longrightarrow> f\<^sup>\<star> "
proof -
 have 1: "\<turnstile>   f;f\<^sup>\<star> \<longrightarrow> empty  \<or> f;f\<^sup>\<star>" by auto
 from 1 show ?thesis using CSEqvOrChopCS by fastforce
qed

lemma CSChopEqvOrChopPlusChop:
 "\<turnstile>   f\<^sup>\<star>; g = (g \<or>  (f;f\<^sup>\<star>) ; g) "
proof -
 have 1: "\<turnstile> f\<^sup>\<star> =  (empty  \<or>  f;f\<^sup>\<star>) " by (rule CSEqvOrChopCS)
 from 1 show ?thesis using EmptyOrChopEqvRule by blast
qed

lemma CSElim:
 assumes "\<turnstile>    empty  \<longrightarrow> g " 
         "\<turnstile> (f \<and>  more ); g \<longrightarrow> g "
 shows   "\<turnstile> f\<^sup>\<star> \<longrightarrow> g "
proof -
 have 1: "\<turnstile> empty \<or> (f \<and> more);g \<longrightarrow> g"
   using assms using Prop02 by blast
 have 2: "\<turnstile> (chopstar f);empty \<longrightarrow> g "
   using ChopstarInductMoreL "1" by blast
 from 2 show ?thesis 
   by (metis ChopEmpty inteq_reflection)
qed

lemma ChopstarImp:
 assumes "\<turnstile> f;(chopstar g) \<or> empty \<longrightarrow> (chopstar g)"
 shows   "\<turnstile> (chopstar f) \<longrightarrow> (chopstar g)"
using assms ChopstarInductL ChopEmpty 
by (metis int_eq int_simps(33) lift_and_com)

lemma CSCSImpCS:
 "\<turnstile>   (f\<^sup>\<star>)\<^sup>\<star>  \<longrightarrow> f\<^sup>\<star> "
proof -
 have 1: "\<turnstile> ((chopstar f);(chopstar f)) \<or> empty \<longrightarrow> (chopstar f)"
   by (meson CSChopCSImpCS EmptyImpCS Prop02)
 from 1 show ?thesis using ChopstarImp by blast
qed

lemma CSImpCSCS:
 "\<turnstile> f\<^sup>\<star> \<longrightarrow> (f\<^sup>\<star>)\<^sup>\<star>"
using ImpCS by auto

lemma CSCSEqvCS:
 "\<turnstile> (f\<^sup>\<star>)\<^sup>\<star>  = f\<^sup>\<star>"
by (simp add: CSCSImpCS CSImpCSCS int_iffI) 

lemma RightEmptyOrChopEqv:
 "\<turnstile>   g;( empty  \<or>  f) = (g \<or>  (g; f)) "
proof -
 have 1: "\<turnstile> g;( empty  \<or>  f) = (g;empty \<or> g;f)  " by (rule ChopOrEqv)
 have 2: "\<turnstile> g;empty  = g " by (rule ChopEmpty)
 from 1 2 show ?thesis by fastforce
qed

lemma RightEmptyOrChopEqvRule:
 assumes "\<turnstile> f = (empty \<or> f1)"
 shows   "\<turnstile> g;f = (g \<or> (g;f1))"
proof -
 have  1: "\<turnstile> f = (empty \<or> f1) " using assms by auto
 hence 2: "\<turnstile> g;f = g;(empty \<or> f1) " by (rule RightChopEqvChop)
 have  3: "\<turnstile> g;(empty \<or> f1) = (g \<or> (g;f1)) " by (rule RightEmptyOrChopEqv)
 from 2 3 show ?thesis by fastforce
qed

lemma ChopPlusEqvOrChopChopPlus:
 "\<turnstile>   (f;f\<^sup>\<star>)  = (f \<or>  f; (f;f\<^sup>\<star>))  "
proof -
 have 1: "\<turnstile> f\<^sup>\<star> =  (empty  \<or>  f;f\<^sup>\<star>) " by (rule CSEqvOrChopCS)
 from 1 show ?thesis by (rule RightEmptyOrChopEqvRule)
qed

lemma CSAndEmptyEqvEmpty:
 "\<turnstile> ((f\<^sup>\<star>) \<and> empty) = empty"
using EmptyImpCS by fastforce

lemma NotAndMoreChopAndEmpty:
 "\<turnstile> \<not>(((f \<and> more);g) \<and> empty) "
by (metis AndChopA ChopEmpty LeftChopImpMoreRule Prop01 empty_d_def int_simps(14) 
          int_simps(25) int_simps(4) inteq_reflection lift_and_com)

lemma NotChopAndMoreAndEmpty:
 "\<turnstile> \<not>((f;(g \<and> more)) \<and> empty) "
by (metis NotEmptyEqvMore Prop01 Prop05 Prop07 RightChopImpMoreRule empty_d_def int_iffD2 
          int_simps(15) inteq_reflection lift_imp_neg)

lemma ChopCSAndEmptyEqvAndEmpty:
  "\<turnstile> ((f;f\<^sup>\<star>) \<and> empty) = (f \<and> empty)"
proof -
 have 1: "\<turnstile> ((f;f\<^sup>\<star>) \<and> empty) = (f\<and> empty);(f\<^sup>\<star> \<and> empty) " 
   using ChopAndEmptyEqvEmptyChopEmpty by blast
 have 2: "\<turnstile> (f\<and> empty);(f\<^sup>\<star> \<and> empty) = (f\<and> empty);empty " 
   using CSAndEmptyEqvEmpty using RightChopEqvChop by blast
 have 3: "\<turnstile> (f\<and> empty);empty = (f\<and> empty)" 
   by (rule ChopEmpty)
 from 1 2 3 show ?thesis by fastforce
qed

lemma AndMoreChopAndMoreEqvAndMoreChop:
 "\<turnstile> ((f \<and> more);g \<and> more) = (f \<and> more);g"
using ChopImpDi DiAndB DiMoreEqvMore by fastforce

lemma ChopPlusEqv:
 "\<turnstile> (f;f\<^sup>\<star>)  = (f \<or>  (f\<and>  more ); (f;f\<^sup>\<star>))"
proof -
 have  1: "\<turnstile> f\<^sup>\<star> =  (empty  \<or>  (f \<and>  more ); f\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 have  2: "\<turnstile> f\<^sup>\<star> =  (empty  \<or> f;f\<^sup>\<star>)" 
   by (rule CSEqvOrChopCS)
 hence 3: "\<turnstile>   (empty  \<or>  f;f\<^sup>\<star>)  = (empty  \<or>  (f \<and>  more );f\<^sup>\<star>)  " 
   using "1" "2" by fastforce
 have  4: "\<turnstile> (f \<and>  more );(f\<^sup>\<star>) = (f \<and>  more );(empty  \<or>  f;f\<^sup>\<star>) " 
   using "2" using RightChopEqvChop by blast 
 hence 5: "\<turnstile>   empty  \<or>  f;f\<^sup>\<star>  = empty  \<or>  (f \<and>  more );(empty  \<or>  f;f\<^sup>\<star>) " 
   using "3" "4" by fastforce
 have  6: "\<turnstile> (f \<and>  more ); ( empty  \<or>  f;f\<^sup>\<star> ) =  
             ((f \<and>  more );  empty  \<or>  (f \<and>  more ); (f;f\<^sup>\<star>))  " 
   using ChopOrEqv by blast
 have  7: "\<turnstile> (f \<and>  more );  empty  = (f \<and>  more) " 
   using ChopEmpty by blast
 have  8: "\<turnstile>  (empty  \<or>  f;f\<^sup>\<star>)  =
              (empty  \<or>  (f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>))" 
   using "5" "6" "7" by (metis "2" "3" inteq_reflection)
 have  9: "\<turnstile>  ((empty  \<or>  f;f\<^sup>\<star>) \<and> more) =  (f;f\<^sup>\<star> \<and> more) " 
   by (auto simp: empty_d_def)
 have 10: "\<turnstile> ((empty  \<or>  (f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>)) \<and> more) =
            (((f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>)) \<and> more) " 
   by (auto simp: empty_d_def)  
 have 11: "\<turnstile> (((f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>)) \<and> more) = 
             ((f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>))" 
   using "10" "6" "7" int_eq 
   using AndMoreChopAndMoreEqvAndMoreChop by fastforce
 have 12: "\<turnstile> (f;f\<^sup>\<star> \<and> more) = ((f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>)) "
   by (metis "10" "11" "8" "9" inteq_reflection)  
 have 13: "\<turnstile> (f;f\<^sup>\<star> \<and> empty) = (f \<and> empty) " 
   by (rule ChopCSAndEmptyEqvAndEmpty)
 have 14: "\<turnstile> ((f \<and>  more ) \<or>  (f \<and>  more ); (f;f\<^sup>\<star>) \<or> (f \<and> empty)) =
             (f \<or>  (f \<and>  more );(f;f\<^sup>\<star>)) " 
   by (auto simp: empty_d_def)
 have 15: "\<turnstile> f;f\<^sup>\<star> = (( f;f\<^sup>\<star> \<and> empty) \<or> ( f;f\<^sup>\<star> \<and> more))" 
   by (auto simp: empty_d_def)
 from 12 13 14 15 show ?thesis by fastforce
qed

lemma ChopPlusImpChopPlus:
 assumes "\<turnstile>   f \<longrightarrow> g "
 shows "\<turnstile>  f;f\<^sup>\<star> \<longrightarrow> g;g\<^sup>\<star> " 
using assms 
by (metis AndChopB CSChopCS ChopImpChop ChopstarImp EmptyImpCS ImpCS Prop01 Prop02 
    Prop05 Prop10 inteq_reflection)

lemma ChopChopPlusImpChopPlus:
 "\<turnstile>   f; (f;f\<^sup>\<star>)  \<longrightarrow> f;f\<^sup>\<star> "
proof -
 have  1: "\<turnstile>  empty  \<or>   more " by (auto simp: empty_d_def)
 hence 2: "\<turnstile> f \<longrightarrow>  empty  \<or>  (f \<and>  more ) " by auto
 hence 3: "\<turnstile> f; (f;f\<^sup>\<star>)   \<longrightarrow> (f;f\<^sup>\<star>)  \<or>  (f \<and>  more );(f;f\<^sup>\<star>) " by (rule EmptyOrChopImpRule)
 have  4: "\<turnstile>  f;f\<^sup>\<star> = (f \<or>  (f\<and>  more ); (f;f\<^sup>\<star>)) " by (rule ChopPlusEqv)
 hence 5: "\<turnstile> (f\<and>  more ); (f;f\<^sup>\<star>)  \<longrightarrow> f;f\<^sup>\<star> " by auto
 from 3 5 show ?thesis using ChopPlusImpCS RightChopImpChop by blast
qed

lemma CSImpCS:
 assumes "\<turnstile> f \<longrightarrow> g "
 shows   "\<turnstile> f\<^sup>\<star> \<longrightarrow> g\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f \<longrightarrow> g  " using assms by auto
 hence 2: "\<turnstile> f;f\<^sup>\<star> \<longrightarrow> g;g\<^sup>\<star> " by (rule ChopPlusImpChopPlus)
 hence 3: "\<turnstile>  empty  \<or>  f;f\<^sup>\<star>  \<longrightarrow>  empty  \<or>   g;g\<^sup>\<star> " by auto
 from 2 3 show ?thesis using CSEqvOrChopCS by (metis inteq_reflection)
qed

lemma ChopPlusIntro:
 assumes "\<turnstile> f  \<longrightarrow> g \<or> (g \<and>  more ); f "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g;g\<^sup>\<star> "
proof -
 have 1: "\<turnstile> f \<and> \<not>  g \<longrightarrow> (g \<and>  more ); f " using assms by auto
 have 2: "\<turnstile> g;g\<^sup>\<star>  = (g \<or>  (g \<and>  more ); (g;g\<^sup>\<star>)) " by (rule ChopPlusEqv)
 have 3: "\<turnstile> f \<and> \<not> (g;g\<^sup>\<star> ) \<longrightarrow>
          (g \<and>  more ); f \<and> \<not> ((g \<and>  more ); (g;g\<^sup>\<star>) ) " using "1" "2" by fastforce
 have 4: "\<turnstile> g \<and>  more  \<longrightarrow>  more " by auto
 from 3 4 show ?thesis using ChopContraB by blast
qed

lemma ChopPlusElim:
 assumes "\<turnstile>   f \<longrightarrow> g "
         "\<turnstile> (f\<and>  more ); g \<longrightarrow> g"
 shows   "\<turnstile> f;f\<^sup>\<star>  \<longrightarrow> g "
proof -
 have 1: "\<turnstile> f \<or> (f \<and> more);g \<longrightarrow> g "
   using assms Prop02 by blast
 have 2: "\<turnstile> f\<^sup>\<star>;f \<longrightarrow> g"
   using ChopstarInductMoreL "1" by blast
 from 2 show ?thesis 
   using ChopplusCommute by fastforce
qed

lemma ChopPlusElimWithoutMore:
 assumes "\<turnstile>   f \<longrightarrow> g"
         "\<turnstile> f; g \<longrightarrow> g " 
 shows   "\<turnstile> f;f\<^sup>\<star> \<longrightarrow> g "
proof -
 have 1: "\<turnstile> f \<longrightarrow> g " using assms by blast
 have 2: "\<turnstile> (f; g) \<longrightarrow> g " using assms by blast
 have 3: "\<turnstile> (f \<and>  more ); g \<longrightarrow> f; g " by (rule AndChopA)
 have 4: "\<turnstile> (f \<and>  more ); g \<longrightarrow> g " using "2" "3" lift_imp_trans  by blast
 from 1 4 show ?thesis using ChopPlusElim by blast
qed

lemma ChopPlusEqvChopPlus:
 assumes "\<turnstile>   f = g "
 shows   "\<turnstile> f;f\<^sup>\<star>  = g;g\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> f \<longrightarrow> g " by auto
 hence 3: "\<turnstile> f;f\<^sup>\<star>  \<longrightarrow> g;g\<^sup>\<star> " by (rule ChopPlusImpChopPlus)
 have  4: "\<turnstile> g \<longrightarrow> f " using "1" by auto
 hence 5: "\<turnstile>  g;g\<^sup>\<star> \<longrightarrow> f;f\<^sup>\<star> " by (rule ChopPlusImpChopPlus)
 from 3 5 show ?thesis by fastforce
qed

lemma CSEqvCS: 
 assumes "\<turnstile>   f = g "
 shows   "\<turnstile> f\<^sup>\<star> = g\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> f;f\<^sup>\<star>  = g;g\<^sup>\<star>" by (rule ChopPlusEqvChopPlus)
 hence 3: "\<turnstile>  (empty  \<or> f;f\<^sup>\<star>)   =  (empty  \<or>  g;g\<^sup>\<star>) " by auto
 from 3 show ?thesis using CSEqvOrChopCS by (metis int_eq)
qed

lemma AndCSA:
 "\<turnstile>   (f \<and> g)\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> "
proof -
 have 1: "\<turnstile> f \<and> g \<longrightarrow> f " by auto
 from 1 show ?thesis using CSImpCS by blast
qed

lemma AndCSB:
 "\<turnstile>   (f \<and> g)\<^sup>\<star> \<longrightarrow> g\<^sup>\<star> "
proof -
 have 1: "\<turnstile> f \<and> g \<longrightarrow> g " by auto
 from 1 show ?thesis using CSImpCS by blast
qed

lemma CSIntro:
 assumes "\<turnstile>   f \<and> more\<longrightarrow> (g \<and>  more ); f "
 shows   "\<turnstile> f  \<and> finite \<longrightarrow> g\<^sup>\<star> " 
proof -
 have   1: "\<turnstile> f \<and>  more  \<longrightarrow> (g \<and>  more ); f" 
   using assms by auto
 have   2: "\<turnstile>  more  = (\<not>  empty) " 
   by (auto simp: empty_d_def)
 have   3: "\<turnstile> f \<and> \<not>  empty  \<longrightarrow> (g \<and>  more ); f " 
   using "1" "2" by fastforce
 have   4: "\<turnstile> g\<^sup>\<star> =  (empty  \<or>  (g \<and>  more ); g\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 hence 41: "\<turnstile> (\<not>(empty  \<or>  (g \<and>  more ); g\<^sup>\<star>)) = (\<not>empty \<and> \<not>((g \<and>  more ); g\<^sup>\<star>)) " 
   by fastforce
 have  411: "\<turnstile> (\<not>empty \<and> \<not>((g \<and>  more ); g\<^sup>\<star>)) = (more \<and> \<not>((g \<and>  more ); g\<^sup>\<star>))" 
   using NotEmptyEqvMore by fastforce
 have  42: "\<turnstile> \<not>( g\<^sup>\<star>)  = (more \<and> \<not>((g \<and>  more ); g\<^sup>\<star>))" 
   using "4" "41" "411" by fastforce
 have  43: "\<turnstile>f \<and> \<not>( g\<^sup>\<star>) \<longrightarrow> f \<and> more \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) " 
   using "42" by fastforce
 have  44: "\<turnstile> f \<and> more \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) \<longrightarrow> (g \<and>  more ); f \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) " 
   using "3" "43" "1" by auto
 have   5: "\<turnstile> f \<and> \<not> ( g\<^sup>\<star>) \<longrightarrow>
              (g \<and>  more ); f \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) " 
   using "43" "44" lift_imp_trans by fastforce
 have   6: "\<turnstile> g \<and>  more  \<longrightarrow>  more  " 
   by auto
 from 5 6 show ?thesis using   ChopContraB by blast
qed

lemma CSElimWithoutMore: 
 assumes "\<turnstile> empty  \<longrightarrow> g "
         "\<turnstile> f; g \<longrightarrow> g "
 shows   "\<turnstile> f\<^sup>\<star>  \<longrightarrow> g "
proof -
 have 1: "\<turnstile>  empty  \<longrightarrow> g" using assms by blast
 have 2: "\<turnstile> f; g \<longrightarrow> g " using assms by blast
 have 3: "\<turnstile> (f \<and>  more ); g \<longrightarrow> f; g " by (rule AndChopA)
 have 4: "\<turnstile> (f \<and>  more ); g \<longrightarrow> g " using "2" "3" lift_imp_trans by blast
 from 1 4 show ?thesis using CSElim by blast
qed

lemma ChopAssocB:
 "\<turnstile> (f;g);h = f;(g;h)"
using ChopAssoc by fastforce

lemma CSChopEqvChopOrRule:
 assumes "\<turnstile>   f= (g\<^sup>\<star>; h) "
 shows   "\<turnstile> f = ((g; f) \<or>  h) "
proof -
 have   1: "\<turnstile> f = (g\<^sup>\<star>; h) " using assms by auto
 have   2: "\<turnstile> g\<^sup>\<star> =  (empty  \<or>  (g; g\<^sup>\<star>)) " by (rule CSEqvOrChopCS)
 hence  3: "\<turnstile> g\<^sup>\<star>; h = (h \<or>  ((g; g\<^sup>\<star>); h)) " by (rule EmptyOrChopEqvRule)
 have   4: "\<turnstile> (g; g\<^sup>\<star>); h =  g; (g\<^sup>\<star>; h)   " by (rule ChopAssocB)
 hence 41: "\<turnstile> g\<^sup>\<star>; h = (h \<or> g; (g\<^sup>\<star>; h))" using "3" by fastforce
 have   5: "\<turnstile> g; f = g; (g\<^sup>\<star>; h) " using "1" by (rule RightChopEqvChop)
 hence  6: "\<turnstile> (g\<^sup>\<star>; h) = (h \<or>  g; f)" using "41" by fastforce
 hence 61: "\<turnstile> (g\<^sup>\<star>; h) = ((g; f) \<or> h) " by auto
 from 1 61 show ?thesis  by fastforce
qed

lemma CSChopIntroRule:
 assumes "\<turnstile> f \<and> \<not>  h \<longrightarrow> g; f "
         "\<turnstile> g \<longrightarrow>  more "
 shows   "\<turnstile> f \<and> finite \<longrightarrow> g\<^sup>\<star>; h"
proof -
 have  1: "\<turnstile> f \<and> \<not>  h \<longrightarrow> g; f " 
   using assms by blast
 have  2: "\<turnstile> g \<longrightarrow>  more " 
   using assms by blast
 hence 3: "\<turnstile> g \<longrightarrow> g \<and>  more " 
   by auto
 hence 4: "\<turnstile> g; f \<longrightarrow> (g \<and>  more ); f " 
   by (rule LeftChopImpChop)
 have  5: "\<turnstile> f \<longrightarrow> (g \<and>  more ); f \<or>  h " 
   using "1" "4" by fastforce
 have  6: "\<turnstile> g\<^sup>\<star> =  (empty  \<or>  (g \<and>  more ); g\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 hence 7: "\<turnstile> (g\<^sup>\<star>); h = (h \<or>  ((g \<and>  more ); g\<^sup>\<star>); h) " 
   by (rule EmptyOrChopEqvRule)
 have  8: "\<turnstile> ((g \<and>  more ); g\<^sup>\<star>); h = (g \<and>  more ); (g\<^sup>\<star>; h) " 
   by (rule ChopAssocB)
 have  9: "\<turnstile> (g\<^sup>\<star>); h = (h \<or>  (g \<and>  more ); (g\<^sup>\<star>; h)) " 
   using "7" "8" by fastforce
 have 10: "\<turnstile> f \<and> \<not> (g\<^sup>\<star>; h) \<longrightarrow> (g \<and> more); f \<and> \<not> ((g \<and> more); (g\<^sup>\<star>; h)) " 
   using "5" "9" by fastforce
 have 11: "\<turnstile> g \<and>  more  \<longrightarrow>  more " 
   by fastforce
 from 10 11 show ?thesis using ChopContraB by blast
qed

lemma DiamondAndEmptyEqvAndEmpty:
 "\<turnstile> (\<diamond> f \<and> empty) = (f \<and> empty)"
 by (metis ChopAndEmptyEqvEmptyChopEmpty EmptyChop FiniteAndEmptyEqvEmpty int_eq sometimes_d_def)


lemma InitAndEmptyEqvAndEmpty:
 "\<turnstile> ((init w) \<and> empty) = (w \<and> empty) "
proof -
 have 1: "\<turnstile> ((init w) \<and> empty) = ((w \<and> empty);#True \<and> empty)" 
   by (metis init_d_def int_eq lift_and_com) 
 have 2: "\<turnstile> ((w \<and> empty);#True \<and> empty) = (w \<and> empty);(#True \<and> empty)" 
   by (meson AndChopA ChopAndA ChopAndEmptyEqvEmptyChopEmpty lift_imp_trans Prop11 Prop12)
 have 3: "\<turnstile> (w \<and> empty);(#True \<and> empty) = (w \<and> empty);empty" 
   using RightChopEqvChop by fastforce
 have 4: "\<turnstile> (w \<and> empty);empty = (w \<and> empty)" 
   using ChopEmpty by blast
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma InitAndNotBoxInitImpNotEmpty:
 "\<turnstile> init w \<and> \<not>( \<box> (init w)) \<longrightarrow> \<not>  empty" 
proof -
 have 1: "\<turnstile> ((init w) \<and> empty) = (w \<and> empty) " 
   by (rule InitAndEmptyEqvAndEmpty)
 have 2: "\<turnstile> (\<not>( \<box> (init w)) \<and> empty) = (\<diamond> (\<not> (init w)) \<and> empty)" 
   by (auto simp: always_d_def)
 have 3: "\<turnstile> (\<diamond> (\<not> (init w)) \<and> empty) = (\<not> (init w) \<and> empty)" 
   by (simp add: DiamondAndEmptyEqvAndEmpty)
 have 4: "\<turnstile> (\<not> (init w)) = (init (\<not> w)) " 
   using Initprop(2) by blast
 have 5: "\<turnstile> (\<not> (init w) \<and> empty) = (\<not> w \<and> empty)" 
   using "4" InitAndEmptyEqvAndEmpty by (metis inteq_reflection)
 have 6: "\<turnstile> (\<not>( \<box> (init w)) \<and> empty) = (\<not> w \<and> empty) " 
   using "2" "3" "5" by fastforce
 have 7: "\<turnstile> \<not>(init w \<and> \<not>( \<box> (init w)) \<and> empty) " 
   using "1" "6" by fastforce  
 from 7 show ?thesis by auto
qed

lemma BoxImpTrueChopAndEmpty: 
 "\<turnstile> \<box> f \<longrightarrow> #True;(f \<and> empty) "
using BoxAndChopImport Finprop(3) by fastforce

lemma BoxInitAndMoreImpBoxInitAndMoreAndFinInit:
 "\<turnstile> \<box>( init w) \<and>  more  \<longrightarrow> (\<box> (init w) \<and>  more ) \<and>  fin ( init w) "
proof -
 have 1: "\<turnstile> fin ( init w) = #True ; (init w \<and> empty)" using FinEqvTrueChopAndEmpty by blast
 have 2: "\<turnstile> \<box>( init w) \<longrightarrow> #True;(init w \<and> empty) " by (rule BoxImpTrueChopAndEmpty)
 from 1 2 show ?thesis by fastforce
qed

lemma CSImpBox:
 assumes "\<turnstile> f \<longrightarrow>  empty  \<or>  ((\<box> (init w) \<and>  more) \<and> finite ); f "
 shows   "\<turnstile> (init w \<and> f) \<and> finite \<longrightarrow> \<box> (init w) "
proof -
 have  1: "\<turnstile> f \<longrightarrow>  empty  \<or>  ((\<box>( init w) \<and>  more) \<and> finite ); f " 
   using assms by auto
 have  2: "\<turnstile> init w \<and> \<not>( \<box> (init w)) \<longrightarrow> \<not>  empty " 
   by (rule InitAndNotBoxInitImpNotEmpty)
 have  3: "\<turnstile> init w \<and> f \<and> \<not>( \<box> (init w)) \<longrightarrow> ((\<box> (init w) \<and>  more) \<and> finite ); f " 
   using "1" "2" by fastforce
 have  4: "\<turnstile> \<box> (init w) \<and>  more  \<longrightarrow> (\<box> (init w) \<and>  more ) \<and>  fin ( init w) " 
   by (rule BoxInitAndMoreImpBoxInitAndMoreAndFinInit)
 have 41: "\<turnstile> (\<box> (init w) \<and>  more) \<and> finite \<longrightarrow> 
             ((\<box> (init w) \<and>  more ) \<and> finite) \<and>  fin ( init w) "
   using "4" by auto
 hence 5: "\<turnstile> ((\<box>( init w) \<and>  more) \<and> finite ); f \<longrightarrow> 
              (((\<box> (init w) \<and>  more ) \<and> finite) \<and>  fin ( init w) ); f " 
   by (rule LeftChopImpChop)
 have  6: "\<turnstile> (((\<box> (init w) \<and>  more ) \<and> finite) \<and>  fin ( init w) ); f =
              ((\<box>( init w) \<and>  more) \<and> finite ); (init w \<and> f) " 
   using AndFinChopEqvStateAndChop by blast
 have  7: "\<turnstile> \<not>( \<box>( init w)) \<longrightarrow> (\<box> (init w)) yields (\<not>( \<box> (init w))) " 
   by (rule NotBoxStateImpBoxYieldsNotBox)
 have  8: "\<turnstile> (\<box>( init w)) yields (\<not> (\<box> (init w))) \<longrightarrow> 
             ((\<box> (init w) \<and>  more) \<and> finite ) yields (\<not>( \<box>( init w))) " 
   using AndYieldsA 
   by (metis AndMoreAndFiniteEqvAndFmore inteq_reflection)
 have  9: "\<turnstile> ((\<box>( init w) \<and>  more) \<and> finite ); (init w \<and> f) \<and> 
              ((\<box>( init w) \<and>  more) \<and> finite ) yields (\<not>( \<box> (init w)))
             \<longrightarrow>  
             ((\<box> (init w) \<and>  more) \<and> finite ); ((init w \<and> f) \<and> \<not> (\<box> (init w))) " 
   by (rule ChopAndYieldsImp)
 have 10: "\<turnstile> (init w \<and> f) \<and> \<not>( \<box> (init w)) \<longrightarrow>
             ((\<box>( init w) \<and>  more) \<and> finite ); ((init w \<and> f) \<and> \<not>( \<box> (init w))) " 
   using "3" "5" "6" "7" "8" "9" by fastforce
 have 11: "\<turnstile> ((\<box>( init w) \<and>  more) \<and> finite ); ((init w \<and> f) \<and> \<not>( \<box> (init w))) \<longrightarrow>
              more ; ((init w \<and> f) \<and> \<not>( \<box> (init w)) ) " 
   by (metis "41" LeftChopImpChop Prop12)
 have 12: "\<turnstile> (init w \<and> f) \<and> \<not>( \<box> (init w)) \<longrightarrow> 
             more ; ((init w \<and> f) \<and> \<not>( \<box> (init w)) ) " 
   using "10" "11" by fastforce
 from 12 show ?thesis using MoreChopContraFiniteB by blast
qed

lemma BoxCSEqvBox:
 "\<turnstile>   (init w \<and> (\<box>( init w))\<^sup>\<star>) = \<box> (init w) "
proof -
 have 1: "\<turnstile> \<box>( init w);\<box>( init w) \<longrightarrow> \<box>( init w)"
   by (simp add: BoxStateChopBoxEqvBox int_iffD1)
 have 2: "\<turnstile> (init w \<and> empty) \<longrightarrow> \<box>( init w)"
   by (simp add: StateAndEmptyImpBoxState)
 have 3: "\<turnstile> (init w \<and> empty) \<or> \<box>( init w);\<box>( init w) \<longrightarrow> \<box>( init w) "
   using "1" "2" by fastforce
 have 4: "\<turnstile> (init w \<and> empty);(\<box>( init w))\<^sup>\<star> \<longrightarrow> \<box> (init w) "
   using ChopstarInductR "3" by blast
 have 5: "\<turnstile> init w \<and> (\<box>( init w))\<^sup>\<star> \<longrightarrow> \<box> (init w) "
   using "4" StateAndEmptyChop by fastforce
 have 11: "\<turnstile> \<box> (init w) \<longrightarrow> (init w) " 
   using BoxElim by blast
 have 12: "\<turnstile> \<box>( init w) \<longrightarrow> (\<box> (init w))\<^sup>\<star> " 
   by (rule ImpCS)
 have 13: "\<turnstile> \<box> (init w) \<longrightarrow> init w \<and> (\<box> (init w))\<^sup>\<star> " 
   using "11" "12" by fastforce
 from 5 13 show ?thesis by fastforce
qed

lemma BoxStateAndCSEqvCS:
 "\<turnstile>   (\<box>( init w) \<and> f\<^sup>\<star> \<and> finite) = (init w \<and> (\<box>( init w) \<and> f)\<^sup>\<star> \<and> finite) "
proof -
 have  1: "\<turnstile> \<box> (init w) \<longrightarrow> init w " 
   using BoxElim by blast
 have  2: "\<turnstile> (f\<^sup>\<star> \<and>  more)  = (f \<and>  more ); f\<^sup>\<star>" 
   by (rule CSAndMoreEqvAndMoreChop)
 have  3: "\<turnstile> (\<box>( init w) \<and> ((f \<and>  more ); f\<^sup>\<star>)) =
            ((\<box> (init w) \<and> f \<and>  more ); (\<box> (init w) \<and> f\<^sup>\<star>)) " 
   by (rule BoxStateAndChopEqvChop)
 have  4: "\<turnstile> \<box> (init w) \<and> f \<and>  more  \<longrightarrow> (\<box> (init w) \<and> f) \<and>  more " 
   by auto
 hence 5: "\<turnstile> (\<box> (init w) \<and> f \<and>  more ); (\<box> (init w) \<and> f\<^sup>\<star>) \<longrightarrow>  
            ((\<box> (init w) \<and> f) \<and>  more ); (\<box> (init w) \<and> f\<^sup>\<star>) " 
   by (rule LeftChopImpChop)
 have  6: "\<turnstile> (\<box>( init w) \<and> f\<^sup>\<star>) \<and>  more  \<longrightarrow>
            ((\<box> (init w) \<and> f) \<and>  more ); (\<box> (init w) \<and> f\<^sup>\<star>) " 
   using "2" "3" "5" by fastforce
 hence 7: "\<turnstile> (\<box> (init w) \<and> f\<^sup>\<star>) \<and> finite \<longrightarrow> (\<box> (init w) \<and> f)\<^sup>\<star>" 
   using CSIntro by blast
 have 71: "\<turnstile> init w \<and> \<box> (init w) \<and> f\<^sup>\<star> \<and> finite \<longrightarrow> init w \<and> (\<box> (init w) \<and> f)\<^sup>\<star> \<and> finite" 
   using "7" by fastforce
 have  8: "\<turnstile> \<box>( init w) \<and> f\<^sup>\<star> \<and> finite \<longrightarrow> init w \<and> (\<box> (init w) \<and> f)\<^sup>\<star> \<and> finite" 
   using "1" "71"  by fastforce
 have 11: "\<turnstile> (\<box> (init w) \<and> f)\<^sup>\<star> \<longrightarrow> (\<box> (init w))\<^sup>\<star> " 
   by (rule AndCSA)
 have 12: "\<turnstile> (init w \<and> (\<box> (init w))\<^sup>\<star>) = \<box> (init w)" 
   by (rule BoxCSEqvBox)
 have 13: "\<turnstile> (\<box> (init w) \<and> f)\<^sup>\<star> \<longrightarrow> f\<^sup>\<star> " 
   by (rule AndCSB)
 have 14: "\<turnstile> init w \<and> (\<box> (init w) \<and> f)\<^sup>\<star> \<longrightarrow> init w \<and> (\<box> (init w))\<^sup>\<star> \<and> f\<^sup>\<star> " 
   using "11" "13" by fastforce
 have 15: "\<turnstile> init w \<and> (\<box> (init w))\<^sup>\<star> \<and> f\<^sup>\<star> \<longrightarrow> \<box> (init w) \<and> f\<^sup>\<star> " 
   using "12" by auto
 have 16: "\<turnstile> init w \<and> (\<box> (init w) \<and> f)\<^sup>\<star> \<longrightarrow> \<box> (init w) \<and> f\<^sup>\<star> " 
   using "14" "15" lift_imp_trans by blast
 from 8 16 show ?thesis by fastforce
qed

lemma BaCSImpCS:
 "\<turnstile>    ba (f \<longrightarrow> g) \<and> finite \<longrightarrow> f\<^sup>\<star> \<longrightarrow> g\<^sup>\<star> "
proof -
 have   1: "\<turnstile> f\<^sup>\<star> =  (empty  \<or>  (f \<and>  more ); f\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 have   2: "\<turnstile> g\<^sup>\<star> =  (empty  \<or>  (g \<and>  more ); g\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 have  21: "\<turnstile> \<not>(g\<^sup>\<star>) =  (\<not>empty  \<and> \<not>( (g \<and>  more ); g\<^sup>\<star>))" 
   using "2" by fastforce
 hence 22: "\<turnstile> \<not>(g\<^sup>\<star>) =  (more  \<and> \<not>( (g \<and>  more ); g\<^sup>\<star>)) " 
   using NotEmptyEqvMore by fastforce
 have   3: "\<turnstile> f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>) \<longrightarrow>
            (empty  \<or> (f \<and>  more ); f\<^sup>\<star>) \<and> more \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) " 
   using "1" "22" by fastforce
 have  31: "\<turnstile> ((empty  \<or> (f \<and>  more ); f\<^sup>\<star>) \<and> more) = ((f \<and>  more ); f\<^sup>\<star> \<and> more)" 
   by (auto simp: empty_d_def)
 have  32: "\<turnstile> f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>) \<longrightarrow> (f \<and>  more ); f\<^sup>\<star> \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>) " 
   using "3" "31" by fastforce
 have   4: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> (f \<and>  more  \<longrightarrow> g \<and>  more ) " 
   by auto
 hence  5: "\<turnstile>  ba (f \<longrightarrow> g) \<longrightarrow>  ba (f \<and>  more  \<longrightarrow> g \<and>  more ) " 
   by (rule BaImpBa)
 have   6: "\<turnstile>  ba (f \<and>  more  \<longrightarrow> g \<and>  more ) \<longrightarrow>
              (f \<and>  more ); f\<^sup>\<star> \<longrightarrow>  (g \<and>  more ); f\<^sup>\<star> " 
   by (rule BaLeftChopImpChop)
 have   7: "\<turnstile>  ba (f \<longrightarrow> g) \<and> (f \<and> more ); f\<^sup>\<star> \<longrightarrow> (g \<and>  more ); f\<^sup>\<star> " 
   using "5" "6" by fastforce
 have   8: "\<turnstile> (g \<and>  more ); f\<^sup>\<star> \<and> \<not> ((g \<and>  more ); g\<^sup>\<star>)
           \<longrightarrow>  (g \<and>  more ); (f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) " 
   by (rule ChopAndNotChopImp)
 have   9: "\<turnstile> (g \<and>  more ); (f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) \<longrightarrow>  more ; (f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) " 
   by (rule AndChopB)
 have  10: "\<turnstile>  ba (f \<longrightarrow> g) \<longrightarrow> more ; (f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) \<longrightarrow>  
              more ; ( ba (f \<longrightarrow> g) \<and> f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) " 
   by (rule BaChopImpChopBa)
 have  11: "\<turnstile>  ba (f \<longrightarrow> g) \<and> f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>) \<longrightarrow>  
              more ; ( ba (f \<longrightarrow> g) \<and> f\<^sup>\<star> \<and> \<not> (g\<^sup>\<star>)) " 
   using "32" "7" "8" "9" "10" by fastforce
 hence 12: "\<turnstile> finite \<longrightarrow> \<not> ( (ba (f \<longrightarrow> g)) \<and> (f\<^sup>\<star>) \<and> (\<not> (g\<^sup>\<star>))) " 
   using MoreChopLoopFiniteB by blast
 from 12 show ?thesis by (simp add: Valid_def)  
qed

lemma BaCSEqvCS:
 "\<turnstile>    ba (f = g) \<and> finite \<longrightarrow> (f\<^sup>\<star> = g\<^sup>\<star>)"
proof -
 have 0: "\<turnstile> (f = g) = ( (f \<longrightarrow> g) \<and>  (g \<longrightarrow> f))"  by fastforce
 have 1: "\<turnstile> ba (f = g) = (ba (f \<longrightarrow> g) \<and> ba (g \<longrightarrow> f))  " by (metis "0" BaAndEqv int_eq) 
 have 2: "\<turnstile> ba (f \<longrightarrow> g) \<and> finite \<longrightarrow> (f\<^sup>\<star> \<longrightarrow> g\<^sup>\<star>)  " by (rule BaCSImpCS)
 have 3: "\<turnstile> ba (g \<longrightarrow> f) \<and> finite \<longrightarrow> (g\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>) " by (rule BaCSImpCS)
 have 4: "\<turnstile> ba (f = g) \<and> finite \<longrightarrow> (f\<^sup>\<star> \<longrightarrow> g\<^sup>\<star>) \<and>  (g\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>) " using "1" "2" "3" by fastforce
 have 5: "\<turnstile> ((f\<^sup>\<star> \<longrightarrow> g\<^sup>\<star>) \<and> (g\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>)) = (f\<^sup>\<star> = g\<^sup>\<star>) " by auto
 from 4 5 show ?thesis by auto
qed

lemma BaAndCSImport:
 "\<turnstile>    ba  f \<and> g\<^sup>\<star> \<and> finite \<longrightarrow> (f \<and> g)\<^sup>\<star> "
proof -
 have  1: "\<turnstile> f \<longrightarrow> (g \<longrightarrow> f \<and> g) " by auto
 hence 2: "\<turnstile>  ba  f \<longrightarrow>  ba  (g \<longrightarrow> f \<and> g)" by (rule BaImpBa)
 have  3: "\<turnstile>  ba  (g \<longrightarrow> f \<and> g) \<and> finite \<longrightarrow> g\<^sup>\<star> \<longrightarrow> (f \<and> g)\<^sup>\<star> " by (rule BaCSImpCS)
 from 2 3 show ?thesis by fastforce
qed

lemma CSSkipImpFinite:
 "\<turnstile> skip\<^sup>\<star> \<longrightarrow> finite"
using CSElimWithoutMore EmptyImpFinite SkipChopFiniteImpFinite by blast

lemma FiniteImpCSSkip:
 "\<turnstile> finite \<longrightarrow> skip\<^sup>\<star>"
using CSIntro
by (metis (no_types, lifting) CSSkipImpFinite ChopAndB  FiniteChopSkipEqvFiniteAndMore 
    FiniteChopSkipEqvSkipChopFinite ImpChopPlus Prop10 Prop12 int_iffD1 inteq_reflection) 

lemma CSSkipEqvFinite:
 "\<turnstile> skip\<^sup>\<star> = finite"
using CSSkipImpFinite FiniteImpCSSkip by fastforce


subsection \<open>Properties of Omega\<close>
 
lemma NotOmegaEmpty:
 "\<turnstile> \<not>((empty)\<^sup>\<omega>)"
proof -
 have 1: "\<turnstile> (empty)\<^sup>\<omega> = (empty \<and> fmore);(empty)\<^sup>\<omega>"
   by (simp add: OmegaUnroll)
 have 2: "\<turnstile> (empty \<and> fmore) = #False"
   using NotFmoreAndEmpty by auto
 have 3: "\<turnstile> #False;(empty)\<^sup>\<omega> = #False"
   by (metis AndInfChopEqvAndInf int_eq int_simps(22))
 from 1 2 3 show ?thesis 
   by (metis TrueW int_simps(3) inteq_reflection)
qed

lemma NotOmegaFalse:
 "\<turnstile> \<not>((#False)\<^sup>\<omega>)"
by (metis ChopImpDi DiIntro NotDiFalse OmegaUnroll int_iffI int_simps(14) 
    int_simps(19) inteq_reflection)

lemma NotOmegaInf:
 "\<turnstile> \<not>((inf)\<^sup>\<omega>)"
proof -
 have 1: "\<turnstile> (inf)\<^sup>\<omega> = (inf \<and> fmore);(inf)\<^sup>\<omega>"
   by (simp add: OmegaUnroll)
 have 2: "\<turnstile> (inf \<and> fmore) = #False"
   using FiniteChopSkipEqvFiniteAndMore FiniteChopSkipEqvSkipChopFinite 
         FmoreEqvSkipChopFinite InfEqvNotFinite by fastforce
 have 3: "\<turnstile> #False;(inf)\<^sup>\<omega> = #False"
   by (metis AndInfChopEqvAndInf int_eq int_simps(22))
 from 1 2 3 show ?thesis 
   by (metis TrueW int_simps(3) inteq_reflection)
qed

lemma OmegaLenPlusOneImpInf:
 "\<turnstile> (len(Suc n))\<^sup>\<omega> \<longrightarrow> inf"
by (metis AndChopB MoreChopLoop OmegaUnroll finite_d_def int_simps(32) inteq_reflection)

lemma InfImpOmegaLenPlusOne:
 "\<turnstile> inf \<longrightarrow> (len(Suc n))\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> inf \<and> #True \<and> \<box>(#True \<longrightarrow>(len(Suc n) \<and> fmore);#True) \<longrightarrow> (len(Suc n))\<^sup>\<omega>"
   using OmegaInduct by blast 
 have 2: "\<turnstile> \<box>(#True \<longrightarrow>(len(Suc n) \<and> fmore);#True) = inf"
   by (auto simp add: Valid_def itl_defs,
       metis enat_min_eq_0_iff ndropn_eq_NNil ndropn_nlast ndropn_nlength nfinite_nlength_enat 
       nlength_NNil the_enat.simps zero_enat_def,
       metis    le_cases   ndropn_nlength nfinite_ndropn_b nfinite_ntaken  ntaken_all,
       metis linear   ndropn_nlength nfinite_ndropn_b nfinite_ntaken ntaken_all,
       metis ndropn_nlength nfinite_ndropn_b nlength_eq_enat_nfiniteD)
 from 1 2 show ?thesis by fastforce
qed

lemma OmegaLenPlusOneEqvInf:
 "\<turnstile> (len(Suc n))\<^sup>\<omega> = inf"
using OmegaLenPlusOneImpInf InfImpOmegaLenPlusOne by fastforce

lemma OmegaSkipEqvInf:
 "\<turnstile> (skip)\<^sup>\<omega> = inf"
proof -
 have 1: "\<turnstile> skip = (len 1)" 
   by (simp add: len_d_def power_d_def) 
     (metis CSSkipEqvFinite ChopEmpty FiniteChopSkipEqvFiniteAndMore 
      FiniteChopSkipEqvSkipChopFinite ImpChopPlus Prop10 Prop12 inteq_reflection) 
 have 2: "\<turnstile> (skip)\<^sup>\<omega> = (len 1)\<^sup>\<omega>"
   using 1 by (metis OmegaUnroll inteq_reflection)
 from 2 show ?thesis using OmegaLenPlusOneEqvInf by fastforce
qed

lemma OmegaTrueImpInf:
 "\<turnstile> (#True)\<^sup>\<omega> \<longrightarrow> inf"
by (metis AndChopB MoreChopLoop OmegaUnroll finite_d_def int_simps(32) inteq_reflection)

lemma InfImpOmegaTrue:
 "\<turnstile> inf \<longrightarrow> (#True)\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> inf \<and> #True \<and> \<box>(#True \<longrightarrow>(#True \<and> fmore);#True) \<longrightarrow> #True\<^sup>\<omega>"
   using OmegaInduct by blast 
 have 2: "\<turnstile> \<box>(#True \<longrightarrow>(#True \<and> fmore);#True) = inf"
   by (auto simp add: Valid_def itl_defs)
      (metis enat_min_eq_0_iff ndropn_eq_NNil ndropn_nlast ndropn_nlength nfinite_nlength_enat
       nlength_NNil the_enat.simps zero_enat_def,
       metis antisym_conv2 canonically_ordered_monoid_add_class.lessE enat.distinct(2) enat_add_sub_same 
       ileI1 less_numeral_extra(1) nlength_eq_enat_nfiniteD one_eSuc one_enat_def  
       zero_le,
       metis ndropn_nlength nfinite_ndropn_b nlength_eq_enat_nfiniteD)
 from 1 2 show ?thesis by fastforce
qed

lemma OmegaTrueEqvInf:
 "\<turnstile> (#True)\<^sup>\<omega> = inf"
using OmegaTrueImpInf InfImpOmegaTrue by fastforce

lemma OmegaMoreImpInf:
 "\<turnstile> (more)\<^sup>\<omega> \<longrightarrow> inf"
by (metis AndChopB MoreChopLoop OmegaUnroll finite_d_def int_simps(32) inteq_reflection)

lemma InfImpOmegaMore:
 "\<turnstile> inf \<longrightarrow> (more)\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> inf \<and> #True \<and> \<box>(#True \<longrightarrow>(more \<and> fmore);#True) \<longrightarrow> more\<^sup>\<omega>"
   using OmegaInduct by blast 
 have 2: "\<turnstile> \<box>(#True \<longrightarrow>(more \<and> fmore);#True) = inf"
   by (auto simp add: Valid_def itl_defs)
      (metis enat_min_eq_0_iff ndropn_eq_NNil ndropn_nlast ndropn_nlength nfinite_nlength_enat 
       nlength_NNil the_enat.simps zero_enat_def,
       metis ndropn_nlength nfinite_ndropn_b nlength_eq_enat_nfiniteD,
       metis add.left_neutral canonically_ordered_monoid_add_class.lessE enat.distinct(2) 
       enat_add_sub_same ileI1 less_numeral_extra(1) nlength_eq_enat_nfiniteD one_eSuc 
       one_enat_def order.not_eq_order_implies_strict zero_enat_def zero_le,
       metis ndropn_nlength nfinite_ndropn_b nlength_eq_enat_nfiniteD)
 from 1 2 show ?thesis by fastforce
qed

lemma OmegaMoreEqvInf:
 "\<turnstile> (more)\<^sup>\<omega> = inf"
using OmegaMoreImpInf InfImpOmegaMore by fastforce

lemma OmegaFiniteImpInf:
 "\<turnstile> (finite)\<^sup>\<omega> \<longrightarrow> inf"
 by (metis AndChopB ChopLoop OmegaUnroll TrueW finite_d_def int_simps(13) int_simps(32) inteq_reflection)

lemma InfImpOmegaFinite:
 "\<turnstile> inf \<longrightarrow> (finite)\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> inf \<and> #True \<and> \<box>(#True \<longrightarrow>(finite \<and> fmore);#True) \<longrightarrow> finite\<^sup>\<omega>"
   using OmegaInduct by blast 
 have 2: "\<turnstile> \<box>(#True \<longrightarrow>(finite \<and> fmore);#True) = inf"
   by (auto simp add: Valid_def itl_defs)
      (metis enat_min_eq_0_iff ndropn_eq_NNil ndropn_nlast ndropn_nlength nfinite_nlength_enat nlength_NNil 
       the_enat.simps zero_enat_def,
       metis antisym_conv2 canonically_ordered_monoid_add_class.lessE enat.distinct(2) enat_add_sub_same 
       ileI1 less_numeral_extra(1)  nlength_eq_enat_nfiniteD one_eSuc one_enat_def 
       zero_le,
       metis ndropn_nlength nfinite_ndropn_b nlength_eq_enat_nfiniteD)
 from 1 2 show ?thesis by fastforce
qed

lemma OmegaFiniteEqvInf:
 "\<turnstile> (finite)\<^sup>\<omega> = inf"
using OmegaFiniteImpInf InfImpOmegaFinite by fastforce

lemma BoxStateAndInfImpOmegaBoxState:
 "\<turnstile> inf \<and> \<box>(init w) \<longrightarrow> (\<box>(init w))\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> inf \<and> (inf \<and> \<box>(init w)) \<and> 
            \<box>((inf \<and> \<box>(init w)) \<longrightarrow> (\<box>(init w) \<and> fmore);(inf \<and> \<box>(init w))) \<longrightarrow> (\<box>(init w))\<^sup>\<omega>"
   using OmegaInduct by blast
 have 10: "\<turnstile>(skip;finite);inf = inf"
   by (metis AndChopA AndMoreAndFiniteEqvAndFmore FiniteChopInfEqvInf FiniteChopSkipEqvFiniteAndMore 
       FiniteChopSkipEqvSkipChopFinite OmegaFiniteEqvInf OmegaUnroll Prop11 inteq_reflection)
 have 11:  "\<turnstile>(inf \<and> \<box> (init w)) = (\<box> (init w) \<and> (skip;finite);inf)"
   using "10" by auto
 have 12: "\<turnstile>(inf \<and> (\<box> (init w) \<and> fmore);\<box> (init w)) = (\<box> (init w) \<and> fmore);(\<box> (init w) \<and> inf)"
   by (meson ChopAndInf Prop04 lift_and_com)
 have 2: "\<turnstile>  (inf \<and> \<box>(init w)) = (inf \<and> (\<box>(init w) \<and> fmore);\<box>(init w))"
   by (metis "11" "12" BoxStateAndChopEqvChop FmoreEqvSkipChopFinite inteq_reflection)
 have 3: "\<turnstile> (inf \<and> (\<box>(init w) \<and> fmore);\<box>(init w)) = (\<box>(init w) \<and> fmore);(inf \<and> \<box>(init w))"
   by (metis ChopAndInf inteq_reflection lift_and_com)
 have 4: "\<turnstile> inf \<and> \<box>(init w) \<longrightarrow> \<box>((inf \<and> \<box>(init w))\<longrightarrow> (\<box>(init w) \<and> fmore);(inf \<and> \<box>(init w)))"
   using "2" "3" by (metis (mono_tags, lifting) BoxGen intD intI inteq_reflection unl_lift2)
 from 1 4 show ?thesis by fastforce
qed
 
lemma OmegaBoxStateImpBoxState:
 "\<turnstile> (\<box>(init w))\<^sup>\<omega> \<and> inf \<longrightarrow>  \<box>(init w)"
proof -
 have 1: "\<turnstile> (\<box>(init w))\<^sup>\<omega> \<longrightarrow> init w"
   by (metis AndChopA BoxEqvAndEmptyOrNextBox OmegaUnroll Prop12 StateAndChop inteq_reflection)
 have 2: "\<turnstile> (\<box>(init w))\<^sup>\<omega> \<longrightarrow> (\<box>(init w) \<and> fmore);((\<box>(init w))\<^sup>\<omega>)"
   by (simp add: OmegaUnroll int_iffD1)
 have 21: " \<turnstile> (\<box>(init w) \<and> fmore) \<longrightarrow> \<circle>(\<box>(init w))"
   by (metis AndChopB BoxStateAndChopEqvChop FmoreEqvSkipChopFinite NextAndEqvNextAndNext 
          Prop12 inteq_reflection next_d_def)
 have 22: " \<turnstile> finite =  (empty \<or>  fmore)"
   by (metis CSEqvOrChopCS CSSkipEqvFinite FmoreEqvSkipChopFinite int_eq)
 have 23: "\<turnstile> (\<box> (init w) \<and> finite) =  ((\<box> (init w) \<and> empty) \<or>  (\<box> (init w) \<and> fmore))"
   using 22 by fastforce
 have 24: "\<turnstile> (\<box> (init w) \<and> empty) = (init w \<and> empty)"
   using BoxEqvAndBox StateAndEmptyImpBoxState by fastforce
 have 25: "\<turnstile> \<circle>(\<box>(init w)) \<and> fmore \<longrightarrow> \<circle>(( (init w) \<and> empty) \<or>  (\<box> (init w) \<and> fmore))"
   using 23 24 by (metis FmoreEqvSkipChopFinite NextAndEqvNextAndNext SkipChopEqvNext int_iffD2 
                   inteq_reflection)
 have 26: "\<And> g. \<turnstile> (\<circle>(( (init w) \<and> empty) \<or>  (\<box> (init w) \<and> fmore)));g =
             (\<circle>(init w \<and> g) \<or> \<circle>((\<box>(init w) \<and> fmore);g))"
   proof -
    fix g :: "'a nellist \<Rightarrow> bool"
    have f1: "\<turnstile>\<circle> (init w \<and> empty \<or> \<box> (init w) \<and> fmore);g = 
               \<circle> ((init w \<and> empty \<or> \<box> (init w) \<and> fmore);g)"
      by (meson NextChop int_eq)
    have "\<turnstile> skip;((init w \<and> empty \<or> \<box> (init w) \<and> fmore);g) = 
            (skip;(init w \<and> g) \<or> skip;((\<box> (init w) \<and> fmore);g))"
      by (metis ChopOrEqvRule OrChopEqv StateAndEmptyChop int_eq)
    then show "\<turnstile> \<circle> (init w \<and> empty \<or> \<box> (init w) \<and> fmore);g = 
                (\<circle> (init w \<and> g) \<or> \<circle> ((\<box> (init w) \<and> fmore);g))"
      by (metis (no_types, opaque_lifting) ChopAssoc Prop04 next_d_def)
    qed
 have 27: "\<turnstile> (\<box> (init w) \<and> fmore) = (\<box> (init w) \<and> skip);(\<box> (init w) \<and> finite)"
   by (metis BoxStateAndChopEqvChop FmoreEqvSkipChopFinite inteq_reflection)
 have 28: "\<turnstile>(init w \<and> empty \<or> \<box> (init w) \<and> fmore) = (\<box> (init w) \<and> finite)"   
   by (metis "23" "24" int_eq)
 have 29: "\<turnstile> (skip;(\<box> (init w) \<and> finite));\<box> (init w) \<^sup>\<omega> = 
             (skip;(init w \<and> \<box> (init w) \<^sup>\<omega>) \<or> skip;((\<box> (init w) \<and> fmore);\<box> (init w) \<^sup>\<omega>))"
   by (metis "26" "28" SkipChopEqvNext int_eq)
 have 3: "\<turnstile> (\<box>(init w) \<and> fmore);((\<box>(init w))\<^sup>\<omega>) \<longrightarrow>
            (\<circle>(init w \<and> (\<box>(init w))\<^sup>\<omega>) \<or> \<circle>((\<box>(init w) \<and> fmore);((\<box>(init w))\<^sup>\<omega>)))"
   by (metis "27" "29" AndChopB LeftChopImpChop inteq_reflection next_d_def)
 have 4: "\<turnstile> (\<circle>(init w \<and> (\<box>(init w))\<^sup>\<omega>) \<or> \<circle>((\<box>(init w) \<and> fmore);((\<box>(init w))\<^sup>\<omega>))) \<longrightarrow>
            \<circle>( (\<box>(init w))\<^sup>\<omega>)"
   by (metis ChopAndB NextImpNext OmegaUnroll Prop02 Prop11 next_d_def)
 have 5: "\<turnstile> (\<box>(init w))\<^sup>\<omega> \<longrightarrow> \<circle>( (\<box>(init w))\<^sup>\<omega>)"
   using "2" "3" "4" by fastforce
 from 1 5 show ?thesis using BoxIntro by (metis Prop01 Prop05 inteq_reflection lift_and_com)
qed

lemma OmegaIntro:
 assumes "\<turnstile> h \<longrightarrow> (f \<and> fmore);h"
 shows   "\<turnstile> h \<and> inf \<longrightarrow> f\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> h \<longrightarrow> (f \<and> fmore);h" using assms by auto
 have 2: "\<turnstile> \<box> (h \<longrightarrow> (f \<and> fmore);h)" by (simp add: BoxGen assms)
 from 1 2 show ?thesis using OmegaInduct by fastforce
qed

lemma OmegaImpRule:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> f\<^sup>\<omega> \<and> inf \<longrightarrow> g\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> (f \<and> fmore) \<longrightarrow> (g \<and> fmore) "
   using assms by auto
 have 2: "\<turnstile> (f \<and> fmore);f\<^sup>\<omega> \<longrightarrow> (g \<and> fmore);f\<^sup>\<omega>"
   using 1 by (simp add: LeftChopImpChop)
 have 3: "\<turnstile> \<box>(f\<^sup>\<omega>\<longrightarrow> (f \<and> fmore);f\<^sup>\<omega>) \<longrightarrow> \<box>(f\<^sup>\<omega> \<longrightarrow> (g \<and> fmore);f\<^sup>\<omega>)"
   by (metis "2" OmegaUnroll int_eq_true int_simps(13) inteq_reflection)
 have 4: "\<turnstile> (inf \<and> f\<^sup>\<omega> \<and> \<box>(f\<^sup>\<omega>\<longrightarrow> (f \<and> fmore);f\<^sup>\<omega>)) \<longrightarrow>
            inf \<and> f\<^sup>\<omega> \<and> \<box>(f\<^sup>\<omega> \<longrightarrow> (g \<and> fmore);f\<^sup>\<omega>)"
   using "3" by fastforce
 have 5: "\<turnstile>  inf \<and> f\<^sup>\<omega> \<and> \<box>(f\<^sup>\<omega> \<longrightarrow> (g \<and> fmore);f\<^sup>\<omega>) \<longrightarrow> g\<^sup>\<omega>"
   using OmegaInduct by blast
 have 6: "\<turnstile> f\<^sup>\<omega>\<longrightarrow> (f \<and> fmore);f\<^sup>\<omega>"
   by (simp add: OmegaUnroll int_iffD1)
 have 7: "\<turnstile>\<box>( f\<^sup>\<omega>\<longrightarrow> (f \<and> fmore);f\<^sup>\<omega>)"
   using "6" by (simp add: BoxGen)  
 from 3 5 7 show ?thesis by fastforce
qed   

lemma OmegaEqvRule:
 assumes "\<turnstile> f = g"
 shows   "\<turnstile> f\<^sup>\<omega> = g\<^sup>\<omega>"
using assms using int_eq by force

lemma AndOmegaA:
 "\<turnstile> (f \<and> g)\<^sup>\<omega> \<and> inf \<longrightarrow> f\<^sup>\<omega>"
by (meson OmegaImpRule Prop12 int_iffD2 lift_and_com)

lemma AndOmegaB:
 "\<turnstile> (f \<and> g)\<^sup>\<omega> \<and> inf \<longrightarrow> g\<^sup>\<omega>"
by (meson OmegaImpRule Prop12 int_iffD2 lift_and_com)

lemma BaOmegaImpOmega:
 "\<turnstile> ba (f \<longrightarrow> g) \<and> inf \<longrightarrow> f\<^sup>\<omega> \<longrightarrow> g\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> ba (f \<longrightarrow> g) \<and> (f \<and> fmore);f\<^sup>\<omega> \<longrightarrow>((f \<longrightarrow> g) \<and> (f \<and> fmore) );((f \<longrightarrow> g) \<and> f\<^sup>\<omega>) "
   using BaAndChopImport by fastforce
 have 2: "\<turnstile> (f \<longrightarrow> g) \<and> (f \<and> fmore) \<longrightarrow> ( g \<and> fmore)"
   by auto
 have 3: "\<turnstile> (f \<longrightarrow> g) \<and> f\<^sup>\<omega> \<longrightarrow> f\<^sup>\<omega>"
   by auto
 have 4: "\<turnstile> ba (f \<longrightarrow> g) \<and> (f \<and> fmore);f\<^sup>\<omega> \<longrightarrow> (g \<and> fmore); f\<^sup>\<omega>" 
   using 1 2 3 
   by (metis (no_types, lifting) AndChopB ChopAndB Prop10 int_eq lift_imp_trans)
 have 5: "\<turnstile> ba (f \<longrightarrow> g) \<and> (f \<and> fmore);f\<^sup>\<omega> \<longrightarrow> ((f \<longrightarrow> g) \<and> (f \<and> fmore));(ba(f \<longrightarrow> g) \<and> f\<^sup>\<omega>) " 
   using BaAndChopImportB by blast
 have 6: "\<turnstile> ((f \<longrightarrow> g) \<and> (f \<and> fmore));(ba(f \<longrightarrow> g) \<and> f\<^sup>\<omega>) \<longrightarrow>
            ((g \<and> fmore));(ba(f \<longrightarrow> g) \<and> f\<^sup>\<omega>)"
   using "2" LeftChopImpChop by blast
 have 7: "\<turnstile> (ba (f \<longrightarrow> g) \<and> f\<^sup>\<omega>) \<longrightarrow>(g \<and> fmore);(ba (f \<longrightarrow> g) \<and> f\<^sup>\<omega>)"
   using  OmegaUnroll "5" "6" by fastforce
 have 8: "\<turnstile> (ba (f \<longrightarrow> g) \<and> f\<^sup>\<omega>) \<and> inf \<longrightarrow> g\<^sup>\<omega>"
   using 7 OmegaIntro by blast
 from 8 show ?thesis by fastforce
qed

lemma BaOmegaEqvOmega:
 "\<turnstile> ba (f = g) \<and> inf \<longrightarrow> (f\<^sup>\<omega> = g\<^sup>\<omega>)"
proof -
 have 0: "\<turnstile> (f = g) = ( (f \<longrightarrow> g) \<and> (g \<longrightarrow> f)) "  by fastforce
 have 1: "\<turnstile> ba (f = g) = (ba (f \<longrightarrow> g) \<and> ba (g \<longrightarrow> f))  " by (metis "0" BaAndEqv int_eq)
 have 2: "\<turnstile> ba (f \<longrightarrow> g) \<and> inf \<longrightarrow> (f\<^sup>\<omega> \<longrightarrow> g\<^sup>\<omega>)" using BaOmegaImpOmega by blast
 have 3: "\<turnstile> ba (g \<longrightarrow> f) \<and> inf \<longrightarrow> (g\<^sup>\<omega> \<longrightarrow> f\<^sup>\<omega>)" using BaOmegaImpOmega by blast
 have 4: "\<turnstile> ba (f = g) \<and> inf \<longrightarrow> (f\<^sup>\<omega> \<longrightarrow> g\<^sup>\<omega>) \<and> (g\<^sup>\<omega> \<longrightarrow> f\<^sup>\<omega>)" using "1" "2" "3" by fastforce
 have 5: "\<turnstile> ((f\<^sup>\<omega> \<longrightarrow> g\<^sup>\<omega>) \<and> (g\<^sup>\<omega> \<longrightarrow> f\<^sup>\<omega>)) = (f\<^sup>\<omega> = g\<^sup>\<omega>) " by auto
 from 4 5 show ?thesis by auto
qed

lemma BaAndOmegaImport:
 "\<turnstile> ba f \<and> g\<^sup>\<omega> \<and> inf \<longrightarrow> (f \<and> g)\<^sup>\<omega>"
proof -
 have 1: "\<turnstile> f \<longrightarrow> (g \<longrightarrow> (f \<and> g))" by auto
 hence 2: "\<turnstile> ba f \<longrightarrow> ba (g \<longrightarrow> f \<and> g)" by (rule BaImpBa)
 have 3: "\<turnstile>  ba  (g \<longrightarrow> f \<and> g) \<and> inf \<longrightarrow> g\<^sup>\<omega> \<longrightarrow> (f \<and> g)\<^sup>\<omega> " by (rule BaOmegaImpOmega)
 from 2 3 show ?thesis by fastforce
qed

lemma InfAndChop:
 "\<turnstile> (inf \<and> (f \<and> fmore);g) = (f \<and> fmore);(g \<and> inf) "
by (meson ChopAndInf Prop04 lift_and_com)

lemma InfImportBox:
 "\<turnstile> (inf \<and> \<box> f) = \<box> (inf \<and> f)"
by (metis BoxAndBoxEqvBoxRule FiniteChopEqvDiamond FiniteChopFiniteEqvFinite InfEqvNotFinite 
     NotDiamondNotEqvBox OrFiniteInf finite_d_def int_eq)
 
lemma InfImportBoxImp:
  "\<turnstile> (inf \<and> \<box> (f \<longrightarrow> g) ) = \<box> ( (inf \<longrightarrow> f) \<longrightarrow> (inf \<and> g))"
proof -
 have 1: "\<turnstile> (inf \<and> (f \<longrightarrow> g)) = ((inf \<longrightarrow> f) \<longrightarrow> (inf \<and> g))"
   by auto
 show ?thesis 
   by (metis "1" InfImportBox inteq_reflection)
qed



subsection \<open>Properties of While\<close>

lemma WhileEqvIf:
 "\<turnstile> ((while  (init w)  do  f) \<and> finite) =  
     ( if\<^sub>i  (init w)  then  (f; ( while  (init w)  do  f))  else   empty  \<and> finite)"
proof -
 have  1: "\<turnstile>  (while  (init w)  do  f \<and> finite) = 
              ((((init w) \<and> f)\<^sup>\<star> \<and>  fin (\<not>  (init w))) \<and> finite) " 
   by (simp add: while_d_def)
 have  2: "\<turnstile> (init w \<and> f)\<^sup>\<star> =  (empty  \<or>  ((init w \<and> f); (init w \<and> f)\<^sup>\<star>)) " 
   by (rule CSEqvOrChopCS)
 have 21: "\<turnstile> (((init w) \<and> f)\<^sup>\<star> \<and>  fin (\<not>  (init w)) \<and> finite) =
             ((empty  \<or>  ((init w \<and> f); (init w \<and> f)\<^sup>\<star>)) \<and>  fin (\<not>  (init w)) \<and> finite)  " 
   using "2" by fastforce
 have 22: "\<turnstile> ((empty  \<or>  ((init w \<and> f); (init w \<and> f)\<^sup>\<star>)) \<and>  fin (\<not>  (init w)) \<and> finite) = 
            (( empty \<and> fin (\<not>(init w)) \<and> finite) \<or> 
             ( ((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite))  " 
   by auto
 have 23: "\<turnstile> ((((init w) \<and> f)\<^sup>\<star> \<and>  fin (\<not>  (init w))) \<and> finite) =
             (( empty \<and> fin (\<not>(init w)) \<and> finite) \<or> 
             ( ((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite))"
   using "21" "22" by auto
 have  3: "\<turnstile>  (empty  \<and>  fin (\<not> ( init w))) = (\<not> ( init w) \<and>  empty) "
   by (metis FinAndEmpty Prop04 lift_and_com)
 hence 31: "\<turnstile>  (empty  \<and>  fin (\<not> ( init w)) \<and> finite) = (\<not> ( init w) \<and>  empty \<and> finite) "
   by auto
 have  32: "\<turnstile> (\<not> ( init w) \<and>  empty \<and> finite) = (\<not> ( init w) \<and>  empty)"
   using FiniteAndEmptyEqvEmpty by auto
 have  33: "\<turnstile> ( empty \<and> fin (\<not>(init w)) \<and> finite) = (\<not> ( init w) \<and>  empty)"
   using "31" "32" by fastforce
 have  34: "\<turnstile> (( empty \<and> fin (\<not>(init w)) \<and> finite) \<or> 
             ( ((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite)) =
             ((\<not> ( init w) \<and>  empty) \<or> 
             ( ((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite))"
   using "23" "33" by auto 
 have  4: "\<turnstile> (init w \<and> f); (init w \<and> f)\<^sup>\<star> = (init w \<and> (f; (init w \<and> f)\<^sup>\<star>)) " 
   by (rule StateAndChop)
 have 41: "\<turnstile> (((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite) =
             (init w \<and> (f; (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite) " 
   using "4" by auto
 have 42: "\<turnstile> (init w \<and> ((f; (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite)) = 
             (init w \<and> ((f; (init w \<and> f)\<^sup>\<star>) \<and>  fin   (init (\<not> w)) \<and> finite)) " 
   using Initprop(2) by (metis StateAndEmptyChop int_eq)
 have  5: "\<turnstile> ((f; ((init w \<and> f)\<^sup>\<star>)) \<and>  (fin  ( init (\<not> w))) \<and> finite)
             = ((f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  (fin  ( init (\<not> w))) \<and> finite)) " 
   using ChopAndFin by fastforce
 hence 49: "\<turnstile> (init w \<and> (f; (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite ) =
              (init  w \<and> (f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  (fin  ( init (\<not> w))) \<and> finite))"
   using "42" by fastforce
 have 50: "\<turnstile>  (((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite) =
             (init  w \<and> (f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  (fin  ( init (\<not> w))) \<and> finite))"
   by (meson "41" "49" Prop04 lift_and_com)
 have 51: "\<turnstile> (init  w \<and> (f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  (fin  ( init (\<not> w))) \<and> finite )) = 
             (init  w \<and> (f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  (fin (\<not> ( init  w))) \<and> finite))  "
   by (metis (no_types) EmptyChop Initprop(2) inteq_reflection) 
 have 52: "\<turnstile> (((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite) =
           (init w \<and> ((f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init  w)) \<and> finite)) ) " 
   using "50" "51" by fastforce
 have 53: "\<turnstile> ((\<not> ( init w) \<and>  empty) \<or> 
             ( ((init w \<and> f); (init w \<and> f)\<^sup>\<star>) \<and>  fin (\<not>  (init w)) \<and> finite)) =
             ((\<not> ( init w) \<and>  empty) \<or> 
             (init w \<and> ((f \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init  w)) \<and> finite)) ))"
   using "52" "34" by auto
 have  6: "\<turnstile> ((f \<and> finite); (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w))) \<and> finite)) = 
             (f \<and> finite);  (while  (init w)  do  f \<and> finite)" 
   by (simp add: while_d_def)
 have 61: "\<turnstile> (init w \<and> ((f \<and> finite); (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init  w))) \<and> finite))) =
            (init w \<and>  ((f \<and> finite);  (while  (init w)  do  f \<and> finite))) " 
   using "6" by auto
 have 62: "\<turnstile> ((\<not> ( init w) \<and>  empty) \<or> 
              (init w \<and> ((f \<and> finite); (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init  w))) \<and> finite)) ))
            =((\<not> ( init w) \<and>  empty ) \<or>  
              (init w \<and> ((f \<and> finite);(while  (init w)  do  f \<and> finite)) ) ) " 
   using  "61" by fastforce
 have  7: "\<turnstile>  (while  (init w)  do  f   \<and> finite)   
           = (((\<not> ( init w) \<and>  empty ) \<or>  
              (init w \<and> (f ;  while  (init w)  do  f) \<and> finite))) " 
   proof -
    have "\<turnstile>(while (init w) do f \<and> finite) = 
        (\<not> init w \<and> empty \<or> (init w \<and> f);(init w \<and> f)\<^sup>\<star> \<and> fin (\<not> init w) \<and> finite)"
      by (metis "1" "23" "34" inteq_reflection)
    then show ?thesis
      by (metis "21" "22" "34" "52" ChopAndFiniteDist int_eq)
   qed  
 have 71: "\<turnstile> ((if\<^sub>i  (init w) then (f; ( while  (init w)  do  f))  else   empty)  \<and> finite)=
              (((\<not> ( init w) \<and>  empty ) \<or>  (init w \<and> (f;  while  (init w)  do  f)) \<and> finite )) " 
   using FiniteAndEmptyEqvEmpty by (auto simp: ifthenelse_d_def) 
 from 7 71 show ?thesis by fastforce
qed

lemma IfAndFiniteDist:
 "\<turnstile> (if\<^sub>i (init w) then (f;g)  else  empty \<and> finite) = 
    (if\<^sub>i (init w) then ((f \<and> finite);(g \<and> finite))  else  empty)"
proof -
 have 1: "\<turnstile> (if\<^sub>i (init w) then (f;g)  else  empty \<and> finite) =
            (( (init w \<and> (f;g)) \<or> (\<not>(init w) \<and> empty)) \<and> finite)"
   by (auto simp: ifthenelse_d_def)
 have 2: "\<turnstile> (( (init w \<and> (f;g)) \<or> (\<not>(init w) \<and> empty)) \<and> finite) =
            (( (init w \<and> (f;g) \<and> finite) \<or> (\<not>(init w) \<and> empty \<and> finite)))"
   by auto
 have 3: "\<turnstile> (init w \<and> (f;g) \<and> finite) = (init w \<and> (f \<and> finite);(g \<and> finite))"
   using ChopAndFiniteDist by fastforce
 have 4: "\<turnstile> (\<not>(init w) \<and> empty \<and> finite) = (\<not>(init w) \<and> empty)"
   using FiniteAndEmptyEqvEmpty by auto   
 have 5: "\<turnstile> (( (init w \<and> (f;g) \<and> finite) \<or> (\<not>(init w) \<and> empty \<and> finite))) =
            ((init w \<and> (f \<and> finite);(g \<and> finite)) \<or> (\<not>(init w) \<and> empty))"
   by (metis "2" "3" "4" inteq_reflection)
 have 6: "\<turnstile> ((init w \<and> (f \<and> finite);(g \<and> finite)) \<or> (\<not>(init w) \<and> empty)) =
            (if\<^sub>i (init w) then ((f \<and> finite);(g \<and> finite))  else  empty)"
   by (auto simp: ifthenelse_d_def)
 from 1 2 5 6 show ?thesis by (metis inteq_reflection)
qed

lemma WhileChopEqvIf:
 "\<turnstile>   ( (while ( init w) do f) \<and> finite); g =  
       if\<^sub>i (init w) then ((f \<and> finite); ( ((while ( init w) do f) \<and> finite) ; g))  else  g "
proof -
 have  1: "\<turnstile>  (while  (init w)  do  f \<and> finite)=  
              (if\<^sub>i  (init w)  then  (f; ( while  (init w)  do  f))  else   empty  \<and> finite)" 
   by (rule WhileEqvIf)
 have 11: "\<turnstile> (if\<^sub>i  (init w)  then  (f; ( while  (init w)  do  f))  else   empty  \<and> finite) =
             (if\<^sub>i  (init w) then  ((f \<and> finite); ( while  (init w)  do  f \<and> finite))  else   empty)"
   using IfAndFiniteDist by fastforce
 have 12: "\<turnstile> (while  (init w)  do  f \<and> finite) =
             (if\<^sub>i  (init w) then  ((f \<and> finite); ( while  (init w)  do  f \<and> finite))  else   empty)"
   using "1" "11" by fastforce
 hence 2: "\<turnstile> ( (while  (init w)  do  f) \<and> finite); g = 
              (if\<^sub>i  (init w) 
               then (((f \<and> finite); ( while  (init w)  do  f \<and> finite));g)  
               else   (empty;g)) " 
   by (rule IfChopEqvRule)
 have  3: "\<turnstile>  empty ; g = g " 
   by (rule EmptyChop)
 have  4: "\<turnstile> (if\<^sub>i  (init w) 
               then (((f \<and> finite); ( while  (init w)  do  f \<and> finite));g)  
               else   (empty;g))  =
              (if\<^sub>i  (init w) 
               then (((f \<and> finite); ( while  (init w)  do  f \<and> finite));g)  
               else   g)" 
   using "3" using inteq_reflection by fastforce
 have  5: "\<turnstile> (((f \<and> finite); ( while  (init w)  do  f \<and> finite));g) =
             ((f \<and> finite);( ( while  (init w)  do  f \<and> finite);g))" 
   by (rule ChopAssocB)
 have  6: "\<turnstile> (if\<^sub>i  (init w) 
               then (((f \<and> finite); ( while  (init w)  do  f \<and> finite));g)  
               else   g) =
              (if\<^sub>i  (init w) 
               then ((f \<and> finite);( ( while  (init w)  do  f \<and> finite);g))  
               else   g)
             " 
   using "5" using inteq_reflection by fastforce
 show ?thesis
   using "2" "4" "6" by fastforce 
qed

lemma WhileChopEqvIfRule:
 assumes "\<turnstile>  f = ( while  (init w)  do  g \<and> finite); h "
 shows   "\<turnstile> f =  if\<^sub>i  (init w)  then  ((g \<and> finite); f)  else  h "
proof -
 have 1: "\<turnstile> f = ( while  (init w)  do  g \<and> finite); h " 
   using assms by auto
 have 2: "\<turnstile> ( while  (init w)  do  g \<and> finite); h =  
            if\<^sub>i  (init w)  then  ((g \<and> finite); (( while  (init w)  do  g \<and> finite); h))  else  h " 
   by (rule WhileChopEqvIf)
 have 3: "\<turnstile> ((g \<and> finite); f) = ((g \<and> finite); (( while  (init w)  do  g \<and> finite); h)) " 
   using "1" by (rule RightChopEqvChop)
 have 4: "\<turnstile> ((g \<and> finite); (( while  (init w)  do  g \<and> finite); h)) = ((g \<and> finite); f) " 
   using "3" by auto
 have 5: "\<turnstile> if\<^sub>i  (init w)  then  ((g \<and> finite); (( while  (init w)  do  g \<and> finite); h))  else  h =
           if\<^sub>i  (init w)  then  ((g \<and> finite); f)  else  h " 
   using "4" using inteq_reflection by fastforce
 from 1 2 5 show ?thesis by fastforce
qed

lemma WhileImpFin:
 "\<turnstile>    while  (init w)  do  f \<longrightarrow>  fin (\<not> ( init w)) "
proof -
 have 1: "\<turnstile> (init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<longrightarrow>  fin (\<not> ( init w))" by auto
 from 1 show ?thesis by (simp add: while_d_def)
qed

lemma WhileEqvEmptyOrChopWhile:
 "\<turnstile>  (while (init w) do f \<and> finite)= 
     ((\<not> (init w) \<and> empty) \<or> (init w \<and> ((f \<and> more) \<and> finite);(while (init w) do f \<and> finite))) "
proof -
 have  1: "\<turnstile> (init w \<and> f)\<^sup>\<star>  =  (empty  \<or>  ((init w \<and> f) \<and> more ); (init w \<and> f)\<^sup>\<star>) " 
   by (rule ChopstarEqv)
 have  2: "\<turnstile> ((init w \<and> f) \<and>  more)  = (init w \<and> (f \<and> more)) " 
   by auto
 hence 3: "\<turnstile> ((init w \<and> f) \<and>  more); (init w \<and> f)\<^sup>\<star>  = (init w \<and> f \<and> more); (init w \<and> f)\<^sup>\<star> " 
   by (rule LeftChopEqvChop)
 have  4: "\<turnstile> (init w \<and> f)\<^sup>\<star>  =  (empty  \<or>  (init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star>) "  
   using "1" "3" by fastforce
 have  5: "\<turnstile> ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite)  =  
             (( empty  \<and>  fin  (\<not> (init w)) \<and> finite) \<or>  
             ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin (\<not> ( init w)) \<and> finite)) " 
   using "1"  "4" by fastforce
 have  51: "\<turnstile>  ((empty  \<and>  fin  (\<not> ( init w))) \<and> finite) = ((\<not> ( init w) \<and>  empty) \<and> finite) " 
   by (metis FinAndEmpty inteq_reflection lift_and_com)
 have  52:"\<turnstile> ((\<not> ( init w) \<and>  empty) \<and> finite) = (\<not> ( init w) \<and>  empty)"
   using EmptyImpFinite by auto  
 have  6: "\<turnstile> ((empty  \<and>  fin  (\<not> ( init w))) \<and> finite) = (\<not> ( init w) \<and>  empty)"
   using "51" "52" by fastforce
 have 60: "\<turnstile> (( empty  \<and>  fin  (\<not> (init w)) \<and> finite) \<or>  
             ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin (\<not> ( init w)) \<and> finite)) =
             ((\<not> ( init w) \<and>  empty) \<or>  
             ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin (\<not> ( init w)) \<and> finite))" 
   using 6 by fastforce
 have 61: "\<turnstile>  ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite)  = 
              ((\<not> ( init w) \<and>  empty) \<or>  
             ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin (\<not> ( init w)) \<and> finite))"
   by (metis "5" "60" inteq_reflection)
 have  70: "\<turnstile> (init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> = (init w \<and> (f \<and>  more ); (init w \<and> f)\<^sup>\<star>) " 
   by (rule StateAndChop)
 have  7: "\<turnstile> ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin  ( init (\<not> w)) \<and> finite) =
             (init w \<and> (f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin  ( init (\<not> w)) \<and> finite)"
   using "70" by auto 
 have 71: "\<turnstile> ((\<not> ( init w) \<and>  empty) \<or>  
             ((init w \<and> f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin (\<not> ( init w)) \<and> finite)) =
             ((\<not> ( init w) \<and>  empty) \<or>  
             (init w \<and> (f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin  ( init (\<not> w)) \<and> finite) )"
   using "7" 
   by (metis (no_types, opaque_lifting) ChopEmpty Initprop(2) inteq_reflection)
 have  8: "\<turnstile> (( (f \<and>  more ); (init w \<and> f)\<^sup>\<star>) \<and>  fin  ( init (\<not> w)) \<and> finite)  =  
            (((f \<and>  more) \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (  init (\<not> w)) \<and> finite))  " 
   using ChopAndFin by fastforce
 have 81: "\<turnstile> fin (  init (\<not> w)) = fin (\<not> ( init w)) " 
   using FinEqvFin Initprop(2) by fastforce
 have 82: "\<turnstile> ((f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) =  
            (((f \<and>  more) \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) ) " 
   using "8" "81" by (metis inteq_reflection)
 have 83: "\<turnstile> (init w \<and> (f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) =
             (init w \<and> ((f \<and>  more) \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) )"
   using "82" by fastforce
 have 84: "\<turnstile> ((\<not> ( init w) \<and>  empty) \<or>  
             (init w \<and> (f \<and>  more ); (init w \<and> f)\<^sup>\<star> \<and> fin  ( init (\<not> w)) \<and> finite) )  =
             ((\<not> ( init w) \<and>  empty) \<or>  
             (init w \<and> ((f \<and>  more) \<and> finite); ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) ) )"
   by (metis "71" "81" "82" int_eq)
 have  9: "\<turnstile> ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite)  =  
            ((\<not> ( init w) \<and>  empty ) \<or>  
                (init w \<and> ((f \<and>  more) \<and> finite); 
                          ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) ) ) " 
   using "84" "61" by (metis "71" inteq_reflection) 
 have 10: "\<turnstile> ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) =
             (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w))) \<and> finite)"
   by auto
 hence 11: "\<turnstile> ((\<not> ( init w) \<and>  empty ) \<or>  
                (init w \<and> ((f \<and>  more) \<and> finite); 
                          ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite) ) ) =
             ((\<not> ( init w) \<and>  empty ) \<or>  
                (init w \<and> ((f \<and>  more) \<and> finite); 
                           (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w))) \<and> finite) ) )"
   by (metis int_simps(1) inteq_reflection)                       
 have 12: "\<turnstile> ((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w)) \<and> finite)  =
             ((\<not> ( init w) \<and>  empty ) \<or>  
                (init w \<and> ((f \<and>  more) \<and> finite); 
                          (((init w \<and> f)\<^sup>\<star> \<and>  fin (\<not> ( init w))) \<and> finite) ) )"
   by (metis "11" "9" inteq_reflection)
 from 12 show ?thesis by (metis "10" inteq_reflection while_d_def)
qed

lemma WhileIntro:
 assumes "\<turnstile>   \<not> ( init w) \<and> f \<longrightarrow>  empty "
         "\<turnstile> init w \<and> f \<longrightarrow> ((g \<and>  more) \<and> finite ); f"
 shows   "\<turnstile> f \<and> finite \<longrightarrow>  while ( init w) do  g "
proof -
 have  1: "\<turnstile>   \<not> ( init w) \<and> f \<longrightarrow>  empty " 
   using assms by blast
 have  2: "\<turnstile> init w \<and> f \<longrightarrow> ((g \<and>  more) \<and> finite ); f " 
   using assms by blast
 have  3: "\<turnstile>  (while ( init w)  do  g \<and> finite)=  
           ((\<not>  (init w) \<and>  empty ) \<or>  
            (init w \<and> ((g \<and>  more) \<and> finite ); (while  (init w)  do  g \<and> finite))) " 
   by (rule WhileEqvEmptyOrChopWhile)
 hence 31: "\<turnstile> \<not> ( while  (init w)  do  g \<and> finite) = 
           (\<not>( (\<not>  (init w) \<and>  empty ) \<or>  
               (init w \<and> ((g \<and>  more) \<and> finite );  (while  (init w)  do  g \<and> finite))))" 
   by fastforce
 hence 32: "\<turnstile> (f \<and> \<not> ( while  (init w)  do  g \<and> finite)) = 
              (f \<and> \<not>( (\<not>(init w) \<and> empty) \<or>  
                      (init w \<and> ((g \<and>  more) \<and> finite );  (while  (init w)  do  g \<and> finite)))) "  
   by fastforce
 have  33: "\<turnstile> (f \<and> \<not>( (\<not>(init w) \<and> empty) \<or> 
                         (init w \<and> ((g \<and> more) \<and> finite);  (while  (init w)  do  g \<and> finite)))) =
              (f \<and> \<not>(\<not>(init w) \<and>  empty ) \<and> 
                     \<not>(init w \<and> ((g \<and>  more) \<and> finite );  (while  (init w)  do  g \<and> finite))) " 
   by auto
 have  34: "\<turnstile> (f \<and> \<not>(\<not>(init w) \<and> empty) \<and> 
                     \<not>((init w) \<and> (((g \<and>  more) \<and> finite );  (while  (init w)  do  g \<and> finite)))) =
               (f \<and>  ( (init w) \<or> more ) \<and> 
                       (\<not>(init w) \<or> \<not>(((g \<and> more) \<and> finite); (while (init w)  do  g \<and> finite)))) "
   by (auto simp: empty_d_def)
 have  35: "\<turnstile> (f \<and> ((init w) \<or> more) \<and> 
                   (\<not>(init w) \<or> \<not>(((g \<and> more) \<and> finite);(while (init w) do g \<and> finite)))) =
              ((f \<and> (init w) \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> (init w) \<and> \<not>(init w)) \<or> 
              (f \<and> more \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> more \<and> \<not>(init w))) "  
   by auto
 have  36: "\<turnstile> (f \<and> \<not> ( while  (init w)  do  g \<and> finite)) =
              ((f \<and> (init w) \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> (init w) \<and> \<not>(init w)) \<or> 
              (f \<and> more \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> more \<and> \<not>(init w)))"
   by (metis "32" "33" "34" "35" int_eq) 
 have  37: "\<turnstile> \<not>(f \<and> more \<and> \<not>(init w))" 
   using "1" by (auto simp: empty_d_def)
 have  38: "\<turnstile> (f \<and> more \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<longrightarrow> 
             (((g \<and>  more) \<and> finite ); f \<and> 
               \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) " 
   using "1" "2" by (auto simp: empty_d_def Valid_def)
 have  39: "\<turnstile> (f \<and> (init w) \<and> \<not> (((g \<and>  more)\<and> finite );  (while ( init w) do  g \<and> finite))) \<longrightarrow> 
             (((g \<and>  more) \<and> finite ); f  \<and> 
               \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite)))" 
   using "2" by auto
 have  40: "\<turnstile> ((f \<and> (init w) \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> (init w) \<and> \<not>(init w)) \<or> 
              (f \<and> more \<and> \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite))) \<or>
              (f \<and> more \<and> \<not>(init w))) \<longrightarrow>
              ((g \<and>  more) \<and> finite ); f \<and> 
                \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite)) " 
   using "39" "38" "37" "38" by fastforce
 have   4: "\<turnstile> f \<and> \<not> ( while  (init w)  do  g \<and> finite) \<longrightarrow>  
              ((g \<and>  more) \<and> finite ); f \<and> 
               \<not> (((g \<and>  more) \<and> finite );  (while ( init w) do  g \<and> finite)) "
   by (meson "36" "40" Prop11 lift_imp_trans) 
 have   50: "\<turnstile> g \<and>  more  \<longrightarrow>  more " 
   by auto
 have   5: "\<turnstile> (g \<and> more) \<and> finite \<longrightarrow> more"
   by (simp add: "50" Prop05 Prop07 finite_d_def)
 have   6: "\<turnstile> f \<and> finite \<longrightarrow>  ( while  (init w)  do  g \<and> finite)"
   using 4 5 ChopContraB by blast
 from 6 show ?thesis by (simp add: Prop12) 
qed

lemma WhileElim:
 assumes "\<turnstile>  \<not> ( init w) \<and>  empty  \<longrightarrow> g "
         "\<turnstile> init w \<and> ((f \<and>  more) \<and> finite ); g \<longrightarrow> g"
 shows   "\<turnstile>  while  (init w) do  f \<and> finite \<longrightarrow> g "
proof -
 have   1: "\<turnstile>  (while ( init w)  do  f \<and> finite)=  
             ((\<not> ( init w) \<and>  empty ) \<or>  
               (init w \<and> ((f \<and>  more) \<and> finite );  (while ( init w)  do  f \<and> finite))) " 
   by (rule WhileEqvEmptyOrChopWhile)
 hence 11: "\<turnstile> ((while ( init w)  do  f \<and> finite) \<and> \<not> g) =
             (((\<not>(init w) \<and> empty) \<or>  
               (init w \<and> ((f \<and> more) \<and> finite);(while (init w) do f \<and> finite))) \<and> \<not> g) " 
   by auto
 have   2: "\<turnstile> \<not> ( init w) \<and>  empty  \<longrightarrow> g " 
   using assms by blast
 hence 21: "\<turnstile> \<not> g \<longrightarrow> \<not>(\<not> ( init w) \<and>  empty)" 
   by auto
 have  22: "\<turnstile>  ((\<not> (init w) \<and> empty) \<or> 
                 (init w \<and> ((f \<and> more) \<and> finite);(while (init w) do f \<and> finite))) \<and> \<not>  g \<longrightarrow>
              (init w \<and> ((f \<and>  more) \<and> finite );  (while ( init w)  do  f \<and> finite))  " 
   using "21" by auto
 have  23: "\<turnstile> (while ( init w)  do  f \<and> finite) \<and> \<not> g \<longrightarrow>
              (init w \<and> ((f \<and>  more) \<and> finite );  (while ( init w)  do  f \<and> finite)) \<and> \<not>  g" 
   using "11" "21" by fastforce
 have   3: "\<turnstile> (init w) \<and> (((f \<and>  more) \<and> finite ); g) \<longrightarrow> g " 
   using assms by blast
 hence 31: "\<turnstile> \<not> g \<longrightarrow> \<not>((init w) \<and> (((f \<and>  more) \<and> finite ); g))" 
   by fastforce
 have  32: "\<turnstile> (init w \<and> ((f \<and>  more) \<and> finite );  (while ( init w)  do  f \<and> finite)) \<and> \<not> g \<longrightarrow> 
             ((((f \<and>  more) \<and> finite );  (while  (init w)  do  f \<and> finite)) \<and> 
                 \<not> (((f \<and>  more) \<and> finite ); g)) \<and> \<not>g" 
   using "31" by auto
 have   4: "\<turnstile>  (while  (init w)  do  f \<and> finite) \<and> \<not>  g \<longrightarrow>  
              (((f \<and>  more) \<and> finite );  (while  (init w)  do  f \<and> finite)) \<and> 
                 \<not> (((f \<and>  more) \<and> finite ); g) " 
   using "23" "32" by fastforce
 have   5: "\<turnstile> (f \<and>  more) \<and> finite  \<longrightarrow>  more " 
   by auto
 from 4 5 show ?thesis using 
   ChopContraB[of "LIFT(while  (init w)  do  f \<and> finite)" "LIFT(g)" "LIFT(((f \<and>  more) \<and> finite ))"]
   by auto
qed

lemma BaWhileImpWhile:
 "\<turnstile>    ba (f \<longrightarrow> g) \<and> finite \<longrightarrow> ( while  (init w)  do  f) \<longrightarrow> ( while  (init w)  do  g)"
proof -
 have  1: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> ((init w \<and> f)  \<longrightarrow> (init w \<and> g)) " 
   by auto
 hence 2: "\<turnstile>  ba (f \<longrightarrow> g)  \<longrightarrow>  ba ((init w \<and> f)  \<longrightarrow> (init w\<and> g)) " 
   using BaImpBa by blast
 have  3: " \<turnstile>  ba ((init w \<and> f)  \<longrightarrow> (init w\<and> g)) \<and> finite \<longrightarrow> ((init w\<and> f)\<^sup>\<star> \<longrightarrow> (init w\<and> g)\<^sup>\<star>) " 
   by (rule BaCSImpCS)
 have  4: "\<turnstile>  ba (f \<longrightarrow> g) \<and> finite \<longrightarrow> ((init w\<and> f)\<^sup>\<star>\<and> fin (\<not> ( init w)) 
             \<longrightarrow> (init w\<and> g)\<^sup>\<star>\<and> fin (\<not> (init w))) " 
   using "2" "3" by fastforce
 from 4 show ?thesis by (simp add: while_d_def)
qed

lemma WhileImpWhile:
 assumes "\<turnstile>   f\<longrightarrow> g "
 shows "\<turnstile> ( while  (init w)  do  f) \<and> finite \<longrightarrow> ( while  (init w)  do  g) "
proof -
 have  1: "\<turnstile> f \<longrightarrow> g " 
   using assms by auto
 hence 2: "\<turnstile>  ba (f \<longrightarrow> g) " 
   by (rule BaGen)
 have  3: "\<turnstile>  ba (f \<longrightarrow> g) \<and> finite \<longrightarrow> ( while  (init w)  do  f) \<longrightarrow> ( while  (init w)  do  g) " 
   by (rule BaWhileImpWhile)
 have  4: "\<turnstile> ba (f \<longrightarrow> g)  \<longrightarrow> ( while  (init w)  do  f \<and> finite) \<longrightarrow> ( while  (init w)  do  g)"
   using "3" by (auto simp: Valid_def)
 from 2 4 show ?thesis using MP by blast 
qed

subsection \<open>Properties of Halt\<close>

lemma WnextAndMoreEqvNext:
 "\<turnstile> (wnext f \<and> more) = \<circle> f" 
proof -
 have 1: "\<turnstile> wnext f = (empty \<or> \<circle> f)" 
   by (simp add: WnextEqvEmptyOrNext)
 have 2: "\<turnstile> \<circle> f \<longrightarrow> more" 
   by (metis DiIntro LeftChopImpMoreRule di_d_def more_d_def next_d_def)
 have 3: "\<turnstile> ((empty \<or> \<circle> f) \<and> more) = \<circle> f " 
   unfolding empty_d_def using "2" by auto
 show ?thesis by (metis "1" "3" int_eq)
qed

lemma BoxStateAndEmptyEqvStateAndEmpty:
 "\<turnstile> (\<box>(empty = (init w)) \<and> empty) = ((init w) \<and> empty) "
proof -
 have 1: "\<turnstile> ((empty = (init w)) \<and> empty) = ((init w) \<and> empty)"
   by force
 have 2: "\<turnstile> (\<box>(empty = (init w)) \<and> empty) \<longrightarrow> ((init w) \<and> empty)"
   using BoxElim by fastforce
 have 3: "\<turnstile> ((init w) \<and> empty) \<longrightarrow> (\<box>(empty = (init w)) \<and> empty)"
   using BoxEqvAndEmptyOrNextBox by fastforce 
 show ?thesis
   by (simp add: "2" "3" int_iffI)
qed

lemma BoxEmptyEqvIStateqvEmptyAndStateOrNotStateNext:
 "\<turnstile> \<box>(empty = (init w)) = ((empty \<and> init w) \<or> (\<not>(init w) \<and> \<circle>(\<box>(empty = (init w))))) "
proof -
 have  1: "\<turnstile> \<box>(empty = (init w)) = 
            ((\<box>(empty = (init w)) \<and> empty) \<or> (\<box>(empty = (init w)) \<and> more))" 
   by (auto simp: empty_d_def)
 have  2: "\<turnstile> (\<box>(empty = (init w)) \<and> empty) = ((init w) \<and> empty) " 
   using BoxStateAndEmptyEqvStateAndEmpty by blast
 have  3: "\<turnstile> \<box>(empty = (init w))  = ((empty = (init w)) \<and> wnext(\<box>(empty = (init w)))) " 
   using BoxEqvAndWnextBox by blast
 hence 4: "\<turnstile> (\<box>(empty = (init w)) \<and> more)  = 
               (((empty = (init w)) \<and> more) \<and> (wnext(\<box>(empty = (init w))) \<and> more)) " 
   by auto
 have  5: "\<turnstile> ((empty = (init w)) \<and> more) = (\<not>(init w) \<and> more) " 
   by (auto simp: empty_d_def)
 have  6: "\<turnstile> (wnext(\<box>(empty = (init w))) \<and> more) = \<circle>(\<box>(empty = (init w)))  " 
   using WnextAndMoreEqvNext by metis
 have  7: "\<turnstile> (\<box>(empty = (init w)) \<and> more) = 
             ((\<not> (init w) \<and> more) \<and> (wnext(\<box>(empty = (init w))) \<and> more) )" 
   using "4" "5" by fastforce
 have  8: "\<turnstile> ((\<not> (init w) \<and> more) \<and> (wnext(\<box>(empty = (init w))) \<and> more) ) =
             ((\<not> (init w)) \<and> (wnext(\<box>(empty = (init w))) \<and> more) )" 
   by auto
 have  9: "\<turnstile> ((\<not> (init w)) \<and> (wnext(\<box>(empty = (init w))) \<and> more) ) =
              ((\<not> (init w)) \<and> \<circle>(\<box>(empty = (init w)))) " 
   using "8" "6" by auto 
 have 10: "\<turnstile> \<box>(empty = (init w)) = (((init w) \<and> empty) \<or> (\<box>(empty = (init w)) \<and> more) ) "
   using "1" "2" by fastforce
 show ?thesis 
   using "10" "7" "9" by fastforce
qed

lemma HaltStateEqvIfStateThenEmptyElseNext:
 "\<turnstile>  halt(  init w) =  if\<^sub>i  (init w)  then   empty   else ( \<circle>( halt ( init w))) "
by (metis BoxEmptyEqvIStateqvEmptyAndStateOrNotStateNext halt_d_def ifthenelse_d_def inteq_reflection 
    lift_and_com) 

lemma HaltChopEqv:
 "\<turnstile>  ((halt ( init w)) ; f) =  (if\<^sub>i  (init w)  then ( f ) else (\<circle>( (halt ( init w)); f))) "
proof -
 have  1: "\<turnstile>  halt(init w) =  
             (if\<^sub>i  (init w)  then   empty   else ( \<circle>( halt ( init w)))) " 
   by (rule HaltStateEqvIfStateThenEmptyElseNext)
 hence 2: "\<turnstile>  ((halt(init w));f) =  
             (if\<^sub>i  (init w)  then   (empty;f)   else ( \<circle>( halt ( init w));f)) " 
   by (rule IfChopEqvRule)
 have  3: "\<turnstile>  empty ; f = f " 
   by (rule EmptyChop)
 have  4: "\<turnstile> (\<circle> (halt ( init w))); f = \<circle>( halt ( init w); f)" 
   by (rule NextChop)
 from 2 3 4 show ?thesis by (metis inteq_reflection)
qed

lemma AndHaltChopImp:
 "\<turnstile>   init w \<and> ( halt ( init w); f) \<longrightarrow> f "
proof -
 have 1: "\<turnstile> halt ( init w); f =  if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) " 
  by (rule HaltChopEqv)
 have 2: "\<turnstile> init w \<and> if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) \<longrightarrow> f" 
   by (auto simp: ifthenelse_d_def)
 from 1 2 show ?thesis by fastforce
qed
 
lemma NotAndHaltChopImpNext:
 "\<turnstile>  \<not> ( init w) \<and> ( halt  (init w); f) \<longrightarrow> \<circle>( halt ( init w); f) "
proof -
 have 1: "\<turnstile>  halt ( init w); f =  if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) " 
   by (rule HaltChopEqv)
 have 2: "\<turnstile> \<not> ( init w) \<and> if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) \<longrightarrow> 
            \<circle>( halt ( init w); f)"  
   by (auto simp: ifthenelse_d_def) 
 from 1 2 show ?thesis by fastforce
qed

lemma NotAndHaltChopImpSkipYields:
 "\<turnstile>  \<not> ( init w) \<and> ( halt ( init w); f) \<longrightarrow>  skip  yields ( halt  (init w); f) "
proof -
 have 1: "\<turnstile> \<not> ( init w) \<and> ( halt ( init w); f) \<longrightarrow> \<circle>( halt ( init w); f) " 
   by (rule NotAndHaltChopImpNext)
 have 2: "\<turnstile> \<circle>( halt ( init w); f) \<longrightarrow>  skip  yields ( halt ( init w); f) " 
   by (rule NextImpSkipYields)
 from 1 2 show ?thesis by fastforce
qed

lemma FiniteChopAndEmptyEqvChopAndEmpty:
 " \<turnstile> ((finite;(f \<and> empty)) \<and> g) = ((g \<and> finite);(f\<and> empty))  "
proof -
 have 1: "\<turnstile> g \<and> finite;(f \<and> empty) \<longrightarrow> fin f"
   by (metis ChopAndA DiamondFin FinAndEmpty Prop01 Prop05 inteq_reflection sometimes_d_def)
 have 2: "\<turnstile> g \<and> finite;(f \<and> empty) \<longrightarrow> (finite \<and> g) \<and> fin f"
   using "1"  by (metis (no_types, lifting) ChopAndB ChopEmpty Prop10 Prop12 int_iffD1 
                  inteq_reflection) 
 have 3: "\<turnstile> ((finite;(f \<and> empty)) \<and> g) \<longrightarrow> ((g \<and> finite);(f\<and> empty))" 
   using "2" using AndFinEqvChopAndEmpty by fastforce
 have 4: "\<turnstile> ((g \<and> finite);(f\<and> empty)) \<longrightarrow> ((finite;(f \<and> empty)) \<and> g)" 
   by (metis AndChopB ChopAndB ChopEmpty Prop12 inteq_reflection)
 from 3 4 show ?thesis by fastforce
qed

lemma WprevEqvEmptyOrPrev:
 "\<turnstile> wprev f = (empty \<or> prev f)"
using nlength_eq_enat_nfiniteD by (auto simp add: Valid_def itl_defs)

lemma NotChopSkipEqvMoreAndNotChopSkip:
 "\<turnstile> (\<not> f);skip = (more \<and> \<not>(f;skip))"
proof -
 have 1: "\<turnstile> wprev f = (empty \<or> prev f)" using WprevEqvEmptyOrPrev by auto
 hence 2: "\<turnstile> (\<not>(wprev f)) = (\<not>(empty \<or> prev f)) " by auto
 have 3: "\<turnstile> \<not>(wprev f) = ((\<not> f);skip) " by (simp add: wprev_d_def prev_d_def)
 have 31: "\<turnstile> (empty \<or> prev f) = (empty \<or> (f;skip))" by (simp add: prev_d_def)
 have 32: "\<turnstile> (empty \<or> (f;skip)) = (\<not>more \<or> \<not>\<not>(f;skip))" by (simp add: empty_d_def)
 have 33: "\<turnstile> (\<not>more \<or> \<not>\<not>(f;skip)) = (\<not>(more \<and> \<not>(f;skip)))" by fastforce
 have 34: "\<turnstile> (empty \<or> prev f) = (\<not>(more \<and> \<not>(f;skip)))" using "31" "32" "33" by (metis int_eq)
 have 4: "\<turnstile> \<not>(empty \<or> prev f) = (more \<and> \<not>(f;skip))" using "34" by fastforce
 from 2 3 4 show ?thesis by fastforce
qed

lemma HaltChopImpNotHaltChopNot:
 "\<turnstile>    halt ( init w); f \<and> finite \<longrightarrow> \<not> ( halt ( init w); (\<not>  f)) "
proof -
 have   1: "\<turnstile>  halt  (init w); f =  if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) " 
   by (rule HaltChopEqv)
 have   2: "\<turnstile> if\<^sub>i  (init w)  then  f  else ( \<circle>( halt ( init w); f)) \<longrightarrow> 
              ( ((init w) \<longrightarrow> f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); f))))"  
   by (rule IfThenElseImp) 
 have   3: "\<turnstile>  halt  (init w); (\<not>f) =  
               if\<^sub>i  (init w)  then  (\<not>f)  else ( \<circle>( halt ( init w); (\<not>f))) " 
   by (rule HaltChopEqv)
 have   4: "\<turnstile> if\<^sub>i  (init w)  then  (\<not>f)  else ( \<circle>( halt ( init w); (\<not>f))) \<longrightarrow> 
              ( ((init w) \<longrightarrow> \<not>f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); (\<not>f)))))"  
   by (rule IfThenElseImp)
 have   5: "\<turnstile> halt  (init w); f \<and> halt  (init w); (\<not>f) \<longrightarrow>
            ( ((init w) \<longrightarrow> f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); f)))) \<and>
            ( ((init w) \<longrightarrow> \<not>f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); (\<not>f))))) " 
   using "1" "2" "3" "4" by fastforce
 have   6: "\<turnstile> ( ((init w) \<longrightarrow> f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); f)))) \<and>
            ( ((init w) \<longrightarrow> \<not>f) \<and> ( \<not>(init w) \<longrightarrow> ( \<circle>( halt ( init w); (\<not>f))))) \<longrightarrow>
            ( \<circle>( halt ( init w); f)) \<and> ( \<circle>( halt ( init w); (\<not>f))) " 
   by auto
 have   7: "\<turnstile> halt  (init w); f \<and> halt  (init w); (\<not>f) \<longrightarrow>
            ( \<circle>( halt ( init w); f)) \<and> ( \<circle>( halt ( init w); (\<not>f))) " 
   using "5" "6" lift_imp_trans by blast
 have   8: "\<turnstile> (( \<circle>( halt ( init w); f)) \<and> ( \<circle>( halt ( init w); (\<not>f)))) =
             \<circle> (halt ( init w); f \<and> halt ( init w); (\<not>f)) " 
   using NextAndEqvNextAndNext by fastforce
 have   9: "\<turnstile> halt  (init w); f \<and> halt  (init w); (\<not>f) \<longrightarrow> 
              \<circle> (halt ( init w); f \<and> halt ( init w); (\<not>f)) " 
   using "7" "8" by fastforce
 hence 10: "\<turnstile> finite \<longrightarrow> \<not>(halt  (init w); f \<and> halt  (init w); (\<not>f)) " 
   using NextLoop by blast
 from 10 show ?thesis by auto
qed

lemma HaltChopImpHaltYields:
 "\<turnstile>    halt ( init w); f \<and> finite \<longrightarrow> ( halt ( init w)) yields  f "
proof -
 have 1: "\<turnstile>  halt ( init w); f \<and> finite \<longrightarrow> \<not> ( halt ( init w); (\<not>  f)) " 
   by (rule HaltChopImpNotHaltChopNot)
 from 1 show ?thesis by (simp add: yields_d_def)
qed

lemma HaltChopAnd:
 "\<turnstile> ( halt (init w)); f \<and> ( halt  (init w)); g \<and> finite\<longrightarrow> ( halt ( init w)); (f \<and> g) "
proof -
 have  1: "\<turnstile> ( halt (init w)); g \<and> finite \<longrightarrow> ( halt (init w))  yields  g " 
   by (rule HaltChopImpHaltYields)
 hence 2: "\<turnstile> ( halt  (init w)); f \<and> ( halt  (init w)); g \<and> finite \<longrightarrow>
             ( halt  (init w)); f \<and> ( halt  (init w)) yields  g "
   by auto
 have  3: "\<turnstile> ( halt  (init w)); f \<and> ( halt  (init w)) yields  g\<longrightarrow>
             ( halt  (init w)); (f \<and> g) " 
   by (rule ChopAndYieldsImp)
 from 2 3 show ?thesis by fastforce
qed

lemma HaltAndChopAndHaltChopImpHaltAndChopAnd:
 "\<turnstile> (halt (init w) \<and> f);f1 \<and> (halt (init w); g) \<and> finite \<longrightarrow> ( halt ( init w) \<and> f); (f1 \<and> g) "
proof -
 have   1: "\<turnstile> f1 \<longrightarrow> \<not>  g \<or>  (f1 \<and> g) " 
   by auto
 hence  2: "\<turnstile> ( halt  (init w) \<and> f); f1 \<longrightarrow>  
             ( halt  (init w) \<and> f); (\<not>  g) \<or>  (( halt  (init w) \<and> f); (f1 \<and> g)) " 
   by (rule ChopOrImpRule)
 have   3: "\<turnstile> ( halt  (init w) \<and> f); (\<not>  g) \<longrightarrow>  halt  (init w); (\<not>  g) " 
   by (rule AndChopA)
 have  31: "\<turnstile> ( halt  (init w) \<and> f); f1 \<longrightarrow> 
             halt  (init w); (\<not>  g) \<or>  (( halt  (init w) \<and> f); (f1 \<and> g)) " 
   using "2""3" by fastforce
 have   4: "\<turnstile>  halt  (init w); g \<and> finite \<longrightarrow> \<not> ( halt  (init w); (\<not>  g)) " 
   by (rule HaltChopImpNotHaltChopNot)
 hence 41: "\<turnstile> ( halt  (init w); (\<not>  g)) \<and> finite \<longrightarrow> \<not>(halt  (init w); g)" 
   by auto
 have  42: "\<turnstile> ( halt  (init w) \<and> f); f1 \<and> finite \<longrightarrow> 
             \<not>( halt  (init w); g) \<or>  (( halt  (init w) \<and> f); (f1 \<and> g)) " 
   using "31" "41" by fastforce
 from 42 show ?thesis by auto
qed

lemma HaltImpBoxYields:
 "\<turnstile>   ( halt  (init w)); f \<and> finite \<longrightarrow> (\<box>(\<not> ( init w)))  yields  (( halt  (init w)); f)"
proof -
 have   1: "\<turnstile> (\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f)) \<longrightarrow>  di  (\<box> (\<not>  (init w))) " 
   by (rule ChopImpDi)
 have   2: "\<turnstile> \<box> (\<not>  (init w)) \<longrightarrow> \<not>  (init w) " 
   by (rule BoxElim)
 hence  3: "\<turnstile>  di (\<box> (\<not>  (init w))) \<longrightarrow>  di (\<not>  (init w)) " 
   by (rule DiImpDi)
 have   4: "\<turnstile>  di   (init (\<not> w)) =   (init (\<not>w)) " 
   by (rule DiState)
 have  41: "\<turnstile>  (init (\<not> w)) =  (\<not> (init w))" 
   using Initprop(2) by fastforce
 have  42: "\<turnstile>  di  (\<not> (init  w)) =   (\<not>(init w)) " 
   using "4" "41" by (metis inteq_reflection)
 have   5: "\<turnstile> ((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f))) \<longrightarrow> \<not> ( init w) " 
   using "1" "2" "42" using "3" by fastforce
 hence 51: "\<turnstile> ( halt  (init w); f) \<and> ((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f))) \<longrightarrow>
            ( halt  (init w); f) \<and> \<not> ( init w) " 
   by fastforce
 have   6: "\<turnstile>  halt  (init w); f = if\<^sub>i (init w)  then  f  else  (\<circle>( halt  (init w); f)) " 
   by (rule HaltChopEqv)
 hence 61: "\<turnstile> (halt  (init w); f \<and> \<not> ( init w)) = 
              ((if\<^sub>i (init w)  then  f  else  (\<circle>( halt  (init w); f))) \<and> \<not> ( init w)) " 
   using "6" by auto
 have  62: "\<turnstile> (if\<^sub>i (init w)  then  f  else  (\<circle>( halt  (init w); f))) \<and> 
              \<not> ( init w) \<longrightarrow> (\<circle>( halt  (init w); f)) "  
   by (auto simp: ifthenelse_d_def) 
 have  63: "\<turnstile> halt  (init w); f \<and> \<not> ( init w) \<longrightarrow> (\<circle>( halt  (init w); f))  " 
   using "61" "62" by fastforce
 have   7: "\<turnstile> ( halt  (init w); f) \<and> (\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f)) \<longrightarrow>
             \<circle>(( halt  (init w)); f) " 
   using "51" "63" using lift_imp_trans by blast
 have   8: "\<turnstile> \<box> (\<not> (init w)) \<longrightarrow>  empty  \<or>  \<circle>(\<box>(\<not>(  init w))) "
   by (metis BoxImpYields WeakNextBoxImpMoreYields WnextEqvEmptyOrNext fmore_d_def int_eq) 
 hence  9: "\<turnstile> ((\<box> (\<not> ( init w))); (\<not> ( halt  (init w); f))) \<longrightarrow>  
            \<not> ( halt  (init w); f) \<or>  \<circle>((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f))) " 
   by (rule EmptyOrNextChopImpRule)
 hence 10: "\<turnstile> (( halt  (init w)); f) \<and> (\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f)) \<longrightarrow>  
            \<circle>((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f))) " 
   by fastforce
 have  11: "\<turnstile> ( halt  (init w)); f \<and> (\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f)) \<longrightarrow> 
           \<circle>(( halt  (init w)); f) \<and> \<circle>((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f))) " 
   using "7" "10" by fastforce
 have  12: "\<turnstile> \<circle>(( halt  (init w)); f) \<and> \<circle>((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f)))
           \<longrightarrow>  \<circle>((( halt (init w)); f) \<and> ((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f)))) " 
   using NextAndEqvNextAndNext by fastforce
 have  13: "\<turnstile> ( halt  (init w)); f \<and> (\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f)) \<longrightarrow>  
             \<circle>((( halt  (init w)); f) \<and> ((\<box>(\<not>  (init w))); (\<not> ( halt  (init w); f)))) " 
   using "11" "12" by fastforce
 hence 14: "\<turnstile> finite \<longrightarrow> \<not> (( halt  (init w)); f \<and> (\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f))) " 
   using NextLoop by blast
 hence 15: "\<turnstile> ( halt  (init w)); f \<and> finite \<longrightarrow> \<not> ((\<box> (\<not>  (init w))); (\<not> ( halt  (init w); f))) " 
   by auto
 from 15 show ?thesis by (simp add: yields_d_def)
qed

subsection \<open>Properties of Groups of chops\<close>

lemma NestedChopImpChop:
 assumes "\<turnstile>   init w \<and> f \<longrightarrow> g; (init w1 \<and> f1) "
         "\<turnstile> init w1 \<and> f1 \<longrightarrow> g1; (init w2 \<and> f2) "
 shows   "\<turnstile> init w \<and> f \<longrightarrow> g; (g1; (init w2 \<and> f2)) "
proof -
 have  1: "\<turnstile> init w \<and> f \<longrightarrow> g; (init w1 \<and> f1) " using assms(1) by auto
 have  2: "\<turnstile> init w1 \<and> f1 \<longrightarrow> g1; (init w2 \<and> f2) " using assms(2) by auto
 hence 3: "\<turnstile> g; (init w1 \<and> f1) \<longrightarrow> g; (g1; (init w2 \<and> f2)) " by (rule RightChopImpChop)
 from 1 3 show ?thesis by fastforce
qed
 

end


